#!/bin/sh

# This script configures the Razorback library.
# Invoke it from the Razorback root.

# 18-Dec-2001. Andras Aszodi

# ==== Functions ====

# NOTE: Some functions set a global variable called RETVAL
# to store their output value.

# query_user():
# queries the user for a variable.
# The positional parameters are:-
# $1 the query string
# $2 default value for the variable
# Sets RETVAL to the new value of the variable
query_user() { 
    echo -n $1 "["$2"]: "
    read RETVAL
    [ "$RETVAL" = "" ] && RETVAL=$2
}

# query_user_yesno():
# queries the user for a Boolean decision.
# The positional parameters are:-
# $1 the query string
# $2 default value for the variable (specify "y" or "n")
# Sets RETVAL to 1 "[yY]", 0 otherwise.
query_user_yesno() { 
    echo -n $1 "(y/n) ["$2"]: "
    read YESNO
    [ "$YESNO" = "" ] && YESNO=$2
    case $YESNO in
    	[yY]) RETVAL=1;;
	*)    RETVAL=0;;
    esac
}

# read_value_fromfile():
# takes a file called $1, and looks in it for an attribute-value pair
# that looks like "attr=value"
# attr is the parameter $2
# if it is found, value is returned in RETVAL
read_value_fromfile() {
    AWKCMD="BEGIN {FS=\"=\" } /^$2=/ { print \$2 }"
    RETVAL=`awk "$AWKCMD" $1`
}

# Tests if the first positional argument is in the user's path
# and exits with a warning if $1 is not found.
cmd_avail() {
    which $1 2>&1 > /dev/null
    if [ $? != 0 ]; then
    	echo "'"$1"' not available. Exiting"; exit 77
    fi
}

# alpha_cxx_licenses():
# Returns the number of currently available C++ licenses on the Alpha in ALPHA_CXXLICNO.
# Sometimes this number is less than the number of CPUs...
alpha_cxx_licenses() {
    USABLE_UNITS=0
    USABLE_UNITS=`/usr/sbin/lmf list full cache for CXXOSF-USER | awk '/Usable/ { print $3 }'`
    if [ $USABLE_UNITS = 0 ]; then
    	echo "No C++ licenses available at the moment. Try later."
	ALPHA_CXXLICNO=0
	return 2
    fi
    ACT_CHARGE=`/usr/sbin/lmf list full cache for CXXOSF-USER | awk '/Activity/ { print $3 }'`
    if [ $ACT_CHARGE = 0 ]; then
    	ALPHA_CXXLICNO=$USABLE_UNITS
    else
    	ALPHA_CXXLICNO=`expr $USABLE_UNITS / $ACT_CHARGE `
    fi
    return 0
}

# ---- Platform-specific Makefile settings ----

# These functions below define the following Makefile macros:-
# $FLAGS, $SONAME, $PIC, $C[C]DSO_CMD, $TMPLOPTS, $TMPLINK
# that are horribly different for each platform.

makefile_settings_irix32() {
    FLAGS="-n32 -mips3 -TARG:proc=r5000 \
-xansi -OPT:Olimit=9000 -D__INLINE_INTRINSICS -D__MATH_HAS_NO_SIDE_EFFECTS \
-D_POSIX_C_SOURCE=199506L -D_XOPEN_SOURCE -D_PTHREADS -D_REENTRANT -U_NOTHREADS \
-DRAZORBACK_LIB_COMPILE"
    SONAME="-soname"
    PIC="-KPIC"
    CDSO_CMD=" ld -n32 -shared"
    CCDSO_CMD=" CC -n32 -shared"
    TMPLOPTS="-DINCLUDE_TMPL_DEFS -no_auto_include"
    TMPLINK="/usr/lib/DCC/edg_prelink"
}

makefile_settings_irix64() {
    FLAGS="-64 -mips4 -TARG:proc=r10000 \
-xansi -OPT:Olimit=9000 -D__INLINE_INTRINSICS -D__MATH_HAS_NO_SIDE_EFFECTS \
-D_POSIX_C_SOURCE=199506L -D_XOPEN_SOURCE -D_PTHREADS -D_REENTRANT -U_NOTHREADS \
-DRAZORBACK_LIB_COMPILE"
    SONAME="-soname"
    PIC="-KPIC"
    CDSO_CMD=" ld -64 -shared"
    CCDSO_CMD=" CC -64 -shared"
    TMPLOPTS="-DINCLUDE_TMPL_DEFS -no_auto_include"
    TMPLINK="/usr/lib/DCC/edg_prelink"
}

makefile_settings_linux() {
    FLAGS="-D_NO_MATHFLOATFUNC -D_REENTRANT -DRAZORBACK_LIB_COMPILE"
    # we redefine CXX to take care of C++ -specific macros
    if [ "$CXX_MAJOR" -ge 3 ]; then
    	CXX="g++ -D_STANDARD_C_PLUS_PLUS -D__USE_STD_IOSTREAM"
    else
    	CXX="g++ -D_STANDARD_C_PLUS_PLUS"
    fi
    SONAME="-soname "
    PIC="-fPIC"
    CDSO_CMD=" ld -shared"
    CCDSO_CMD=$CDSO_CMD
    TMPLOPTS="-DINCLUDE_TMPL_DEFS"
    TMPLINK="echo"
}

makefile_settings_alpha() {
    # we redefine both CC and CXX
    CC="cc -std1"
    CXX="cxx -std ansi -D_STANDARD_C_PLUS_PLUS -D__USE_STD_IOSTREAM"
    FLAGS="-D_INLINE_INTRINSICS -D_REENTRANT -D_XOPEN_SOURCE=500 -DRAZORBACK_LIB_COMPILE"
    SONAME="-soname "
    PIC=""
    CDSO_CMD="ld -shared -no_archive -expect_unresolved '*'"
    CCDSO_CMD=$CDSO_CMD
    TMPLOPTS="-DINCLUDE_TMPL_DEFS -noimplicit_include"
    TMPLINK="echo"
}

# ==== End of functions ====

# ---- The major and minor version numbers ----

MAJOR=2
MINOR=0

# Works only if invoked from the Razorback root.
# This is necessary for navigating the hierarchy without getting lost...

if [ $0 != "admin/configure.sh" ]; then
    echo "Please invoke me from the Razorback root. Exiting"; exit 1
fi
RAZORBACK_ROOT=`pwd`

echo "***************************************************************"
echo "*                                                             *"
echo "* Welcome to the Razorback configuration script (Version $MAJOR.$MINOR) *"
echo "*                                                             *"
echo "***************************************************************"
echo

# ---- Obtaining architecture information ----

ARCH=`admin/getarch.sh`
if [ "$ARCH" = "UNSUPPORTED" ]; then
    echo "Sorry, this architecture is unsupported. Exiting"; exit 5
fi
if [ "$ARCH" = "irix64" ]; then
    echo "This machine is running 64-bit IRIX."
    query_user_yesno "Configure for 32-bit" "n"
    if [ "$RETVAL" != 0 ]; then
    	ARCH="irix32"
    fi
fi

# if there is a Makefile.defs file from a previous configuration run,
# check if the architecture matches and ask the user whether to use it
# or do a configuration from scratch
MAKEDEFS="Makefile.defs"
if [ -f "$MAKEDEFS" ]; then
    read_value_fromfile $MAKEDEFS "ARCH"
    OLDARCH=$RETVAL
    if [ "$OLDARCH" = "$ARCH" ]; then
    	read_value_fromfile $MAKEDEFS "ARCHDIR"
	OLDARCHDIR=$RETVAL
	if [ -f "$OLDARCHDIR/$OLDARCH/ldpath.csh" ]; then
	    echo "Razorback has already been configured for"
	    echo "the architecture \""$OLDARCH"\""
	    echo "and installation directory \""$OLDARCHDIR"\""
	    query_user_yesno "Do you want to keep this configuration" "y"
	    if [ "$RETVAL" != 0 ]; then
	    	echo "OK, configuration kept."
		echo "Type 'make' to see how to proceed."
	    	exit 0	    # old config kept, exit OK
	    fi
	fi
    fi
fi

# On some multiprocessor architectures it is possible to run Make in parallel.
# We figure out how many processors we have and set up Make accordingly.
# Note: the SGI make will do 2 threads unless the PARALLEL env var is set.
# Note: There is no parallel Make for the Alphas as far as I know.
MAKECOMMAND="make"
case "$ARCH" in
    irix* )
    	CPUNO=`sysconf | grep AVAIL_PROCESSORS | nawk '{print $2}'`
	if [ $CPUNO -ge 2 ]; then
	    MAKECOMMAND="make -P"
	fi
    ;;
    linux )
    	CPUNO=`grep 'processor' /proc/cpuinfo | wc -l`
	if [ $CPUNO -ge 2 ]; then
	    MAKECOMMAND="make -j $CPUNO"
	fi
    ;;
    alpha )
    	CPUNO=`/usr/sbin/psrinfo -n | nawk '{print $7}'`
	# if GNU make is there, then use it in parallel
	# there seems to be no native parallel make
	type gmake > /dev/null
	if [ "$?" = 0 ] && [ "$CPUNO" -ge 2 ]; then
	    # check the no. of available C++ licenses
	    alpha_cxx_licenses
	    if [ $ALPHA_CXXLICNO = 0 ]; then
	    	echo "Sorry, have to exit now"
		exit 11
	    fi
	    if [ $ALPHA_CXXLICNO -ge $CPUNO ]; then
	    	MAKECOMMAND="gmake -j $CPUNO"
	    else
	    	echo "The number of C++ licenses ("$ALPHA_CXXLICNO")"
		echo "is less than the number of CPUs ("$CPUNO")"
	    	MAKECOMMAND="gmake -j $ALPHA_CXXLICNO"
	    fi
	fi
    ;;
    *)
    	CPUNO=1
    ;;
esac

echo
echo "Your architecture is \""$ARCH"\", number of CPUs is "$CPUNO
echo

# ---- Obtaining compiler information ----

# Compiler name and availability: look for native compilers
case $ARCH in
    irix* ) CC="cc"; CXX="CC" ;;
    linux ) CC="gcc"; CXX="g++" ;;
    alpha ) CC="cc"; CXX="cxx" ;;
esac

# C++ versions. Some of them are hopeless, we bail out on those
# Check compiler availability and version
# We look for native compilers on non-Linux platforms
# on Linux, for C/C++ we use GNU, and the Intel Fortran Compiler for F90/95
case $ARCH in 
    irix*) cmd_avail "cc"; cmd_avail "CC"
    	CXX_VERSION=`CC -version 2>&1| awk '{print $4}'`
	CXX_MAJOR=`echo $CXX_VERSION | awk 'BEGIN {FS="."} {print $1}'`
    	if [ "$CXX_MAJOR" -lt 7 ]; then
    	    echo "IRIX C++ compiler version "$CXX_VERSION "too old, exiting"
	    exit 11
	fi
	# The V7.3 compiler returned a version string "7.30"
	# its bugfixes return "7.3.1.xm" (found out on 17-Sep-2002)
	CXX_MINOR=`echo $CXX_VERSION | awk 'BEGIN {FS="."} {print $2}'`
    	if [ "$CXX_MINOR" -lt 3 ] || [ "$CXX_MINOR" -ge 10 -a "$CXX_MINOR" -le 29 ]; then
    	    echo "IRIX C++ compiler version "$CXX_VERSION "too old, exiting"
	    exit 11
	fi
	;;
    alpha) cmd_avail "cc"; cmd_avail "cxx"
    	CXX_VERSION=`cxx -V | awk 'BEGIN {FS="[ V]"} /Compaq/ {print $4}'`
    	;;
    linux) cmd_avail "gcc"; cmd_avail "g++"
    	CXX_VERSION=`g++ -v 2>&1|awk '/version/ {print $3}'`
	CXX_MAJOR=`echo $CXX_VERSION | awk 'BEGIN {FS="."} {print $1}'`
	CXX_MINOR=`echo $CXX_VERSION | awk 'BEGIN {FS="."} {print $2}'`
    	if [ "$CXX_MAJOR" -le 2 ] && [ "$CXX_MINOR" -le 91 ]; then
    	    echo "GNU C++ compiler version "$CXX_VERSION "too old, exiting"
	    exit 12
	fi
	;;
esac
echo
echo "C++ version: "$CXX_VERSION
echo

# ---- Decide on the optimisation level ----

query_user_yesno "Compile optimised version" "y"
OPTIMISE=$RETVAL

# ---- Get the architecture-specific directory ----

echo
echo "The libraries for the architecture \""$ARCH"\""
echo "will be copied to a directory called \"<ARCHDIR>/"$ARCH"/lib\""
echo "<ARCHDIR> must exist and be writable by you (your user ID is "$USER")"
echo "the subdirectory \""$ARCH"/lib\" in <ARCHDIR> will be created by me."
echo

ARCHDIR=""
DEFARCHDIR=$RAZORBACK_ROOT/arch
mkdir -p $DEFARCHDIR
while ( [ "$ARCHDIR" = "" ] || [ ! -w $ARCHDIR ] )
do
    query_user "Full path to <ARCHDIR> directory" $DEFARCHDIR
    ARCHDIR=$RETVAL
done
mkdir -p $ARCHDIR/$ARCH/lib/shared
echo
echo "Installation root is "$ARCHDIR
echo

INCLDIR=$RAZORBACK_ROOT/include

# ---- Set up compilation directories ----

COBJDIR=$RAZORBACK_ROOT/c/obj/$ARCH
mkdir -p $COBJDIR
if [ $? != 0 ]; then
    echo "Cannot create C object directory \""$COBJDIR"\", exiting..."
    exit 21
fi
CCOBJDIR=$RAZORBACK_ROOT/cc/obj/$ARCH
mkdir -p $CCOBJDIR
if [ $? != 0 ]; then
    echo "Cannot create C++ object directory \""$CCOBJDIR"\", exiting..."
    exit 22
fi

# ---- Construct platform-dependent Makefile.defs ----

# First the Makefile flags & settings have to be prepared.
# This is horrendously platform-dependent, the gory details
# are in the makefile_settings_<ARCH> functions.
case $ARCH in
    irix32 ) makefile_settings_irix32 ;;
    irix64 ) makefile_settings_irix64 ;;
    linux )  makefile_settings_linux ;;
    alpha )  makefile_settings_alpha ;;
esac
if [ "$OPTIMISE" = 1 ]; then
    FLAGS="-O2 "$FLAGS
else
    FLAGS="-g "$FLAGS
fi
# Now we have $FLAGS, $SONAME, $PIC, $C[C]DSO_CMD, $TMPLOPTS, $TMPLINK defined
# and can generate Makefile.defs
# this is used both by the Top makefile and the C and C++ makefiles.

cat <<EOF >$RAZORBACK_ROOT/Makefile.defs
# Razorback platform-specific Makefile definitions
# generated automatically by `basename $0`
# on `date`
# Do not edit unless you know what you are doing!
#
# Architecture
ARCH=$ARCH
# Make command
MAKE=$MAKECOMMAND
# Razorback library version
MAJOR=$MAJOR
MINOR=$MINOR
# Compilers
CC=$CC
CXX=$CXX
# Compiler flags
CFLAGS=$FLAGS
CCFLAGS=$FLAGS
# Header locations
INCLUDE=$INCLDIR
CC_INCLUDE=$INCLDIR
# DSO related settings
SONAME=$SONAME
PIC=$PIC
CDSO_CMD=$CDSO_CMD
CCDSO_CMD=$CCDSO_CMD
# Template settings
TMPLOPTS=$TMPLOPTS
TMPLINK=$TMPLINK
# Object directories
COBJDIR=$COBJDIR
CCOBJDIR=$CCOBJDIR
# Installation directory
ARCHDIR=$ARCHDIR
EOF
echo
echo "Wrote platform-dependent Makefile definitions to:-"
echo "\""$RAZORBACK_ROOT/Makefile.defs"\""
echo

# ---- Writing the DSO path-setting csh source file ----

SHAREDLIBDIR=$ARCHDIR/$ARCH/lib/shared
LDPATH_SCRIPT=$ARCHDIR/$ARCH/ldpath.csh
cat <<EOF >$LDPATH_SCRIPT
# Setting the DSO path for Razorback under the C shell
# generated automatically by `basename $0`
# on `date`
# source, not execute
if ( \$?LD_LIBRARY_PATH ) then
        set LDPATH = \${LD_LIBRARY_PATH}
endif
set RAZORPATH = $SHAREDLIBDIR
if ( ! \$?LDPATH ) then
    set LDPATH = \${RAZORPATH}
else
    echo \${LDPATH} | egrep -s "\${RAZORPATH}"
    if ( \$status ) then
        set LDPATH = \${RAZORPATH}:\${LDPATH}
    endif
endif
setenv LD_LIBRARY_PATH \${LDPATH}
unset LDPATH
unset RAZORPATH
EOF
echo
echo "Wrote DSO path-setting csh source to:-"
echo "\""$LDPATH_SCRIPT"\""
echo
echo "======================================================"
echo
echo "Configuration complete."
echo "Now type 'make' to see how to proceed."
