/* ==== FUNCTIONS gor.c ==== */

/* The GOR-III secondary structure prediction method. */

/* ANSI C, 1. Dec. 1994. Andris Aszodi */

/* ---- HEADER ---- */

#include "gor.h"

/* ---- DEFINITIONS ---- */

#define LINELEN 240
#define GAP '-'
#define GOR_POS 17
#define GOR_HALFPOS 8

/* --- PROTOTYPES ---- */

static int get_index(const char *Aalist, char Aa);

/* ==== FUNCTIONS ==== */

/*
 * read_gortable(Fname):
 * reads a GOR table from the ASCII file Fname.
 * The format is rather strict in this case: the first line is the
 * description of the table (Gtbl.Descr), the second
 * line must have the format "# States=x", where x is a positive 
 * integer, stored in Gtbl.Statno (the number of states), 
 * the third line must contain an all-uppercase string
 * (the amino acid permutation Gtbl.Aaperm), its length is Gtbl.Aano. 
 * The rest of the file must contain
 * the tables for each state, starting with a line that serves 
 * as a descriptor (like "Alpha"), stored in Gtbl.States[i].Name, 
 * the first character of this string will be the one-letter code
 * for the structural state. This is followed by a line containing
 * a single integer, the decision constant.
 * The following Gtbl.Aano lines must
 * contain the info. table rows for each amino acid (GOR_POS
 * entries), the lines follow each other in the same order as the
 * AA codes in Gtbl.Aaperm. Rudimentary error checking is done.
 * Return value: the complete Gtbl structure. A read error
 * is indicated by Gtbl.Aano==0 or Gtbl.Statno==0.
 */
Gortable_ read_gortable(const char *Fname)
{
    FILE *Gorf;
    Gortable_ Gtbl={NULL, NULL, 0, 0, NULL};
    char Line[LINELEN];
    int s, i, j;
    char Endl;
    
    if (NULL==(Gorf=fopen(Fname, "r")))
    {
	fprintf(stderr, "! Unable to open GOR table %s\n", Fname);
	return(Gtbl);
    }
    
    /* read name, states, perm */
    if (NULL==fgets(Line, LINELEN, Gorf))
    {
	fprintf(stderr, "! Missing GOR table header in %s\n", Fname);
	fclose(Gorf); return(Gtbl);
    }
    *strchr(Line, '\n')='\0';    /* remove trailing \n */

    Gtbl.Descr=(char *) calloc(strlen(Line)+1, sizeof(char));
    strcpy(Gtbl.Descr, Line);
    if (1!=fscanf(Gorf, "# States=%d\n", &(Gtbl.Statno)) 
	|| Gtbl.Statno<=0)
    {
	fprintf(stderr, "! Invalid or missing States number in %s\n", Fname);
	fclose(Gorf); return(Gtbl);
    }
    fscanf(Gorf, "%s\n", Line); 
    if ((Gtbl.Aano=strlen(Line))>26)
    {
	fprintf(stderr, "! Too many AAs (%d) in perm in %s\n",
	    Gtbl.Aano, Fname);
	fclose(Gorf); return(Gtbl);
    }
    Gtbl.Aaperm=(char *) calloc(Gtbl.Aano+1, sizeof(char));
    strcpy(Gtbl.Aaperm, Line);
    
    /* get the tables for each state */
    Gtbl.States=(State_ *) calloc(Gtbl.Statno, sizeof(State_));
    for (s=0; s<Gtbl.Statno; s++)
    {
	/* read and store the name of the s-th state */
	fscanf(Gorf, "%s\n", Line); 
	Gtbl.States[s].Name=(char *) calloc(strlen(Line)+1, sizeof(char));
	strcpy(Gtbl.States[s].Name, Line);

	/* get the decision constant */
	if (1!=fscanf(Gorf, "%d\n", &(Gtbl.States[s].Dc)))
	{
	    fprintf(stderr, "! Missing decision constant, 0 assumed [state %d (%s)] in %s\n", 
		s, Gtbl.States[s].Name, Fname);
	    Gtbl.States[s].Dc=0.0;
	}
	
	/* read and store the table */
	Gtbl.States[s].Table=(int **) calloc(Gtbl.Aano, sizeof(int *));
	for (i=0; i<Gtbl.Aano; i++)
	{
	    Gtbl.States[s].Table[i]=(int *) calloc(GOR_POS, sizeof(int));
	    for (j=0; j<GOR_POS; j++)
		if (1!=fscanf(Gorf, "%4d", Gtbl.States[s].Table[i]+j)) break;
	    for (Endl='\0'; Endl!='\n';) Endl=getc(Gorf);
	}
    }
    fclose(Gorf);
    return(Gtbl);
}
/* END of read_gortable */

/*
 * gor_predict(Seq, Gtbl, Info):
 * carries out a GOR prediction on a sequence Seq
 * using the tables in Gtbl. Gaps are skipped. The prediction
 * string will contain the first char of the structural state
 * for each residue with the highest probability and GAP
 * characters for gaps.
 * If Info!=NULL, then the detailed information values are 
 * returned in it. Info should be allocated as an array of int pointers
 * long enough to hold the sequence positions. Its pointers
 * will be set within and will point to int arrays Gtbl.Statno
 * long. The info entries are in the same order as in Gtbl.
 * Return value: the string holding the predicted structure codes.
 * If Gtbl was invalid (see above) then NULL is returned.
 */
char *gor_predict(const char *Seq, Gortable_ Gtbl, int *Info[])
{
    int *Sts;
    char *Pred=NULL;
    register int s, i, j, k, ix, Slen=strlen(Seq);
    
    /* a simple error check */
    if (!Gtbl.Statno || !Gtbl.Aano)
    {
	fprintf(stderr, "! Statno=%d, Aano=%d: suspicious, I'd better not predict...\n", 
	    Gtbl.Statno, Gtbl.Aano);
	return(NULL);
    }
    
    if (Info==NULL) /* no detailed output of inform. values */
	Sts=(int *) calloc(Gtbl.Statno, sizeof(int));
    Pred=(char *) calloc(Slen+1, sizeof(char));

    /* main cycle: walk over the sequence positions */
    for (i=0; i<Slen; i++)
    {
	if (Info!=NULL)
	    Sts=(int *) calloc(Gtbl.Statno, sizeof(int));
	if (Seq[i]==GAP)
	{
	    Pred[i]=GAP;    /* don't predict gaps */
	    if (Info!=NULL) Info[i]=Sts;	/* store 0-s */
	    continue;
	}
	
	/* init the state sum with the current pos for all states */
	ix=get_index(Gtbl.Aaperm, Seq[i]);
	for (s=0; s<Gtbl.Statno; s++)
	{
	    Sts[s]=(ix>=0)? Gtbl.States[s].Table[ix][GOR_HALFPOS]: 0;
	    Sts[s]-=Gtbl.States[s].Dc;	/* minus decision constant */
	}
	
	/* walk back */
	for (j=k=-1; j>=-GOR_HALFPOS && i+k>=0; k--)
	{
	    if (Seq[i+k]==GAP) continue;    /* skip gaps */
	    ix=get_index(Gtbl.Aaperm, Seq[i+k]);
	    for (s=0; s<Gtbl.Statno; s++)
		Sts[s]+=(ix>=0)? Gtbl.States[s].Table[ix][GOR_HALFPOS-j]: 0;
	    j--;
	}
	
	/* walk forward */
	for (j=k=1; j<=GOR_HALFPOS && i+k<Slen; k++)
	{
	    if (Seq[i+k]==GAP) continue;    /* skip gaps */
	    ix=get_index(Gtbl.Aaperm, Seq[i+k]);
	    for (s=0; s<Gtbl.Statno; s++)
		Sts[s]+=(ix>=0)? Gtbl.States[s].Table[ix][GOR_HALFPOS-j]: 0;
	    j++;
	}
	
	/* find the highest value for this pos and store pred. code */
	Pred[i]=pred_char(Gtbl, Sts);    /* 1st char of state name */
	if (Info!=NULL) Info[i]=Sts;	/* store details */
    }
    
    if (Info==NULL) free(Sts);
    return(Pred);
}
/* END of gor_predict */

/*
 * gor_pred(Seq, Gtbl, Info):
 * carries out a GOR prediction on a sequence Seq
 * using the tables in Gtbl. Gaps are skipped.
 * The detailed information values are returned in Info[][]
 * which should be allocated prior to the call. Info has as many
 * rows as the length of Seq[] and as Gtbl.Statno columns.
 * The Info entries for each pos are in the same order as in Gtbl.
 * No error check is done here: Gtbl is assumed to be correct.
 * This routine is a version of the gor_predict() function which
 * is more general. Here no prediction code is generated.
 */
void gor_pred(const char *Seq, Gortable_ Gtbl, int *Info[])
{
    register int s, i, j, k, ix, Slen=strlen(Seq);
    
    /* main cycle: walk over the sequence positions */
    for (i=0; i<Slen; i++)
    {
	if (Seq[i]==GAP)    /* don't predict gaps */
	{
	    for (s=0; s<Gtbl.Statno; s++)
		Info[i][s]=0;	/* store 0-s */
	    continue;
	}
	
	/* init the state sum with the current pos for all states */
	ix=get_index(Gtbl.Aaperm, Seq[i]);
	for (s=0; s<Gtbl.Statno; s++)
	{
	    Info[i][s]=(ix>=0)? Gtbl.States[s].Table[ix][GOR_HALFPOS]: 0;
	    Info[i][s]-=Gtbl.States[s].Dc;	/* minus decision constant */
	}
	
	/* walk back */
	for (j=k=-1; j>=-GOR_HALFPOS && i+k>=0; k--)
	{
	    if (Seq[i+k]==GAP) continue;    /* skip gaps */
	    ix=get_index(Gtbl.Aaperm, Seq[i+k]);
	    for (s=0; s<Gtbl.Statno; s++)
		Info[i][s]+=(ix>=0)? Gtbl.States[s].Table[ix][GOR_HALFPOS-j]: 0;
	    j--;
	}
	
	/* walk forward */
	for (j=k=1; j<=GOR_HALFPOS && i+k<Slen; k++)
	{
	    if (Seq[i+k]==GAP) continue;    /* skip gaps */
	    ix=get_index(Gtbl.Aaperm, Seq[i+k]);
	    for (s=0; s<Gtbl.Statno; s++)
		Info[i][s]+=(ix>=0)? Gtbl.States[s].Table[ix][GOR_HALFPOS-j]: 0;
	    j++;
	}
	
    }	    /* for i */
}
/* END of gor_pred */

/* ---- MISCELLANEOUS FUNCTIONS ---- */

/*
 * pred_char(Gtbl, Infos):
 * returns a 1-letter code for the predicted conformation
 * at a given position. Infos[] holds the information values for each
 * conformation defined in Gtbl. The code is the lower-case version
 * of the 1st char of the state names in Gtbl.
 * Returns the code 'c' (for coil) if Gtbl is empty.
 */
char pred_char(Gortable_ Gtbl, const int Infos[])
{
    register int s, sb, Smax;
    
    /* a simple (silent) error check */
    if (!Gtbl.Statno || !Gtbl.Aano)
	return('c');
    
    /* find the highest value for this pos and store pred. code */
    Smax=INT_MIN;
    for (s=0; s<Gtbl.Statno; s++)
    {
	if (Infos[s]>Smax)
	{ Smax=Infos[s]; sb=s; }
    }
    return(tolower(Gtbl.States[sb].Name[0]));    /* 1st char of state name */
}
/* END of pred_char */

/*
 * clean_gortable(Gtbl):
 * frees up all memory allocated to Gtbl.
 */
void clean_gortable(Gortable_ *Gtbl)
{
    int s, i;
    
    for (s=0; s<Gtbl->Statno; s++)
    {
	for (i=0; i<Gtbl->Aano; i++) free(Gtbl->States[s].Table[i]);
	free(Gtbl->States[s].Table);
	free(Gtbl->States[s].Name);
    }
    free(Gtbl->States); free(Gtbl->Aaperm); free(Gtbl->Descr);
}
/* END of clean_gortable */

/* get_index: returns the position of the character Aa in string
 * Aalist or -1 if Aa was not found. Aa is converted to uppercase.
 * Aalist is an "amino acid permutation string" used in the
 * input of relatedness matrices where the amino acid codes are
 * usually not in alphabetic order.
 */
static int get_index(const char *Aalist, char Aa)
{
    char *Pos;
    
    Aa=toupper(Aa);
    Pos=strchr(Aalist, Aa);
    return((Pos==NULL)? -1: Pos-Aalist);
}
/* END of get_index */

#undef LINELEN
#undef GAP
#undef GOR_POS
#undef GOR_HALFPOS

/* ==== END OF FUNCTIONS gor.c ==== */
