/* ==== FUNCTIONS pdbread.c ==== */

/* Protein Data Bank I/O routines. */

/* ANSI C, 19-Oct-1998. Andris Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h> /* SEEK_SET on Suns */
#include <time.h>

/* ---- INCLUDE FILES ---- */

#include "pdbread.h"

/* ---- DEFINITIONS ---- */

#define LINELEN 82  /* line buffer length: 80 chars, newline and \0 */

/* Some math libraries lack the float equivalents of the
 * usual double functions (eg. float sqrtf(float) in addition
 * to double sqrt(double). Define NO_MATHFLOATFUNC in these cases.
 */
#ifdef NO_MATHFLOATFUNC
#define sqrtf sqrt
#endif

/* ---- ALLOCATION MACROS ---- */

/* The following "general alloc" macros are just shorthand for the
 * tedious C memory allocation syntax.
 */
#define PDB_MALLOC(PTR, TYPE) (PTR)=(TYPE*) pdb_malloc(sizeof(TYPE))
#define PDB_CALLOC(PTR, TYPE, SIZE) (PTR)=(TYPE*) pdb_calloc((SIZE), sizeof(TYPE))
#define PDB_REALLOC(PTR, TYPE, SIZE) \
    (PTR)=(TYPE*) pdb_realloc((PTR), (SIZE)*sizeof(TYPE))
#define PDB_FREE(PTR) if ((PTR)!=NULL) { free(PTR); (PTR)=NULL; }

/* ---- FILE-SCOPE GLOBALS ---- */

static long Lastpos=0L;	/* filepos where last good line was found */
static char Line[LINELEN];  /* input/output line buffer */

/* ---- PROTOTYPES ---- */

/* ---- Input ---- */

static char *multicard_string(FILE *Pdb, const char *Record);
static void get_seqres(FILE *Pdb, Pdbentry_ *Entry);
static int read_seqres(const char *Line, char *Seq);

static void read_hetinfo(FILE *Pdb, Pdbentry_ *Entry);
static int locate_het(const Hetgroup_ *Hets, int Hetsno, const Str3_ Hetid);
static int locate_hetnamformul(const Hetnamformul_ *Hnfs, int Hetsno, const Str3_ Hetid);
static char *read_hettag(FILE *Pdb, const char *Rectag, Str3_ Hetid, int *Compnum);

static Secstr_ *get_secstr(FILE *Pdb, int *Secno);
static Ssbond_ *get_ssbonds(FILE *Pdb, int *Ssbondno);
static void read_sites(FILE *Pdb, Pdbentry_ *Entry);

static void read_atom(const char *Line, Atom_ *Atom);
static void get_allatoms(FILE *Pdb, Pdbentry_ *Entry);
static int read_residue(const char *Str, Residue_ *Res);

static Chain_ *locate_chain(const Chain_ *Chains, int Chainno,
	char Chid, int Modelno);
static Hetgroup_ *locate_hetgroup(const Hetgroup_ *Hets, int Hetsno,
    const Atom_ *Atom, int Modelno);
static void init_chain(Chain_ *Chain, char Chid, int Modelno);
static Seqinfo_ *hetname_inseq(const Pdbentry_ *Entry, const Atom_ *Atom);
static void copy_hets(Pdbentry_ *Entry, int Modelno);
static void finish_coords(Chain_ *Chain);

static void get_connects(FILE *Pdb, Pdbentry_ *Entry);
static void read_connect(const char *Line, Connect_ *Connect);

static int ascend_res(const void *Sec1, const void *Sec2);
static int get_record(FILE *Pdb, char *Line, const char *Label);

/* ---- Memory management ---- */

static void *pdb_malloc(size_t Size);
static void *pdb_calloc(size_t Elemno, size_t Size);
static void *pdb_realloc(void *Ptr, size_t Size);
static void *expand_array(void *Array, int Len, int Chunk, int Elsize);

static void free_chain(Chain_ *Chain);
static void free_het(Hetgroup_ *Het);
static void free_hetnamformul(Hetnamformul_ *Het);
static void free_seq(Seqinfo_ *Seqinfo);
static void free_site(Site_ *Site);

/* ---- Output ---- */

static void write_multicard(FILE *Pdb, const char *Tag, const char *Cmp);
static int write_seqres(FILE *Pdb, const Seqinfo_ *Seqinfo);
static void write_het(FILE *Pdb, const Hetgroup_ *Het);
static void write_hetnam(FILE *Pdb, const Hetnamformul_ *Het);
static void write_formul(FILE *Pdb, const Hetnamformul_ *Het);
static int write_site(FILE *Pdb, const Site_ *Site);
static void write_atom(FILE *Pdb, const Atom_ *Atom);
static void write_residue(char *Str, const Residue_ *Res);
static void write_connect(FILE *Pdb, const Connect_ *Connect);

static void print_card(FILE *Pdb, char *Card);
static int merge_secstr(Secstr_ *Secs, int Size, const Chain_ *Chain);
static int secstr_diff(const Secstr_ *S1, const Secstr_ *S2);
static int merge_ssbond(Ssbond_ *Ssbs, int Size, const Chain_ *Chain);
static int ssbond_diff(const Ssbond_ *S1, const Ssbond_ *S2);
static int secstr_outorder(const void *Ssb1, const void *Ssb2);
static int ssbond_outorder(const void *Ssb1, const void *Ssb2);
static long set_pdbdate(Pdbentry_ *Entry);
static const char *current_date();

/* ==== FUNCTIONS ==== */

/* aa_code31: converts 3-letter all-uppercase amino acid abbreviations
 * into 1-char codes. Returns 'X' if it cannot translate Aa3.
 * In addition to the 20 standard AA-s, the following translations
 * are accepted:
 * ALB -> 'A' (beta-alanine)
 * CYH, CSH, CSS, CYX -> 'C' (cysteine and cystine)
 * HYP, PR0, PRZ -> 'P' (hydroxyproline and various Pro synonyms)
 * ILU -> 'I' (isoleucine)
 * TRY-> 'W' (tryptophan)
 * UNK -> 'U' (unknown)
 * any other code is translated to X.
 */
char aa_code31(const char *Aa3)
{
    static char *Names3="ALAALBARGASNASPASXCYSCYHCSHCSSCYXGLNGLUGLXGLYHISILEILULEULYSMETPHEPROPR0PRZHYPSERTHRTRPTRYTYRVALUNKXXX";
    static char *Letters1="AARNDBCCCCCQEZGHIILKMFPPPPSTWWYVUX";
    char *Code;
    
    if (strlen(Aa3)<3) return('X'); /* 1,2-letter codes not accepted */
    Code=strstr(Names3, Aa3);
    if (Code==NULL) return('X');    /* not found */
    else return(Letters1[(Code-Names3)/3]);
}
/* END of aa_code31 */   

/* aa_code13: converts the 1-letter codes to 3-letter standards.
 * Almost the inverse of aa_code31(). Non-standard translations:
 * J->JJJ, O->OOO, U->UNK, X->XXX
 */
char *aa_code13(char Aa1)
{
    static char *Aacode[26]={"ALA","ASX","CYS","ASP","GLU","PHE",
	    "GLY","HIS","ILE","JJJ","LYS","LEU",
	    "MET","ASN","OOO","PRO","GLN","ARG",
	    "SER","THR","UNK","VAL","TRP","XXX",
	    "TYR","GLX"};
    
    if (Aa1<'A' || Aa1>'Z') return(Aacode[23]);	/* 'X' for out-of-range */
    else return(Aacode[Aa1-'A']);
}
/* END of aa_code13 */

/* nu_code31(): converts the "3-letter" nucleotide specifications
 * (right-justified space+letter strings like "  A" or " +I")
 * into their 1-letter equivalent. Returns 'X' if it did not work.
 */
char nu_code31(const char *Nu3)
{
    static char *Names3="  A +A  C +C  G +G  I +I  T +T  U +U";
    static char *Letters1="AACCGGIITTUU";
    char *Code;
    
    if (strlen(Nu3)<3) return('X'); /* 1,2-letter codes not accepted */
    Code=strstr(Names3, Nu3);
    if (Code==NULL) return('X');    /* not found */
    else return(Letters1[(Code-Names3)/3]);
}
/* END of nu_code31 */   

/* ---- PDB INPUT ---- */

/* read_pdb(): reads a PDB file called Pdbfn and puts the result
 * in a Pdbentry_ which is created inside and returned.
 * If something went wrong then NULL is returned.
 */
Pdbentry_ *read_pdb(const char *Pdbfn)
{
    static const unsigned int SECS_CHUNK=8, SSBS_CHUNK=4;
    
    Pdbentry_ *Entry=NULL;	    /* the complete PDB entry */
    FILE *Pdb;
    Secstr_ *Secs=NULL;  /* temp array for sec. structure elements */
    Ssbond_ *Ssbs=NULL;	    /* temp array for S-S bonds */
    Chain_ *Chains;    /* temp ptr to chains */
    char *Cptr;	    /* temp char ptr */
    int i,j, Sno=0, Ssbno=0;

    /* open PDB file */
    if (NULL==(Pdb=fopen(Pdbfn, "r")))
    {
	fprintf(stderr, "? read_pdb(): Cannot open %s\n", Pdbfn);
	return(NULL);
    }
    
    /* go to the beginning of file */
    rewind(Pdb); Lastpos=0L;
    
    /* allocate the entry record */
    if (NULL==(Entry=alloc_pdbentry()))
    {
	fputs("? get_pdb(): Out of memory\n", stderr);
	return(NULL);
    }
    
    /* get the HEADER card */
    if (get_record(Pdb, Line, "HEADER"))
    {
	strncpy(Entry->Header, Line+10, 40); Entry->Header[40]='\0';
	strncpy(Entry->Date, Line+50, 9); Entry->Date[9]='\0';
	strncpy(Entry->Pdbcode, Line+62, 4); Entry->Pdbcode[4]='\0';
    }
    else Entry->Header[0]=Entry->Date[0]=Entry->Pdbcode[0]='\0';   /* missing HEADER */
    
    /* get TITLE cards */
    Entry->Title=multicard_string(Pdb, "TITLE ");
    
    /* get the COMPND cards */
    Entry->Compound=multicard_string(Pdb, "COMPND");

    /* get the SOURCE cards */
    Entry->Source=multicard_string(Pdb, "SOURCE");
    
    /* KEYWDS would be mandatory but skipped */
    
    /* get the EXPDTA card */
    Entry->Expdta[0]='\0';
    if (!get_record(Pdb, Line, "EXPDTA"))
	strcpy(Entry->Expdta, "X-RAY DIFFRACTION"); /* default technique */
    else
	sscanf(Line+10, "%60c", Entry->Expdta);
    
    /* get the AUTHOR cards, do not parse them any further */
    Entry->Author=multicard_string(Pdb, "AUTHOR");
    
    /* REVDAT would be mandatory but skipped */
    
    /* get resolution from REMARK 2, store -1.0 for "NOT APPLICABLE" */
    Entry->Resol=-1.0;
    while (get_record(Pdb, Line, "REMARK"))
    {
	if (strncmp(Line, "REMARK   2", 10)) continue;	/* not second remark */
	Cptr=strstr(Line, "RESOLUTION.");
	if (Cptr==NULL) continue;
	if (NULL!=strstr(Line, "NOT APPLICABLE"))
	{ Entry->Resol=-1.0; break; }
	if (sscanf(Cptr+11, "%f", &(Entry->Resol)))   /* found resolution */
	    break;
    }
    
    /* REMARK 3 would be mandatory but skipped (refinement programs) */
    
    /* all other REMARKs are ignored */
    
    /* DBREF ignored */
    
    /* Get "theoretical" sequence data from the SEQRES records.
     * Note that there might be discrepancies between the sequence
     * declared here and the sequence actually resolved (which
     * comes from the ATOM records). Every Chain_ struct
     * has a ptr to its corresponding theoretical sequence
     * (for MODEL/ENDMDL, there might be many chains for the 
     * same theoretical sequence).
     * SEQADV etc. are ignored. 
     */
    get_seqres(Pdb, Entry);
    
    /* get HET cards */
    read_hetinfo(Pdb, Entry);

    /* get secondary structures (HELIX,SHEET,TURN) in bulk */
    Secs=get_secstr(Pdb, &Sno);

    /* read S-S bond records in bulk */
    Ssbs=get_ssbonds(Pdb, &Ssbno);
    
    /* read SITE records */
    read_sites(Pdb, Entry);
    
    /* crystallographics skipped: spacegroup, matrices etc.... */
    
    /* get coordinate section: ATOM, HETATM,... */
    get_allatoms(Pdb, Entry);
    
    /* read CONECT records */
    get_connects(Pdb, Entry);

    fclose(Pdb);
    
    /* separate secondary structure records by chains */
    for (i=0; i<Sno; i++)
    {
	/* locate the chain where the i-th secstr record belongs:
	 * a given secstr may belong to separate chains as in
	 * NMR structural models e.g. 3CI2
	 */
	for (j=0, Chains=Entry->Chains; j<Entry->Chainno; j++, Chains++)
	    if (Chains->Chid==Secs[i].Chid) /* found, make room,store */
	    {
		Chains->Secs=
			(Secstr_ *)expand_array(Chains->Secs, Chains->Secsno, 
			SECS_CHUNK, sizeof(Secstr_));
		Chains->Secs[Chains->Secsno++]=Secs[i];
	    }
    }
    PDB_FREE(Secs);
    
    /* separate S-S bonds by chains and build interchain S-S bonds */
    for (i=0; i<Ssbno; i++)
    {
	if (Ssbs[i].Ch1!=Ssbs[i].Ch2)	/* interchain */
	{
	    Entry->Ssbs=
		    (Ssbond_ *)expand_array(Entry->Ssbs, Entry->Ssbno,
		    SSBS_CHUNK, sizeof(Ssbond_));
	    Entry->Ssbs[Entry->Ssbno++]=Ssbs[i];
	}
	else	    /* intrachain */
	{
	    /* locate the chain(s) with the same ID */
	    for (j=0, Chains=Entry->Chains; j<Entry->Chainno; j++, Chains++)
		if (Chains->Chid==Ssbs[i].Ch1)
		{
		    Chains->Ssbs=
			    (Ssbond_ *)expand_array(Chains->Ssbs, Chains->Ssbno,
			    SSBS_CHUNK, sizeof(Ssbond_));
		    Chains->Ssbs[Chains->Ssbno++]=Ssbs[i];
		}
	}
    }
    PDB_FREE(Ssbs);
    
    return(Entry);
}
/* END of read_pdb */

/* ---- Initialisation ---- */

/* alloc_pdbentry(): creates and returns a clean Pdbentry_
 * or NULL if allocation failed.
 */
Pdbentry_ *alloc_pdbentry()
{
    Pdbentry_ *Entry=NULL;
    
    PDB_MALLOC(Entry, Pdbentry_);
    if (Entry!=NULL)
    {
	Entry->Header[0]='\0';
	strcpy(Entry->Date, "xxx");
	set_pdbdate(Entry); /* sets current date */
	strcpy(Entry->Pdbcode, "0XXX");
	Entry->Title=NULL; Entry->Compound=NULL;
	Entry->Source=NULL; Entry->Author=NULL;
	strcpy(Entry->Expdta, "X-RAY DIFFRACTION");
	Entry->Resol=-1.0;  /* "not applicable" */
	Entry->Seqs=NULL; Entry->Seqno=0;
	Entry->Chains=NULL; Entry->Ssbs=NULL;
	Entry->Chainno=Entry->Ssbno=0;
	Entry->Hets=NULL; Entry->Hetsno=0;
	Entry->Hetnams=NULL; Entry->Hetnamsno=0;
	Entry->Hetforms=NULL; Entry->Hetformsno=0;
	Entry->Sites=NULL; Entry->Siteno=0;
	Entry->Connects=NULL; Entry->Conno=0;
    }
    return(Entry);
}
/* END of alloc_pdbentry() */

/* ---- Multiline description strings ---- */

/* multicard_string(): get a string which is distributed
 * over several consecutive cards. This applies to "TITLE ", "COMPND", 
 * "SOURCE", "AUTHOR".
 * Returns a freshly allocated and '\0'-terminated string.
 */
static char *multicard_string(FILE *Pdb, const char *Record)
{
    int i;
    char *Str=NULL;
    
    for (i=1; get_record(Pdb, Line, Record); i++)
    {
	PDB_REALLOC(Str, char, 60*i+1);
	strncpy(Str+60*(i-1), Line+10, 60);
	Str[60*i]='\0';	/* not appended */
    }
    return(Str);
}
/* END of multicard_string() */

/* ---- Global sequence information ---- */

/* get_seqres(): read all SEQRES cards from *Pdb and construct an array
 * of Seqinfo_ records in *Entry. These are held globally because
 * there might be several chains sharing the same sequence information
 * (as in NMR models). Note that these sequences do not necessarily
 * match the private sequence information held in each Chain_
 * which corresponds to the resolved residues. 
 */
#define RESNO_PERLINE 13
static void get_seqres(FILE *Pdb, Pdbentry_ *Entry)
{
    static char SEQRESBUFFER[3*RESNO_PERLINE+1];
    
    Seqinfo_ *Seqs=NULL, *Curseqinfo=NULL;
    int Seqno=0, Sernum, Len, Seen;
    
    while(get_record(Pdb, Line, "SEQRES"))
    {
	/* get serial number and length */
	Sernum=0; sscanf(Line+8, "%2d", &Sernum);
	sscanf(Line+13, "%4d", &Len);
	if (!Sernum) continue;	/* unknown sequence, ignore */
	if (Sernum==1)
	{
	    Seqs=(Seqinfo_*) expand_array(Seqs, Seqno, 8, sizeof(Seqinfo_));
	    Curseqinfo=Seqs+Seqno;
	    Seqno++;
	    
	    /* init new Seqinfo_ record */
	    Curseqinfo->Chid=Line[11];
	    Curseqinfo->Len=0;
	    PDB_CALLOC(Curseqinfo->Seq, char, 3*Len+1);
	}
	
	/* read the SEQRES card into the buffer */
	Seen=read_seqres(Line, SEQRESBUFFER);
	if (Len<Curseqinfo->Len+Seen)
	{
	    fprintf(stderr, "\n? get_seqres(): Expected %d, found %d\n",
		    Len, Curseqinfo->Len+Seen);
	    PDB_REALLOC(Curseqinfo->Seq, char, 3*(Curseqinfo->Len+Seen)+1);
	}
	strcpy(Curseqinfo->Seq+3*Curseqinfo->Len, SEQRESBUFFER);
	Curseqinfo->Len+=Seen;
    }
    if (Seqno) PDB_REALLOC(Seqs, Seqinfo_, Seqno);
    Entry->Seqs=Seqs; Entry->Seqno=Seqno;
}
/* END of get_seqres() */

/* read_seqres(): reads one SEQRES card presented in Line into
 * the sequence array Seq which should be 40 chars long, ie.
 * RESNO_PERLINE 3-letter residue names + terminating '\0'. 
 * Returns the number of residues read (0..13).
 */
static int read_seqres(const char *Line, char *Seq)
{
    int i, i3, i4;
    
    Line+=19;	/* offset of the first residue name */
    for (i=i3=i4=0; i<RESNO_PERLINE; i++, i3+=3, i4+=4)
    {
	if (!strncmp(Line+i4, "   ", 3)) break;    /* blank is the end */
	else strncpy(Seq+i3, Line+i4, 3);
    }
    Seq[i3]='\0';
    return(i);
}
/* END of read_seqres() */
#undef RESNO_PERLINE

/* ---- Hetero group information ---- */

/* read_hetinfo(): get the hetero group information
 * from Pdb and store it in Entry. Waters need special treatment.
 */
static void read_hetinfo(FILE *Pdb, Pdbentry_ *Entry)
{
    static const int HETS_CHUNK=8;	/* hetgroup array pre-alloc size */
    Hetgroup_ *Hets=NULL, *Hetptr;   /* temp array and ptr for heteros */
    Hetnamformul_ *Hnfs=NULL;
    Str3_ Hetid;
    int Hetsno, Hetnamsno, Hetformsno, i, j, Compnum, Atomno;
    char *Cptr=NULL;

    for (Hetsno=0; get_record(Pdb, Line, "HET   "); Hetsno++)
    {
	Hets=(Hetgroup_*) expand_array(Hets, Hetsno, HETS_CHUNK, sizeof(Hetgroup_));
	Hetptr=Hets+Hetsno;
	strncpy(Hetptr->Het.Name, Line+7, 3); Hetptr->Het.Name[3]='\0';
	Hetptr->Het.Chid=Line[12];
	sscanf(Line+13, "%4d", &(Hetptr->Het.Resno));
	Hetptr->Het.Rid=Line[17];
	sscanf(Line+20, "%5d     %40c", &Atomno, Hetptr->Descr);
	
	Hetptr->Atoms=NULL; Hetptr->Atomno=Hetptr->Biopolyno=0;
	Hetptr->Solvent=PDB_NOSOLVENT; Hetptr->Modelno=-1;
    }
    /* size may grow by 1 later if there is water in the entry */
    
    /* get HETNAM cards */
    Hetnamsno=0;
    PDB_CALLOC(Hnfs, Hetnamformul_, Hetsno);	/* cannot be more: water has no HETNAM */
    for (i=0; i<Hetsno; i++) Hnfs[i].Name=NULL;
    while (NULL!=(Cptr=read_hettag(Pdb, "HETNAM", Hetid, NULL)))
    {
	i=locate_het(Hets, Hetsno, Hetid);
	if (i>=0)
	{
	    j=locate_hetnamformul(Hnfs, Hetnamsno, Hetid);
	    if (j>=0)
	    {
	    	fprintf(stderr, 
		    "\n? read_hetinfo(): duplicate HETNAM: group \"%s\"\n", Hetid);
	    	free(Cptr);
		continue;
	    }
		
	    PDB_CALLOC(Hnfs[Hetnamsno].Name, char, strlen(Cptr)+1);
	    strcpy(Hnfs[Hetnamsno].Name, Cptr);
	    strcpy(Hnfs[Hetnamsno].Hetid, Hetid);
	    Hnfs[Hetnamsno].Compnum=0;	/* indicates HETNAM */
	    Hetnamsno++;
	}
	else
	{
	    fprintf(stderr, 
		"\n? read_hetinfo(): HETNAM found for unknown group \"%s\"\n", Hetid);
	    free(Cptr);
	}
    }
    
    /* transfer HETNAM records to Entry */
    if (!Hetnamsno)
    {
	PDB_FREE(Hnfs); Entry->Hetnams=NULL;    /* none found */
    }
    else
    {
	if (Hetnamsno<Hetsno)
	    PDB_REALLOC(Hnfs, Hetnamformul_, Hetnamsno);
	Entry->Hetnams=Hnfs;
    }
    Entry->Hetnamsno=Hetnamsno;
    
    /* HETSYN cards are skipped */
    
    /* get FORMUL cards: same as HETNAM, really */
    Hetformsno=0;
    
    /* I have thought that there are at most Hetsno+1
    * formulae (all hetgroups plus water).
    * However, 1AGA and 1CAR contain duplicate FORMUL records.
    * Both are very old files and let us hope such
    * things will not happen again.
    */
    PDB_CALLOC(Hnfs, Hetnamformul_, Hetsno+1);
    for (i=0; i<Hetsno+1; i++) Hnfs[i].Name=NULL;   /* paranoid */
    while (NULL!=(Cptr=read_hettag(Pdb, "FORMUL", Hetid, &Compnum)))
    {
	i=locate_het(Hets, Hetsno, Hetid);
	if (i>=0)
	{
	    j=locate_hetnamformul(Hnfs, Hetformsno, Hetid);
	    if (j>=0)
	    {
	    	fprintf(stderr, 
		    "\n? read_hetinfo(): duplicate FORMUL: group \"%s\"\n", Hetid);
	    	free(Cptr);
		continue;
	    }
		
	    Hnfs[Hetformsno].Name=Cptr;
	    strcpy(Hnfs[Hetformsno].Hetid, Hetid);
	    Hnfs[Hetformsno].Compnum=Compnum;	/* indicates FORMUL */
	    Hetformsno++;
	    
	    /* hack to indicate non-aqueous solvents which have a HET
	     * record, are not UNX, have a * at pos. 19: like MOH in 9PAP
	     * (see locate_hetgroup() later)
	     */
	    if (Cptr[0]=='*' && strncmp(Hetid, "UNX", 3))
		Hets[i].Solvent=PDB_SOLVENT;
	}
	else	/* name not found */
	{
	    if (Cptr[0]!='*')
	    {
		fprintf(stderr, 
		    "\n? read_hetinfo(): FORMUL found for unknown group \"%s\"\n", Hetid);
		free(Cptr);
	    }
	    else
	    {
		/* Water! Add an extra Hetgroup_ */
		if (Hetsno % HETS_CHUNK == 0)	/* pre-alloc'd array was full */
		    PDB_REALLOC(Hets, Hetgroup_, Hetsno+1);
		Hetptr=Hets+Hetsno;
		strncpy(Hetptr->Het.Name, Hetid, 3);
		Hetptr->Het.Name[3]='\0';
		Hetptr->Het.Chid=' ';
		Hetptr->Het.Resno=0; Hetptr->Het.Rid=' '; 
		Hetptr->Descr[0]='\0'; Hetptr->Solvent=PDB_WATER;
		Hetptr->Modelno=-1;
		Hetptr->Atoms=NULL; Hetptr->Atomno=Hetptr->Biopolyno=0;
		Hetsno++;

		/* record formula for those who do not know what H2O is :-) */
		Hnfs[Hetformsno].Name=Cptr;
		strcpy(Hnfs[Hetformsno].Hetid, Hetid);
		Hnfs[Hetformsno].Compnum=Compnum;	/* indicates FORMUL */
		Hetformsno++;
	    }
	}
    }
    if (Hetsno) PDB_REALLOC(Hets, Hetgroup_, Hetsno);   /* adjust final size */
    Entry->Hets=Hets; Entry->Hetsno=Hetsno;	/* add HET array */
    
    /* transfer FORMUL array to Entry */
    if (!Hetformsno)
    {
	PDB_FREE(Hnfs); Entry->Hetforms=NULL;
    }
    else Entry->Hetforms=Hnfs;
    Entry->Hetformsno=Hetformsno;
}
/* END of read_hetinfo() */

/* locate_het(): returns the index of a Hetgroup_ record 
 * in the Hets[] array (Hetsno long) which contains the 
 * description of the hetero group identified through HetID.
 * Returns -1 if not found.
 */
static int locate_het(const Hetgroup_ *Hets, int Hetsno, const Str3_ Hetid)
{
    register int i;
    
    if (Hets==NULL || !Hetsno) return(-1);
    for (i=0; i<Hetsno; i++)
    {
	if (!strncmp(Hets[i].Het.Name, Hetid, 3)) return(i);
    }
    return(-1);
}
/* END of locate_het() */

/* locate_hetnamformul(): returns the index of a Hetnamformul_ record 
 * in the Hnfs[] array (Hetsno long) which contains the 
 * description of the hetero group identified through HetID.
 * Returns -1 if not found.
 */
static int locate_hetnamformul(const Hetnamformul_ *Hnfs, int Hetsno, const Str3_ Hetid)
{
    register int i;
    
    if (Hnfs==NULL || !Hetsno) return(-1);
    for (i=0; i<Hetsno; i++)
    {
	if (!strncmp(Hnfs[i].Hetid, Hetid, 3)) return(i);
    }
    return(-1);
}
/* END of locate_hetnamformul() */

/* read_hettag(): reads the HETNAM and FORMUL cards 
 * (Rectag says which) belonging to one HET group
 * (this can span multiple lines). If successful, then returns
 * a newly created string containing the full description and
 * sets the Hetid string. If there were no Rectag cards, then
 * NULL is returned and Hetid is not modified.
 * Also returns the component number in *Compnum (for FORMUL only).
 */
static char *read_hettag(FILE *Pdb, const char *Rectag, Str3_ Hetid, int *Compnum)
{
    int i, Idoffs, Namoffs, Namwidth;
    char *Nam=NULL;
    long Oldlastpos;
    
    /* read the first */
    if (!get_record(Pdb, Line, Rectag))
	return(NULL);   /* none found */

    /* set the various offsets, description widths
     * and format strings. Why could not these be the same ?!
     */
    if (!strcmp(Rectag, "HETNAM"))
    {
        Idoffs=11; Namoffs=15; Namwidth=55;
    }
    else if (!strcmp(Rectag, "FORMUL"))
    {
        Idoffs=12; Namoffs=18; Namwidth=52;
	sscanf(Line+8, "%2d", Compnum);	/* component number from FORMUL */
    }
    else
    {
        fprintf(stderr, "\n? read_hettag(\"%s\"): Tag not implemented\n", Rectag);
	return(NULL);
    }

    PDB_CALLOC(Nam, char, Namwidth+1);
    strncpy(Hetid, Line+Idoffs, 3); Hetid[3]='\0';
    strncpy(Nam, Line+Namoffs, Namwidth); Nam[Namwidth]='\0';
    
    /* look for continuation */
    for (i=2; Oldlastpos=Lastpos, get_record(Pdb, Line, Rectag); i++)
    {
	if (strncmp(Line+16, "  ", 2)	    /* no continuation number */
		|| strncmp(Line+Idoffs, Hetid, 3))	/* other Hetgroup */
	{
	    Lastpos=Oldlastpos;	    /* skip back one line */
	    break;		    /* and do not carry on */
	}
	
	/* extend the string */
	PDB_REALLOC(Nam, char, Namwidth*i+1);
	strncpy(Nam+Namwidth*(i-1), Line+Namoffs, Namwidth);
	Nam[Namwidth*i]='\0';
    }
    return(Nam);
}
/* END of read_hettag() */

/* ---- Secondary structure ---- */

/* get_secstr(): read all protein secondary structure records
 * into one big array. These will be assigned later into 
 * chains.
 * Returns NULL if none is found. Also returns the length of 
 * the array in *Secno.
 */
static Secstr_ *get_secstr(FILE *Pdb, int *Secno)
{
    static const int SECS_CHUNK=32; /* pre-alloc chunk size */
    
    Secstr_ *Secs=NULL;
    Str3_ Aa3;
    int Sno=0;
    
    while (get_record(Pdb, Line, "HELIX"))	/* read HELIX entries */
    {
	Secs=(Secstr_*) expand_array(Secs, Sno, SECS_CHUNK, sizeof(Secstr_));
	Secs[Sno].Sectype=PDB_HELIX;
	sscanf(Line+7, "%d %s", &(Secs[Sno].No), Secs[Sno].Id);
	Secs[Sno].Chid=Line[19];
	sscanf(Line+15, "%3s", Aa3);
	Secs[Sno].Begaa=aa_code31(Aa3);
	sscanf(Line+21, "%d%c", &(Secs[Sno].Beg), &(Secs[Sno].Begrid));
	sscanf(Line+27, "%3s", Aa3);
	Secs[Sno].Endaa=aa_code31(Aa3);
	sscanf(Line+33, "%d%c", &(Secs[Sno].End), &(Secs[Sno].Endrid));
	sscanf(Line+38, "%d", &(Secs[Sno].Type));
	Sno++;
    }
    while (get_record(Pdb, Line, "SHEET"))	/* read SHEET entries */
    {
	Secs=(Secstr_*) expand_array(Secs, Sno, SECS_CHUNK, sizeof(Secstr_));
	Secs[Sno].Sectype=PDB_SHEET;
	sscanf(Line+7, "%d %s%d", &(Secs[Sno].No), Secs[Sno].Id, &(Secs[Sno].Strandno));
	Secs[Sno].Chid=Line[21];
	sscanf(Line+17, "%3s", Aa3);
	Secs[Sno].Begaa=aa_code31(Aa3);
	sscanf(Line+22, "%d%c", &(Secs[Sno].Beg), &(Secs[Sno].Begrid));
	sscanf(Line+28, "%3s", Aa3);
	Secs[Sno].Endaa=aa_code31(Aa3);
	sscanf(Line+33, "%d%c", &(Secs[Sno].End), &(Secs[Sno].Endrid));
	sscanf(Line+38, "%d", &(Secs[Sno].Type));
	
	/* registration information is "blank" for the first strand
	* or if missing (this happens quite often).
	* Will be indicated by Secs[Sno].Other==0.
	*/
	Line[70]='\0';	/* terminate before old-style decoration */
	if (!Secs[Sno].Type || NULL==strpbrk(Line+41, "0123456789"))
	{
	    Secs[Sno].Thisaa=Secs[Sno].Thisrid=Secs[Sno].Thisat[0]='\0';
	    Secs[Sno].Otheraa=Secs[Sno].Otherid=Secs[Sno].Otherat[0]='\0';
	    Secs[Sno].This=Secs[Sno].Other=0;
	    Secs[Sno].Otherchid='\0';
	}
	else /* read registration */
	{
	    /* atom name, residue name, pos and insert ID for the
	     * registration point on THIS strand. For atom names, 
	     * see the hack comment at the ATOM record read below
	     */
	    if (Line[41]==' ') sscanf(Line+42, "%3s", Secs[Sno].Thisat);
	    else sscanf(Line+41, "%4s", Secs[Sno].Thisat);
	    sscanf(Line+45, "%3s",Aa3);
	    Secs[Sno].Thisaa=aa_code31(Aa3);
	    sscanf(Line+50, "%d%c", &(Secs[Sno].This), &(Secs[Sno].Thisrid));

	    /* same + chain ID for the OTHER strand */
	    if (Line[56]==' ') sscanf(Line+57, "%3s", Secs[Sno].Otherat);
	    else sscanf(Line+56, "%4s", Secs[Sno].Otherat);
	    sscanf(Line+60, "%3s", Aa3);
	    Secs[Sno].Otheraa=aa_code31(Aa3);
	    sscanf(Line+64, "%c%d%c", &(Secs[Sno].Otherchid), 
		    &(Secs[Sno].Other), &(Secs[Sno].Otherid));
	}
	Sno++;
    }
    while (get_record(Pdb, Line, "TURN"))	/* read TURN entries */
    {
	Secs=(Secstr_*) expand_array(Secs, Sno, SECS_CHUNK, sizeof(Secstr_));
	Secs[Sno].Sectype=PDB_TURN;
	sscanf(Line+7, "%d %s", &(Secs[Sno].No), Secs[Sno].Id);
	sscanf(Line+15, "%3s", Aa3);
	Secs[Sno].Chid=Line[19];
	Secs[Sno].Begaa=aa_code31(Aa3);
	sscanf(Line+20, "%4d%c", &(Secs[Sno].Beg), &(Secs[Sno].Begrid));
	sscanf(Line+26, "%3s", Aa3);
	Secs[Sno].Endaa=aa_code31(Aa3);
	sscanf(Line+31, "%4d%c", &(Secs[Sno].End), &(Secs[Sno].Endrid));
	Secs[Sno].Type=0;
	Sno++;
    }
    
    /* sort secondary structure records in ascending residue order */
    if (Sno)
    {
	PDB_REALLOC(Secs, Secstr_, Sno);
	qsort(Secs, Sno, sizeof(Secstr_), ascend_res);
    }
    *Secno=Sno;
    return(Secs);
}
/* END of get_secstr() */

/* ---- Disulfide bond input ---- */

/* get_ssbonds(): reads the disulfide bond descriptions from Pdb.
 * Returns a newly allocated array of Ssbond_ records or NULL if
 * no S-S bonds have been found.
 * Also returns the length of the array in *Ssbondno.
 */
static Ssbond_ *get_ssbonds(FILE *Pdb, int *Ssbondno)
{
    Ssbond_ *Ssbs=NULL;
    int Ssbno;
    
    for (Ssbno=0; get_record(Pdb, Line, "SSBOND"); Ssbno++)
    {
	Ssbs=(Ssbond_*) expand_array(Ssbs, Ssbno, 16, sizeof(Ssbond_));
	sscanf(Line+7, "%3d", &(Ssbs[Ssbno].No));
	sscanf(Line+15, "%c%4d%c", 
	    &(Ssbs[Ssbno].Ch1), &(Ssbs[Ssbno].Pos1), &(Ssbs[Ssbno].Rid1));
	sscanf(Line+29, "%c%4d%c", 
	    &(Ssbs[Ssbno].Ch2), &(Ssbs[Ssbno].Pos2), &(Ssbs[Ssbno].Rid2));
    }
    if (Ssbno) PDB_REALLOC(Ssbs, Ssbond_, Ssbno);
    *Ssbondno=Ssbno;
    return(Ssbs);
}
/* END of get_ssbonds() */

/* ---- Site input ---- */

/* read_sites(): reads the SITE records from PDB into Entry. */
static void read_sites(FILE *Pdb, Pdbentry_ *Entry)
{
    static const int SITE_CHUNK=16;
    Site_ *Sites=NULL, *Cursite=NULL;
    Residue_ *Curres;
    int Siteno=0, Rsno=-1, Ridx, i;
    
    while (get_record(Pdb, Line, "SITE  "))
    {
	/* Start new site if sequence number is 1, says the PDB Guide.
	* In 4CPV and 5CPV (at least), the numbering is wrong,
	* so we check if all site members have been seen.
	* Works only if the number at pos. 16 is correct. 29-Sep-1998.
	*/
	if (Rsno==-1)	/* has been reset or no sites yet */
	{
	    Sites=(Site_*) expand_array(Sites, Siteno, SITE_CHUNK, sizeof(Site_));
	    Cursite=Sites+Siteno;   /* mark current site */
	    strncpy(Cursite->Name, Line+11, 3); /* store site name */
	    Cursite->Name[3]='\0';
	    sscanf(Line+15, "%d", &(Cursite->Resno));	/* no. of residues comprising site */
	    PDB_CALLOC(Cursite->Res, Residue_, Cursite->Resno);
	    Curres=Cursite->Res; Rsno=0;
	    Siteno++; 
	}
	
	/* Read the residue descriptions from the current Line
	* and put them into *Curres (hanging off *Cursite).
	* There are 4 at most in each line.
	* Rsno counts the residues seen in the current site.
	*/
	for (i=0, Ridx=18; 
		i<4 && Rsno<Cursite->Resno; 
		i++, Ridx+=11, Curres++, Rsno++)
	{
	    if (!read_residue(Line+Ridx, Curres))
		break;
	}
	if (Rsno==Cursite->Resno) Rsno=-1;  /* all residues seen, reset */
    }
    
    /* transfer the site array to Entry */
    /* if (Siteno) PDB_REALLOC(Sites, Site_, Siteno); 4-Oct-1999 */
    Entry->Sites=Sites; Entry->Siteno=Siteno;
}
/* END of read_sites() */

/* ---- Coordinate input ---- */

/* read_atom(): reads the data from an ATOM or HETATM card
 * (miraculously the format is the same for both). Line is
 * the full card string, the atom record to be read into is *Atom.
 */
static void read_atom(const char *Line, Atom_ *Atom)
{
    int Tempint;
    
    /* get atom number as in PDB */
    sscanf(Line+6,"%d", &(Atom->Atno));
    
    /* an ugly hack: in the PDB format, atom names are built of
     * 4 chars: 
     * 1-2=chem.symbol,  RIGHT justified;
     * 3=remoteness (alpha, optional)
     * 4=branch designator (numeric, optional)
     * When the 1st pos (Line[12]) is blank, we read in a 3-char
     * string only in order not to pick up the (optional) Alt character. 
     * Otherwise a 4-char string is read. My .Id strings are
     * always left-justified internally
     */
    if (Line[12]==' ') sscanf(Line+13, "%3s", Atom->Id);
    else sscanf(Line+12, "%4s", Atom->Id);
    
    /* get alternate position */
    Atom->Alt=Line[16];

    /* get residue description, work out what it is and set
     * the 1-letter code if possible and the residue type.
     */
    read_residue(Line+17, &(Atom->Res));
    if (!strncmp(Line, "ATOM  ", 6))	/* try to assign type if it is ATOM */
    {
	Atom->Res1=aa_code31(Atom->Res.Name);  /* try to convert to 1-letter AA code */
	if (Atom->Res1!='X')
	    Atom->Type=PDB_PROTEIN; /* belongs to an acceptable amino acid */
	else
	{
	    Atom->Res1=nu_code31(Atom->Res.Name);   /* try to convert to 1-letter nucleotide code */
	    switch(Atom->Res1)
	    {
		case 'T': Atom->Type=PDB_DNA; break;	/* thymine in DNA only */
		case 'U': Atom->Type=PDB_RNA; break; /* uracil in RNA only */
		case 'X': Atom->Type=PDB_UNKNOWN; break; /* how to get polysacch? */
		default: Atom->Type=PDB_NACID; break;	/* DNA or RNA */
	    }
	}
    }
    else    /* came from HETATM */
    {
	Atom->Type=PDB_HETERO;
	Atom->Res1='X';
    }
    
    /* read coordinates, occupancy and B-factor */
    sscanf(Line+30,"%f%f%f%f%f",
	Atom->X, Atom->X+1, Atom->X+2,
	&(Atom->Occu), &(Atom->Bfact));

    /* Read segment ID and element symbol. In the old format,
     * there was some redundant info here like "0LFS1234".
     * We test for the presence of a max 4-digit integer at pos 77-80, 
     * if it is there then old-fashioned format is assumed and
     * the Segid and Element fields will be set to all-space.
     */
    strcpy(Atom->Segid, "    ");
    strcpy(Atom->Element, "  ");
    Atom->Charge=0;
    if (!sscanf(Line+76, "%4d", &Tempint))
    {
	/* new fields: get segment ID and element symbol */
	sscanf(Line+72, "%4s%2s", Atom->Segid, Atom->Element);
	
	/* get the charge: usually 0, indicated by spaces */
	if (Line[78]==' ') Atom->Charge=0;
	else
	{
	    sscanf(Line+78, "%1d", &(Atom->Charge));
	    if (Line[79]=='-') Atom->Charge*=-1;	/* sign after number */
	}
    }
}
/* END of read_atom() */

/* get_allatoms(): starts at the current input position and
 * reads in all coordinate records (ATOM and HETATM). Chain_
 * structs will be allocated for every biopolymer chain (deduced
 * from the ATOM cards).
 */
static void get_allatoms(FILE *Pdb, Pdbentry_ *Entry)
{
    static const int ATOM_CHUNK=1024, HETATM_CHUNK=128;	    /* prealloc chunks */
    
    Chain_ *Chains=NULL, *Tmpchain, *Curchain;    /* temp array for chains, and ptrs */
    Atom_ Tempatom;	/* temporary atom */
    Hetgroup_ *Hetptr;	/* temporary hetgroup ptr */
    int i, Chainno=0, Modelno=-9999;
    long Prevlastpos;
    
    if (ftell(Pdb)!=Lastpos)
	fseek(Pdb, Lastpos, SEEK_SET);    /* go where last was found */
    Line[0]='\0';
    while (Prevlastpos=Lastpos, NULL!=fgets(Line, LINELEN, Pdb))
    {
	if (!strncmp(Line, "ATOM  ", 6))    /* std biopolymer chain */
	{
	    read_atom(Line, &Tempatom);
	    Tmpchain=locate_chain(Chains, Chainno, Tempatom.Res.Chid, Modelno);
	    if (Tmpchain==NULL)
	    {
		/* must add new chain */
		Chains=(Chain_*) expand_array(Chains, Chainno, 8, sizeof(Chain_));
		Curchain=Chains+Chainno;
		Chainno++;
		init_chain(Curchain, Tempatom.Res.Chid, Modelno);
	    }
	    else Curchain=Tmpchain;
	    
	    /* add the atom to the current chain */
	    Curchain->Atoms=
		(Atom_*) expand_array(Curchain->Atoms, Curchain->Atomno, 
					ATOM_CHUNK, sizeof(Atom_));
	    Curchain->Atoms[Curchain->Atomno]=Tempatom;
	    Curchain->Type|=Tempatom.Type;  /* overall type set */
	    Curchain->Atomno++;
	    continue;
	}
	if (!strncmp(Line, "HETATM", 6))    /* heteroatom */
	{
	    read_atom(Line, &Tempatom);
	    
	    /* find the hetero group for this heteroatom */
	    Hetptr=locate_hetgroup(Entry->Hets, Entry->Hetsno, &Tempatom, Modelno);
	    if (Hetptr==NULL)
	    {
		fprintf(stderr, "\n? get_allatoms(): Inconsistent PDB file, no HET card for:\n%s\n",
		    Line);
		continue;   /* be tough and do not store orphan heteros */
	    }
	    
	    /* check if this heteroatom might be part of a biopolymer bkbone */
	    Tmpchain=locate_chain(Chains, Chainno, Tempatom.Res.Chid, Modelno);
	    if (Tmpchain!=NULL)
	    {
		/* really part of backbone? */
		if (NULL==hetname_inseq(Entry, &Tempatom))
		    Tmpchain=NULL; /* was just a chance Chid match */
	    }
	    
	    if (Tmpchain==NULL)	    /* belongs to a "standalone" hetero */
	    {
		/* add atom to hetero group */
		Hetptr->Atoms=(Atom_*) 
		    expand_array(Hetptr->Atoms, Hetptr->Atomno, 
				HETATM_CHUNK, sizeof(Atom_));
		Hetptr->Atoms[Hetptr->Atomno]=Tempatom;
		Hetptr->Atomno++;
	    }
	    else    /* add heteroatom to biopolymer chain */
	    {
		Curchain=Tmpchain;
		Curchain->Atoms=
		    (Atom_*) expand_array(Curchain->Atoms, Curchain->Atomno, 
					ATOM_CHUNK, sizeof(Atom_));
		Curchain->Atoms[Curchain->Atomno]=Tempatom;
		Curchain->Type|=Tempatom.Type;  /* overall type set */
		Curchain->Atomno++;
		Hetptr->Biopolyno++; /* count occurrence but don't store */
	    }
	    continue;
	}
	if (!strncmp(Line, "MODEL ", 6))
	{
	    sscanf(Line+10, "%4d", &Modelno);
	    if (Entry->Hetsno)
		copy_hets(Entry, Modelno);
	    continue;
	}
	if(!strncmp(Line, "ENDMDL", 6))
	{
	    Modelno=-9999;  /* reset model number */
	    continue;
	}
	
	/* these must either be present or are optional,
	 * but we ignore them. Uncomment if want to bother later.
	 */
#if 0
	if(!strncmp(Line, "TER   ", 6)) continue;
	if(!strncmp(Line, "ANISOU", 6) || !strncmp(Line, "SIGUIJ", 6))
	    continue;
#endif
	
	if (!strncmp(Line, "CONECT", 6))
	{
	    /* connections will be processed later,
	     * so make a backstep and finish coordinate reading
	     */
	    fseek(Pdb, Prevlastpos, SEEK_SET);
	    break;
	}
	
	/* only junk is left, no connections, finish up */
	if (!strncmp(Line, "MASTER", 6) ||
	    !strncmp(Line, "END   ", 6))
		break;
    }	    /* while */
    
    /* finishing touches */
    for (i=0; i<Chainno; i++)
	finish_coords(Chains+i);
    for (i=0; i<Entry->Hetsno; i++)
	if (Entry->Hets[i].Atomno)
	    PDB_REALLOC(Entry->Hets[i].Atoms, Atom_, Entry->Hets[i].Atomno);
    Entry->Chains=Chains;
    Entry->Chainno=Chainno;
}
/* END of get_allatoms() */

/* read_residue(): reads a residue description from Str into Res.
 * The format corresponds to the residue descriptions in SITE, 
 * TURN, ATOM/HETATM; idiotically, the description in the HET card
 * has an extra space before the chain identifier character
 * and therefore cannot be processed by this routine.
 * Returns 0 on failure, non-0 if OK.
 */
static int read_residue(const char *Str, Residue_ *Res)
{
    if (!strncmp(Str, "   ", 3)) return(0);   /* blank, no good */
    strncpy(Res->Name, Str, 3); Res->Name[3]='\0';
    Res->Chid=Str[4];
    if (1!=sscanf(Str+5, "%4d", &(Res->Resno))) return(0);  /* blank again */
    Res->Rid=Str[9];
    return(1);	/* OK */
}
/* END of read_residue() */

/* locate_chain(): tries to find a chain in the Chains array (Chainno long)
 * whose ID is Chid and whose model number is Modelno (provided Modelno
 * is non-negative: if <0 then it is ignored).
 * Returns a ptr to it if it is found, or NULL on failure.
 */
static Chain_ *locate_chain(const Chain_ *Chains, int Chainno,
	char Chid, int Modelno)
{
    Chain_ *Ptr=NULL;
    int i;
    
    if (Chains==NULL || !Chainno) return(NULL);
    for (i=0, Ptr=(Chain_*)Chains; i<Chainno; Ptr++, i++)
    {
	 if (Modelno<0 && Ptr->Chid==Chid) return(Ptr);   /* found */
	 if (Modelno>=0 && Ptr->Modelno==Modelno 
	    && Ptr->Chid==Chid) return(Ptr);   /* also found */
    }
    return(NULL);   /* not found */
}
/* END of locate_chain() */

/* init_chain(): inits the *Chain structure so that it contains
 * the prescribed chain ID (Chid) and model number (Modelno).
 */
static void init_chain(Chain_ *Chain, char Chid, int Modelno)
{
    Chain->Atoms=NULL; Chain->Atomno=0;
    Chain->Secs=NULL; Chain->Secsno=0;
    Chain->Ssbs=NULL; Chain->Ssbno=0;
    Chain->Resno=0;
    Chain->Chid=Chid; Chain->Modelno=Modelno;
    Chain->Type=PDB_UNKNOWN;
    Chain->Seq=NULL;
}
/* END of init_chain() */

/* locate_hetgroup(): returns a ptr to the Hetgroup_ record
 * in the Hets[] array (Hetsno long) which contains the 
 * description of the hetero group (identified through the
 * Res struct subfield of *Atom). For solvents we require
 * that only the groupname match, Resno and insertion codes are ignored.
 * If Modelno>0 then it will be taken into account as well.
 * Returns NULL if not found.
 */
static Hetgroup_ *locate_hetgroup(const Hetgroup_ *Hets, int Hetsno,
    const Atom_ *Atom, int Modelno)
{
    int i;
    Hetgroup_ *Hptr;
    
    if (Hets==NULL || !Hetsno) return(NULL);
    for (i=0, Hptr=(Hetgroup_ *)Hets; i<Hetsno; i++, Hptr++)
    {
	if (!strncmp(Hptr->Het.Name, Atom->Res.Name, 3)
	    && (Hptr->Solvent!=PDB_NOSOLVENT
		|| Hptr->Het.Resno<-9	/* see 1AGA, 1CAR */
		|| (Hptr->Het.Resno==Atom->Res.Resno 
		    && Hptr->Het.Rid==Atom->Res.Rid
		    && Hptr->Het.Chid==Atom->Res.Chid))
	    && (Modelno<1 || (Modelno>=1 && Modelno==Hptr->Modelno)))
		return(Hptr);
    }
    return(NULL);
}
/* END of locate_hetgroup() */

/* hetname_inseq(): scans the global sequence information array in *Entry,
 * and checks if the residue name of *Atom (which is supposed to be
 * a heteroatom) is contained in the sequence of the chain whose
 * chain ID is in the atom.
 * If found, then a ptr is returned to the info record containing it, 
 * otherwise NULL is returned.
 */
static Seqinfo_ *hetname_inseq(const Pdbentry_ *Entry, const Atom_ *Atom)
{
    Seqinfo_ *Seqptr;
    int i, j, j3;
    
    /* find the sequence with the right chain ID */
    for (Seqptr=Entry->Seqs, i=0; i<Entry->Seqno; Seqptr++, i++)
	if (Seqptr->Chid==Atom->Res.Chid) break;
    if (i>=Entry->Seqno) return(NULL);	/* not found */
    
    for (j=j3=0; j<Seqptr->Len; j++, j3+=3)
	if (!strncmp(Seqptr->Seq+j3, Atom->Res.Name, 3)) break;
    return((j>=Seqptr->Len)? NULL: Seqptr);
}
/* END of hetname_inseq() */

/* copy_hets(): makes a copy of the hetgroup records (except water)
 * and appends them at the end of the Hets array in *Entry.
 * The copied group records receive Modelno as their model number.
 * When invoked with Modelno==1, then only certain preparations
 * are made. The whole complicated procedure is necessary when
 * there are more MODEL/ENDMDL structures in the PDB entry.
 * Modelno>=1 is assumed but not checked.
 */
static void copy_hets(Pdbentry_ *Entry, int Modelno)
{
    static int Firsthetsno=0;
    
    Hetgroup_ *Hets=Entry->Hets;
    int i, k, Hetsno=Entry->Hetsno;
    
    if (Modelno==1)	/* preparation */
    {
	Firsthetsno=Entry->Hetsno;
	
#if 0
	if (Hets[Firsthetsno-1].Solvent==PDB_WATER) /* uncomment if h2o is not copied */
	    Firsthetsno--;
#endif
	
	for (i=0; i<Firsthetsno; i++)	/* init model numbers */
	    Hets[i].Modelno=1;
	return;
    }
    
    /* Copy the "original" hetgroups up to Firsthetsno
     * after the current end, fill in the model number
     */
    PDB_REALLOC(Hets, Hetgroup_, Hetsno+Firsthetsno);	/* make room */
    for (i=0, k=Hetsno; i<Firsthetsno; i++, k++)
    {
	Hets[k]=Hets[i];    /* copy what is to be copied */
	strcpy(Hets[k].Descr, Hets[i].Descr);	/* but carefully */
	Hets[k].Atoms=NULL; Hets[k].Atomno=0;	/* no atoms yet */
	Hets[k].Modelno=Modelno;    /* give them the new model number */
    }
    Entry->Hets=Hets;
    Entry->Hetsno=Hetsno+Firsthetsno;
}
/* END of copy_hets() */

/* finish_coords(): finishes up the assignments done to a molecule *Chain
 * after the coordinate section has been read.
 * Reallocates the atom array, cleans up the type assignment, 
 * counts residues, builds one-letter actual sequence array etc.
 */
static void finish_coords(Chain_ *Chain)
{
    short Tpresent=0, Upresent=0;   /* for nucleic acid typing */
    int i, r, Oldresno=-9999;
    char Oldrid='\0';
    Atom_ *Curatom=NULL;
    
    PDB_REALLOC(Chain->Atoms, Atom_, Chain->Atomno);
    PDB_CALLOC(Chain->Seq, char, Chain->Atomno);    /* can't have more residues :-) */
    for (i=r=0, Curatom=Chain->Atoms; i<Chain->Atomno; i++, Curatom++)
    {
	if (Curatom->Res.Resno!=Oldresno || Curatom->Res.Rid!=Oldrid)
	{
	    /* new residue */
	    Chain->Seq[r]=Curatom->Res1;    /* store 1-letter residue ID */
	    Oldresno=Curatom->Res.Resno;
	    Oldrid=Curatom->Res.Rid;
	    r++;
	}
	if (Chain->Type|PDB_NACID)  /* see if only T or only U is present */
	{
	    if (Curatom->Type==PDB_DNA) Tpresent++;
	    if (Curatom->Type==PDB_RNA) Upresent++;
	}
    }
    PDB_REALLOC(Chain->Seq, char, r+1);
    Chain->Seq[r]='\0';
    Chain->Resno=r;
    if (Chain->Type|PDB_NACID && !(Tpresent && Upresent))
    {
	Chain->Type&=~PDB_NACID;    /* switch off original nucl assignment */
	if (Tpresent) Chain->Type|=PDB_DNA;
	if (Upresent) Chain->Type|=PDB_RNA;
    }
}
/* END of finish_coords() */

/* ---- Connectivities ---- */

/* get_connects(): reads all connectivity records from Pdb
 * and puts them into *Entry.
 */
static void get_connects(FILE *Pdb, Pdbentry_ *Entry)
{
    static const int CONN_CHUNK=128;
    
    Connect_ *Conns=NULL;
    int Conno=0;
    
    for (Conno=0; get_record(Pdb, Line, "CONECT"); Conno++)
    {
	Conns=(Connect_*) expand_array(Conns, Conno, CONN_CHUNK, sizeof(Connect_));
	read_connect(Line, Conns+Conno);
    }
    
    PDB_REALLOC(Conns, Connect_, Conno);
    Entry->Connects=Conns;
    Entry->Conno=Conno;
}
/* END of get_connects() */

/* read_connect(): parses one CONECT record in Line and fills
 * the connection structure *Connect.
 */
static void read_connect(const char *Line, Connect_ *Connect)
{
    static const char SPACE5[6]="     ";
    int i, ix;
    
    /* clean */
    for (i=0; i<4; i++)
	Connect->Bonded[i]=Connect->Hbonded[i]=0;
    Connect->Saltbrdg[0]=Connect->Saltbrdg[1]=0;
    
    sscanf(Line+6, "%5d", &(Connect->Atomno));
    for (i=0, ix=11; i<4; i++, ix+=5)
    {
	if (strncmp(Line+ix, SPACE5, 5))
	    sscanf(Line+ix, "%5d", Connect->Bonded+i);
    }
    
    /* the 2:1 2:1 order is more than bizarre here.
     * We also have to skip spaces etc....
     */
    ix=31;
    if (strncmp(Line+ix, SPACE5, 5))
	sscanf(Line+ix, "%5d", Connect->Hbonded);
    ix=36;
    if (strncmp(Line+ix, SPACE5, 5))
	sscanf(Line+ix, "%5d", Connect->Hbonded+1);
    ix=41;
    if (strncmp(Line+ix, SPACE5, 5))
	sscanf(Line+ix, "%5d", Connect->Saltbrdg);
    ix=46;
    if (strncmp(Line+ix, SPACE5, 5))
	sscanf(Line+ix, "%5d", Connect->Hbonded+2);
    ix=51;
    if (strncmp(Line+ix, SPACE5, 5))
	sscanf(Line+ix, "%5d", Connect->Hbonded+3);
    ix=56;
    if (strncmp(Line+ix, SPACE5, 5))
	sscanf(Line+ix, "%5d", Connect->Saltbrdg+1);
}
/* END of read_connect() */

/* ---- Auxiliaries ---- */

/* get_record: returns the first record (in Line) from Pdb which
 * begins with the string Label. Line is assumed to be longer than
 * LINELEN. Stores the file position where the last matching
 * line was found in Lastpos (global).
 * Return value: 1 if a matching record was found, 0 if no more
 * records were found. In this latter case, Line=="".
 */
static int get_record(FILE *Pdb, char *Line, const char *Label)
{
    int Lablen=strlen(Label);
    
    if (ftell(Pdb)!=Lastpos)
	fseek(Pdb, Lastpos, SEEK_SET);    /* go where last was found */
    Line[0]='\0';	/* zero line */
    while (NULL!=fgets(Line, LINELEN, Pdb))
	if (!strncmp(Line, Label, Lablen))
	{ 
	    Lastpos=ftell(Pdb); 	/* found */
	    Line[LINELEN-1]='\0';
	    return(1);
	}
    return(0);	    /* not found */
}
/* END of get_record */

/* ascend_res: for sorting Secstr_ records in increasing order.
 * Alphabetical ordering for chain identifiers, and increasing
 * residue number (beginning) order within the same chain.
 */
static int ascend_res(const void *Sec1, const void *Sec2)
{
    register Secstr_ *S1=(Secstr_ *)Sec1, *S2=(Secstr_ *)Sec2;
    
    if (S1->Chid<S2->Chid) return(-1);
    else if (S1->Chid==S2->Chid) 
    {
	if (S1->Beg<S2->Beg) return(-1);
	else if (S1->Beg>S2->Beg) return(1);
	else return(0);
    }
    else return(0);
}
/* END of ascend_res */

/* ---- Memory management ---- */

/* pdb_malloc(), pdb_calloc(), pdb_realloc():
 * these functions are called by the PDB allocation macros.
 * They perform some useful checks before allocation.
 */
static void *pdb_malloc(size_t Size)
{
    return (Size? malloc(Size): NULL);
}

static void *pdb_calloc(size_t Elemno, size_t Size)
{
    return ((Elemno && Size)? calloc(Elemno, Size): NULL);
}

static void *pdb_realloc(void *Ptr, size_t Size)
{
    if (!Size) { free(Ptr); return NULL; }
    if (Ptr==NULL && Size) return malloc(Size);
    return realloc(Ptr, Size);	/* normal case */
}

/* expand_array(): expands the array Array by a fixed "chunk" Chunk
 * if the current number of elements (Len) is an exact
 * multiple of this chunk. Elsize specifies the size
 * of the array element type.
 * Returns a (possibly new) ptr to the array.
 */
static void *expand_array(void *Array, int Len, int Chunk, int Elsize)
{
    if (Array==NULL)	/* start from scratch */
    {
	if (Len)
	    fprintf(stderr, "\n? expand_array(NULL, Len=%d, %d, %d): Len should be 0\n",
		    Len, Chunk, Elsize);
	return(calloc(Chunk, Elsize));
    }
    if (Len % Chunk == 0)   /* grow! */
	return(realloc(Array, (Len+Chunk)*Elsize));
    else return(Array);
}
/* END of expand_array() */

/* ---- "Destructors" ---- */

/* free_pdb: frees up memory allocated to Entry and its arrays.
 * Does not set Entry to NULL. (Destructor in C++)
 */
void free_pdb(Pdbentry_ *Entry)
{
    register int i;
    
    if (Entry==NULL) return;
    
    for (i=0; i<Entry->Chainno; i++)
	free_chain(Entry->Chains+i);
    free(Entry->Chains);
    for (i=0; i<Entry->Hetsno; i++)
	free_het(Entry->Hets+i);
    free(Entry->Hets);
    for (i=0; i<Entry->Hetnamsno; i++)
	free_hetnamformul(Entry->Hetnams+i);
    free(Entry->Hetnams);
    for (i=0; i<Entry->Hetformsno; i++)
	free_hetnamformul(Entry->Hetforms+i);
    free(Entry->Hetforms);
    for (i=0; i<Entry->Seqno; i++)
	free_seq(Entry->Seqs+i);
    free(Entry->Seqs);
    free(Entry->Ssbs);
    for (i=0; i<Entry->Siteno; i++)
	free_site(Entry->Sites+i);
    free(Entry->Sites);
    free(Entry->Connects);
    free(Entry->Title); free(Entry->Compound);
    free(Entry->Source); free(Entry->Author);
    
    free(Entry);
}
/* END of free_pdb() */

static void free_chain(Chain_ *Chain)
{
    if (Chain==NULL) return;
    free(Chain->Atoms); free(Chain->Secs);
    free(Chain->Ssbs); free(Chain->Seq);
}
/* END of free_chain() */

static void free_het(Hetgroup_ *Het)
{
    if (Het==NULL) return;
    free(Het->Atoms);
}
/* END of free_het() */

static void free_hetnamformul(Hetnamformul_ *Hnf)
{
    if (Hnf==NULL) return;
    free(Hnf->Name);
}
/* END of free_hetnamformul() */

static void free_seq(Seqinfo_ *Seqinfo)
{
    if (Seqinfo==NULL) return;
    free(Seqinfo->Seq);
}
/* END of free_seq() */

static void free_site(Site_ *Site)
{
    if (Site==NULL) return;
    free(Site->Res);
}
/* END of free_site() */

/* atom_dist(): returns the distance between two Atom_ entries. */
float atom_dist(const Atom_ *At1, const Atom_ *At2)
{
    register float Dx, D=0;
    register int i;

    D=0.0;
    for (i=0; i<3; i++)
    {
	Dx=At1->X[i]-At2->X[i];
	D+=Dx*Dx;
    }
    return(sqrtf(D));
}
/* END of atom_dist() */

/* ---- OUTPUT ---- */

/* write_pdb(): writes the atomic coordinate information in Entry
 * to the file Pdbfn in PDB format. 
 * Remno remarks are supplied in a char array Remarks[].
 */
void write_pdb(const char *Pdbfn, Pdbentry_ *Entry, 
	char *Remarks[], int Remno)
{
    /* index for the counters in the MASTER record */
    typedef enum {REM, FTN, HET, HLX, SHT, TRN, 
		SIT, TRS, ATM, TER, CON, SQR} Masteridx_ ;
    
    FILE *Pdb;		/* output file */
    Chain_ *Chains;	/* temp ptrs to Entry's (sub)fields */
    Atom_ *Atoms=NULL;
    Hetgroup_ *Hetptr=NULL;
    Hetnamformul_ *Hnf=NULL;
    Secstr_ *Secs=NULL;	/* merged list of secondary structures */
    Ssbond_ *Ssbs=NULL;	/* merged list of S-S bonds */
    register int i, j, Se, Seplus, Ss, Ssplus;
    int Chno, Atomno, Ssbno, Secsno, Atomcnt;
    int Mc[12]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};	/* record counter */
    Masteridx_ Mx;

    /* check data quickly and open file */
    if (Entry==NULL) return;    /* possibly empty */
    if (NULL==(Pdb=fopen(Pdbfn, "w")))
    {
	fprintf(stderr, "\n? write_pdb(): Cannot open \"%s\"\n", Pdbfn);
	return;
    }
    
    /* compose header if non-empty */
    if (Entry->Header[0]!='\0')
    {
    	set_pdbdate(Entry);
    	sprintf(Line, "HEADER    %-40s%9s   %4s",
	    Entry->Header, Entry->Date, Entry->Pdbcode);
    	print_card(Pdb, Line);
    }

    /* title, compound and source description (may span several lines) */
    write_multicard(Pdb, "TITLE ", Entry->Title);
    write_multicard(Pdb, "COMPND", Entry->Compound);
    write_multicard(Pdb, "SOURCE", Entry->Source);

    /* experimental technique */
    sprintf(Line, "EXPDTA    %-60.60s", Entry->Expdta);
    print_card(Pdb, Line);
    
    /* authors */
    write_multicard(Pdb, "AUTHOR", Entry->Author);
    
    /* REMARKs: the PDB is making a mess again by
     * attaching specific meanings to certain REMARK numbers.
     * I keep only REMARK 2 for the resolution data, 
     * print REMARK 4 for format compatibility, 
     * REMARK 5 to warn that this file did not come from
     * the PDB but from me, and then put the user-supplied
     * user remarks as REMARK 99 (range 6-99 is allowed
     * for this purpose).
     */
    
    /* print resolution */
    if (Entry->Resol<0)
	strcpy(Line, "REMARK   2 RESOLUTION. NOT APPLICABLE.");
    else
	sprintf(Line, "REMARK   2 RESOLUTION. %.2f ANGSTROMS.", Entry->Resol);
    print_card(Pdb, Line);
    
    /* add format compliance and warning */
    sprintf(Line, "REMARK   4 %4s COMPLIES WITH FORMAT V. 2.1, %11s", 
	Entry->Pdbcode, current_date());
    print_card(Pdb, Line);
    strcpy(Line, "REMARK   5 CAUTION: FORMATTED BY A PROGRAM BY A. ASZODI!");
    print_card(Pdb, Line);
    Mc[REM]=3;

    /* add user-supplied remarks starting at 99 */
    for (j=0; j<Remno; j++, Mc[REM]++)
    {
	if (strlen(Remarks[j])>=62) continue;	/* skip longs */
	sprintf(Line,"REMARK  99 %s",Remarks[j]);
	print_card(Pdb, Line);
    }

    /* SEQRES */
    for (j=0; j<Entry->Seqno; j++)
	Mc[SQR]+=write_seqres(Pdb, Entry->Seqs+j);

    /* hetero groups: HET, HETNAM, FORMUL */
    for (j=0; j<Entry->Hetsno; j++)
    {
	Hetptr=Entry->Hets+j;
	
	/* do not print the HET card for water
	 * and print the cards only once if there
	 * were multiple models in the entry
	 */
	if (Hetptr->Solvent==PDB_WATER || Hetptr->Modelno>1) continue;
	write_het(Pdb, Hetptr);
    }
    Mc[HET]=Entry->Hetsno;
    for (j=0; j<Entry->Hetnamsno; j++)
    {
	Hnf=Entry->Hetnams+j;
	write_hetnam(Pdb, Hnf);
    }
    for (j=0; j<Entry->Hetformsno; j++)
    {
	Hnf=Entry->Hetforms+j;	/* FORMUL is given for water, too */
	write_formul(Pdb, Hnf);
    }

    /* the secstr, S-S bond and H-bond records are distributed
     * among the chains and have to be pooled for output. The
     * records are counted here, temp space is allocated, and
     * after copying the records are sorted. For MODELs,
     * every chain contains (hopefully) the same secstr records,
     * these will be merged
     */
    Secsno=0; Ssbno=Entry->Ssbno;
    Chains=Entry->Chains;
    for (Chno=0; Chno<Entry->Chainno; Chno++)
    {
	Secsno+=Chains[Chno].Secsno;
	Ssbno+=Chains[Chno].Ssbno;
    }
    PDB_CALLOC(Secs, Secstr_, Secsno);
    PDB_CALLOC(Ssbs, Ssbond_, Ssbno);
    for (Se=Ss=0, Chno=0; Chno<Entry->Chainno; Chno++)
    {
	Seplus=merge_secstr(Secs, Se, Chains+Chno);
	Se+=Seplus;
	Ssplus=merge_ssbond(Ssbs, Ss, Chains+Chno);
	Ss+=Ssplus;
    }
    Secsno=Se;
    memcpy(Ssbs+Ss, Entry->Ssbs, sizeof(Ssbond_)*Entry->Ssbno); /* interchain S-S */
    Ssbno=Ss+Entry->Ssbno;
    qsort(Secs, Secsno, sizeof(Secstr_), secstr_outorder);
    qsort(Ssbs, Ssbno, sizeof(Ssbond_), ssbond_outorder);
    
    /* list sec. struct. info */
    for (i=0; i<Secsno; i++)
	if (Secs[i].Sectype==PDB_HELIX)	    /* helices */
	{
	    sprintf(Line, "HELIX  %3d %3s %3s %c %4d%c %3s %c %4d%c%2d", 
		Secs[i].No, Secs[i].Id, aa_code13(Secs[i].Begaa), 
		Secs[i].Chid, Secs[i].Beg, Secs[i].Begrid, 
		aa_code13(Secs[i].Endaa), Secs[i].Chid, Secs[i].End,
		Secs[i].Endrid, Secs[i].Type);
	    print_card(Pdb, Line); Mc[HLX]++;
	}
	else if(Secs[i].Sectype==PDB_SHEET)	/* sheets */
	{
	    sprintf(Line, "SHEET  %3d %3s%2d %3s %c%4d%c %3s %c%4d%c%2d ", 
		Secs[i].No, Secs[i].Id, Secs[i].Strandno,
		aa_code13(Secs[i].Begaa), Secs[i].Chid,
		Secs[i].Beg, Secs[i].Begrid, 
		aa_code13(Secs[i].Endaa), Secs[i].Chid, Secs[i].End,
		Secs[i].Endrid, Secs[i].Type);
		
	    /* print registration if present */
	    if (Secs[i].Type && Secs[i].Other)
	    {
		sprintf(Line+41, (strlen(Secs[i].Thisat)==4)? "%4s": " % -3.3s", 
		    Secs[i].Thisat);
		sprintf(Line+45, "%3s %c%4d%c ",  
		    aa_code13(Secs[i].Thisaa),
		    Secs[i].Chid, Secs[i].This, Secs[i].Thisrid);
		sprintf(Line+56, (strlen(Secs[i].Otherat)==4)? "%4s": " % -3.3s", 
		    Secs[i].Otherat);
		sprintf(Line+60, "%3s %c%4d%c",  
		    aa_code13(Secs[i].Otheraa),
		    Secs[i].Otherchid, Secs[i].Other, Secs[i].Otherid);
	    }
	    print_card(Pdb, Line); Mc[SHT]++;
	}
	else	/* turns */
	{
	    sprintf(Line, "TURN   %3d %3s %3s %c%4d%c %3s %c%4d%c", 
		Secs[i].No, Secs[i].Id, aa_code13(Secs[i].Begaa), 
		Secs[i].Chid, Secs[i].Beg, Secs[i].Begrid, 
		aa_code13(Secs[i].Endaa), Secs[i].Chid, Secs[i].End,
		Secs[i].Endrid);
	    print_card(Pdb, Line); Mc[TRN]++;
	}
    PDB_FREE(Secs);
    
    /* list disulfide bridges */
    for (i=0; i<Ssbno; i++)
    {
	sprintf(Line, "SSBOND %3d CYS %c %4d%c   CYS %c %4d%c", 
	    Ssbs[i].No, 
	    Ssbs[i].Ch1, Ssbs[i].Pos1, Ssbs[i].Rid1, 
	    Ssbs[i].Ch2, Ssbs[i].Pos2, Ssbs[i].Rid2);
	print_card(Pdb, Line);
    }
    PDB_FREE(Ssbs);
    
    /* list SITE records */
    for (i=0; i<Entry->Siteno; i++)
	Mc[SIT]+=write_site(Pdb, Entry->Sites+i);
    
    /* list atomic coordinates for each chain (ATOM, MODEL/ENDMDL, TER) */
    Atomcnt=1;
    for (Chno=0; Chno<Entry->Chainno; Chno++)
    {
	if (Chains[Chno].Modelno>0)
	{
	    sprintf(Line, "MODEL     %4d", Chains[Chno].Modelno);
	    print_card(Pdb, Line);
	}
	Atoms=Chains[Chno].Atoms; Atomno=Chains[Chno].Atomno;
	for (i=0; i<Atomno; Atomcnt++,i++)
	{
	    write_atom(Pdb, Atoms+i);
	    Mc[ATM]++;
	    
	    /* TER should come at end of chain */
	    if (i==Atomno-1)
	    {
		sprintf(Line, "TER   %5d      %3s %c%4d",
		    ++Atomcnt, Atoms[i].Res.Name,
		    Atoms[i].Res.Chid,Atoms[i].Res.Resno);
		Mc[TER]++;
		print_card(Pdb, Line);
	    }
	}
	
	/* check if there are HET records with the same chain ID,
	 * list them right after the biopolymer
	 */
	for (i=0; i<Entry->Hetsno; i++)
	{
	    Hetptr=Entry->Hets+i;
	    if (Hetptr->Het.Chid!=Chains[Chno].Chid ||
		(Chains[Chno].Modelno>0 && Chains[Chno].Modelno!=Hetptr->Modelno))
		continue;
	    for (j=0; j<Hetptr->Atomno; Atomcnt++, j++)
	    {
		write_atom(Pdb, Hetptr->Atoms+j);
		Mc[ATM]++;
	    }
	}
	if (Chains[Chno].Modelno>0)
	{
	    strcpy(Line, "ENDMDL");
	    print_card(Pdb, Line);
	}
    }
    
    /* list hetgroups with chain IDs different from the biopolymer chain IDs */
    for (i=0; i<Entry->Hetsno; i++)
    {
	Hetptr=Entry->Hets+i;
	if (NULL!=locate_chain(Chains, Entry->Chainno, Hetptr->Het.Chid, -1))
	    continue;
	for (j=0; j<Hetptr->Atomno; Atomcnt++, j++)
	{
	    write_atom(Pdb, Hetptr->Atoms+j);
	    Mc[ATM]++;
	}
    }
    
    /* list CONECT records */
    for (j=0; j<Entry->Conno; j++)
	write_connect(Pdb, Entry->Connects+j);
    Mc[CON]=Entry->Conno;
    
    /* finish listing with MASTER and END */
    strcpy(Line, "MASTER    ");
    for (Mx=REM; Mx<=SQR; Mx++)
	sprintf(Line+10+5*Mx, "%5d", Mc[Mx]);
    print_card(Pdb, Line);
    strcpy(Line, "END");
    print_card(Pdb, Line);
    fclose(Pdb);
}
/* END of write_pdb() */

/* ---- Output ---- */

/* write_multicard(): writes COMPND or SOURCE cards
 * which can span several lines.
 */
static void write_multicard(FILE *Pdb, const char *Tag, const char *Cmp)
{
    int i, k;
    if (Cmp==NULL) return;  /* nothing to write */
    for (i=1, k=strlen(Cmp); k>0; i++, Cmp+=60, k-=60)
    {
	if (i==1) sprintf(Line, "%6s    %-60.60s", Tag, Cmp);
	else sprintf(Line, "%6s  %2d%-60.60s", Tag, i, Cmp);
	print_card(Pdb, Line);
    }
}
/* END of write_multicard() */

/* write_seqres(): writes *Seqinfo to Pdb.
 * Returns the number of cards generated.
 */
static int write_seqres(FILE *Pdb, const Seqinfo_ *Seqinfo)
{
    int Sernum, i, ix;
    
    for (Sernum=1; 13*(Sernum-1)<=Seqinfo->Len; Sernum++)
    {
	sprintf(Line, "SEQRES  %2d %c %4d ",
	    Sernum, Seqinfo->Chid, Seqinfo->Len);
	for (i=13*(Sernum-1), ix=18; i<13*Sernum; i++, ix+=4)
	{
	    Line[ix]=' '; 
	    strncpy(Line+ix+1, Seqinfo->Seq+3*i, 3);
	}
	Line[ix]='\0';
	print_card(Pdb, Line);
    }
    return(Sernum);
}
/* END of write_seqres() */

/* write_het(): writes a HET card for Het to Pdb. */
static void write_het(FILE *Pdb, const Hetgroup_ *Het)
{
    sprintf(Line, "HET    %-3s  %c%4d%c  %5d     %40s", 
	Het->Het.Name, Het->Het.Chid, 
	Het->Het.Resno, Het->Het.Rid, 
	Het->Atomno+Het->Biopolyno, Het->Descr);
    print_card(Pdb, Line);
}
/* END of write_het() */

/* write_hetnam(): writes the HETNAM card for Het to Pdb. */
static void write_hetnam(FILE *Pdb, const Hetnamformul_ *Het)
{
    int i, k;
    char *Txt=Het->Name;
    
    if (Txt==NULL) return;  /* no HETNAM record */
    
    for (i=1, k=strlen(Txt); k>0; i++, Txt+=55, k-=55)
    {
	if (i==1) sprintf(Line, "HETNAM     %-3.3s %-55.55s", Het->Hetid, Txt);
	else sprintf(Line, "HETNAM  %2d %-3.3s %-55.55s", i, Het->Hetid, Txt);
	print_card(Pdb, Line);
    }
}
/* END of write_hetnam() */

/* write_formul(): writes the FORMUL card for Het to Pdb. */
static void write_formul(FILE *Pdb, const Hetnamformul_ *Het)
{
    int i, k;
    char *Txt=Het->Name;
    
    if (Txt==NULL) return;  /* no FORMUL record */
    for (i=1, k=strlen(Txt); k>0; i++, Txt+=52, k-=52)
    {
	if (i==1) sprintf(Line, "FORMUL  %2d  %-3.3s %-52.52s",
	    Het->Compnum, Het->Hetid, Txt);
	else sprintf(Line, "FORMUL  %2d  %-3.3s %2d%-52.52s",
	    Het->Compnum, Het->Hetid, i, Txt);
	print_card(Pdb, Line);
    }
}
/* END of write_formul() */

/* write_site(): writes a complete site entry
 * (can span more than 1 card) to Pdb.
 * Returns the number of cards written.
 */
static int write_site(FILE *Pdb, const Site_ *Site)
{
    int Cardno, Sernum, i, ix;
    
    Cardno=Site->Resno/4;
    if (Site->Resno % 4) Cardno++;  /* last incomplete card */
    for (Sernum=1; Sernum<=Cardno; Sernum++)
    {
	sprintf(Line, "SITE   %3d %-3.3s %2d",
	    Sernum, Site->Name, Site->Resno);
	for (i=4*(Sernum-1), ix=17; i<4*Sernum && i<Site->Resno; i++, ix+=11)
	{
	    Line[ix]=' ';
	    write_residue(Line+ix+1, Site->Res+i);
	}
	print_card(Pdb, Line);
    }
    return(Sernum);
}
/* END of write_site() */

/* write_atom(): writes an ATOM or HETATM record to Pdb. */
static void write_atom(FILE *Pdb, const Atom_ *Atom)
{
    sprintf(Line, "%6s%5d ",
	(Atom->Type==PDB_HETERO)? "HETATM" : "ATOM  ", Atom->Atno);
    sprintf(Line+12, (strlen(Atom->Id)==4)? "%4s": " %-3.3s", 
	Atom->Id);
    Line[16]=Atom->Alt;
    write_residue(Line+17, &(Atom->Res));
    sprintf(Line+27, "   %8.3f%8.3f%8.3f%6.2f%6.2f",
	Atom->X[0], Atom->X[1], Atom->X[2],
	Atom->Occu,Atom->Bfact);

    /* new atom description format at end */
    sprintf(Line+72, "%-4.4s%2s", Atom->Segid, Atom->Element);
    if (Atom->Charge!=0)
    {
	sprintf(Line+78, "%1d%c", 
	    Atom->Charge, (Atom->Charge>0)? '+':'-');
    }
    print_card(Pdb, Line);
}
/* END of write_atom() */

/* write_residue(): writes the residue information from *Res
 * to Str. Inverse of read_residue(), do not use for HET records.
 */
static void write_residue(char *Str, const Residue_ *Res)
{
    sprintf(Str, "%3s %c%4d%c",
	Res->Name, Res->Chid, Res->Resno, Res->Rid);
}
/* END of write_residue() */

/* write_connect(): writes the *Connect structure into Pdb. */
static void write_connect(FILE *Pdb, const Connect_ *Connect)
{
    int i, ix;
    
    memset(Line, ' ', LINELEN-1);
    sprintf(Line, "CONECT%5d", Connect->Atomno);
    for (i=0, ix=11; i<4; i++, ix+=5)
    {
	if (Connect->Bonded[i])
	    sprintf(Line+ix, "%5d", Connect->Bonded[i]);
	else
	    strncpy(Line+ix, "     ", 5);
    }
    
    if (Connect->Hbonded[0])
	sprintf(Line+31, "%5d", Connect->Hbonded[0]);
    else
	strncpy(Line+31, "     ", 5);
    if (Connect->Hbonded[1])
	sprintf(Line+36, "%5d", Connect->Hbonded[1]);
    else
	strncpy(Line+36, "     ", 5);
    if (Connect->Saltbrdg[0])
	sprintf(Line+41, "%5d", Connect->Saltbrdg[0]);
    else
	strncpy(Line+41, "     ", 5);
    if (Connect->Hbonded[2])
	sprintf(Line+46, "%5d", Connect->Hbonded[2]);
    else
	strncpy(Line+46, "     ", 5);
    if (Connect->Hbonded[3])
	sprintf(Line+51, "%5d", Connect->Hbonded[3]);
    else
	strncpy(Line+51, "     ", 5);
    if (Connect->Saltbrdg[1])
	sprintf(Line+56, "%5d", Connect->Saltbrdg[1]);
    else
	strncpy(Line+56, "     ", 5);
    print_card(Pdb, Line);
}
/* END of write_connect() */

/* ---- Secondary structure and S-S bond merging ---- */

/* merge_secstr(): add the secondary structure elements stored in *Chain
 * to the end of the Size-long array Secs. Only new elements are added.
 * Return value: the number of elements added.
 */
static int merge_secstr(Secstr_ *Secs, int Size, const Chain_ *Chain)
{
    int Sfrom, Sto, Secsno=Chain->Secsno, Added=0;
    short Diff;
    
    /* check all secstr elements in Chain */
    for (Sfrom=0; Sfrom<Secsno; Sfrom++)
    {
	/* is the present element already in Secs? [0..Size-1] */
	Diff=1;
	for (Sto=0; Diff && Sto<Size; Sto++)
	{
	    Diff=secstr_diff(Secs+Sto, Chain->Secs+Sfrom);
	}
	
	/* new element, store */
	if (Diff)
	{
	    Secs[Size]=Chain->Secs[Sfrom];
	    Size++; Added++;
	}
    }
    return Added;
}
/* END of merge_secstr() */

/* secstr_diff(): return 0 if the two secstr records *S1 and *S2
 * are equal, 1 if they are different.
 */
static int secstr_diff(const Secstr_ *S1, const Secstr_ *S2)
{
    /* different secondary structure types */
    if (S1->Sectype!=S2->Sectype) return 1;

    /* check fields common to all secstr types */
    if (S1->No!=S2->No || strncmp(S1->Id, S2->Id, 4)
	    || S1->Beg!=S2->Beg || S1->End!=S2->End
	    || S1->Chid!=S2->Chid
	    || S1->Begaa!=S2->Begaa || S1->Endaa!=S2->Endaa
	    || S1->Begrid!=S2->Begrid || S2->Endrid!=S2->Endrid
	    || S1->Type!=S2->Type) return 1;

    /* beta strands have some more things to be compared */
    if (S1->Sectype==PDB_SHEET
	    &&
	    (S1->Strandno!=S2->Strandno
	    || strncmp(S1->Thisat, S2->Thisat, 4)
	    || strncmp(S1->Otherat, S2->Otherat, 4)
	    || S1->This!=S2->This || S1->Other!=S2->Other
	    || S1->Thisrid!=S2->Thisrid || S1->Otherid!=S2->Otherid
	    || S1->Otherchid!=S2->Otherchid))
	return 1;
    return 0;	/* they are indeed the same... */
}
    
/* merge_ssbond(): add the disulfide bond records stored in *Chain
 * to the end of the Size-long array Ssbs. Only new elements are added.
 * Return value: the number of elements added.
 */
static int merge_ssbond(Ssbond_ *Ssbs, int Size, const Chain_ *Chain)
{
    int Sfrom, Sto, Ssbno=Chain->Ssbno, Added=0;
    short Diff;
    
    /* check all S-S bonds in Chain */
    for (Sfrom=0; Sfrom<Ssbno; Sfrom++)
    {
	/* is the present element already in Secs? [0..Size-1] */
	Diff=1;
	for (Sto=0; Diff && Sto<Size; Sto++)
	{
	    Diff=ssbond_diff(Ssbs+Sto, Chain->Ssbs+Sfrom);
	}
	
	/* new element, store */
	if (Diff)
	{
	    Ssbs[Size]=Chain->Ssbs[Sfrom];
	    Size++; Added++;
	}
    }
    return Added;
}
/* END of merge_ssbond() */

/* ssbond_diff(): return 0 if the two S-S bond records *S1 and *S2
 * are equal, 1 if they are different.
 */
static int ssbond_diff(const Ssbond_ *S1, const Ssbond_ *S2)
{
    /* check fields */
    if (S1->Pos1!=S2->Pos1 || S1->Pos2!=S2->Pos2
	    || S1->Ch1!=S2->Ch1 || S1->Ch2!=S2->Ch2
	    || S1->Rid1!=S2->Rid1 || S2->Rid2!=S2->Rid2)
    	return 1;
    return 0;	/* they are indeed the same... */
}
/* END of ssbond_diff() */
    
/* ---- Entry sorting for output ---- */

/* secstr_outorder: sorts the secstruct elements into PDB output order:
 * helix-sheet-turn, and same types are sorted
 * separately according to the order number (No).
 */
static int secstr_outorder(const void *Sec1, const void *Sec2)
{
    register Secstr_ *S1=(Secstr_ *)Sec1, *S2=(Secstr_ *)Sec2;
    register int Ord;
    
    Ord=(int)(S1->Sectype)-(int)(S2->Sectype);
    if (Ord) return(Ord);	/* helix-sheet-turn */
    else if (S1->Sectype==PDB_SHEET)    /* both are sheets */
    {
	Ord=strcmp(S1->Id, S2->Id); /* sort by alphabetical ID order */
	if (Ord) return(Ord);
	else return(S1->No-S2->No); /* same sheet, different strands */
    }
    else return(S1->No-S2->No);	    /* both helices or turns */
}
/* END of secstr_outorder */

/* ssbond_outorder: for sorting the S-S bond records according
 * their No field.
 */
static int ssbond_outorder(const void *Ssb1, const void *Ssb2)
{
    register Ssbond_ *S1=(Ssbond_ *)Ssb1, *S2=(Ssbond_ *)Ssb2;
    return(S1->No-S2->No);
}
/* END of ssbond_outorder */

/* ---- Other auxiliaries ---- */

/* print_card(): prints the LINELEN-long string buffer Line
 * to Pdb. Line is padded with spaces at its end.
 */
static void print_card(FILE *Pdb, char *Line)
{
    int Len=strlen(Line);
    memset(Line+Len, ' ', LINELEN-1-Len);
    fprintf(Pdb, "%s\n", Line);
}
/* END of print_card() */

/* set_pdbdate(): checks the Data field of the PDB structure
 * pointed to by Entry. If the Data field contains a valid date
 * in the format "DD-MMM-YY" then no change is made and 0
 * is returned. If Data doesn't contain a valid date string, 
 * then the current date will be used to fill it in and
 * a non-0 value (the current system time) is returned.
 */
static long set_pdbdate(Pdbentry_ *Entry)
{
    static const char Months[]="JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC";
    
    time_t Curtime;
    struct tm *Tmp;
    int Day, Year;
    char Month[4];
    
    /* valid date? */
    if (strlen(Entry->Date)==9 &&
	3==sscanf(Entry->Date, "%2d-%3s-%2d", &Day, Month, &Year) &&
	1<=Day && Day<=31 &&
	NULL!=strstr(Months, Month) &&
	0<=Year && Year<=99)
	    return(0);
    
    /* invalid date, set correct value */
    Curtime=time(NULL);
    Tmp=localtime(&Curtime);
    strncpy(Month, Months+3*(Tmp->tm_mon), 3); Month[3]='\0';
    sprintf(Entry->Date, "%02d-%3s-%02d", Tmp->tm_mday, Month, 
	    Tmp->tm_year % 100);    /* y2k bug :-) */
    return(Curtime);
}
/* END of set_pdbdate() */

/* current_date(): returns a ptr to an internal string
 * which contains the current date in DD-MMM-YYYY format.
 * The string is overwritten between calls.
 */
static const char *current_date()
{
    static const char Months[]="JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC";
    static char Curdate[12];
    
    time_t Curtime;
    struct tm *Tmp;
    
    Curtime=time(NULL);
    Tmp=localtime(&Curtime);
    sprintf(Curdate, "%02d-", Tmp->tm_mday);
    strncpy(Curdate+3, Months+3*(Tmp->tm_mon), 3);
    sprintf(Curdate+6, "-%04d", 1900+Tmp->tm_year);
    return((const char*) Curdate);
}
/* END of current_date() */

#undef LINELEN
#undef PDB_MALLOC
#undef PDB_CALLOC
#undef PDB_REALLOC
#undef PDB_FREE

/* ==== END OF FUNCTIONS pdbread.c ==== */

