/* ==== FUNCTIONS matrect.c ==== */

/* Simple rectangular matrix module. */

/* 16-Nov-2000. Andras Aszodi */

/* ==== STANDARD HEADERS ==== */

#include <string.h>

/* ==== MODULE HEADERS ==== */

#include "matrect.h"

/* ==== PROTOTYPES ==== */

static void grow_matrect(Matrect_ *M, unsigned int Newrno, unsigned int Newcno);
static const char *next_sscan(const char* Inputstr);

/* ==== FUNCTIONS ==== */

/*
 * init_matrect(M, Rno, Cno):
 * Sets up the matrix *M to contain Rno rows and Cno columns.
 * If Rno==0 or Cno==0 then they will be set to 1.
 * This must be the first operation performed on a Matrect_ struct.
 * Disaster strikes if *M is initialised more than once.
 */
void init_matrect(Matrect_ *M, unsigned int Rno, unsigned int Cno)
{
    double **Mat=NULL;
    register unsigned int i, Eno;
    
    /* if something goes wrong, an empty matrix should be the result */
    if (M==NULL) return;
    M->Mat=NULL; M->Rno=M->Cno=M->Rlen=M->Elen=0;
    if (!Rno)
    {
	fprintf(stderr, "\n? init_matrect(M, %u, %u): Zero size, set to 1\n",
		Rno, Cno);
	Rno=1;
    }
    if (!Cno)
    {
	fprintf(stderr, "\n? init_matrect(M, %u, %u): Zero size, set to 1\n",
		Rno, Cno);
	Cno=1;
    }
    
    Eno=Rno*Cno;
    Mat=(double **) calloc(Rno, sizeof(double *));  /* array of row ptrs */
    
    Mat[0]=(double *) calloc(Eno, sizeof(double));  /* contig. array of elems */
    
    for (i=1; i<Rno; ++i)  /* set up row ptrs */
	Mat[i]=Mat[i-1]+Cno;
    
    M->Mat=Mat; M->Rno=Rno; M->Cno=Cno;
    M->Rlen=Rno; M->Elen=Eno;
}
/* END of init_matrect(M, Rno, Cno) */

/*
 * array_matrect(M, Array):
 * Assumes that Array contains a matrix in row major order
 * and copies its contents into M.
 * It is your responsibility to make sure that Array has the right size
 * and the appropriate layout. This routine assumes no responsibility
 * and performs no dim checks. M must have been initialised and
 * set to the desired size before calling this routine.
 * The only support you can expect that nothing is done if
 * either of the parameters is NULL.
 */
void array_matrect(Matrect_ *M, const double *Array)
{
    if (M==NULL || Array==NULL) return;     /* do nothing */
    memcpy(M->Mat[0], Array, sizeof(double)*M->Elen);
}
/* END of array_matrect(M, Array) */

/*
 * resize_matrect(M, Newrno, Newcno):
 * changes the size of *M so that Newrno rows and Newcno columns can be stored.
 * Does nothing if the dimensions are the same as before.
 * If *M was empty then it will be initialised.
 * Newly added rows/cols will be initialised to 0.0.
 * If Newrno or Newcno are 0, then they will be set to 1.
 */
void resize_matrect(Matrect_ *M, unsigned int Newrno, unsigned int Newcno)
{
    unsigned int Minrno;
    int Cdiff;
    register unsigned int i, j;
    double *Newrowptr;
    
    if (M==NULL) return;
    if (M->Mat==NULL)
    {
	init_matrect(M, Newrno, Newcno);
	return;
    }
    if (!Newrno)
    {
	fprintf(stderr, "\n? resize_matrect(M, %u, %u): Zero size, set to 1\n",
		Newrno, Newcno);
	Newrno=1;
    }
    if (!Newcno)
    {
	fprintf(stderr, "\n? resize_matrect(M, %u, %u): Zero size, set to 1\n",
		Newrno, Newcno);
	Newcno=1;
    }
    if (Newrno==M->Rno && Newcno==M->Cno) return;   /* no change */
    
    /* grow arrays if necessary */
    grow_matrect(M, Newrno, Newcno);
    
    /* move the elements up if Cno grows: do nothing if Cno shrinks */
    Minrno=(Newrno<M->Rno)? Newrno: M->Rno;
    Cdiff=Newcno-M->Cno;	/* each row grows by this much, <0 for shrinkage */
    if (Cdiff>0)
    {
	/* move rows to create new space */
	for (i=Minrno-1; i>=1; --i)
	{
	    Newrowptr=M->Mat[i]+Cdiff*i;
	    memmove(Newrowptr, M->Mat[i], M->Cno*sizeof(double));
	    M->Mat[i]=Newrowptr;
	}
	
	/* zero newly created elements */
	for (i=0; i<M->Rno; ++i)
	    for (j=M->Cno; j<Newcno; ++j)
		M->Mat[i][j]=0.0;
    }

    /* set new row pointers if rows were added, zero new rows */
    for (i=M->Rno; i<Newrno; ++i)
    {
	M->Mat[i]=M->Mat[i-1]+Newcno;
	for (j=0; j<Newcno; ++j)
	    M->Mat[i][j]=0.0;
    }
    
    /* done */
    M->Rno=Newrno; M->Cno=Newcno;
}
/* END of resize_matrect(M, Newrno, Newcno) */

/*
 * copy_matrect(From, To):
 * Copies the contents of *From to *To. *To will be initialised
 * and/or resized if necessary.
 */
void copy_matrect(const Matrect_ *From, Matrect_ *To)
{
    register unsigned int i;
    
    /* various pathological cases */
    if (From==To) return;
    if (From==NULL || To==NULL)
    {
	fprintf(stderr, "\n? copy_matrect(%p,%p): NULL pointer(s), do nothing\n",
		(void*)From, (void*)To);
	return;
    }
    
    /* make room in target if necessary */
    grow_matrect(To, From->Rno, From->Cno);
    
    /* do the actual copying */
    memcpy(To->Mat[0], From->Mat[0], From->Rno*From->Cno*sizeof(double));
    if (From->Rno!=To->Rno || From->Cno!=To->Cno)
    {
	for (i=1; i<From->Rno; ++i) /* set row ptrs in target */
	    To->Mat[i]=To->Mat[i-1]+From->Cno;
	To->Rno=From->Rno; To->Cno=From->Cno;   /* physlen not copied :-) */
    }
}
/* END of copy_matrect(From, To) */

/*
 * read_matrect(In, M):
 * reads a matrix from the file In into *M.
 * Format is the "nice format" produced by write_matrect().
 * Returns 0 on error, the number of lines processed otherwise.
 */
int read_matrect(FILE *In, Matrect_ *M)
{
#define LINELEN 120
    static char Line[LINELEN];
    
    Matrect_ Temp;
    unsigned int Lineno=0, Dashlen, Cidx=0, Newcidx, Cxtemp, 
            Ridx=0, Rxtemp, Rexpect;
    short Colidx=0, Topdelim, Chunkdelim, Botdelim=0;	/* "Booleans" */
    int Retval;
    char *Eolnptr;
    const char *Scanptr;
    double Tempcoord;
    
    /* read input line by line */
    init_matrect(&Temp, 1, 1);
    while (NULL!=fgets(Line, LINELEN, In))
    {
        ++Lineno;
        /* skip empty lines or comments */
        if (!Line[0] || Line[0]=='\n' || Line[0]=='#')
            continue;
        
        /* chop off terminating \n if present */
        Eolnptr=strrchr(Line, '\n');
        if (Eolnptr!=NULL) *Eolnptr='\0';
        
        /* look for column index line (on top)
         * the next chunk will be between Cidx..Newcidx */
        if (!Colidx)
        {
            for (Newcidx=Cidx, Scanptr=Line;
		    *Scanptr && sscanf(Scanptr, "%u", &Cxtemp) && Cxtemp==Newcidx;
		    ++Newcidx)
		Scanptr=next_sscan(Scanptr);
            if (Newcidx>Cidx)
            {
                resize_matrect(&Temp, Ridx? Ridx: 1, Newcidx);
                Colidx=1;
                Topdelim=0; 
            }
            continue;
        }
        
        /* col idx line seen, check for top delimiter "----" */
        if (!Topdelim)
        {
            Dashlen=strspn(Line, "-");
            if (Dashlen==strlen(Line))
            {
                Topdelim=1;  /* found top delimiter */
                Chunkdelim=0;
                Rexpect=0;  /* expected row number */
            }
            continue;
        }
        
        /* check for bottom delimiter "===" if rows have been seen */
        if (Ridx && Topdelim && !Botdelim)
        {
            Dashlen=strspn(Line, "=");
            if (Dashlen==strlen(Line))
            {
                Botdelim=1;    /* found bottom delimiter */
                break;  /* we're done */
            }
        }
        
        /* check for chunk delimiter "---" if rows have been seen */
        if (Ridx && Topdelim && !Chunkdelim)
        {
            Dashlen=strspn(Line, "-");
            if (Dashlen==strlen(Line))
            {
                Chunkdelim=1;    /* found chunk delimiter */
                Colidx=0;   /* look for next column index line */
                Cidx=Newcidx;
                continue;
            }
        }
        
        /* get the rows: look for data line "x | yyyy zzzz ..." */
        Retval=sscanf(Line, "%u", &Rxtemp);
        if (!Retval || Rxtemp!=Rexpect)
        {
            fprintf(stderr, "\n? read_matrect(): Row idx %u expected, %u found at line %u\n",
		    Rexpect, Rxtemp, Lineno);
            break;
        }
        ++Rexpect;  /* next row index to be expected */
        if (Rexpect>Ridx)   /* first chunk */
        {
            Ridx=Rexpect;   /* count rows */
            resize_matrect(&Temp, Ridx, Newcidx);   /* make room for new row */
        }
	
	/* find the '|' separator */
        Scanptr=strchr(Line, '|');
        if (Scanptr==NULL)
        {
            fprintf(stderr, "\n? read_matrect(): Separator \'|\' not found at line %u\n",
		    Lineno);
            break;
        }
        
        /* read the matrix elements in the [Rxtemp] row chunk */
        for (Cxtemp=Cidx, Scanptr+=1; 
		*Scanptr && sscanf(Scanptr, "%lg", &Tempcoord) && Cxtemp<Newcidx; 
		++Cxtemp)
	{
            Temp.Mat[Rxtemp][Cxtemp]=Tempcoord;
	    Scanptr=next_sscan(Scanptr);
	}
        if (Cxtemp!=Newcidx)
        {
            fprintf(stderr, "\n? read_matrect(): Data field problem (line %u)\n",
		    Lineno);
            break;
        }
    }       /* while */
    
    /* input was OK if bottom delimiter has been seen */
    if (Botdelim)
    {
	copy_matrect(&Temp, M);
	return Lineno;
    }
    else return 0;
#undef LINELEN
}
/* END of read_matrect(In, M) */

/* 
 * write_matrect(Out, M, Linewidth, Width, Prec): 
 * lists M to the open stream Out with entries in scientific format,
 * Prec digits precision. If a row takes up more than 80 chars,
 * then the matrix is cut up nicely.
 */
void write_matrect(FILE *Out, const Matrect_ *M, int Prec)
{
    static const int LINEWIDTH=80;
    char Entrys[10],Cols[8],Rows[8];	/* format strings and len determ */
    int i,j,k,Sizew,Width,Chunk,Jbeg,Items,Ulinelen;
    
    if (M==NULL)
    {
	fputs("\n? write_matrect(Out, NULL, ...)\n", stderr);
	return;
    }
    
    sprintf(Cols,"%u",M->Cno);	/* get printed width of Size */
    Sizew=strlen(Cols);
    Width=Prec+7;
    if (Sizew>Width) Width=Sizew;
    sprintf(Entrys,"%%-%d.%de ",Width,Prec);	/* make format strs */
    sprintf(Cols,"%%-%dd ",Width);
    sprintf(Rows,"%%%dd | ",Sizew);
    Items=(LINEWIDTH-Sizew-3)/(Width+1);	/* columns per chunk */

    /* main cycle: print chunks of the matrix */
    for (Jbeg=0,Chunk=(M->Cno-1)/Items+1; Chunk>0; Jbeg+=Items,Chunk--)
    {
	/* underline length */
	Ulinelen=(Chunk>1)? Items*(Width+1)+Sizew+3:
		(M->Cno-Jbeg)*(Width+1)+Sizew+3;
	/* print head line */
	for (k=0; k<Sizew+3; k++) putc(' ', Out);
	for (j=Jbeg; j<M->Cno && j<Jbeg+Items; j++)
	    fprintf(Out, Cols, j);	/* col numbers */
	putc('\n', Out);
	for (k=0; k<Ulinelen; k++) putc('-', Out);
	putc('\n', Out);

	/* print chunks for all rows */
	for (i=0; i<M->Rno; i++)
	{
	    fprintf(Out, Rows, i); 	/* row idx */
	    for (j=Jbeg; j<M->Cno && j<Jbeg+Items; j++)
		fprintf(Out, Entrys, M->Mat[i][j]);
	    putc('\n', Out);
	}

	/* chunk separator */
	for (k=0; k<Ulinelen; k++) putc('=', Out);
	fputs("\n\n", Out);
    }
}
/* END of write_matrect(Out, M, Linewidth, Width, Prec) */

/*
 * free_matrect(M):
 * frees up all storage associated with *M.
 * *M will be kept, it will be set to an invalid empty state.
 * This must be the last operation performed on a Matrect_ struct.
 */
void free_matrect(Matrect_ *M)
{
    if (M==NULL || M->Mat==NULL) return;
    free(M->Mat[0]); free(M->Mat);
    M->Mat=NULL;
    M->Rno=M->Cno=M->Rlen=M->Elen=0;
}
/* END of free_matrect(M) */

/* ---- Auxiliaries ---- */

/*
 * grow_matrect(M, Newrno, Newcno):
 * Resizes the arrays in *M if Newrno or Newcno require that they
 * are to be grown over their previous physical size.
 */
static void grow_matrect(Matrect_ *M, unsigned int Newrno, unsigned int Newcno)
{
    register unsigned int Newsize, Neweno, i;
    
    /* grow the row ptr array */
    if (Newrno>M->Rlen)
    {
	Newsize=(Newrno>2*M->Rlen)? Newrno: 2*M->Rlen;
	M->Mat=(double **) realloc(M->Mat, Newsize*sizeof(double*));
	M->Rlen=Newsize;
    }
    
    /* grow the data array */
    Neweno=Newrno*Newcno;
    if (Neweno>M->Elen)
    {
	Newsize=(Neweno>2*M->Elen)? Neweno: 2*M->Elen;
	M->Mat[0]=(double *) realloc(M->Mat[0], Newsize*sizeof(double));
	M->Elen=Newsize;
	
	/* reassign row ptrs as realloc might have moved data */
	for (i=1; i<(Newrno>M->Rno? Newrno: M->Rno); ++i)
	    M->Mat[i]=M->Mat[i-1]+M->Cno;
    }
}
/* END of grow_matrect(M, Newrno, Newcno) */

/*
 * next_sscan(Inputstr):
 * Assuming that Inputstr is a concatenation of segments
 * consisting of zero or more whitespace chars followed by
 * one or more non-whitespace chars, a pointer is returned
 * pointing past the first non-ws region, ie. where input
 * would stop if Inputstr were read by sscanf.
 * Inputstr is assumed to be \0-terminated.
 */
static const char *next_sscan(const char* Inputstr)
{
    static const char *Ws=" \t\n";  /* whitespace */
    unsigned int Span1, Span2;
    
    if (Inputstr==NULL) return NULL;
    Span1=strspn(Inputstr, Ws);     /* ws segment at begin */
    Span2=strcspn(Inputstr+Span1, Ws);	/* non-ws following */
    return (Inputstr+Span1+Span2);
}
/* END of next_sscan(Inputstr) */

/* ==== END OF FUNCTIONS matrect.c ==== */
