/* ==== FUNCTIONS mattri.c ==== */

/* Simple symmetric (lower triangular) matrix module. */

/* 16-Nov-2000. Andras Aszodi */

/* ==== STANDARD HEADERS ==== */

#include <string.h>

/* ==== MODULE HEADERS ==== */

#include "mattri.h"

/* ==== PROTOTYPES ==== */

static void grow_mattri(Mattri_ *M, unsigned int Newsize);
static const char *next_sscan(const char* Inputstr);

/* ==== FUNCTIONS ==== */

/*
 * init_mattri(M, Size):
 * Sets up the matrix *M to contain Size rows and columns.
 * If Size==0, it will be set to 1.
 * This must be the first operation performed on a Mattri_ struct.
 * Disaster strikes if *M is initialised more than once.
 */
void init_mattri(Mattri_ *M, unsigned int Size)
{
    double **Mat=NULL;
    register unsigned int i, Eno;
    
    /* if something goes wrong, an empty matrix should be the result */
    if (M==NULL) return;
    M->Mat=NULL; M->Size=M->Rlen=M->Elen=0; 
    if (!Size)
    {
	fputs("\n? init_mattri(M, 0): Size set to 1\n", stderr);
	Size=1;
    }
    
    Eno=(Size*(Size+1))/2;
    Mat=(double **) calloc(Size, sizeof(double *));  /* array of row ptrs */
    
    Mat[0]=(double *) calloc(Eno, sizeof(double));  /* contig. array of elems */
    
    for (i=1; i<Size; ++i)  /* set up row ptrs */
	Mat[i]=Mat[i-1]+i;
    
    M->Mat=Mat; M->Size=M->Rlen=Size;
    M->Elen=Eno;
}
/* END of init_mattri(M, Size) */

/*
 * array_mattri(M, Array):
 * Assumes that Array contains a lower triangular matrix
 * in row major order and copies its contents into M.
 * It is your responsibility to make sure that Array has the right size
 * and the appropriate layout. This routine assumes no responsibility
 * and performs no dim checks. M must have been initialised and
 * set to the desired size before calling this routine.
 * The only support you can expect that nothing is done if
 * either of the parameters is NULL.
 */
void array_mattri(Mattri_ *M, const double *Array)
{
    if (M==NULL || Array==NULL) return;     /* do nothing */
    memcpy(M->Mat[0], Array, sizeof(double)*M->Elen);
}
/* END of array_mattri(M, Array) */

/*
 * resize_mattri(M, Newsize):
 * changes the size of *M so that Newsize rows and columns can be stored.
 * Does nothing if the dimensions are the same as before.
 * Newly added rows/cols will be initialised to 0.0.
 * If Newsize==0, then it will be set to 1.
 */
void resize_mattri(Mattri_ *M, unsigned int Newsize)
{
    register unsigned int i, j;
    
    if (M==NULL) return;
    if (M->Mat==NULL)
    {
	init_mattri(M, Newsize);
	return;
    }
    if (!Newsize)
    {
	fputs("\n? resize_mattri(M, 0): Newsize set to 1\n", stderr);
	Newsize=1;
    }
    if (Newsize==M->Size) return;   /* no change */
    
    /* grow arrays if necessary */
    grow_mattri(M, Newsize);
    
    /* set new row pointers if rows were added, zero new rows */
    for (i=M->Size; i<Newsize; ++i)
    {
	M->Mat[i]=M->Mat[i-1]+i;
	for (j=0; j<=i; ++j)
	    M->Mat[i][j]=0.0;
    }
    
    /* done */
    M->Size=Newsize;
}
/* END of resize_mattri(M, Newsize) */

/*
 * copy_mattri(From, To):
 * Copies the contents of *From to *To. *To will be initialised
 * and/or resized if necessary.
 */
void copy_mattri(const Mattri_ *From, Mattri_ *To)
{
    register unsigned int i;
    
    /* various pathological cases */
    if (From==To) return;
    if (From==NULL || To==NULL)
    {
	fprintf(stderr, "\n? copy_mattri(%p,%p): NULL pointer(s), do nothing\n",
		(void*)From, (void*)To);
	return;
    }
    
    /* make room in target if necessary */
    grow_mattri(To, From->Size);
    
    /* do the actual copying */
    memcpy(To->Mat[0], From->Mat[0], (From->Size*(From->Size+1)*sizeof(double))/2);
    if (From->Size!=To->Size)
    {
	for (i=1; i<From->Size; ++i) /* set row ptrs in target */
	    To->Mat[i]=To->Mat[i-1]+i;
	To->Size=From->Size;   /* physical lengths not copied :-) */
    }
}
/* END of copy_mattri(From, To) */

/*
 * read_mattri(In, M):
 * reads a matrix from the file In into *M.
 * Format is the "nice format" produced by write_mattri().
 * Returns 0 on error, the number of lines processed otherwise.
 */
int read_mattri(FILE *In, Mattri_ *M)
{
#define LINELEN 120
    static char Line[LINELEN];
    
    Mattri_ Temp;
    unsigned int Lineno=0, Dashlen, Cidx=0, Newcidx, Cxtemp, 
            Ridx=0, Rxtemp, Rexpect, Newsize, Maxcx;
    short Colidx=0, Topdelim, Chunkdelim, Botdelim=0;	/* "Booleans" */
    int Retval;
    char *Eolnptr;
    const char *Scanptr;
    double Tempcoord;
    
    /* read input line by line */
    init_mattri(&Temp, 0);
    while (NULL!=fgets(Line, LINELEN, In))
    {
        ++Lineno;
        /* skip empty lines or comments */
        if (!Line[0] || Line[0]=='\n' || Line[0]=='#')
            continue;
        
        /* chop off terminating \n if present */
        Eolnptr=strrchr(Line, '\n');
        if (Eolnptr!=NULL) *Eolnptr='\0';
        
        /* look for column index line (on top)
         * the next chunk will be between Cidx..Newcidx */
        if (!Colidx)
        {
            for (Newcidx=Cidx, Scanptr=Line;
		    *Scanptr && sscanf(Scanptr, "%u", &Cxtemp) && Cxtemp==Newcidx;
		    ++Newcidx)
		Scanptr=next_sscan(Scanptr);
            if (Newcidx>Cidx)
            {
		if (Ridx) Newsize=(Ridx>Newcidx)? Ridx: Newcidx;
		else Newsize=(1>Newcidx)? 1: Newcidx;
                resize_mattri(&Temp, Newsize);
                Colidx=1;
                Topdelim=0; 
            }
            continue;
        }
        
        /* col idx line seen, check for top delimiter "----" */
        if (!Topdelim)
        {
            Dashlen=strspn(Line, "-");
            if (Dashlen==strlen(Line))
            {
                Topdelim=1;  /* found top delimiter */
                Chunkdelim=0;
                Rexpect=0;  /* expected row number */
            }
            continue;
        }
        
        /* check for bottom delimiter "===" if rows have been seen */
        if (Ridx && Topdelim && !Botdelim)
        {
            Dashlen=strspn(Line, "=");
            if (Dashlen==strlen(Line))
            {
                Botdelim=1;    /* found bottom delimiter */
                break;  /* we're done */
            }
        }
        
        /* check for chunk delimiter "---" if rows have been seen */
        if (Ridx && Topdelim && !Chunkdelim)
        {
            Dashlen=strspn(Line, "-");
            if (Dashlen==strlen(Line))
            {
                Chunkdelim=1;    /* found chunk delimiter */
                Colidx=0;   /* look for next column index line */
                Cidx=Newcidx;
                continue;
            }
        }
        
        /* get the rows: look for data line "x | yyyy zzzz ..." */
        Retval=sscanf(Line, "%u", &Rxtemp);
        if (!Retval || Rxtemp!=Rexpect)
        {
            fprintf(stderr, "\n? read_mattri(): Row idx %u expected, %u found at line %u\n",
		    Rexpect, Rxtemp, Lineno);
            break;
        }
        ++Rexpect;  /* next row index to be expected */
        if (Rexpect>Ridx)   /* first chunk */
        {
            Ridx=Rexpect;   /* count rows */
	    Newsize=(Ridx>Newcidx)? Ridx: Newcidx;
            resize_mattri(&Temp, Newsize);   /* make room for new row */
        }
	
	/* find the '|' separator */
        Scanptr=strchr(Line, '|');
        if (Scanptr==NULL)
        {
            fprintf(stderr, "\n? read_mattri(): Separator \'|\' not found at line %u\n",
		    Lineno);
            break;
        }
        
        /* read the matrix elements in the [Rxtemp] row chunk */
        for (Cxtemp=Cidx, Scanptr+=1; 
		*Scanptr && sscanf(Scanptr, "%lg", &Tempcoord) && Cxtemp<=Rxtemp; 
		++Cxtemp)
	{
            Temp.Mat[Rxtemp][Cxtemp]=Tempcoord;
	    Scanptr=next_sscan(Scanptr);
	}
	Maxcx=(Rxtemp>Newcidx)? Newcidx+1: Rxtemp+1;
        if (Cxtemp<Maxcx)   /* didn't read enough data */
        {
            fprintf(stderr, "\n? read_mattri(): Data field problem (line %u)\n",
		    Lineno);
            break;
        }
    }       /* while */
    
    /* input was OK if bottom delimiter has been seen */
    if (Botdelim)
    {
	copy_mattri(&Temp, M);
	return Lineno;
    }
    else return 0;
#undef LINELEN
}
/* END of read_mattri(In, M) */

/* 
 * write_mattri(Out, M, Linewidth, Width, Prec): 
 * lists M to the open stream Out with entries in scientific format,
 * Prec digits precision. If a row takes up more than 80 chars,
 * then the matrix is cut up nicely.
 */
void write_mattri(FILE *Out, const Mattri_ *M, int Prec)
{
    static const int LINEWIDTH=80;
    
    char Entrys[10],Cols[8],Rows[8];	/* format strings and len determ */
    int i,j,k,Width,Sizew,Chunk,Jbeg,Items,Ulinelen;
    
    if (M==NULL)
    {
	fputs("\n? write_mattri(Out, NULL, ...)\n", stderr);
	return;
    }
    
    sprintf(Cols,"%u",M->Size);	/* get printed width of Size */
    Sizew=strlen(Cols);
    Width=Prec+7;
    if (Sizew>Width) Width=Sizew;
    sprintf(Entrys,"%%-%d.%de ",Width,Prec);	/* make format strs */
    sprintf(Cols,"%%-%dd ",Width);
    sprintf(Rows,"%%%dd | ",Sizew);
    Items=(LINEWIDTH-Sizew-3)/(Width+1);	/* columns per chunk */

    /* main cycle: print chunks of the matrix */
    for (Jbeg=0,Chunk=(M->Size-1)/Items+1; Chunk>0; Jbeg+=Items,Chunk--)
    {
	/* underline length */
	Ulinelen=(Chunk>1)? Items*(Width+1)+Sizew+3:
		(M->Size-Jbeg)*(Width+1)+Sizew+3;
	/* print head line */
	for (k=0; k<Sizew+3; k++) putc(' ', Out);
	for (j=Jbeg; j<M->Size && j<Jbeg+Items; j++)
	    fprintf(Out, Cols, j);	/* col numbers */
	putc('\n', Out);
	for (k=0; k<Ulinelen; k++) putc('-', Out);
	putc('\n', Out);

	/* print chunks for all rows */
	for (i=0; i<M->Size; i++)
	{
	    fprintf(Out, Rows, i); 	/* row idx */
	    for (j=Jbeg; j<=i && j<Jbeg+Items; j++)
		fprintf(Out, Entrys, M->Mat[i][j]);
	    putc('\n', Out);
	}

	/* chunk separator */
	for (k=0; k<Ulinelen; k++) putc('=', Out);
	fputs("\n\n", Out);
    }
}
/* END of write_mattri(Out, M, LINEWIDTH, Width, Prec) */

/*
 * free_mattri(M):
 * frees up all storage associated with *M.
 * *M will be kept, it will be in an invalid empty state.
 * This must be the last operation performed on a Mattri_ struct.
 */
void free_mattri(Mattri_ *M)
{
    if (M==NULL || M->Mat) return;
    free(M->Mat[0]); free(M->Mat);
    M->Mat=NULL;
    M->Size=M->Rlen=M->Elen=0;
}
/* END of free_mattri(M) */

/* ---- Auxiliaries ---- */

/*
 * grow_mattri(M, Newsize):
 * Resizes the arrays in *M if Newsize requires that they
 * are to be grown over their previous physical size.
 */
static void grow_mattri(Mattri_ *M, unsigned int Newsize)
{
    register unsigned int Newrlen, Neweno, i;
    
    /* grow the row ptr array */
    if (Newsize>M->Rlen)
    {
	Newrlen=(Newsize>2*M->Rlen)? Newsize: 2*M->Rlen;
	M->Mat=(double **) realloc(M->Mat, Newrlen*sizeof(double*));
	M->Rlen=Newrlen;
    }
    
    /* grow the data array */
    Neweno=(Newsize*(Newsize+1))/2;
    if (Neweno>M->Elen)
    {
	if (Neweno<2*M->Elen) Neweno=2*M->Elen; /* grow at least 2x */
	M->Mat[0]=(double *) realloc(M->Mat[0], Neweno*sizeof(double));
	M->Elen=Neweno;
	
	/* reassign row ptrs as realloc might have moved data */
	for (i=1; i<Newsize; ++i)
	    M->Mat[i]=M->Mat[i-1]+i;
    }
}
/* END of grow_mattri(M, Newsize) */

/*
 * next_sscan(Inputstr):
 * Assuming that Inputstr is a concatenation of segments
 * consisting of zero or more whitespace chars followed by
 * one or more non-whitespace chars, a pointer is returned
 * pointing past the first non-ws region, ie. where input
 * would stop if Inputstr were read by sscanf.
 * Inputstr is assumed to be \0-terminated.
 */
static const char *next_sscan(const char* Inputstr)
{
    static const char *Ws=" \t\n";  /* whitespace */
    unsigned int Span1, Span2;
    
    if (Inputstr==NULL) return NULL;
    Span1=strspn(Inputstr, Ws);     /* ws segment at begin */
    Span2=strcspn(Inputstr+Span1, Ws);	/* non-ws following */
    return (Inputstr+Span1+Span2);
}
/* END of next_sscan(Inputstr) */

/* ==== END OF FUNCTIONS matrect.c ==== */
