/* ==== FUNCTIONS cmdlong.c ==== */

/* Long command-line options. */

/* ANSI C, 14-Oct-1999. Andris */

/* ---- STANDARD HEADERS ---- */

#include <string.h>

/* ---- MODULE HEADER ---- */

#include "cmdlong.h"

/* ---- TYPEDEFS ---- */

/* Cmdlongtype_: enum for the possible option variable types.
 * CMDLONG_VOID is only a placeholder to make sure the other enums
 * are not 0: used for input format error checking. 
 */
typedef enum {CMDLONG_VOID=0, CMDLONG_BOOL, CMDLONG_YESNO, CMDLONG_CHAR, 
    	CMDLONG_LONG, CMDLONG_DBL, CMDLONG_STR} Cmdlongtype_;

/* Cmdlong_: struct for storing the option string and value. */
typedef struct
{
    char *Optstr;   /* the full option string w/o initial "-" or "--" */
    int Uniqlen;    /* the "unique length" of Optstr for abbreviations */
    Cmdlongtype_ Type;	/* the type of the option variable */
    Cmdlongvar_ Val;   /* the value of the option variable */
    int Argvidx;     /* index of option in argv[], 0 if not found */
} Cmdlong_;

/* ---- STATIC GLOBALS ---- */

Cmdlong_ *Opts=NULL;	/* array storing option descriptions */
int Optsno=0;	    	/* length of Opts array */

/* ---- PROTOTYPES ---- */

static int make_cmdlong(const char *Opttok, Cmdlong_ *Newopt);
static int unique_length(Cmdlong_ *Newopt);
static int uniq_len(const char *Str1, const char *Str2);

static int parse_optargv(int argc, char *argv[], int Fx);
static int locate_option(const char *Opt);

/* ==== FUNCTIONS ==== */

/*
 * cmdlong_addopts(Descr): parses the long option description string Descr
 * and adds the descriptions to the internal Opts array.
 * Descr should contain the following description tokens separated
 * by one or more spaces or tab chars:-
 * "opt": boolean option -opt or --opt, or
 * "opt=x": option with a mandatory argument
 * where the argument type is
 *     'y' for "yes|no" switches ("on|off" also accepted, case-insensitive)
 *     'c' for characters,
 *     'd' for integers (actually longs),
 *     'f' for floats (actually doubles)
 *     's' for strings.
 * The function may be invoked more than once (but this is unnecessary).
 * To avoid memory leaks, call remove_cmdlongs() before exiting your program.
 * Return value: the number of successfully added options (0 on error).
 */
int cmdlong_addopts(const char *Descr)
{
    static const int OPTS_CHUNK=4;  /* prealloc size */
    
    char *Des=NULL; /* local copy of Descr for strtok() */
    char *Opttok;   /* ptr to an option token */
    int Added=0;    /* counts successful additions */
    Cmdlong_ Newopt;	/* storage for new option */
    
    /* input sanity check */
    if (Descr==NULL || ! *Descr)
    {
	fprintf(stderr, "\n? cmdlong_addopts(\"%s\"): empty\n", Descr);
	return(Added);
    }
    
    /* poor man's strdup() */
    Des=(char *) calloc(strlen(Descr)+1, sizeof(char));
    strcpy(Des, Descr);

    /* get first token */
    Opttok=strtok(Des, " \t");
    if (Opttok==NULL) Opttok=Des;   /* one token only, no delimiter */
    
    /* process tokens */
    do
    {
	if (!make_cmdlong(Opttok, &Newopt))
	    continue;	/* could not be parsed for some reason, skip */
	
	/* make room for new option record */
	if (Optsno % OPTS_CHUNK == 0)
	    Opts=(Cmdlong_ *) realloc(Opts, (Optsno+OPTS_CHUNK)*sizeof(Cmdlong_));
	if (!unique_length(&Newopt)) /* adjust unique lengths */
	{
	    fprintf(stderr, "\n? cmdlong_addopts(\"%s\"): Option \"%s\" not unique\n",
		    Descr, Newopt.Optstr);
	    free(Newopt.Optstr);
	    continue;
	}
	
	Opts[Optsno]=Newopt;	/* shallow copy will suffice */
	Optsno++; Added++;
    }
    while (NULL!=(Opttok=strtok(NULL, " \t")));
    
    free(Des);
    return(Added);
}
/* END of cmdlong_addopts(Descr) */

/*
 * make_cmdlong(Opttok, Newopt): parses Opttok (for the format, see 
 * the comments to cmdlong_addopts(...)) and inits the option
 * structure in *Newopt accordingly. In particular,
 * Newopt->Optstr will be allocated and copied to here.
 * Quite dangerous, to be used by cmdlong_addopts() only
 * where an immediate shallow copy is done.
 * Return value: 0 on error, 1 if OK.
 */
static int make_cmdlong(const char *Opttok, Cmdlong_ *Newopt)
{
    int Toklen=strlen(Opttok), Optlen;
    char *Lasteq;
    
    /* find last '=' */
    Lasteq=strrchr(Opttok, '=');
    if (Lasteq==NULL)	/* Boolean option, no argument */
    {
	/* poor man's strdup() again */
	Newopt->Optstr=(char *) calloc(Toklen+1, sizeof(char));
	strcpy(Newopt->Optstr, Opttok);
	Newopt->Type=CMDLONG_BOOL;
	Newopt->Val.Bool=0;
    }
    else
    {
	/* check type description char after last '=' */
	switch(*(Lasteq+1))
	{
	    case 'y':
		Newopt->Type=CMDLONG_YESNO;
		Newopt->Val.Bool=0;
	    break;
	    case 'c':
		Newopt->Type=CMDLONG_CHAR;
		Newopt->Val.Char='\0';
	    break;
	    case 'd':
		Newopt->Type=CMDLONG_LONG;
		Newopt->Val.Long=0;
	    break;
	    case 'f':
		Newopt->Type=CMDLONG_DBL;
		Newopt->Val.Dbl=0.0;
	    break;
	    case 's':
		Newopt->Type=CMDLONG_STR;
		Newopt->Val.Str=NULL;
	    break;
	    default:
		    fprintf(stderr, "\n? make_cmdlong(\"%s\"): Unknown arg type \'%c\'\n",
		    	    Opttok, *(Lasteq+1));
	    	    return(0);
	}   /* END switch */
	
	/* copy actual option string only */
	Optlen=Lasteq-Opttok;
	Newopt->Optstr=(char *) calloc(Optlen+1, sizeof(char));
	strncpy(Newopt->Optstr, Opttok, Optlen);
    }	/* END if Lasteq...else */
    
    Newopt->Uniqlen=1;	/* at least one char is needed :-) */
    Newopt->Argvidx=0;	/* not seen yet */
    return(1);	/* OK */
}
/* END of make_cmdlong(Opttok, Newopt) */

/*
 * unique_length(Newopt): compares the option string Newopt->Optstr
 * to the other option strings already in the static Opts array
 * and updates their unique lengths if necessary.
 * Return value: 0 if Newopt->Optstr has been seen before, 1 if OK.
 */
static int unique_length(Cmdlong_ *Newopt)
{
    int i, Unilen;
    
    if (!Optsno) return(1); /* one char is enough :-) */
    
    for (i=0; i<Optsno; i++)
    {
	Unilen=uniq_len(Opts[i].Optstr, Newopt->Optstr);
	if (Unilen<0)	/* same! */
	    return(0);
	if (Opts[i].Uniqlen<Unilen) Opts[i].Uniqlen=Unilen;
	if (Newopt->Uniqlen<Unilen) Newopt->Uniqlen=Unilen;
    }
    return(1);
}
/* END of unique_length(Newopt) */
	
/*
 * uniq_len(Str1, Str2): returns the unique length between Str1, 2
 * which is the length that ensures that strncmp(Str1, Str2, Length)
 * returns !=0. If the two strings are equal then -strlen(Str1)
 * is returned.
 */
static int uniq_len(const char *Str1, const char *Str2)
{
    int Unilen;
    
    if (Str1==NULL || Str2==NULL
	    || (! *Str1 && ! *Str2))
	return(0);
    for (Unilen=1; *Str1 && *Str2 && *Str1 == *Str2; Str1++, Str2++, Unilen++);
    return((! *Str1 && ! *Str2)? -Unilen: Unilen);
}
/* END of uniq_len(Str1, Str2) */

/*
 * cmdlong_getopts(argc, argv): reads the command-line argument array argv
 * and parses all options. Parsing stops at "-", "--" or at the first
 * non-option argument. Invoke once before querying the options.
 * A warning is issued if an unknown option is detected. Malformed
 * options (with bad or missing values etc.) will be treated as absent.
 * Return value: the index of the argv[] element containing the first
 * non-option value, a negative value if there are no non-options left.
 */
int cmdlong_getopts(int argc, char *argv[])
{
    int Fx, Consume;
    
    /* paranoia */
    if (!Optsno)
    {
	fputs("\n? cmdlong_getopts(): called before cmdlong_addopts()\n", stderr);
	return(-argc+1);
    }
    
    /* locate all options in argv[]: stop at first non-option */
    for (Fx=1; Fx<argc; Fx++)
    {
	if (argv[Fx][0]!='-')	/* first non-option */
	    break;
	if (!strcmp(argv[Fx], "-") || !strcmp(argv[Fx], "--"))	/* end-of-opts */
	    break;
	
	/* how many argv[] items are used by the option? */
	Consume=parse_optargv(argc, argv, Fx);
	if (!Consume)
	{
	    fprintf(stderr, "\n? %s: Unknown option \"%s\"\n",
		    argv[0], argv[Fx]);
	    continue;
	}
	Fx+=(Consume-1);    /* move to next (skip one or two) */
    }
    return((Fx<argc)? Fx: -Fx);
}
/* END of cmdlong_getopts(argc, argv) */

/*
 * parse_optargv(argc, argv, Fx): parses the option at argv[Fx]
 * and its argument (if present) into the internal Opts array.
 * Returns 0 on error, 1 if Fx should be incremented by 1 to
 * get the next option (ie. for "-opt=val" or "-opt"),
 * 2 if Fx should be incremented by 2 (when the option
 * and its value were separated by a space).
 */
static int parse_optargv(int argc, char *argv[], int Fx)
{
    char *Opt, *Valstr, *Endconv;
    int Idx, Retval;
    Cmdlongtype_ Err;
    long Ltmp;
    double Dtmp;
    
    /* walk over initial "--": argv[Fx] is guaranteed to begin with '-' */
    Opt=argv[Fx]+1;
    if (*Opt=='-') Opt++;
    
    /* find option record */
    Idx=locate_option(Opt);
    if (Idx<0)  /* not found */
	return(0);
    
    if (Opts[Idx].Type==CMDLONG_BOOL)	/* we're almost done */
    {
	Opts[Idx].Val.Bool=1;	/* set */
	Opts[Idx].Argvidx=Fx;
	return(1);
    }
    
    /* figure out if "-opt=val" or "-opt" "val" */
    Valstr=strrchr(argv[Fx], '=');
    if (Valstr==NULL)	/* "-opt val" */
    {
	if (Fx==argc-1)
	{
	    fprintf(stderr, "\n? %s: Value missing for option \"%s\"\n",
		    argv[0], argv[Fx]);
	    return(1);
	}
	Valstr=argv[Fx+1];
	Retval=2;   /* must skip both "-opt" and "val" */
    }
    else
    {
	Valstr++;   /* next after '=' */
	if (! *Valstr)  /* points to end-of-string or something bad */
	{
	    fprintf(stderr, "\n? %s: Value missing for option \"%s\"\n",
		    argv[0], argv[Fx]);
	    return(1);
	}
	Retval=1;   /* skip only "-opt=val" */
    }
    
    /* convert value from Valstr and store if successful */
    Err=CMDLONG_VOID;
    switch(Opts[Idx].Type)
    {
	case CMDLONG_YESNO:
	    Opts[Idx].Val.Bool=
		    (!strncasecmp(Valstr, "yes", 3)
		    || !strncasecmp(Valstr, "on", 2));
	break;
	case CMDLONG_CHAR:
	    Opts[Idx].Val.Char=(*Valstr);
	break;
	case CMDLONG_LONG:
	    Ltmp=strtol(Valstr, &Endconv, 10);
	    if (*Endconv!='\0') Err=CMDLONG_LONG;
	    else Opts[Idx].Val.Long=Ltmp;
	break;
	case CMDLONG_DBL:
	    Dtmp=strtod(Valstr, &Endconv);
	    if (*Endconv!='\0') Err=CMDLONG_DBL;
	    else Opts[Idx].Val.Dbl=Dtmp;
	break;
	case CMDLONG_STR:
	    Opts[Idx].Val.Str=Valstr;    /* shallow string copy */
	break;
    }	/* END switch */
    
    switch (Err)
    {
	case CMDLONG_LONG:
	    fprintf(stderr, "\n? %s: option \"%s\": arg \"%s\" is not \'long\'\n",
		    argv[0], Opt, Valstr);
    	break;
	case CMDLONG_DBL:
	    fprintf(stderr, "\n? %s: option \"%s\": arg \"%s\" is not \'double\'\n",
		    argv[0], Opt, Valstr);
    	break;
	default:
		Opts[Idx].Argvidx=Fx;	/* accept good match */
	break;
    }	/* END switch */
    
    return(Retval);
}
/* END of parse_optstr(argc, argv, Fx) */

/*
 * bool_option(Optname): returns 1 if the Boolean option Optname
 * has been seen on the command line, 0 otherwise.
 */
int bool_option(const char *Optname)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx>=0 && Opts[Idx].Type!=CMDLONG_BOOL)
    {
	fprintf(stderr, "\n? bool_option(\"%s\"): type not Boolean\n", Optname);
	Idx=-1; /* not found */
    }
    return((Idx<0)? 0: Opts[Idx].Argvidx);
}
/* END of bool_option(Optname) */

/*
 * TYPE_option(Optname, Defval): returns the value for the option Optname
 * if it was correctly supplied on the command line, or the corresponding
 * default value Defval otherwise. TYPE=(yesno|char|long|dbl|str).
 */
int yesno_option(const char *Optname, int Defval)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx>=0 && Opts[Idx].Type!=CMDLONG_YESNO)
    {
	fprintf(stderr, "\n? yesno_option(\"%s\"): type not \"yes|no\"\n", Optname);
	Idx=-1; /* not found */
    }
    if (Idx<0 || !Opts[Idx].Argvidx) return (Defval);
    else return (Opts[Idx].Val.Bool);
}

char char_option(const char *Optname, char Defval)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx>=0 && Opts[Idx].Type!=CMDLONG_CHAR)
    {
	fprintf(stderr, "\n? char_option(\"%s\"): type not char\n", Optname);
	Idx=-1; /* not found */
    }
    if (Idx<0 || !Opts[Idx].Argvidx) return (Defval);
    else return (Opts[Idx].Val.Char);
}

long long_option(const char *Optname, long Defval)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx>=0 && Opts[Idx].Type!=CMDLONG_LONG)
    {
	fprintf(stderr, "\n? long_option(\"%s\"): type not long\n", Optname);
	Idx=-1; /* not found */
    }
    if (Idx<0 || !Opts[Idx].Argvidx) return (Defval);
    else return (Opts[Idx].Val.Long);
}

double dbl_option(const char *Optname, double Defval)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx>=0 && Opts[Idx].Type!=CMDLONG_DBL)
    {
	fprintf(stderr, "\n? dbl_option(\"%s\"): type not double\n", Optname);
	Idx=-1; /* not found */
    }
    if (Idx<0 || !Opts[Idx].Argvidx) return (Defval);
    else return (Opts[Idx].Val.Dbl);
}

const char *str_option(const char *Optname, const char *Defval)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx>=0 && Opts[Idx].Type!=CMDLONG_STR)
    {
	fprintf(stderr, "\n? str_option(\"%s\"): type not str\n", Optname);
	Idx=-1; /* not found */
    }
    if (Idx<0 || !Opts[Idx].Argvidx) return (Defval);
    else return (Opts[Idx].Val.Str);
}

/* END of TYPE_option(Optname, Defval) */

/*
 * check_option(Optname, Value): returns non-0 if option Optname
 * has been seen on the command line, 0 otherwise.
 * If Value!=NULL, then the value of the option is returned in *Value.
 * It is the caller's responsibility to convert the *Value union
 * to something less dangerous.
 */
int check_option(const char *Optname, Cmdlongvar_ *Value)
{
    int Idx;
    
    Idx=locate_option(Optname);
    if (Idx<0) return(0);   /* nonexistent Optname, silent return */
    if (Opts[Idx].Argvidx && Value!=NULL)
    	*Value=Opts[Idx].Val;
    return(Opts[Idx].Argvidx);
}
/* END of check_option(Optname, Value) */

/*
 * locate_option(Opt): tries to find the option string Opt (w/o leading '-')
 * among the options in Opts[].
 * Return value: the index in Opts if found, -1 if not found.
 * In the latter case, a warning is also printed.
 */
static int locate_option(const char *Opt)
{
    int Idx, Optlen, Ambi=0;
    
    Optlen=strlen(Opt);
    for (Idx=0; Idx<Optsno; Idx++)
    {
	if (Optlen<Opts[Idx].Uniqlen 
		&& !strncmp(Opt, Opts[Idx].Optstr, Optlen))
	    Ambi++;
	if (!strncmp(Opt, Opts[Idx].Optstr, Opts[Idx].Uniqlen))
	{
	    Ambi=0; 	/* found a full match, after all */
	    break;
	}
    }
    if (Ambi)
    {
	fprintf(stderr, "\n? locate_option(\"%s\"): Ambiguous (%d possibilities)\n",
		Opt, Ambi);
	Idx=-1;
    }
    if (Idx>=Optsno)
    {
	fprintf(stderr, "\n? locate_option(\"%s\"): Not found\n", Opt);
	Idx=-1;
    }
    return(Idx);
}
/* END of locate_option(Opt) */

/*
 * help_string(Optname): returns a description of option Optname
 * in a static string. Returns NULL if Optname was not found.
 * Printing does not end w/ a newline so it can be embedded nicely.
 * Note that only a simplified syntax is printed: options printed
 * with a single '-' although "--" is also accepted, and the option
 * value separator is '=' although whitespace is also OK.
 * The returned string is overwritten between calls.
 */
const char *help_string(const char *Optname)
{
    static char Helpstr[128];	/* very generous size */
    
    int Idx, Len, Uniqlen;
    const char *Optstr;
    char *Lastpos=Helpstr;
    
    Idx=locate_option(Optname);
    if (Idx<0) return(NULL);
    
    Optstr=Opts[Idx].Optstr;
    Len=strlen(Optstr);
    Uniqlen=Opts[Idx].Uniqlen;

    memset(Helpstr, 0, 128);
    *Lastpos++ ='-';
    strncpy(Lastpos, Optstr, Uniqlen); 
    Lastpos+=Uniqlen;  
    if (Len>Uniqlen)
    {
	*Lastpos++ = '[';
	strncpy(Lastpos, Optstr+Uniqlen, Len-Uniqlen);
	Lastpos+=Len-Uniqlen;
	*Lastpos++ = ']';
    }
    switch(Opts[Idx].Type)
    {
	case CMDLONG_YESNO:
	    strncpy(Lastpos, "=<\"yes|no\">", 11);
	    Lastpos+=11;
	break;
	case CMDLONG_CHAR:
	    strncpy(Lastpos, "=<char>", 7);
	    Lastpos+=7;
	break;
	case CMDLONG_LONG:
	    strncpy(Lastpos, "=<long>", 7);
	    Lastpos+=7;
	break;
	case CMDLONG_DBL:
	    strncpy(Lastpos, "=<double>", 9);
	    Lastpos+=9;
	break;
	case CMDLONG_STR:
	    strncpy(Lastpos, "=<string>", 9);
	    Lastpos+=9;
	break;
    }
    *Lastpos='\0';
    
    return (const char *) Helpstr;
}
/* END of help_string(Optname) */

/*
 * remove_cmdlongs(): frees up the internal Opts array.
 * Call it once before exiting to avoid memory leaks.
 */
void remove_cmdlongs(void)
{
    register size_t i;
    if (Optsno)
    {
	for (i=0; i<Optsno; ++i)
	    free(Opts[i].Optstr);
	free(Opts); Opts=NULL;
	Optsno=0;
    }
}
/* END of remove_cmdlongs() */

/* ==== END OF FUNCTIONS cmdlong.c ==== */
