/* ====	FUNCTIONS dirfile.c ==== */

/* Directory and filename manipulation routines. */

/* ANSI C + UNIX, 19-Feb-1998, Andris Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>

/* ---- MODULE HEADER ---- */

#include "dirfile.h"

/* ==== FUNCTIONS ==== */

/* fname_split(): splits the filename Fullname into Path, Name and Ext.
 * Allocates and returns a string which is essentially Fullname
 * but with appropriate \0-s inserted. So from
 * Fullname --> "dir1/dir2/dir3/name.ext"
 * you get
 * "dir1/dir2/dir3/\0name\0.ext"
 *  ^                ^     ^
 *  |                |     |
 * *Path,          *Name  *Ext    pointing to the appropriate positions
 * as shown. The returned string should be saved into a char ptr
 * in any case otherwise you might not be able to free it up when
 * *Path is NULL or not set.
 * If any of Path, Name or Ext is NULL, then the corresponding
 * pointer is not set. If all three are NULL, nothing is done and
 * NULL is returned with a warning. If there is no path, then *Path=NULL, 
 * if there is no extension, then *Ext=NULL on return. Filenames beginning
 * with a dot are treated as such: there is no ".ext" without a "name".
 */
char *fname_split(const char *Fullname, char **Path, char **Name, char **Ext)
{
    char *Slashpos=NULL, *Namepos=NULL, *Dotpos=NULL, *Newname=NULL, *Cur=NULL;
    int Newlen, Pathlen, Namelen, Extlen;
    
    /* sanity */
    if (Fullname==NULL || !strlen(Fullname))
    {
	fputs("\n? fname_split(): No valid full name\n", stderr);
	return(NULL);
    }
    if (Path==NULL && Name==NULL && Ext==NULL)
    {
	fputs("\n? fname_split(): All component ptrs are NULL\n", stderr);
	return(NULL);
    }
    
    /* locate the slash */
    Slashpos=strrchr(Fullname, '/');
    if (Slashpos!=NULL)
    {
	Pathlen=Slashpos-Fullname+1; Namepos=Slashpos+1;
    }
    else
    {
	Pathlen=0; Namepos=(char *)Fullname;
    }
    
    /* locate the dot, get the name and ext lens */
    Dotpos=strrchr(Namepos, '.');
    if (Dotpos!=NULL && Dotpos==Namepos)    /* "dir/.name" */
	Dotpos=NULL;	/* this is not a valid extension then */
    if (Dotpos==NULL)
    {
	Extlen=0; Namelen=strlen(Namepos);  /* no extension */
    }
    else
    {
	Extlen=strlen(Dotpos);
	Namelen=Dotpos-Namepos;
    }

    /* calc the size needed, make the output string */
    Newlen=Pathlen? Pathlen+1: 0;
    Newlen+=Namelen;
    if (Extlen) Newlen+=Extlen+1;   
    Newname=(char *) calloc(Newlen+1, sizeof(char));
    
    /* copy the path */
    if (Pathlen)
    {
	strncpy(Newname, Fullname, Pathlen);
	if (Path!=NULL) *Path=Newname;
	Cur=Newname+Pathlen+1;
    }
    else
    {
	Cur=Newname;
	if (Path!=NULL) *Path=NULL;
    }
    
    /* copy the name: it is always there */
    strncpy(Cur, Namepos, Namelen);
    if (Name!=NULL) *Name=Cur;
    Cur+=Namelen+1;
    
    /* copy the extension */
    if (Extlen)
    {
	strncpy(Cur, Dotpos, Extlen);
	if (Ext!=NULL) *Ext=Cur;
    }
    else
    {
	if (Ext!=NULL) *Ext=NULL;
    }
    
    return(Newname);
}
/* END of fname_split() */

/* change_ext: changes the extension of the filename Name to Ext.
 * An extension is the part of the filename after the last dot
 * ('.') after the last slash ('/') in it. Ext should begin with a dot.
 * If Name has no extension, then Ext is simply appended to it.
 * Return value: the modified Name (allocated within).
 */
char *change_ext(const char *Name, const char *Ext)
{
    char *New, *Pos, *Slash;
    int Len;
    
    Pos=strrchr(Name, '.'); /* find last dot */
    Slash=strrchr(Name, '/');	/* find last slash */
    if (Slash==NULL)	/* no directory name */
	Len=(Pos==NULL || Pos==Name)? strlen(Name): Pos-Name;
    else    /* last dot may be in a directory name */
	Len=(Pos==NULL || Slash-Pos>=-1)? strlen(Name): Pos-Name;
    New=(char *) calloc(Len+strlen(Ext)+1, sizeof(char));
    strncpy(New, Name, Len); New[Len]='\0';
    strcat(New, Ext);
    return(New);
}
/* END of change_ext */

/* mkdir_p(): creates the path Path with rwxrwxrwx permissions,
 * spiced by the current umask of the process. Does the same
 * job as the 'mkdir -p' UNIX command.
 * Return value: 0 if OK, or the UNIX error number (errno)
 * if something went wrong.
 */
int mkdir_p(const char *Path)
{
    static const mode_t DIR_MODE=S_IRWXU|S_IRWXG|S_IRWXO;   /* mode 777 */
    char *P=NULL, *Cur=NULL, *C=NULL;
    struct stat Statbuf;
    int Plen;
    
    /* don't do anything if Path is empty */
    if (Path==NULL || !(Plen=strlen(Path))) return(0);
    
    /* make a working copy of Path, append a slash
     * at the end if there is none
     */
    if (Path[Plen-1]!='/') Plen++;
    P=(char *) calloc(Plen+1, sizeof(char));
    strcpy(P, Path);
    P[Plen-1]='/';
    
    for(Cur=(*P=='/')? P+1:P;	/* skip leading slash */
	NULL!=(C=strchr(Cur, '/')); Cur=C+1)
    {
	*C='\0';    /* truncating P up to C */
	errno=0;    /* reset error code */
	
	/* The idea is: if P (up to C) can be stat()-ed,
	 * then obviously the path is OK and we can continue.
	 * Otherwise, if the error was that P did not exist, 
	 * then create it.
	 */
	if (stat(P, &Statbuf)<0)    /* error during access */
	{
	    if (errno==ENOENT)	/* does not exist --> create */
	    {
		errno=0;
		if (mkdir(P, DIR_MODE)<0)   /* create path component now */
		{
		    fprintf(stderr, "\n? mkdir_p(): Problem creating dir \"%s\"\n", P);
		    perror(NULL);	    /* something went wrong */
		    free(P); return(errno);
		}
	    }
	    else    /* real trouble, bail out */
	    {
		fprintf(stderr, "\n? mkdir_p(): Problem creating dir \"%s\"\n", P);
		perror(NULL);
		free(P); return(errno);
	    }
	}
	
	/* replace slash and trundle on */
	*C='/';
    }	    /* for */
    free(P);
    return(0);
}
/* END of mkdir_p() */

/* ==== END OF FUNCTIONS dirfile.c ==== */
