#ifndef PDF2_H
#define PDF2_H

/* ==== HEADER pdf2.h ==== */

/**
 * \file pdf2.h
 * Routine collection for estimating bivariate probability density functions. */

/* 9. Nov. 1993. Andras Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

/* ---- GLOBAL TYPES ---- */

#ifdef __cplusplus
namespace RazorBack {
#endif

/**
 * Pdfrec2_ : this structure holds the variables for estimating a 
 * joint PDF of 2 random variables.
 * X1, 2 are the indep. vars, Y is the PDF, Yint holds the integer counts
 * of the values falling in each bin. The bins are centred around the
 * values in X1[] (Bno1 long) and X2[] (Bno2 long). Y, Yint are Bno1 x Bno2 arrays.
 * Wid1, 2 are the half-bin widths for the 1st and 2nd variables, 
 * N is the total number of values processed.
 */
typedef struct
{
    double *X1, *X2;
    double **Y;
    unsigned long **Yint;
    double Wid1, Wid2;
    unsigned long N, Bno1, Bno2;
}
Pdfrec2_ ;

/* ---- PROTOTYPES ---- */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * init_pdf2(Binno1, Binno2, Binwidth1, Binwidth2, Low1, Up1, Low2, Up2, Pdf):
 * sets up Pdf for accepting raw data. Binno1,2 are the
 * bin numbers, Binwidth1, 2 are the bin widths for variables 1 and 2.
 * The range will be Low[12]..Up[12] (everything outside is ignored.
 * It is assumed that Pdf is either empty or has been cleared.
 * Return value: 0 if OK, -1 if some of the parameters make no sense, 
 * -2 if memory could not be allocated.
 */
int init_pdf2(int Binno1, int Binno2, double Binwidth1, double Binwidth2, 
	    double Low1, double Up1, double Low2, double Up2, 
	    Pdfrec2_ *Pdf);

/**
 * add_pdf2(Value1, Value2, Pdf):
 * enters the value pair Value1,2 into Pdf. All bins in Pdf->Yint[][]
 * which lie closer to Value1, 2 than the half of the bin widths get
 * incremented by 1 (the "sliding bin" idea by Willie) and the 
 * total data count Pdf->N is incremented by 1.
 * Return value: 0 if OK, -1 if Value-s are too small, 1 if they are
 * too large.
 */
int add_pdf2(double Value1, double Value2, Pdfrec2_ *Pdf);

/**
 * eval_pdf2(Pdf):
 * calculates the PDF and puts the estimated values 
 * in Pdf->Y[][] (a normalisation).
 * Return value: 0 if OK, -1 if no data were in Pdf.
 */
int eval_pdf2(Pdfrec2_ *Pdf);

/**
 * remove_pdf2(Pdf): frees up the arrays in Pdfrec_ and resets all
 * variables to 0. Does not remove Pdf itself!
 * This would be a C++ destructor.
 */
void remove_pdf2(Pdfrec2_ *Pdf);

#ifdef __cplusplus
} }
#endif

/* ==== END OF HEADER pdf2.h ==== */

#endif	/* PDF2_H */
