/* ==== FUNCTIONS cdf.c ==== */

/* Routine collection for estimating cumulative distribution functions. */

/* ANSI C, IRIX 4.0.5, 1. Oct. 1993. Andris Aszodi */

/* ---- HEADER ---- */

#include "cdf.h"

/* ---- PROTOTYPES ---- */

static int get_index(double X[], double V, int Len);

/* ==== FUNCTIONS ==== */

/* 
 * init_cdf(Binno, Low, Up, Cdf):
 * initialises Cdf to collect data between Low and Up
 * into Binno bins. X[] is initialised to
 * uniform sampling but the user can overwrite it after the call
 * to init_cdf(), the only rule being that if i<j then X[i]<=X[j].
 * If Low>Up, they are
 * silently swapped. In C++ this would be a constructor.
 * Return value: 0 if OK, -1 if some of the parameters were silly, 
 * -2 if no more memory could be allocated.
 */
int init_cdf(int Binno, double Low, double Up, Cdfrec_ *Cdf)
{
    double Step;
    int d;
    
    /* error checks */
    Cdf->X=Cdf->Y=NULL; Cdf->Yint=NULL;
    Cdf->N=Cdf->Bno=0;
    if (Low>Up)  /* swap Low and Up */
    { Step=Low; Low=Up; Up=Step; }
    if (Binno<=0 || Low==Up) return(-1);
    
    /* create arrays */
    Cdf->X=(double *) calloc(Binno, sizeof(double));
    Cdf->Y=(double *) calloc(Binno, sizeof(double));
    Cdf->Yint=(int *) calloc(Binno, sizeof(int));
    if (Cdf->X==NULL || Cdf->Y==NULL || Cdf->Yint==NULL)
	return(-2);	/* mem allocation error */
    
    /* init variables */
    Step=(Up-Low)/(Binno-1);
    for (d=0; d<Binno; d++)
	{ Cdf->X[d]=Low+d*Step; Cdf->Yint[d]=0; }
    Cdf->Bno=Binno;
    return(0);
}
/* END of init_cdf */

/* 
 * add_cdf(Value, Cdf):
 * enters the value Value into Cdf. The bin which is
 * the first that is higher than Value is incremented by 1.
 * Return value: 0 if OK, 1 if it is too large.
 */
int add_cdf(double Value, Cdfrec_ *Cdf)
{
    register int Dup, Binno=Cdf->Bno;
    
    /* out-of range check */
    if (Value>=Cdf->X[Binno-1]) return(1);
    
    /* get bin position and count */
    Dup=get_index(Cdf->X, Value, Binno);
    (Cdf->Yint[Dup])++; Cdf->N++;
    return(0);
}
/* END of add_cdf */

/*
 * get_index(X, V, Len):
 * find the index kh which satisfies X[kh-1]<V<=X[kh]
 * (X[] is Len entries long, assumed to hold values in
 * ascending order).
 * Return value: kh or 0 if V<X[0] or Len-1 if V>X[Len-1].
 */
static int get_index(double X[], double V, int Len)
{
    register int kl, k, kh;
    
    /* out of range: return first or last index */
    if (V<X[0]) return(0);
    if (V>X[Len-1]) return(Len-1);
    
    /* search */
    kl=0; kh=Len;
    while (kl<kh-1)
    {
	k=(kl+kh)/2;
	if (V==X[k]) { kh=k; break; }
	if (V<X[k]) kh=k; else kl=k;
    }
    return(kh);
}
/* END of get_index */

/*
 * eval_cdf(Cdf):
 * calculates the CDF and puts the estimated values 
 * in Cdf->Y[] (a normalisation). Leaves the original Cdf->Yint[]
 * intact.
 * Return value: 0 if OK, -1 if no data were in Cdf.
 */
int eval_cdf(Cdfrec_ *Cdf)
{
    register int d, Cum;
    register double Total=Cdf->N;   /* convert to real */
    
    if (!(Cdf->N)) return(-1);	/* no data */
    Cum=0;
    for (d=0; d<Cdf->Bno; d++)
    {
	Cum+=Cdf->Yint[d];
	Cdf->Y[d]=Cum/Total;
    }
    return(0);
}
/* END of eval_cdf */

/*
 * remove_cdf(Cdf):
 * frees up the 3 arrays in Cdfrec_ and resets all
 * variables to 0. Does not remove Cdf itself!
 * This would be a C++ destructor.
 */
void remove_cdf(Cdfrec_ *Cdf)
{
    free(Cdf->X); free(Cdf->Y); free(Cdf->Yint);
    Cdf->N=Cdf->Bno=0;
}
/* END of remove_cdf */

/* ==== END OF FUNCTIONS cdf.c ==== */
