/* ==== FUNCTIONS clu.c ==== */

/* Simple hierarchical clustering. */

/* 10-Nov-2000. Andras Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <string.h>

/* ---- MODULE HEADER ---- */

#include "clu.h"

/* ---- PROTOTYPES ---- */

static void init_clu(Clu_ *Ptclu, const Mattri_ *Dist, int Thing);
static int clu_mem(const Clu_ *Clu, int Thing);
static Clu_ *clu_dup(const Clu_ *Clu);
static void clu_remove(Clu_ *Clu);

static int test_merge(Clu_ *Ptclu1, Clu_ *Ptclu2,
    const Mattri_ *Dist, double *Ptlid, double *Ptsed);
static void merge_clus(Clu_ *Ptclu1, Clu_ *Ptclu2, double Lid, double Sed);

/* ==== FUNCTIONS ==== */

/* 
 * make_clus(Dist, Cluno):
 * performs the clustering. The distances between the
 * things to be clustered are in the Dist matrix: the [i][j]:th
 * entry corresponds to the (non-negative) distance between the i and j.
 * Only the lower triangle is used, i>=j. Self-distances are 0.0.
 * Return value: a forest of Clus_ struct trees describing the clusters
 * (allocated within). The no. of cluster trees is returned in *Cluno.
 * Returns NULL on error.
 */
Clu_ *make_clus(const Mattri_ *Dist, int *Cluno)
{
    Clu_ *Clus=NULL;
    int a, b, c, Thingno, Cno, Besta, Bestb;
    double Lid, Sed, Bestlid, Bestsed;
    
    if (Dist==NULL || Cluno==NULL)
    {
	fputs("\n? make_clus(): NULL argument\n", stderr);
	return NULL;
    }
    Thingno=Dist->Size;
    Cno=Thingno;
    
    /* init the clusters: there are Thingno of them, holding 1 thing each */
    Clus=(Clu_ *)calloc(Thingno, sizeof(Clu_));
    for (c=0; c<Cno; c++)
	init_clu(Clus+c, Dist, c);
    
    do	/* try to merge the clusters */
    {
	Bestlid=DBL_MAX; 
	for (a=0; a<Cno; a++)
	    for (b=a+1; b<Cno; b++)
	    {
		if(test_merge(Clus+a, Clus+b, Dist, &Lid, &Sed))
		{
		    if (Lid<Bestlid)	/* pick closest pair */
		    {
			Besta=a; Bestb=b;
			Bestlid=Lid; Bestsed=Sed;
		    }
		}
	    }
	if (Bestlid<DBL_MAX)
	{
	    merge_clus(Clus+Besta, Clus+Bestb, Bestlid, Bestsed);
	    if (Bestb<Cno-1)    /* compact Clu: overlap/copy */
		memmove(Clus+Bestb, Clus+Bestb+1, (Cno-Bestb-1)*sizeof(Clu_));
	    Cno--;
	}
    } while(Bestlid<DBL_MAX);
    
    *Cluno=Cno;
    Clus=(Clu_ *)realloc(Clus, Cno*sizeof(Clu_));
    return(Clus);
}
/* END of make_clus(Dist, Cluno) */

/* init_clu(Ptclu, Dist, Thing):
 * inits the cluster pointed to by Ptclu to hold
 * the Thing-th of Thingno items. No checks. ("constructor")
 */
static void init_clu(Clu_ *Ptclu, const Mattri_ *Dist, int Thing)
{
    double Sed=DBL_MAX;
    int i, Thingno=Dist->Size;
    
    /* allocated so that it could hold all things if necessary */
    Ptclu->Members=(int *) calloc(Thingno, sizeof(int));
    Ptclu->Members[0]=Thing; Ptclu->No=1;
    Ptclu->Lid=0.0; /* this is the self-distance */
    Ptclu->Sub1=Ptclu->Sub2=NULL;   /* no subclusters */
    
    /* get the smallest external distance: scan lower triangle only */
    for (i=0; i<Thing; i++)
	if (Dist->Mat[Thing][i]<Sed) Sed=Dist->Mat[Thing][i];
    for (i=Thing+1; i<Thingno; i++)
	if (Dist->Mat[i][Thing]<Sed) Sed=Dist->Mat[i][Thing];
    Ptclu->Sed=Sed;
}
/* END of init_clu(Ptclu, Dist, Thing) */

/* 
 * test_merge(Ptclu1, Ptclu2, Dist, Ptlid, Ptsed): 
 * check if the clusters pointed to by
 * Ptclu1 and Ptclu2 can be merged, using the distances in Dist. The clusters
 * can be merged if their union still retains the property that
 * the largest distance between its members is smaller than
 * the smallest distance between any of its members and all
 * other things outside. If merging is possible, then
 * a nonzero number is returned and the largest internal distance
 * and the smallest external distance in the union are returned
 * in *Ptlid and *Ptsed, respectively.
 * If they cannot be merged, then nothing is done and 0 is returned.
 */
static int test_merge(Clu_ *Ptclu1, Clu_ *Ptclu2,
    const Mattri_ *Dist, double *Ptlid, double *Ptsed)
{
    double Dij, Lid, Sed=DBL_MAX;
    int i, j, k, m, Thingno=Dist->Size;
    
    /* get smallest external distance for the union */
    for (k=0; k<Ptclu1->No; k++)
    {
	i=Ptclu1->Members[k];	/* k-th member of 1st cluster is i */
	for (j=0; j<i; j++)	/* get i:j distance in i:th row */
	{
	    /* skip if j is in 1 or 2 */
	    if (clu_mem(Ptclu1, j) || clu_mem(Ptclu2, j))
		continue;

	    if (Dist->Mat[i][j]<Sed) Sed=Dist->Mat[i][j];
	}
	for (j=i+1; j<Thingno; j++)	/* get i:j distance in i:th col */
	{
	    /* skip if j is in 1 or 2 */
	    if (clu_mem(Ptclu1, j) || clu_mem(Ptclu2, j))
		continue;

	    if (Dist->Mat[j][i]<Sed) Sed=Dist->Mat[j][i];
	}
    }
    
    /* get largest internal distance for the union */
    Lid=(Ptclu1->Lid>Ptclu2->Lid)? Ptclu1->Lid: Ptclu2->Lid;
    for (k=0; k<Ptclu1->No; k++)
    {
	i=Ptclu1->Members[k];
	for (m=0; m<Ptclu2->No; m++)
	{
	    j=Ptclu2->Members[m];
	    Dij=(i>=j)? Dist->Mat[i][j]: Dist->Mat[j][i];
	    if (Dij>Lid) Lid=Dij;
	}
    }
    
    /* can they be merged? */
    *Ptlid=Lid; *Ptsed=Sed;
    return(Sed<DBL_MAX && Lid<Sed);
}
/* END of test_merge(Ptclu1, Ptclu2, Dist, Ptlid, Ptsed) */

/* merge_clus(Ptclu1, Ptclu2, Lid, Sed): 
 * merges the clusters pointed to by
 * Ptclu1 and Ptclu2, using the pre-calculated Lid and Sed values.
 * *Ptclu1 is updated with *Ptclu2's members, the latter's
 * storage is freed.
 * Also, Ptclu1's Sub1, 2 ptrs will point to the copies of
 * the original Ptclu1, 2 clusters. 
 */
static void merge_clus(Clu_ *Ptclu1, Clu_ *Ptclu2, double Lid, double Sed)
{
    /* merge: here make sure non-terminals have 2 sub-nodes */
    Clu_ *Oldptclu1=clu_dup(Ptclu1), *Oldptclu2=clu_dup(Ptclu2);
    
    memcpy(Ptclu1->Members+Ptclu1->No, Ptclu2->Members, Ptclu2->No*sizeof(int));
    Ptclu1->No+=Ptclu2->No;
    Ptclu1->Lid=Lid; Ptclu1->Sed=Sed;
    Ptclu1->Sub1=Oldptclu1; Ptclu1->Sub2=Oldptclu2;	/* store subclusters */
    free(Ptclu2->Members); 
}
/* END of merge_clus(Ptclu1, Ptclu2, Lid, Sed) */

/* 
 * clu_mem(Clu, Thing): 
 * returns 1 if *Clu contains Thing, 0 otherwise.
 */
static int clu_mem(const Clu_ *Clu, int Thing)
{
    int k;
    for (k=0; k<Clu->No && Clu->Members[k]!=Thing; k++);
    return(Clu->Members[k]==Thing);
}

/* 
 * clu_dup(Clu):
 * duplicates the cluster pointed to by Clu and returns
 * a ptr to the new cluster. The Clu->Members array will be
 * duplicated and copied, the subclusters won't (only
 * the Sub1, 2 pointers).
 */
static Clu_ *clu_dup(const Clu_ *Clu)
{
    Clu_ *Newclu=(Clu_ *) malloc(sizeof(Clu_));
    *Newclu=*Clu;   /* use built-in ANSI memberwise copy */
    
    /* duplicate the member array */
    Newclu->Members=(int *) calloc(Newclu->No, sizeof(int));
    memcpy(Newclu->Members, Clu->Members, Clu->No*sizeof(int));
    return(Newclu);
}
/* END of clu_dup(Clu) */

/* 
 * free_clusters(Clu, Cluno):
 * frees up the cluster "forest" (i.e. the
 * array of cluster trees generated by make_clus(...)).
 */
void free_clusters(Clu_ Clus[], int Cluno)
{
    int i;
    for (i=0; i<Cluno; i++) clu_remove(Clus+i);
    free(Clus);
}
/* END of free_clusters(Clus, Cluno) */

/*
 * clu_remove(Clu):
 * deletes the Members array of the cluster
 * pointed to by Clu. Recursively calls itself for the subclusters
 * as well if they exist and removes them, too.
 */
static void clu_remove(Clu_ *Clu)
{
    free(Clu->Members);	/* removes members */
    if (Clu->Sub1!=NULL)
    {
	clu_remove(Clu->Sub1);
	free(Clu->Sub1);    /* remove subcluster itself! */
    }
    if (Clu->Sub2!=NULL)
    {
	clu_remove(Clu->Sub2);
	free(Clu->Sub2);
    }
}
/* END of clu_remove(Clu) */

/*
 * print_clus(Clu, Out):
 * prints the whole cluster tree with root pointed
 * to by Clu to Out.
 */
void print_clus(const Clu_ *Clu, FILE *Out)
{
    static const int NODEWIDTH=11;   /* "-[+x.xe+00]" */
    static char *Hor=NULL, *Vert=NULL;    /* connecting lines etc. */
    static int Len=0;
    
    if (Clu==NULL) return;  /* bail out */
    
    /* init */
    if (!Len)
    {
	Hor=(char *) calloc(1, sizeof(char));	/* automatic \0-termination */
	Vert=(char *) calloc(1, sizeof(char));
    }
    
    /* leaf */
    if (Clu->Sub1==NULL && Clu->Sub2==NULL)
    {
	fprintf(Out, "%s-%d\n%s\n", Hor, Clu->Members[0], Vert);
	return;
    }
    
    /* non-leaf: always has two sub-nodes */
    if (Clu->Sub1!=NULL)    /* go to "left" */
    {
	Len+=NODEWIDTH;
	Hor=(char *) realloc(Hor, (Len+1)*sizeof(char));
	Vert=(char *) realloc(Vert, (Len+1)*sizeof(char));
	Hor[Len]=Vert[Len]='\0';
	sprintf(Hor+Len-NODEWIDTH, "-[%8.1e]", Clu->Lid);
	strcpy(Vert+Len-NODEWIDTH, "     |     ");
	print_clus(Clu->Sub1, Out);
    
	/* go to "right" */
	strcpy(Hor, Vert);
	strcpy(Hor+Len-NODEWIDTH, "     ^-----");
	strcpy(Vert+Len-NODEWIDTH, "           ");
	print_clus(Clu->Sub2, Out);

	Len-=NODEWIDTH;	    /* shrink back to previous */
	if (Len)
	{
	    Hor=(char *) realloc(Hor, (Len+1)*sizeof(char));
	    Vert=(char *) realloc(Vert, (Len+1)*sizeof(char));
	    Hor[Len]=Vert[Len]='\0';
	}
	else	/* at topmost leaf */
	{
	    free(Hor); free(Vert);
	    Hor=Vert=NULL; Len=0;
	}
    }
    return;
}
/* END of print_clus(Clu, Out) */

/* ==== END OF FUNCTIONS clu.c ==== */
