/* ==== FUNCTIONS golsav.c ==== */

/* Golay-Savitzky data and first derivative smoothing. */

/* ANSI C, 16-Jul-1998, Andras Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdlib.h>
#include <stdio.h>

/* ---- MODULE HEADER ---- */

#include "golsav.h"

/* ---- PROTOTYPES ---- */

static void gol_sav1(const double *Data, int Len, int Formula,
    	    double *Sm, double *Smder);

/* ==== FUNCTIONS ==== */

/*
 * gol_sav(Data, Len, Formula, Cyc, Sm, Smder): smooths the Len-long
 * array Data Cyc times using the Golay-Savitzky formula. Formula
 * is 5,7,9 or 11 (the number of G-S points): Formula<5 is rounded to 5,
 * Formula>11 down to 11, the numbers 6, 8, 10 will be rounded up.
 * If Sm!=NULL, then the smoothed points will be put into it.
 * If Smder!=NULL, then the smoothed derivatives will be put into it.
 * It is an error if both are NULL at the same time.
 * Return value: the number of smoothing cycles performed.
 */
int gol_sav(const double *Data, int Len, int Formula, int Cyc, 
    	    double *Sm, double *Smder)
{
    double *Temp1=NULL, *Temp2=NULL, *Swap;
    int k;
    
    /* sanity checks */
    Len=abs(Len); Formula=abs(Formula); Cyc=abs(Cyc);
    if (Data==NULL || !Len)
    {
	fprintf(stderr, "\n? gol_sav(Data=%p, Len=%d, ...): No data\n",
		Data, Len);
	return(0);
    }
    if (!Cyc)
    {
	fputs("\n? gol_sav(..., Cyc=0, ...): No cycles\n", stderr);
	return(0);
    }
    if (Sm==NULL && Smder==NULL)
    {
	fputs("\n? gol_sav(..., Sm=NULL, Smder=NULL): No place for output\n", stderr);
	return(0);
    }

    /* set the formula */
    if (Formula % 2 == 0) Formula++;	/* must be odd */
    if (Formula<5) Formula=5;
    if (Formula>11) Formula=11;
    
    if (Len<Formula)
    {
	fprintf(stderr, "\n? gol_sav(..., Len=%d, Formula=%d...): Not enough data\n",
		Len, Formula);
	return(0);
    }
    
    /* single smoothing cycle: result goes directly to Sm and/or Smder */
    if (Cyc==1)
    {
	gol_sav1(Data, Len, Formula, Sm, Smder);
	return(1);
    }
    
    /* more cycles */
    Temp1=(double *) calloc(Len, sizeof(double));
    Temp2=(double *) calloc(Len, sizeof(double));
    
    /* smooth the data themselves Cyc-1 times */
    gol_sav1(Data, Len, Formula, Temp1, NULL);	/* first Data->Temp1 */
    
    /* then Temp1->Temp2 */
    for (k=2; k<Cyc; k++)
    {
	gol_sav1(Temp1, Len, Formula, Temp2, NULL);
	Swap=Temp2; Temp2=Temp1; Temp1=Swap;
    }
    
    /* last smoothing goes to output arrays as required */
    gol_sav1(Temp1, Len, Formula, Sm, Smder);
    free(Temp1); free(Temp2);
    return(Cyc);
}
/* END of gol_sav(Data, Len, Formula, Cyc, Sm, Smder) */

/*
 * gol_sav1(Data, Len, Formula, Sm, Smder): performs one cycle
 * of Golay-Savitzky smoothing. The parameters are the same as
 * in gol_sav(...). Formula=(5|7|9|11) is expected.
 * No checks are done.
 */
static void gol_sav1(const double *Data, int Len, int Formula,
    	    double *Sm, double *Smder)
{
    /* smooth coefficients and norm factors (last element) */
    static const double
	    Coeff5[4]={-3.0, 12.0, 17.0, 35.0},
	    Coeff7[5]={-2.0, 3.0, 6.0, 7.0, 21.0},
	    Coeff9[6]={-21.0, 14.0, 39.0, 54.0, 59.0, 231.0},
	    Coeff11[7]={-36.0, 9.0, 44.0, 69.0, 84.0, 89.0, 429.0};
    
    /* end coefficients */
    static const double
	    End5[2][6]={
	    	    {207.0, 12.0, -18.0, 12.0, -3.0, 210.0},
    	    	    {2.0, 27.0, 12.0, -8.0, 2.0, 35.0}
		},
	    End7[3][8]={
		    {39.0, 8.0, -4.0, -4.0, 1.0, 4.0, -2.0, 42.0},
		    {8.0, 19.0, 16.0, 6.0, -4.0, -7.0, 4.0, 42.0},
		    {-4.0, 16.0, 19.0, 12.0, 2.0, -4.0, 1.0, 42.0}
		},
	    End9[4][10]={
		    {85.0, 28.0, -2.0, -12.0, -9.0, 0.0, 8.0, 8.0, -7.0, 99.0},
		    {56.0, 65.0, 56.0, 36.0, 12.0, -9.0, -20.0, -14.0, 16.0, 198.0},
		    {-28.0, 392.0, 515.0, 432.0, 234.0, 12.0, -143.0, -140.0, 112.0, 1386.0},
		    { -56.0, 84.0, 144.0, 145.0, 108.0, 54.0, 4.0, -21.0, 0.0, 462.0}
		},
	    End11[5][12]={
		    {113.0, 48.0, 8.0, -12.0, - 17.0, -12.0, -2.0, 8.0, 13.0, 8.0, -12.0, 143.0},
		    {48.0, 41.0, 32.0, 22.0, 12.0,  3.0, -4.0, -8.0, -8.0, -3.0, 8.0, 143.0},
		    {24.0, 96.0, 123.0, 116.0, 86.0, 44.0, 1.0, -32.0, -44.0, -24.0, +39.0, 429.0},
		    {-72.0, 132.0, 232.0, 251.0, 212.0, 138.0, 52.0, -23.0, -64.0, -48.0, 48.0, 858.0},
		    {- 51.0, 36.0, 86.0, 106.0, 103.0, 84.0, 56.0, 26.0, 1.0, -12.0, -6.0, 429.0}
		};
        
    /* smooth 1st deriv coefficients */
    static const double
	    Dcoeff5[3]={1.0, -8.0, 12.0},
	    Dcoeff7[4]={22.0, -67.0, -58.0, 252.0},
	    Dcoeff9[5]={86.0, -142.0, -193.0, -126.0, 1188.0},
	    Dcoeff11[6]={300.0, -294.0, -532.0, -503.0, -296.0, 5148.0};
    
    /* 1st deriv end coefficients */
    static const double
	    Dend5[2][6]={
	    	    {-125.0, 136.0, 48.0, -88.0, 29.0, 84.0},
		    {-57.0, -3.0, 36.0, 39.0, -15.0, 126.0}
		},
	    Dend7[3][8]={
		    {-257.0, 122.0, 185.0, 72.0, -77.0, -122.0, 77.0, 252.0},
		    {-122.0, 17.0, 62.0, 48.0, 10.0, -17.0, 2.0, 252.0},
		    {-29.0, -46.0, -19.0, 24.0, 55.0, 46.0, -31.0, 252.0}
		},
	    Dend9[4][10]={
		    {-6118.0, 854.0, 3593.0, 3366.0, 1440.0, -918.0, -2441.0, -1862.0, 2086.0, 8316.0},
		    {-3556.0, -49.0, 1538.0, 1737.0, 1080.0, 99.0, -674.0, -707.0, 532.0, 8316.0},
		    {-1582.0, -658.0, 29.0, 486.0, 720.0, 738.0, 547.0, 154.0, -434.0, 8316.0},
		    {-196.0, -973.0, -934.0, -387.0, 360.0, 999.0, 1222.0, 721.0, -812.0, 8316.0}
		},
	    Dend11[5][12]={
		    {-1425.0, -102.0, 589.0, 791.0, 647.0, 300.0, -107.0, -431.0, -529.0, -258.0, 525.0, 2574.0},
		    {-1860.0, -294.0, 572.0, 889.0, 808.0, 480.0, 56.0, -313.0, -476.0, -282.0, 420.0, 5148.0},
		    {-525.0, -174.0, 49.0, 167.0, 203.0, 180.0, 121.0, 49.0, -13.0, -42.0, -15.0, 2574.0},
		    {-420.0, -366.0, -244.0, -83.0, 88.0, 240.0, 344.0, 371.0, 292.0, 78.0, -300.0, 5148.0},
		    {15.0, -174.0, -227.0, -181.0, -73.0, 60.0, 181.0, 253.0, 239.0, 102.0, -195.0, 2574.0}
		};       

    const double *Coeff, *Dcoeff, *End, *Dend;
    int k, i, j, F2=Formula/2; /* note integer div */
    double Smval;
    
    /* choose coefficient table */
    switch(Formula)
    {
	case 5: 
	    Coeff=Coeff5; Dcoeff=Dcoeff5;
	    break;
	case 7:
	    Coeff=Coeff7; Dcoeff=Dcoeff7;
	    break;
	case 9:
	    Coeff=Coeff9; Dcoeff=Dcoeff9;
	    break;
	case 11:
	    Coeff=Coeff11; Dcoeff=Dcoeff11;
	    break;
    }
    
    /* do the smoothing */
    if (Sm!=NULL)
    {
	for (i=0; i<Len-Formula+1; i++)
	{
	    Smval=0.0;
	    for (k=0; k<F2; k++)
		Smval+=Coeff[k]*(Data[i+k]+Data[i+Formula-1-k]);	/* symm */
	    Smval+=Coeff[F2]*Data[i+F2];
	    Smval/=Coeff[F2+1];
	    Sm[i+F2]=Smval;
	}
	
	/* endpoints */
	for (j=0; j<F2; j++) 
	{
	    /* choose end coeff table */
	    switch(Formula)
	    {
		case 5: End=End5[j]; break;
		case 7: End=End7[j]; break;
		case 9: End=End9[j]; break;
		case 11: End=End11[j]; break;
	    }
	    
	    /* beginning of data */
	    Smval=0.0;
	    for (k=0; k<Formula; k++)
		Smval+=End[k]*Data[k];
	    Smval/=End[Formula];
	    Sm[j]=Smval;
	    
	    /* end of data */
	    Smval=0.0;
	    for (k=0; k<Formula; k++)
		Smval+=End[k]*Data[Len-k-1];
	    Smval/=End[Formula];
	    Sm[Len-j-1]=Smval;
	}
    }
    
    /* do the derivative smoothing */
    if (Smder!=NULL)
    {
	for (i=0; i<Len-Formula+1; i++)
	{
	    Smval=0.0;
	    for (k=0; k<F2; k++)
		Smval+=Dcoeff[k]*(Data[i+k]-Data[i+Formula-1-k]);	/* antisymm */
	    /* central value not used */
	    Smval/=Dcoeff[F2];
	    Smder[i+F2]=Smval;
	}
	
	/* endpoints */
	for (j=0; j<F2; j++) 
	{
	    /* choose end coeff table */
	    switch(Formula)
	    {
		case 5: Dend=Dend5[j]; break;
		case 7: Dend=Dend7[j]; break;
		case 9: Dend=Dend9[j]; break;
		case 11: Dend=Dend11[j]; break;
	    }
	    
	    /* beginning of data */
	    Smval=0.0;
	    for (k=0; k<Formula; k++)
		Smval+=Dend[k]*Data[k];
	    Smval/=Dend[Formula];
	    Smder[j]=Smval;
	    
	    /* end of data */
	    Smval=0.0;
	    for (k=0; k<Formula; k++)
		Smval+=-Dend[k]*Data[Len-k-1];    /* note sign */
	    Smval/=Dend[Formula];
	    Smder[Len-j-1]=Smval;
	}
    }
}
/* END of gol_sav1(Data, Len, Formula, Sm, Smder) */
	    
/* ==== END OF FUNCTIONS golsav.c ==== */
