/* ==== kuiperkolms.c ==== */

/*
 * \file
 * Kolmogorov-Smirnov statistics with Kuiper's modification.
 * \author Andras Aszodi
 * \date 2-Jul-2002.
 */

/* ==== STANDARD HEADERS ==== */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* ==== MODULE HEADERS ==== */

#include "kuiperkolms.h"

/* ==== PROTOTYPES ==== */

static double kks_prob(double Vstat, double Neff);
static int dbl_comp(const void* vp1, const void* vp2);

/* ==== FUNCTIONS ==== */

/**
 * Calculates the Kuiper-modified Kolmogorov-Smirnov statistics probability
 * by comparing the SORTED discrete abscissa values in Data (n long)
 * to the known distribution given by the CDF Cdf. The parameters of Cdf
 * are optionally supplied in Cdfpars, use NULL if there are none.
 */
double kolms_vstat1(const double *Data, int n, Cdftype_ Cdf, const void* Cdfpars)
{
    double Dplus=0.0, Dminus=0.0, Dprev=0.0, Dincr=1.0/n,
	Dcur=0.0, Cdfval, Dp, Dm, Vstat;
    int i;
    
    for (i=0; i<n; ++i)
    {
	Cdfval=Cdf(Data[i], Cdfpars);
	Dcur+=Dincr;
	Dp=Dprev-Cdfval; Dm=-Dp;
	if (Dp>Dplus) Dplus=Dp;
	if (Dm>Dminus) Dminus=Dm;
	Dp=Dcur-Cdfval; Dm=-Dp;
	if (Dp>Dplus) Dplus=Dp;
	if (Dm>Dminus) Dminus=Dm;
	Dprev=Dcur;
    }
    Vstat=Dplus+Dminus;
    return kks_prob(Vstat, n);
}
/* END of vstat1(Data,n,Cdf,Cdfpars) */

/**
 * Calculates the Kuiper-modified Kolmogorov-Smirnov statistics probability
 * by comparing the SORTED discrete abscissa values in Data1 (n1 long)
 * to the SORTED discrete abscissa values in Data2 (n2 long).
 */
double kolms_vstat2(const double *Data1, int n1, const double *Data2, int n2)
{
    double Dplus=0.0, Dminus=0.0, D1=0.0, D2=0.0, 
	Dincr1=1.0/n1, Dincr2=1.0/n2, Dp, Dm, Vstat, Neff;
    int i1,i2;
    
    for (i1=i2=0; i1<n1 && i2<n2; /* incr within cycle */ )
    {
	/* stepping */
	if (Data1[i1]<Data2[i2])
	{
	    ++i1; D1+=Dincr1;
	}
	else 
	{
	    ++i2; D2+=Dincr2;
	}
	
	Dp=D1-D2; Dm=-Dp;
	if (Dp>Dplus) Dplus=Dp;
	if (Dm>Dminus) Dminus=Dm;
    }
    Vstat=Dplus+Dminus;
    Neff=n1*n2/(n1+n2);
    return kks_prob(Vstat, Neff);
}
/* END of vstat2(Data1,n1,Data2,n2) */

/* Calculates the approximate Kuiper-KS probability. */
static double kks_prob(double Vstat, double Neff)
{
    static const double TERMEPS=0.001, SUMEPS=1.0e-8;
    
    double Arg, Term, ia, Prevterm=0.0, Sum=0.0;
    int i;
    
    Neff=sqrt(Neff);
    Arg=(Neff+0.155+0.24/Neff)*Vstat;
    if (Arg<0.4) return 1.0;
    
    for (i=1; i<=100; ++i)
    {
	ia=2.0*i*i*Arg*Arg;
	Term=(2.0*ia-1.0)*exp(-ia);
	if (fabs(Term)<TERMEPS*fabs(Prevterm) || fabs(Term)<SUMEPS*Sum)
	    break;  /* converged */
	Sum+=Term;
	Prevterm=Term;
    }
    return 2.0*Sum;
}
/* END of kks_prob(Vstat, Neff) */

/* ---- Auxiliaries ---- */

/** Sorts the n-long double array d into ascending order. */
void ascend_dblsort(double *d, int n)
{
    qsort(d, n, sizeof(double), dbl_comp);
}

/* Ascending-order comparison function for qsort. */
static int dbl_comp(const void* vp1, const void* vp2)
{
    const double *dp1=(const double*)vp1;
    const double *dp2=(const double*)vp2;    
    double d=*dp2-*dp1;
    if (d>0.0) return -1;
    else if (d<0.0) return 1;
    else return 0;
}

/* ==== END OF FUNCTIONS kuiperks.c ==== */
