/* ==== FUNCTIONS paramest.c ==== */

/* This module is based on my parameter estimation library
 * written in Hungary in 1990. The algorithms come originally
 * from Valko+Vajda. 
 */

/* 16-Nov-2000. Andras Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

/* ---- MODULE HEADERS ---- */

#include "paramest.h"

/* ---- PROTOTYPES ---- */

static int posdef_inv(Mattri_ *Mat);

/* ==== FUNCTIONS ==== */

/*
 * lin_reg(X,Y,Wgt,Sres,Par,Sd,Corr,Dstat):
 * Multiple weighed least-square linear regression.
 * Contains an optional ridge parameter Rid for the regularization of
 * the matrix X'WX if the problem is ill-conditioned.
 * The parameters are:-
 * X: ptr to matrix of independent vars (Nm x Nx)
 * Y: ptr to vector of dependent variable (Nm)
 * Wgt: ptr to weight vector (Nm), if NULL then uniform weighting is used
 * Sres: residual deviation
 * Par: vector of estimated parameters (Nx, size set automatically)
 * Sd: standard deviation of params (Nx, size set automatically)
 * Corr: correlation matrix, lower tridiag (Nx*Nx, size set automatically)
 * Dstat: D-statistics (not calculated if Dstat==NULL), good if 1.7<Dstat<2.3
 * Return value: the degrees of freedom or a negative value on errors.
 */
int lin_reg(const Matrect_ *X, const Vect_ *Y, const Vect_ *Wgt, 
	    double *Sres, Vect_ *Par, Vect_ *Sd, Mattri_ *Corr,
	    double *Dstat)
{
    double **Xmeas, **Correl, **Xtx, **Xtxold;
    double *Ymeas, *W, *P, *Sdev, *Dy, *Xty, *Norm;
    Mattri_ Xtxmat, Xtxoldmat;
    unsigned int Nx, Nm, im, ip, jp, Nf;
    double Q, Trace, D, Rid, Dn, Dd;
    double Temp1, Temp2;
    char Inv;
    
    /* checks */
    if (X==NULL || Y==NULL || Sres==NULL
	    || Par==NULL || Sd==NULL|| Corr==NULL)
    {
	fputs("\n? lin_reg(): Only Wgt or Dstat may be NULL\n", stderr);
	return -1;
    }
    Nm=X->Rno; Nx=X->Cno;
    if (Nm<=Nx)
    {
	fprintf(stderr, "\n? lin_reg(): Too few (%u) measurements for %u parameters\n",
		Nm, Nx);
	return -2;
    }
    Nf=Nm-Nx;
    if (Y->Len!=Nm)
    {
	fprintf(stderr, "\n? lin_reg(): Y size (%u) should be %u\n",
		Y->Len, Nm);
	return -3;
    }
    if (Wgt!=NULL && Wgt->Len!=Nm)
    {
	fprintf(stderr, "\n? lin_reg(): Wgt size (%u) should be %u, uniform weights used\n",
		Wgt->Len, Nm);
	Wgt=NULL;
    }
    
    /* set up shorthand pointers */
    Xmeas=X->Mat; Ymeas=Y->Vec;
    W=(Wgt!=NULL)? Wgt->Vec: NULL;
    P=Par->Vec; Sdev=Sd->Vec;
    
    /* adjust output variable sizes */
    resize_vect(Par, Nx); P=Par->Vec;
    resize_vect(Sd, Nx); Sdev=Sd->Vec;
    resize_mattri(Corr, Nx); Correl=Corr->Mat;
    
    /* initialise auxiliaries */
    Dy=(double *) calloc(Nm, sizeof(double));
    Xty=(double *) calloc(Nx, sizeof(double));
    Norm=(double *) calloc(Nx, sizeof(double));
    init_mattri(&Xtxmat, Nx); Xtx=Xtxmat.Mat;
    init_mattri(&Xtxoldmat, Nx); Xtxold=Xtxoldmat.Mat;

    /* calculate Xtx=X'WX and Xty=X'WY from observed data */
    for (im = 0; im < Nm; im++)
    {
	for (ip = 0; ip < Nx; ip++) 
	{
	    for (jp = 0; jp <=ip ; jp++)
	    {
		Temp1=Xmeas[im][ip]*Xmeas[im][jp];
		if (W!=NULL) Temp1*=W[im];
	    	Xtx[ip][jp]+=Temp1;
	    }
	    
	    Temp2=Xmeas[im][ip]*Ymeas[im];
	    if (W!=NULL) Temp2*=W[im];
	    Xty[ip] += Temp2;
	}
    }

    /* normalization of X'WX */
    Trace = 0.0;
    for (ip = 0; ip < Nx; ip++)
    	Trace += Xtx[ip][ip];
    Trace /= (Nx * 1000.0);   /* no norm with smaller pivots than this */
    for (ip = 0; ip < Nx; ip++)
    	Norm[ip] =(Xtx[ip][ip] > Trace)? sqrt(Xtx[ip][ip]): 1.0;

    for (ip = 0; ip < Nx; ip++)
    for (jp = 0; jp <=ip; jp++)	/* preserve normalized Xtx */
    {
	Xtx[ip][jp]/=(Norm[ip] * Norm[jp]);
    	Xtxold[ip][jp]=Xtx[ip][jp];
    }

    /* inversion; if Xtx is singular then Rid is increased up to 1 */
    Rid = 0.0; Inv=0;
    while (!Inv)
    {
	if (!(Inv=posdef_inv(&Xtxmat)))   /* matrix inversion */
	{
	    for (ip = 0; ip < Nx; ip++)	/* get back old Xtx */
	    for (jp = 0; jp <=ip; jp++)	/* with ridge added to diag */
	    	Xtx[ip][jp]=(ip==jp)? Xtxold[ip][jp]+Rid: Xtxold[ip][jp];
	    if (Rid < 1.0)
	    	Rid += 0.01;
	}
    }
    
    /* solution */
    for (ip = 0; ip < Nx; ip++)
    {
	D = 0.0;
	for (jp = 0; jp <= ip; jp++)
	    D += Xtx[ip][jp] / Norm[jp] * Xty[jp];
	for (jp = ip+1; jp < Nx; jp++)
	    D += Xtx[jp][ip] / Norm[jp] * Xty[jp];
	P[ip] = D / Norm[ip];
    }

    /* get the quality function and the D-stat for serial correlation */
    Q = 0.0;
    if (Dstat!=NULL)
    {
    	Dn = 0.0;   /* D-statistic numerator */
    	Dd = 0.0;   /* and denominator */
    }
    for (im = 0; im < Nm; im++)
    {
	D = 0.0;
	for (ip = 0; ip < Nx; ip++)   /* estimated Y */
	    D += Xmeas[im][ip] * P[ip];
	Dy[im] = Ymeas[im] - D;
	Temp1 = Dy[im];
	if (W!=NULL)
	    Q += W[im] * Temp1 * Temp1;
	else
	    Q += Temp1*Temp1;
	
	if (Dstat!=NULL)    /* calc D-statistics */
	{
	    if (im)   /* numerator for Dstat */
	    {
		Temp1 = Dy[im] - Dy[im - 1];
		Dn += Temp1 * Temp1;
	    }
	    Temp1 = Dy[im];
	    Dd += Temp1 * Temp1;   /*denominator for Dstat*/
    	}
    }

    /* set output parameters */
    *Sres = sqrt(Q / Nf);   /* residual deviation */
    if (Dstat!=NULL)
    	*Dstat = Dn / Dd;
  
    /* get the cov matrix */
    for (ip = 0; ip < Nx; ip++)
    {
	for (jp = 0; jp <= ip ; jp++)
	    Correl[ip][jp] = Q * Xtx[ip][jp] / Norm[ip] / Norm[jp] / Nf;
	Sdev[ip] = sqrt(Correl[ip][ip]);
    }

    /* norm to correlation matrix */
    for (ip=0; ip<Nx; ip++)
    {
	Correl[ip][ip]=1.0;
	for (jp=0; jp<ip; jp++)
	    Correl[ip][jp]/=(Sdev[ip]*Sdev[jp]);
    }

    /* cleanup */
    free(Dy); free(Xty); free(Norm);
    free_mattri(&Xtxmat); free_mattri(&Xtxoldmat);
    return Nf;
}
/* END of lin_reg(X,Y,Wgt,Sres,Par,Sd,Corr,Dstat) */

/*
 * lms_fit(x,y,N,b,a,sigb,siga,sigab):
 * no-frills linear regression algorithm to fit y=a+bx
 * without weighing and the other extras.
 * Use this routine for quick-and-dirty fits. So dirty, in fact,
 * that there are no sanity checks on the parameters...
 * The parameters:-
 *     x,y       - independent and dependent variables (both N long)
 *     b         - slope
 *     a         - intercept
 *     siga,sigb - S.D. of intercept & slope, respectively
 *     sigab     - cross-correlation
 * Return value: the error of fit.
 */
double lms_fit(const double *x, const double *y, unsigned int N,
    	    double *b, double *a, double *Sigb, double *Siga, double *Sigab)
{ 
    double sumsig=0.0, delta, sumx=0.0, sumy=0.0, sumx2=0.0, sumxy=0.0, Sig;
    register unsigned int i;
    double TEMP;

    for (i = 0; i < N; i++)
    {
	sumx += x[i];
	sumy += y[i];
	sumx2 += x[i] * x[i];
	sumxy += x[i] * y[i];
    }
    
    delta = sumx2 * N - sumx * sumx;
    *a = (sumx2 * sumy - sumx * sumxy) / delta;
    *b = (N * sumxy - sumx * sumy) / delta;
    for (i = 0; i < N; i++)
    {
	TEMP = y[i] - *b * x[i] - *a;
	sumsig += TEMP * TEMP;
    }
    Sig = sumsig / (N - 2);
    *Sigb = sqrt(Sig * N / delta);
    *Siga = sqrt(Sig * sumx2 / delta);
    *Sigab = Sig * sumx / delta;
    return Sig;
}  
/* END of lms_fit(x,y,N,b,a,sigb,siga,sigab) */

/*
 * nonlin_reg(Funct, Itmax, Steplim, X, Y, Wgt, ItNo, Sres, Par, Sd, Corr):
 * fits the function Y=F(X,P) where X and Y are vectors, P
 * is the parameter vector using the Gauss-Newton-Marquardt method.
 * F should be supplied by the user: see comments at the declaration
 * of the type NonlinregFunct. The algorithm requires an initial guess of
 * the parameter values supplied in Par. Par must also be correctly dimensioned.
 * Due to the parameter normalization used, the
 * estimated parameters keep the sign of the corresponding initial
 * guess. In case of extremely high Q values and bad convergence
 * it is advisable to try sign changes.
 *
 * Note on vector and matrix dimensions:-
 * In the explanation below, Nm is the number of measurements,
 * Nx is the dimension of the independent variables,
 * Ny is the dimension of the dependent variables,
 * Np is the number of parameters.
 * These are supplied to the routine indirectly as the row and column sizes
 * of the X, Y matrices and the dimension of Par.
 *
 * The parameters:-
 * Funct: the nonlinear function (X, P)-> Y
 *  	this operates on traditional C vectors for speed and does no dim checks.
 *  	It is your responsibility to make sure that Funct knows about
 *  	Nx, Np and Ny being the lengths of its args X, P and Y, respectively.
 * ItMax: max. no. of iterations
 * StepLim: smallest stepsize allowed
 * X: matrix of X values (indep. vars, Nm x Nx)
 * Y: matrix of Y values (dep. variable, Nm x Ny)
 * W: weight matrix (Nm x Ny), if NULL then uniform weighting is used
 * ItNo: actual no. of iterations done
 * Sres: residual deviation
 * Par: vector (Np-long) containing the initial parameter guess on input
 *  	and the final parameter estimate on output
 * Sd: vector of estimated parameters and S.D. (Np, size adjusted within)
 * Corr: correlation matrix (Np x Np, size adjusted within)
 * Return value: the degrees of freedom or a negative number on error.
 */
int nonlin_reg(Nonlinregfunct_ Funct, 
	       unsigned int ItMax, double StepLim, 
	       const Matrect_ *X, const Matrect_ *Y, const Matrect_ *Wgt,
	       unsigned int *ItNo, double *Sres, 
	       Vect_ *Par, Vect_ *Sd, Mattri_ *Corr)
{
    double **Xmeas, **Ymeas, **W, **Jac, **Jtj, **Jtjold, **Correl;
    double *Dy, *F, *Fd;	/* Ny long */
    double *P, *Sdev, *Pnew, *Pd, *JtDy, *Db, *Norm; /* Np long */
    Matrect_ Jacobian;
    Mattri_ Jtjmat, Jtjoldmat;
    Vect_ Badp;     /* storage for param combinations that screw Funct */
    register unsigned int Ny, Np, Nm, im, iy, ip, jp, Rep;
    int Nf, Fretval=0;
    double Q, Qold, Lm, StLen, StFac, Dly, Dp, D, Trace;
    char Inv, Grow;
    register double Temp1;
    
    /* checks */
    if (Funct==NULL || X==NULL || Y==NULL
	    || ItNo==NULL || Sres==NULL
	    || Par==NULL || Sd==NULL || Corr==NULL)
    {
	fputs("\n? nonlin_reg(): NULL pointer argument(s)\n", stderr);
	return -1;
    }
    Nm=X->Rno;
    if (Nm!=Y->Rno)
    {
	fprintf(stderr, "\n? nonlin_reg(): X<->Y Measurement no. mismatch: %u!=%u\n",
		Nm, Y->Rno);
	return -2;
    }
    Ny=Y->Cno;
    if (Wgt!=NULL)  /* user-defined weights */
    {
	if (Nm!=Wgt->Rno)
	{
	    fprintf(stderr, "\n? nonlin_reg(): X<->W Measurement no. mismatch: %u!=%u\n",
		    Nm, Wgt->Rno);
	    return -3;
	}
	if (Ny!=Wgt->Cno)
	{
	    fprintf(stderr, "\n? nonlin_reg(): Y<->W dep.var.no. mismatch: %u!=%u\n",
		    Ny, Wgt->Cno);
	    return -4;
	}
    }
    Np=Par->Len;    /* no. of params */
    Nf = Nm * Ny - Np;   /* degrees of freedom */
    if (Nf<=0)
    {
	fprintf(stderr, "\n? nonlin_reg(): %d degrees of freedom (Nm=%u, Ny=%u, Np=%u)\n",
		Nf, Nm, Ny, Np);
	return -5;
    }

    /* create Jacobian and auxiliary matrices */
    init_matrect(&Jacobian, Ny, Np);
    Jac=Jacobian.Mat;	/* shorthand */
    init_mattri(&Jtjmat, Np); init_mattri(&Jtjoldmat, Np);
    Jtj=Jtjmat.Mat; Jtjold=Jtjoldmat.Mat;
    init_vect(&Badp, Np);
    
    /* set shorthand pointers to some params */
    Xmeas=X->Mat; Ymeas=Y->Mat;
    W=(Wgt!=NULL)? Wgt->Mat: NULL;
    P=Par->Vec;
    
    /* adjust output sizes and set shorthand ptrs */
    resize_vect(Sd, Np); Sdev=Sd->Vec;
    resize_mattri(Corr, Np); Correl=Corr->Mat;

    /* old-style init for internal arrays */
    Dy=(double *) calloc(Ny, sizeof(double));
    F=(double *) calloc(Ny, sizeof(double));
    Fd=(double *) calloc(Ny, sizeof(double));
    Pnew=(double *) calloc(Np, sizeof(double));
    Pd=(double *) calloc(Np, sizeof(double));
    JtDy=(double *) calloc(Np, sizeof(double));
    Db=(double *) calloc(Np, sizeof(double));
    Norm=(double *) calloc(Np, sizeof(double));

    *ItNo = 0;
    Qold = 0.0;   /* 0 iterations so far; old quality function is 0 */
    Lm = 0.01;   /* Marquardt's lambda */
    Rep = 0;   /* total no. of backsteps */

    /* calculate MLS quality for initial parameters */
    for (im = 0; im < Nm; im++)  /* do for all observ. points */
    {
	Fretval=Funct(Xmeas[im], P, F);
	if (Fretval)
	{
	    memcpy(Badp.Vec, P, sizeof(double)*Np);
	    goto FUNCT_ERROR;
	}
	for (iy = 0; iy < Ny; iy++)
	{
	    Dly = Ymeas[im][iy] - F[iy];
	    Temp1= Dly*Dly;
	    if (W!=NULL) Temp1*=W[im][iy];
	    Qold += Temp1;
	}
    }
  
    /* main iteration cycle */
    do
    {   
	(*ItNo)++;
	/* zeroing */
	for (ip = 0; ip < Np; ip++)
	{
	    for (jp = 0; jp <= ip; jp++)
		Jtj[ip][jp] = 0.0;
	    JtDy[ip] = 0.0;
	}

	/* calculate Jtj=J'WJ and JtDy=J'W(Y-F(X,P)) from observed data
	in a stepwise manner */
	for (im = 0; im < Nm; im++) /* do for all observ. points */
	{
	    Fretval=Funct(Xmeas[im], P, F);
	    if (Fretval)
	    {
		memcpy(Badp.Vec, P, sizeof(double)*Np);
		goto FUNCT_ERROR;
	    }
	    for (iy = 0; iy < Ny; iy++)   /* diff. between observ. & calc. */
	    	Dy[iy] = Ymeas[im][iy] - F[iy];

	    /* numerical derivation for the Jacobian */
	    for (ip = 0; ip < Np; ip++)
	    {
		Dp = 0.001 * fabs(P[ip]) + 1.0e-6;   /* step for derivation */
		memcpy(Pd, P, sizeof(double)*Np);
		Pd[ip] += Dp;   /* shift one of the P-s */
		Fretval=Funct(Xmeas[im], Pd, Fd);
		if (Fretval)
		{
		    memcpy(Badp.Vec, Pd, sizeof(double)*Np);
		    goto FUNCT_ERROR;
		}

		/* F values at shifted pos */
		for (iy = 0; iy < Ny; iy++)
		Jac[iy][ip] = (Fd[iy] - F[iy]) / Dp * P[ip];   /* normalized Jac */
	    }

	    /* now J'WJ and the "right-hand side" JtDy */
	    for (iy = 0; iy < Ny; iy++)
	    {
		for (ip = 0; ip < Np; ip++)
		{
		    for (jp = 0; jp <= ip; jp++)
		    {
			Temp1=Jac[iy][ip] * Jac[iy][jp];
			if (W!=NULL) Temp1*=W[im][iy];
		    	Jtj[ip][jp] += Temp1;
		    }
		    Temp1=Jac[iy][ip] * Dy[iy];
		    if (W!=NULL) Temp1*=W[im][iy];
		    JtDy[ip] += Temp1;
		}
	    }
	}   	/* for im */

	/* normalization of J'WJ */
	Trace = 0.0;
	for (ip = 0; ip < Np; ip++)
	    Trace += Jtj[ip][ip];
	Trace = Trace / (Np * 1000.0);   /* no norm with smaller pivots than this */
	for (ip = 0; ip < Np; ip++)
	{
	    if (Jtj[ip][ip] > Trace)
	    	Norm[ip] = sqrt(Jtj[ip][ip]);
	    else
	    	Norm[ip] = 1.0;
	}
	for (ip = 0; ip < Np; ip++)
	{
	    for (jp = 0; jp <=ip; jp++)   /* preserve Jtj in Jtjold */
	    {
		Temp1=Jtj[ip][jp] / (Norm[ip] * Norm[jp]);
	    	Jtjold[ip][jp]=Jtj[ip][jp] = Temp1;
	    }
	}

	/* the inner cycle; when Q grows, repeat step with larger Lm */
	do
	{
	    /* inversion; if Jtj is singular then Lm is increased */
	    for (Inv=0; !Inv; )
	    {
		for (ip = 0; ip < Np; ip++)   /* Marquardt */
		    Jtj[ip][ip] = Jtjold[ip][ip] + Lm;
		if (!(Inv=posdef_inv(&Jtjmat)))   /* matrix inversion */
		{
		    Rep++;  /* backstep */
		    copy_mattri(&Jtjoldmat, &Jtjmat);	/* get back Jtj */
		    Lm = (Lm==0.0)? 0.01: 10.0*Lm;  /* increase Lambda */
		}
	    }

	    /* get the step relative to the old param.set, i.e. Db=Dp/P */
	    for (ip = 0; ip < Np; ip++)
	    {
		D = 0.0;
		for (jp = 0; jp <= ip; jp++)
		    D += Jtj[ip][jp] / Norm[jp] * JtDy[jp];
		for (jp = ip+1; jp < Np; jp++)
		    D += Jtj[jp][ip] / Norm[jp] * JtDy[jp];
		Db[ip] = D / Norm[ip];
	    }

	    /* reduce stepsize if it would result in sign change */
	    StLen = 0.0; StFac = 1.0;
	    for (ip = 0; ip < Np; ip++)
	    {
		if (StFac * Db[ip] < -0.95)
		    StFac = -0.95 / Db[ip];
		Temp1 = Db[ip];
		StLen += Temp1 * Temp1;
	    }
	    StLen *= StFac * StFac;

	    /* get the new parameters */
	    for (ip = 0; ip < Np; ip++)
	    	Pnew[ip] = P[ip] * (1.0 + StFac * Db[ip]);

	    /* new quality function */
	    Q = 0.0;
	    for (im = 0; im < Nm; im++)
	    {
		Fretval=Funct(Xmeas[im], Pnew, F);
		if (Fretval)
		{
		    memcpy(Badp.Vec, Pnew, sizeof(double)*Np);
		    goto FUNCT_ERROR;
		}
		for (iy = 0; iy < Ny; iy++)
		{
		    Dly = Ymeas[im][iy] - F[iy];
		    Temp1=Dly*Dly;
		    if (W!=NULL) Temp1*=W[im][iy];
		    Q += Temp1;
		}
	    }
	    
	    /* increase Marquardt's lambda if Q grows, decrease otherwise */
	    if (Grow = (Q >= Qold)) /* grew */
	    {		/* one-step cycle finished */
		Lm=(Lm > 0.0)? 10.0*Lm: 0.01;
		Rep++;
	    } 
	    else /* successful step */
	    {		
		Lm=(Lm > 1e-6)? 0.1*Lm: 0.0;
		Qold = Q;
		memcpy(P, Pnew, sizeof(double)*Np);
	    }
	} 
	while (Grow && StLen >= StepLim / 100); /* inner cycle */
    } 
    while (StLen >= StepLim && *ItNo <= ItMax);
    /* big iteration cycle finished */
  
    *Sres = sqrt(Qold / Nf);   /* residual deviation */

    /* get the cov matrix from the inverse without the Lm */
    copy_mattri(&Jtjoldmat, &Jtjmat);
    Lm = 0.0;
    do 
    {
	Inv=!posdef_inv(&Jtjmat);
	/* check the main diagonal */
	if (!Inv) 
	{
	    for (ip = 0; ip < Np; ip++)
	    	Inv = (Inv || Jtj[ip][ip] < 0.0);
	}
	if (Inv) /* must get it invertible somehow */
	{
	    copy_mattri(&Jtjoldmat, &Jtjmat);	/* get back */
	    Lm=(Lm == 0.0)? 1.0e-8: 2.0*Lm; /* cautious Marquardt */
	    for (ip = 0; ip < Np; ip++)
	    	Jtj[ip][ip] += Lm;
	}
    } while (Inv);

    /* SD and correlation matrix */
    for (ip = 0; ip < Np; ip++)
    {
	for (jp = 0; jp < ip + 1; jp++)
	    Correl[ip][jp] = Qold * Jtj[ip][jp] / (Norm[ip] * Norm[jp] * Nf);
	Sdev[ip] = fabs(P[ip] * sqrt(Correl[ip][ip]));
    }

    for (ip=0; ip<Np; ip++)
    	Norm[ip]=sqrt(Correl[ip][ip]);
    for (ip=0; ip<Np; ip++)
    {
	Correl[ip][ip]=1.0;
	for (jp=0; jp<ip; jp++)
	    Correl[ip][jp]/=(Norm[ip]*Norm[jp]);
    }
    /* end of calculations */
    
    /* User function evaluation errors land us here through goto */
    /* this is the poor man's exception handler :-) */
    FUNCT_ERROR:
    if (Fretval)
    {
	fprintf(stderr,
		"\n? nonlin_reg(): Funct eval returned %d\n",
		Fretval);
	fputs("Parameter vector:\n", stderr);
	write_vect(stderr, &Badp, 6);
    }
    
    /* cleanup */
    free_matrect(&Jacobian);
    free_mattri(&Jtjmat); free_mattri(&Jtjoldmat);
    free_vect(&Badp);
    free(Dy); free(F); free(Fd);
    free(Pnew); free(Pd); free(JtDy); free(Db); free(Norm);
  
    return Fretval? Fretval: Nf;
}
/* END of nonlin_reg(Funct, Itmax, Steplim, X, Y, Wgt, ItNo, Sres, Par, Sd, Corr) */

/*
 * tcrit_95(Nf):
 * returns the value of the Nf degrees-of-freedom t-distribution
 * at 95% significance level. Use for confidence interval calculations.
 */
double tcrit_95(unsigned int Nf)
{
  double t, TEMP;

  if (Nf <= 5)
    t = 6.415 + (5 - Nf) *
	  (0.289 + (4 - Nf) * (0.0575 + (3 - Nf) * (0.022 + (2 - Nf) * 0.020552)));
  else if (Nf <= 30)
    t = 7.6278 - 0.2316 * Nf + 0.00421 * Nf * Nf - 0.186 * sin(0.2214 * Nf) -
	0.0116 * sin(0.4428 * Nf) + 0.0186 * cos(0.4428 * Nf);
  else if (Nf < 40)
    t = 4.4067 + 0.0296 * (30 - Nf);
  else if (Nf < 60)
    t = 4.1108 + 0.021 * (40 - Nf);
  else
    t = 3.6888 + 0.0116 * (60 - Nf);
  t = exp(t);
  if (Nf == 7)
    t += 2;
  if (Nf < 32 && ((1L << Nf) & 0x100a100L) != 0)
    t++;
  modf(t, &TEMP);
  return (TEMP / 1000);   /* round to x.xxx */
}
/* END of tcrit_95(Nf) */

/* 
 * posdef_inv(Mat): 
 * inverts the positive definite symmetric matrix given in *Mat.
 * Only the lower part is used and as the inverse
 * is symmetric, too, it is returned in the same lower triangle format
 * also in A (original values are overwritten). 
 * Return value: 1 if OK, 0 if singular.
 */
static int posdef_inv(Mattri_ *Mat)
{
#ifndef DBL_MIN
    static const double LIM=1.0e-30;
#else
    static const double LIM=100.0*DBL_MIN;
#endif
    
    register unsigned int i, j, k, N=Mat->Size;
    double **A=Mat->Mat;	/* convenience ptr */
    double *H;
    double At;

    H=(double *) calloc(N, sizeof(double));

    /* main elimination cycle */
    for (k = N; k >= 1; k--)
    {  
	if (fabs(A[0][0]) < LIM)
	{ free(H); return 0 ; }    /* singularity */
	H[N - 1] = 1 / A[0][0];
	for (i = 2; i <= N; i++)
	{
	    At = A[i - 1][0] * H[N - 1];
	    H[i - 2] = (i>k)? At: -At;
	    for (j = 2; j <= i; j++)
	    	A[i - 2][j - 2] = A[i - 1][j - 1] + A[i - 1][0] * H[j - 2];
	}
	for (i = 0; i < N; i++)
	    A[N - 1][i] = H[i];
    }
    free(H); return 1;
}
/* END of posdef_inv(Mat) */

/* ==== END OF FUNCTIONS paramest.c ==== */
