/* ====	FUNCTIONS pdf.c ==== */

/* Routine collection for estimating probability density functions. */

/* ANSI C, IRIX 4.0.5, 29. Sept. 1993. Andris Aszodi */

/* ---- HEADER ---- */

#include "pdf.h"

/* ---- LOCAL DEFINITIONS ---- */

/* LOW_IDX and UP_IDX are two symbolic constants which decide
 * whether get_index() should return the lower or the upper
 * index of X[kl]<=V<=X[kh]
 */
#define LOW_IDX 0
#define UP_IDX 1

/* ---- PROTOTYPES ---- */

static int get_index(double X[], double V, int Len, int Idx);

/* ==== FUNCTIONS ==== */

/* init_pdf: initialises Pdf to collect data between Low and Up
 * into Binno bins, each Binwidth wide. X[] is initialised to
 * uniform sampling but the user can overwrite it after the call
 * to init_pdf(), the only rule being that if i<j then X[i]<=X[j].
 * If Low>Up, they are
 * silently swapped. In C++ this would be a constructor.
 * Return value: 0 if OK, -1 if some of the parameters were silly, 
 * -2 if no more memory could be allocated.
 */
int init_pdf(int Binno, double Binwidth, double Low, double Up, 
		Pdfrec_ *Pdf)
{
    double Step;
    int d;
    
    /* error checks */
    Pdf->X=Pdf->Y=NULL; Pdf->Yint=NULL; Pdf->Wid=0.0;
    Pdf->N=Pdf->Bno=0;
    if (Low>Up)  /* swap Low and Up */
    { Step=Low; Low=Up; Up=Step; }
    if (Binno<=0 || Binwidth<=0.0 || Low==Up) return(-1);
    
    /* create arrays */
    Pdf->X=(double *) calloc(Binno, sizeof(double));
    Pdf->Y=(double *) calloc(Binno, sizeof(double));
    Pdf->Yint=(int *) calloc(Binno, sizeof(int));
    if (Pdf->X==NULL || Pdf->Y==NULL || Pdf->Yint==NULL)
	return(-2);	/* mem allocation error */
    
    /* init variables */
    Step=(Up-Low)/(Binno-1);
    for (d=0; d<Binno; d++)
	{ Pdf->X[d]=Low+d*Step; Pdf->Yint[d]=0; }
    Pdf->Wid=Binwidth/2.0; Pdf->Bno=Binno;
    return(0);
}
/* END of init_pdf */

/* add_pdf: enters the value Value into Pdf. All bins in Pdf->Yint[]
 * which lie closer to Value than the half of the bin width get
 * incremented by 1 (the "sliding bin" idea by Willie) and the 
 * total data count Pdf->N is incremented by 1.
 * Return value: 0 if OK, -1 if Value is too small, 1 if it is 
 * too large.
 */
int add_pdf(double Value, Pdfrec_ *Pdf)
{
    register int Dlow, d, Dup, Binno=Pdf->Bno;
    register double Whalf=Pdf->Wid;
    
    /* out-of range checks */
    if (Value+Whalf<Pdf->X[0]) return(-1);
    if (Value-Whalf>=Pdf->X[Binno-1]) return(1);
    
    /* get bin begin and end positions */
    Dlow=get_index(Pdf->X, Value-Whalf, Binno, LOW_IDX);
    Dup=get_index(Pdf->X, Value+Whalf, Binno, UP_IDX);
    
    /* enter counts in the bins concerned */
    for (d=Dlow+1; d<Dup; d++) (Pdf->Yint[d])++;
    Pdf->N++;
    return(0);
}
/* END of add_pdf */

/* get_index: find the indices kl and kh=kl+1
 * which satisfy X[kl]<=V<=X[kh]
 * (X[] is Len entries long, assumed to hold values in
 * ascending order).
 * Return value: kl if Idx==LOW_IDX, or kh if Idx==UP_IDX
 * or 0 if V<X[0] or Len-1 if V>X[Len-1].
 */
static int get_index(double X[], double V, int Len, int Idx)
{
    register int kl, k, kh;
    
    /* out of range: return first or last index */
    if (V<X[0]) return(0);
    if (V>X[Len-1]) return(Len-1);
    
    /* search */
    kl=0; kh=Len;
    while (kl<kh-1)
    {
	k=(kl+kh)/2;
	if (V==X[k]) { kh=kl=k; break; }
	if (V<X[k]) kh=k; else kl=k;
    }
    return((Idx==LOW_IDX)? kl: kh);
}

/* eval_pdf: calculates the PDF and puts the estimated values 
 * in Pdf->Y[] (a normalisation).
 * Return value: 0 if OK, -1 if no data were in Pdf.
 */
int eval_pdf(Pdfrec_ *Pdf)
{
    int d;
    double Step, Sl, Sh, 
	Whalf=Pdf->Wid, Low=Pdf->X[0], Up=Pdf->X[Pdf->Bno-1];
    
    if (!(Pdf->N)) return(-1);	/* no data */
    for (d=0; d<Pdf->Bno; d++)
    {
	Sl=(Pdf->X[d]-Whalf<Low)? Low: Pdf->X[d]-Whalf;
	Sh=(Pdf->X[d]+Whalf>Up)? Up: Pdf->X[d]+Whalf;
	Step=Sh-Sl;
	Pdf->Y[d]=Pdf->Yint[d]/(Step*Pdf->N);
    }
    return(0);
}
/* END of eval_pdf */

/* remove_pdf: frees up the 3 arrays in Pdfrec_ and resets all
 * variables to 0. Does not remove Pdf itself!
 * This would be a C++ destructor.
 */
void remove_pdf(Pdfrec_ *Pdf)
{
    free(Pdf->X); free(Pdf->Y); free(Pdf->Yint);
    Pdf->Wid=0.0; Pdf->N=Pdf->Bno=0;
}
/* END of remove_pdf */

#undef LOW_IDX
#undef UP_IDX

/* ==== END OF FUNCTIONS pdf.c ==== */
