/* ====	FUNCTIONS pdf2.c ==== */

/* Routine collection for estimating probability density functions. */

/* 9. Nov. 1993. Andris Aszodi */

/* ---- HEADER ---- */

#include "pdf2.h"

/* ---- LOCAL DEFINITIONS ---- */

/* LOW_IDX and UP_IDX are two symbolic constants which decide
 * whether get_index() should return the lower or the upper
 * index of X[kl]<=V<=X[kh]
 */
#define LOW_IDX 0
#define UP_IDX 1

/* TYPE_DOUBLE and TYPE_LONG choose matrix types */
#define TYPE_DOUBLE 2
#define TYPE_ULONG 3

/* ---- PROTOTYPES ---- */

static void **alloc_rectmat(unsigned long Rowno, unsigned long Colno, 
			    int Type);
static int get_index(double X[], double V, int Len, int Idx);

/* ==== FUNCTIONS ==== */

/*
 * init_pdf2(Binno1, Binno2, Binwidth1, Binwidth2, Low1, Up1, Low2, Up2, Pdf):
 * sets up Pdf for accepting raw data. Binno1,2 are the
 * bin numbers, Binwidth1, 2 are the bin widths for variables 1 and 2.
 * The range will be Low[12]..Up[12] (everything outside is ignored.
 * It is assumed that Pdf is either empty or has been cleared.
 * Return value: 0 if OK, -1 if some of the parameters make no sense, 
 * -2 if memory could not be allocated.
 */
int init_pdf2(int Binno1, int Binno2, double Binwidth1, double Binwidth2, 
	    double Low1, double Up1, double Low2, double Up2, 
	    Pdfrec2_ *Pdf)
{
    double Step1, Step2;
    int d1, d2;
    
    /* error checks */
    Pdf->X1=Pdf->X2=NULL; 
    Pdf->Y=NULL; Pdf->Yint=NULL; Pdf->Wid1=Pdf->Wid2=0.0;
    Pdf->N=Pdf->Bno1=Pdf->Bno2=0;
    /* swap Low-s and Up-s if necessary */
    if (Low1>Up1)  
    { Step1=Low1; Low1=Up1; Up1=Step1; }
    if (Low2>Up2)  
    { Step2=Low2; Low2=Up2; Up2=Step2; }
    if (Binno1<=0 || Binwidth1<=0.0 || Low1==Up1 || 
	Binno2<=0 || Binwidth2<=0.0 || Low2==Up2) return(-1);
    
    /* create arrays */
    Pdf->X1=(double *) calloc(Binno1, sizeof(double));
    Pdf->X2=(double *) calloc(Binno2, sizeof(double));
    Pdf->Y=(double **) alloc_rectmat(Binno1, Binno2, TYPE_DOUBLE);
    Pdf->Yint=(unsigned long **) alloc_rectmat(Binno1, Binno2, TYPE_ULONG);
    if (Pdf->X1==NULL || Pdf->X2==NULL || Pdf->Y==NULL || Pdf->Yint==NULL)
	return(-2);	/* mem allocation error */
    
    /* init variables */
    Step1=(Up1-Low1)/(Binno1-1);
    Step2=(Up2-Low2)/(Binno2-1);
    for (d1=0; d1<Binno1; d1++)
	Pdf->X1[d1]=Low1+d1*Step1;
    for (d2=0; d2<Binno2; d2++)
	Pdf->X2[d2]=Low2+d2*Step2;
    for (d1=0; d1<Binno1; d1++)
	for (d2=0; d2<Binno2; d2++)
	    Pdf->Yint[d1][d2]=0;
	    
    Pdf->Wid1=Binwidth1/2.0; Pdf->Wid2=Binwidth2/2.0;
    Pdf->Bno1=Binno1; Pdf->Bno2=Binno2;
    return(0);
}
/* END of init_pdf2 */

/* alloc_rectmat: allocates storage for a Rowno x Colno matrix
 * of Type type (cf. the TYPE_* definitions above). The returned void
 * pointer-to-pointer should be typecast properly like the [mc]alloc()
 * returned pointers.
 * Return value: a void ** pointer if OK, or NULL if allocation
 * has failed.
 */
static void **alloc_rectmat(unsigned long Rowno, unsigned long Colno, 
			    int Type)
{
    double **Dpp=NULL;
    unsigned long **Ulpp=NULL;
    register int i;
    
    switch(Type)
    {
	case TYPE_DOUBLE:
	Dpp=(double **) calloc(Rowno, sizeof(double *));
	if (Dpp==NULL) return(NULL);
	for (i=0; i<Rowno; i++)
	{
	    Dpp[i]=(double *) calloc(Colno, sizeof(double));
	    if (Dpp[i]==NULL) return(NULL);
	}
	return((void **) Dpp);
	break;	/* unnecessary */
	
	case TYPE_ULONG:
	default:
	Ulpp=(unsigned long **) calloc(Rowno, sizeof(unsigned long *));
	if (Ulpp==NULL) return(NULL);
	for (i=0; i<Rowno; i++)
	{
	    Ulpp[i]=(unsigned long *) calloc(Colno, sizeof(unsigned long));
	    if (Ulpp[i]==NULL) return(NULL);
	}
	return((void **) Ulpp);
	break;
    }	    /* switch */
}
/* END of alloc_rectmat */

/*
 * add_pdf2(Value1, Value2, Pdf):
 * enters the value pair Value1,2 into Pdf. All bins in Pdf->Yint[][]
 * which lie closer to Value1, 2 than the half of the bin widths get
 * incremented by 1 (the "sliding bin" idea by Willie) and the 
 * total data count Pdf->N is incremented by 1.
 * Return value: 0 if OK, -1 if Value-s are too small, 1 if they are
 * too large.
 */
int add_pdf2(double Value1, double Value2, Pdfrec2_ *Pdf)
{
    register int Dlow1, d1, Dup1, Binno1=Pdf->Bno1,
	    Dlow2, d2, Dup2, Binno2=Pdf->Bno2;
    register double Whalf1=Pdf->Wid1, Whalf2=Pdf->Wid2;
    
    /* out-of range checks */
    if (Value1+Whalf1<Pdf->X1[0] || Value2+Whalf2<Pdf->X2[0])
	return(-1);
    if (Value1-Whalf1>=Pdf->X1[Binno1-1] || 
	Value2-Whalf2>=Pdf->X2[Binno2-1] ) return(1);
    
    /* get bin begin and end positions */
    Dlow1=get_index(Pdf->X1, Value1-Whalf1, Binno1, LOW_IDX);
    Dup1=get_index(Pdf->X1, Value1+Whalf1, Binno1, UP_IDX);
    Dlow2=get_index(Pdf->X2, Value2-Whalf2, Binno2, LOW_IDX);
    Dup2=get_index(Pdf->X2, Value2+Whalf2, Binno2, UP_IDX);
    
    /* enter counts in the bins concerned */
    for (d1=Dlow1+1; d1<Dup1; d1++)
	for (d2=Dlow2+1; d2<Dup2; d2++)
	    (Pdf->Yint[d1][d2])++;
    Pdf->N++;
    return(0);
}
/* END of add_pdf2 */

/*
 * get_index(X, V, Len, Idx):
 * find the indices kl and kh=kl+1
 * which satisfy X[kl]<=V<=X[kh]
 * (X[] is Len entries long, assumed to hold values in
 * ascending order).
 * Return value: kl if Idx==LOW_IDX, or kh if Idx==UP_IDX
 * or 0 if V<X[0] or Len-1 if V>X[Len-1].
 */
static int get_index(double X[], double V, int Len, int Idx)
{
    register int kl, k, kh;
    
    /* out of range: return first or last index */
    if (V<X[0]) return(0);
    if (V>X[Len-1]) return(Len-1);
    
    /* search */
    kl=0; kh=Len;
    while (kl<kh-1)
    {
	k=(kl+kh)/2;
	if (V==X[k]) { kh=kl=k; break; }
	if (V<X[k]) kh=k; else kl=k;
    }
    return((Idx==LOW_IDX)? kl: kh);
}

/*
 * eval_pdf2(Pdf):
 * calculates the PDF and puts the estimated values 
 * in Pdf->Y[][] (a normalisation).
 * Return value: 0 if OK, -1 if no data were in Pdf.
 */
int eval_pdf2(Pdfrec2_ *Pdf)
{
    register int d1, d2;
    double Step1, Sl1, Sh1, Step2, Sl2, Sh2, 
	Whalf1=Pdf->Wid1, Whalf2=Pdf->Wid2,
	Low1=Pdf->X1[0], Up1=Pdf->X1[Pdf->Bno1-1], 
	Low2=Pdf->X2[0], Up2=Pdf->X2[Pdf->Bno2-1];
    
    if (!(Pdf->N)) return(-1);	/* no data */
    for (d1=0; d1<Pdf->Bno1; d1++)
    {
	Sl1=(Pdf->X1[d1]-Whalf1<Low1)? Low1: Pdf->X1[d1]-Whalf1;
	Sh1=(Pdf->X1[d1]+Whalf1>Up1)? Up1: Pdf->X1[d1]+Whalf1;
	Step1=Sh1-Sl1;
	for (d2=0; d2<Pdf->Bno2; d2++)
	{
	    Sl2=(Pdf->X2[d2]-Whalf2<Low2)? Low2: Pdf->X2[d2]-Whalf2;
	    Sh2=(Pdf->X2[d2]+Whalf2>Up2)? Up2: Pdf->X2[d2]+Whalf2;
	    Step2=Sh2-Sl2;
	    Pdf->Y[d1][d2]=Pdf->Yint[d1][d2]/(Step1*Step2*Pdf->N);
	}
    }
    return(0);
}
/* END of eval_pdf2 */

/*
 * remove_pdf2(Pdf): frees up the arrays in Pdfrec_ and resets all
 * variables to 0. Does not remove Pdf itself!
 * This would be a C++ destructor.
 */
void remove_pdf2(Pdfrec2_ *Pdf)
{
    register int i;
    
    free(Pdf->X1); free(Pdf->X2);
    for (i=0; i<Pdf->Bno1; i++)
    { free(Pdf->Y[i]); free(Pdf->Yint[i]); }
    free(Pdf->Y); free(Pdf->Yint);
    Pdf->Wid1=Pdf->Wid2=0.0; Pdf->N=Pdf->Bno1=Pdf->Bno2=0;
}
/* END of remove_pdf2 */

#undef LOW_IDX
#undef UP_IDX
#undef TYPE_DOUBLE
#undef TYPE_ULONG

/* ==== END OF FUNCTIONS pdf2.c ==== */
