/* ==== FUNCTIONS random.c ==== */

/*
 * Uniformly distributed random numbers using the 32-bit generator of
 * L'Ecuyer, P. Efficient and portable combined random number
 * generators, C.A.C.M., vol. 31, 742-749, June 1988.
 * Period is very long, about 2.3e18.
 * Plus the Box/Mueller method for standard normal variates.
 * \author Andras Aszodi
 * \date 20-Nov-2001.
 */

/* ==== STANDARD HEADERS ==== */

#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <unistd.h>
#include <time.h>

/* ==== FUNCTIONS ==== */

/* ---- Seeds ---- */

static int Seed1, Seed2;

/*
 * Initialises the random number generator with two seeds.
 * If S1==0, then the value returned by getpid() is used instead.
 * If S2==0, then the value returned by time(NULL) is used instead.
 */
void uniform_init(int S1, int S2)
{
    Seed1=!S1? getpid(): S1;
    Seed2=!S2? time(NULL): S2;
} 
/* END of init_uniform(S1, S2) */

/* Returns a uniformly distributed integer in the range [0 .. 2147483563]. */
int int_uniform(void)
{
    /* magic constants */
    static const int C1=53668u, C2=40014u, C3=12211u, C4=2147483563u,
	    	D1=52774u, D2=40692u, D3=3791u, D4=2147483399u;
    register int k, z, s1, s2;
    
    s1=Seed1; s2=Seed2; /* calc locally */
    
    k = s1 / C1;
    s1 = C2 * (s1 - k * C1) - k * C3;
    if (s1<0) s1 += C4;

    k = s2 / D1;
    s2 = D2 * (s2 - k * D1) - k * D3;
    if (s2<0) s2 += D4;

    z = s1 - s2;
    if (z<1) z += 2147483562;
    
    Seed1=s1; Seed2=s2;     /* save for next invocation */
    return z;
}
/* END of int_uniform() */

/* Returns a uniformly distributed real value between [0.0 .. 1.0]. */
double dbl_uniform(void)
{
    static const double FACTOR=4.6566130573e-10;    /* 1.0/2147483563.0 */
    return FACTOR*int_uniform();
}
/* END of dbl_uniform() */

/* Returns a random number distributed as Standard Normal. */
double random_gauss(void)
{
    static const double EPSILON
#ifdef DBL_EPSILON
	    =DBL_EPSILON;	/* from <float.h> */
#else
    	    =2e-15;	/* a small number */
#endif

    static short Spare=0;   	/* 1 if saved value is available */
    static double Spareval; 	/* stores saved value */
    
    register double F, R, U1, U2;
    
    if (Spare)	/* we had a value from a previous call: return it */
    {
	Spare=0;
	return Spareval;
    }
    
    /* make two values, save one, return the other */
    do  /* get two random numbers within the unit circle */
    {
	U1=2.0*dbl_uniform()-1.0; U2=2.0*dbl_uniform()-1.0;
	R=U1*U1+U2*U2;
    }
    while (R>=1.0 || R<=EPSILON);
    F=sqrt(-2.0*log(R)/R);
    Spareval=U1*F;	/* spare value to be returned next time */
    Spare=1;
    return (U2*F);
}
/* END of random_gauss() */

