/* ==== FUNCTIONS spl.c ==== */

/* For the fit and evaluation of third-order splines.
 * This routine collection is different from its predecessor
 * "spline.c" in that it is more 'object-oriented'. A spline
 * structure Splrec_ holds the necessary auxiliary arrays
 * and evaluation functions operate on this record.
 */

/* Output from p2c, the Pascal-to-C translator */
/* From input file "spline.p" */

/* ANSI C, 10. Sept. 1997. Andris Aszodi */

/* ---- HEADER ---- */

#include "spl.h"

/* ---- PROTOTYPES ---- */

static double int_0x(const Splrec_ *Spl, double xi);

/* ==== FUNCTIONS ==== */

/*
 * init_spl(n, Spl):
 * initialises a Spl record to hold n>0 data points.
 * After this call, the user can fill up the arrays
 * Spl.X and Spl.Y as he likes.
 * NOTE: If called with a Spl record that has been used before, 
 * call remove_spl() to free up the arrays before calling init_spl(), 
 * otherwise memory leaks may occur.
 * Return value: 0 if OK, -1 if n<=0, -2 if there was not
 * enough memory left.
 */
int init_spl(long n, Splrec_ *Spl)
{
    if (n<=0) return(-1);   /* silly */
    
    Spl->X=(double *) calloc(n, sizeof(double));
    Spl->Y=(double *) calloc(n, sizeof(double));
    Spl->Y2=(double *) calloc(n, sizeof(double));
    Spl->Yin=(double *) calloc(n, sizeof(double));
    if (Spl->X==NULL || Spl->Y==NULL 
	|| Spl->Y2==NULL || Spl->Yin==NULL)
	return(-2);	/* out of memory */
    Spl->N=n;
    return(0);
}
/* END of init_spl */

/*
 * fit_spl(yp1, ypn, Spl):
 * fits a cubic spline to a series of data points.
 * yp1,ypn: user-supplied first derivatives at the 1st and last
 * points. If they are >1e30 then the second derivatives are
 * assumed to be 0 (natural spline)
 * Spl: is the spline record (assumed to be initialised by a call to
 * init_spl() above and the arrays Spl->X[], Spl->Y[] are
 * filled up by the user).
 * Return value: 0 if OK, <0 if something went wrong.
 */
int fit_spl(double yp1, double ypn, Splrec_ *Spl)
{
  register long i, k, n=Spl->N;
  double p, Qn, Sig, Un, dx;
  double *u;

    /* init */
    u=(double *) calloc(n, sizeof(double));
    if (u==NULL) return(-2);	/* out of memory */
    
  if (yp1 > 0.99e30) {
    Spl->Y2[0] = 0.0;
    u[0] = 0.0;
  } else {
    Spl->Y2[0] = -0.5;
    u[0] = 3.0 / (Spl->X[1] - Spl->X[0]) * ((Spl->Y[1] - Spl->Y[0]) / (Spl->X[1] - Spl->X[0]) - yp1);
  }
  for (i = 2; i < n; i++) {
    Sig = (Spl->X[i - 1] - Spl->X[i - 2]) / (Spl->X[i] - Spl->X[i - 2]);
    p = Sig * Spl->Y2[i - 2] + 2.0;
    Spl->Y2[i - 1] = (Sig - 1.0) / p;
    u[i - 1] = (Spl->Y[i] - Spl->Y[i - 1]) / (Spl->X[i] - Spl->X[i - 1]) -
	       (Spl->Y[i - 1] - Spl->Y[i - 2]) / (Spl->X[i - 1] - Spl->X[i - 2]);
    u[i - 1] = (6.0 * u[i - 1] / (Spl->X[i] - Spl->X[i - 2]) - Sig * u[i - 2]) / p;
  }
  if (ypn > 0.99e30) {
    Qn = 0.0;
    Un = 0.0;
  } else {
    Qn = 0.5;
    Un = 3.0 / (Spl->X[n - 1] - Spl->X[n - 2]) *
	 (ypn - (Spl->Y[n - 1] - Spl->Y[n - 2]) / (Spl->X[n - 1] - Spl->X[n - 2]));
  }
  Spl->Y2[n - 1] = (Un - Qn * u[n - 2]) / (Qn * Spl->Y2[n - 2] + 1.0);
  for (k = n - 1; k >= 1; k--)
    Spl->Y2[k - 1] = Spl->Y2[k - 1] * Spl->Y2[k] + u[k - 1];
    
  /* calculating the integral */
  Spl->Yin[0] = 0.0;
  for (k = 2; k <= n; k++) {
    dx = Spl->X[k - 1] - Spl->X[k - 2];
    Spl->Yin[k - 1] = dx * (Spl->Y[k - 1] + Spl->Y[k - 2]) / 2 -
		 dx * dx * dx * (Spl->Y2[k - 1] + Spl->Y2[k - 2]) / 24 + Spl->Yin[k - 2];
  }
  free(u);
  return(0);	/* OK */
}
/* END of fit_spl */

/*
 * eval_spl(Spl, xi, Der1, Der2, Der3, Integ):
 * evaluates the spline *Spl at point xi. If
 * xi is out of range, it is silently replaced by the respective
 * lower or upper limit. The derivatives are returned in
 * *Der1..*Der3, the definite integral between Spl->X[0] and xi
 * is returned in *Integ if the corresponding pointers are not NULL.
 */
double eval_spl(const Splrec_ *Spl, 
		double xi, double *Der1, double *Der2,
		double *Der3, double *Integ)
{
  register long kl, kh, k, n=Spl->N;
  register double a, b, h, yh, yl, y2h, y2l, dx, yi;

  /* locate the position of xi */
  if (xi > Spl->X[n - 1]) xi=Spl->X[n-1];
  if (xi < Spl->X[0]) xi=Spl->X[0];

  /* find the smallest interval containing xi */
  kl = 1;
  kh = n;
  while (kh - kl > 1) {
    k = (kh + kl) / 2;
    if (Spl->X[k - 1] > xi)
      kh = k;
    else
      kl = k;
  }
  h = Spl->X[kh - 1] - Spl->X[kl - 1];   /* length of the interval */
  if (h == 0.0)
  {
      if (Der1!=NULL) (*Der1)=0.0;  /* zero interval length */
      if (Der2!=NULL) (*Der2)=0.0;
      if (Der3!=NULL) (*Der3)=0.0;
      if (Integ!=NULL) (*Integ)=0.0;
	return(0.0);
  }
  a = (Spl->X[kh - 1] - xi) / h;
  b = 1 - a;
  yh = Spl->Y[kh - 1];
  yl = Spl->Y[kl - 1];
  y2h = Spl->Y2[kh - 1];
  y2l = Spl->Y2[kl - 1];
      /* value */
  yi = a * yl + b * yh +
	((a * a * a - a) * y2l + (b * b * b - b) * y2h) * h * h / 6.0;
      /*derivatives */
    if (Der1!=NULL)
	*Der1 = (yh - yl) / h + h * ((1 - 3 * a * a) * y2l + (3 * b * b - 1) * y2h) / 6;
    if (Der2!=NULL)
	*Der2 = a * y2l + b * y2h;
    if (Der3!=NULL)
	*Der3 = (y2h - y2l) / h;
    if (Integ!=NULL)
    {
	dx = xi - Spl->X[kl - 1];
	*Integ = dx * (yi + yl) / 2 - dx * dx * dx * (a * y2l + b * y2h + y2l) / 24 + Spl->Yin[kl - 1];
    }
    return(yi);
}
/* END of eval_spl */

/*
 * integ_spl(Spl, Low, Up):
 * returns the definite integral of the spline Spl
 * between Low and Up or 0.0 if range error.
 */
double integ_spl(const Splrec_ *Spl, double Low, double Up)
{
    register int n=Spl->N;

  if (Low < Spl->X[0] || Low > Spl->X[n - 1] || Up < Spl->X[0] || Up > Spl->X[n - 1]) {
    fprintf(stderr, "Integration limit(s) %e ... %e out of range [%e ... %e]!\n", 
	Low, Up, Spl->X[0], Spl->X[n-1]);
    return(0.0);
  }
  if (Low <= Up)
    return (int_0x(Spl, Up) - int_0x(Spl, Low));
  else
    return (int_0x(Spl, Low) - int_0x(Spl, Up));
}
/* END of integ_spl */

/* int_0x: returns the definite integral of Spl between X[0] and xi. */
static double int_0x(const Splrec_ *Spl, double xi)
{
  long kl, kh, k, n=Spl->N;
  double a, b, h, yi, yh, yl, Der2, y2h, y2l, dx;

  /* find the smallest interval containing xi */
  kl = 1;
  kh = n;
  while (kh - kl > 1) {
    k = (kh + kl) / 2;
    if (Spl->X[k - 1] > xi)
      kh = k;
    else
      kl = k;
  }
  h = Spl->X[kh - 1] - Spl->X[kl - 1];   /* length of the interval */
  if (h == 0.0) {
    return(0.0);
  }
  a = (Spl->X[kh - 1] - xi) / h;
  b = 1 - a;
  yh = Spl->Y[kh - 1];
  yl = Spl->Y[kl - 1];
  y2h = Spl->Y2[kh - 1];
  y2l = Spl->Y2[kl - 1];
  yi = a * yl + b * yh +
       ((a * a * a - a) * y2l + (b * b * b - b) * y2h) * h * h / 6.0;
      /* value */
  Der2 = a * y2l + b * y2h;
  dx = xi - Spl->X[kl - 1];
  return (dx * (yi + yl) / 2 - dx * dx * dx * (Der2 + y2l) / 24 + Spl->Yin[kl - 1]);
}
/* END of int_0x */

/*
 * remove_spl(Spl):
 * frees up the arrays associated with Spl. Does not
 * remove Spl itself!
 */
void remove_spl(Splrec_ *Spl)
{
    free(Spl->X); Spl->X=NULL;
    free(Spl->Y); Spl->Y=NULL;
    free(Spl->Y2); Spl->Y2=NULL;
    free(Spl->Yin); Spl->Yin=NULL;
    Spl->N=0;
}
/* END of remove_spl */

/* ==== END OF FUNCTIONS spline.c ==== */
