/* ==== FUNCTIONS stat2.c ==== */

/* One- and two-variable statistics: mean, SD and correlation
 * for double data.
 */

/* ANSI C, 25. Aug. 1997. Andras Aszodi */

/* ---- HEADER ---- */

#include "stat2.h"

/* ---- PRIVATE CONSTANTS ---- */

static const double EPSILON=1e-10;

/* ==== FUNCTIONS ==== */

/* ---- UNIVARIATE STATISTICS ---- */

/* init_stat: initialises the record St. */
void init_stat(Statrec_ *St)
{
    St->Sx=St->Sx2=St->Min=St->Max=0.0;
    St->N=0;
}
/* END of init_stat */

/* add_stat: enters the data X into St. */
void add_stat(double X, Statrec_ *St)
{
    St->Sx+=X; St->Sx2+=X*X;
    if (!(St->N)) { St->Min=St->Max=X; }    /* first item */
    else
    {
	if (X<St->Min) St->Min=X;
	if (X>St->Max) St->Max=X;
    }
    (St->N)++;
}
/* END of add_stat */

/* eval_stat: returns the average or SD 
 * depending on the Op code.
 */
double eval_stat(const Statrec_ *St, Statops_ Op)
{
    register double Xavg, Sdnum;
    
    if (!(St->N))
    {
	fputs("? eval_stat(): No data points\n", stderr);
	return(0.0);
    }
    switch (Op)
    {
	case STAT_AVG:
	return(St->Sx/St->N);

	case STAT_SD:
	Xavg=St->Sx/St->N;
	Sdnum=St->Sx2-St->N*Xavg*Xavg;
	return(Sdnum<EPSILON? 0.0: sqrt(Sdnum/St->N));

	case STAT_MIN: return(St->Min);
	
	case STAT_MAX: return(St->Max);
	
	case STAT_NOOP:
	default:
	return(0.0);	
    }
}
/* END of eval_stat */

/* ---- BIVARIATE STATISTICS ---- */

/* init_stat2: initialises the record St. */
void init_stat2(Stat2rec_ *St)
{
    St->Sx=St->Sx2=St->Sy=St->Sy2=St->Sxy=0.0;
    St->Minx=St->Maxx=St->Miny=St->Maxy=0.0;
    St->N=0;
}
/* END of init_stat2 */

/* add_stat2: enters the data pair (X,Y) into St. */
void add_stat2(double X, double Y, Stat2rec_ *St)
{
    St->Sx+=X; St->Sy+=Y; St->Sxy+=X*Y;
    St->Sx2+=X*X; St->Sy2+=Y*Y;
    if (!(St->N))
    {
	St->Minx=St->Maxx=X;
	St->Miny=St->Maxy=Y;
    }
    else
    {
	if (X<St->Minx) St->Minx=X;
	if (X>St->Maxx) St->Maxx=X;
	if (Y<St->Miny) St->Miny=Y;
	if (Y>St->Maxy) St->Maxy=Y;
    }
    (St->N)++;
}
/* END of add_stat2 */

/* eval_stat2: returns the average, SD or correlation values
 * depending on the Op code.
 */
double eval_stat2(const Stat2rec_ *St, Stat2ops_ Op)
{
    register double Xavg, Yavg, Sdnum;
    
    if (!(St->N))
    {
	fputs("? eval_stat2(): No data points\n", stderr);
	return(0.0);
    }
    switch (Op)
    {
	case STAT2_AVGX:
	return(St->Sx/St->N);

	case STAT2_AVGY:
	return(St->Sy/St->N);
	
	case STAT2_SDX:
	Xavg=St->Sx/St->N;
	Sdnum=St->Sx2-St->N*Xavg*Xavg;
	return(Sdnum<EPSILON? 0.0: sqrt(Sdnum/St->N));

	case STAT2_SDY:
	Yavg=St->Sy/St->N;
	Sdnum=St->Sy2-St->N*Yavg*Yavg;
	return(Sdnum<EPSILON? 0.0: sqrt(Sdnum/St->N));
	
	case STAT2_MINX: return(St->Minx);
	
	case STAT2_MAXX: return(St->Maxx);
	
	case STAT2_MINY: return(St->Miny);
	
	case STAT2_MAXY: return(St->Maxy);
	
	case STAT2_CORR:
	Xavg=St->Sx/St->N;
	Yavg=St->Sy/St->N;
	return((St->Sxy-St->N*Xavg*Yavg)/St->N/
	    (eval_stat2(St,STAT2_SDX)*eval_stat2(St, STAT2_SDY)));
	
	case STAT2_NOOP:
	default:
	return(0.0);	
    }
}
/* END of eval_stat2 */

/* ==== END OF FUNCTIONS stat2.c ==== */
