/* ==== FUNCTIONS statprob.c ==== */

/*
 * Probability functions for statistics: regularized incomplete gamma and beta.
 * \author Andras Aszodi
 * \date 2-Jul-2002.
 */

/* ==== STANDARD HEADERS ==== */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* ==== MODULE HEADERS ==== */

#include "statprob.h"

/* ==== PROTOTYPES ==== */

static double igamma_series(double a, double x, int *Err);
static double igamma_contfrac(double a, double x, int *Err);
static double ibeta_contfrac(double a, double b, double x, int *Err);

static double init_contfrac(double *Nr, double *Dr, double d0);
static double eval_contfrac(double *Nr, double *Dr, double nk, double dk);

/* ---- CONSTS ---- */

static const unsigned int STATPROB_MAXITER=100;
static const double STATPROB_EPSILON=1.0e-12;
static const double STATPROB_SMALL=1.0e-30;

/* ==== FUNCTIONS ==== */

/* ---- Distributions for simple tests ---- */

/**
 * Returns the tail of the standard normal distribution at z.
 */
double stdnorm_tail(double z)
{
    static const double SQRT2=1.414213562373;
    register double Cdf;
    
    Cdf=0.5*((z<0.0)? erfc(-z/SQRT2): (1.0+erf(z/SQRT2)));
    return (1.0-Cdf);
}
/* END of stdnorm_tail(z) */

/**
 * Returns the significance level for the t test.
 * If the value returned by t_test() is low, then the probability is large
 * that t is high.
 * An error code is returned in *Err if Err!=NULL and something went wrong.
 */
double t_test(unsigned int df, double t, int *Err)
{
    register double x=df/(df+t*t);
    return ibeta_regul(0.5*df, 0.5, x, Err);
}
/* END of t_test(df, t, Err) */

/**
 * Returns the significance level for the test that the
 * F ratio is larger than 1.0.
 * If the value returned by F_test() is low, then the probability is large
 * that F is high.
 * An error code is returned in *Err if Err!=NULL and something went wrong.
 */
double F_test(unsigned int df1, unsigned int df2, double F, int* Err)
{
    register double x=df2/(df2+df1*F);
    return ibeta_regul(0.5*df2, 0.5*df1, x, Err);
}
/* END of F_test(df1, df2, F, Err) */

/**
 * Returns the significance level for the test that the Chi2 is exceeded
 * for a correct model. The less the probability, the better. :-)
 * An error code is returned in *Err if Err!=NULL and something went wrong.
 */
double chi2_test(unsigned int df, double Chi2, int* Err)
{
    return 1.0-igamma_regul(0.5*df, 0.5*Chi2, Err);
}
/* END of chi2_test(df, Chi2, Err) */

/* ---- Special functions ---- */

/* ---- regularized incomplete gamma ---- */

/*
 * Calculates and returns the regularized incomplete gamma function
 * (1/Gamma(a))*INTEGRAL[exp(-t)*t^(a-1) dt, 0, x].
 * If Err!=NULL, then an appropriate error code is returned
 * in *Err if something went wrong (see the Statproberr_ typedef above).
 * igamma_regul(dof/2, chi^2/2) gives the probability that the
 * correct model's chi^2 value (with dof degrees of freedom)
 * is less than chi^2. See also complement_incomplgamma_regul(...) below.
 */
double igamma_regul(double a, double x, int *Err)
{
    double G;

    if (x < 0.0 || a <= 0.0)
    {
	fprintf(stderr, "\n? igamma_regul(%.2e, %.2e): Invalid arguments\n", a, x);
	if (Err!=NULL) *Err=STATPROB_BADARGS;
	return 0.0;
    }
    if (x==0.0) return 0.0;
    
    G=(x < (a+1.0))? igamma_series(a, x, Err): 1.0-igamma_contfrac(a, x, Err);
    return G;
}
/* END of igamma_regul(a, x, Err) */

/*
 * This is 1-igamma_regul(a, x).
 * complement_incomplgamma_regul(dof/2, chi^2/2) gives the probability that the chi2
 * value is exceeded even for a correct model: the less the better :-)
 */
double complement_incomplgamma_regul(double a, double x, int *Err)
{
    double G1;

    if (x < 0.0 || a <= 0.0)
    {
	fprintf(stderr, "\n? complement_incomplgamma_regul(%.2e, %.2e): Invalid arguments\n", a, x);
	if (Err!=NULL) *Err=STATPROB_BADARGS;
	return(0.0);
    }
    if (x==0.0) return 1.0;
    
    G1=(x < (a+1.0))? 1.0-igamma_series(a, x, Err): igamma_contfrac(a, x, Err);
    return G1;
}
/* END of complement_incomplgamma_regul(a, x, Err) */

/*
 * igamma_series(a,x): series approximation
 * of the incomplete gamma function.
 */
static double igamma_series(double a, double x, int *Err)
{
    register unsigned n;
    register double Sum, Term, An, Igamma;

    if (Err!=NULL) *Err=STATPROB_OK;

    Sum=1.0; Term=1.0;
    for (n=1, An=a+1.0; n<STATPROB_MAXITER; ++n, An+=1.0)
    {
	Term*=(x/An);
	if (fabs(Term)<STATPROB_EPSILON*fabs(Sum))
	    break;  	/* converged */
	Sum+=Term;
    }
    Igamma=exp(-x+a*log(x)-lgamma(a+1.0))*Sum;
    
    if (n >= STATPROB_MAXITER)
    {
    	fprintf(stderr, "\n? igamma_series(%.2e, %.2e): No convergence in %d iterations\n",
	    a, x, STATPROB_MAXITER);
	if (Err!=NULL) *Err=STATPROB_NOCONV;
    }
    return Igamma;
}
/* END of igamma_series(a,x) */

/*
 * igamma_contfrac(a,x): continued fraction approximation
 * of the regularized incomplete gamma function.
 */
static double igamma_contfrac(double a, double x, int *Err)
{
    register unsigned int i;
    register double Frac, Oldfrac, Factor, Ni, Di, Igamma;
    double Nr, Dr;

    if (Err!=NULL) *Err=STATPROB_OK;
    /* setting up the continued fraction, d0==0.0 */
    Oldfrac=init_contfrac(&Nr, &Dr, 0.0);
    
    /* do the first step, convergence not tested */
    Oldfrac*=eval_contfrac(&Nr, &Dr, 1.0, x+1.0-a);

    for (i=1; i<STATPROB_MAXITER; ++i)
    {
	/* do the recurrence */
	Ni=i*(a-i);
	Di=x-a+2.0*i+1.0;
	Factor=eval_contfrac(&Nr, &Dr, Ni, Di);
	Frac=Factor*Oldfrac;
	if (fabs(Factor-1.0)<STATPROB_EPSILON)
	    break;  	/* converged */
	Oldfrac=Frac;
    }
    Igamma=exp(-x+a*log(x)-lgamma(a))*Frac;
    
    if (i >= STATPROB_MAXITER)
    {
	fprintf(stderr, "\n? igamma_contfrac(%.2e, %.2e): No convergence in %d iterations\n",
		a, x, STATPROB_MAXITER);
	if (Err!=NULL) *Err=STATPROB_NOCONV;
    }
    return Igamma;
}
/* END of igamma_contfrac(a,x) */

/* ---- Incomplete beta ---- */

/*
 * The regularized incomplete beta function.
 * If x is outside the range [0..1] then -1.0 is returned
 * and *Err is set to STATPROB_BADARGS if Err!=NULL.
 */
double ibeta_regul(double a, double b, double x, int *Err)
{
    double Prefactor, Ibeta;
    
    if (x<0.0 || x>1.0)
    {
	fprintf(stderr, "\n? ibeta_regul(): x=%.2e is out of range [0.0 .. 1.0]\n", x);
	if (Err!=NULL) *Err=STATPROB_BADARGS;
	return -1.0;
    }
    if (x==0.0 || x==1.0) return x;
    
    /* do the prefactor with the builtin ln(gamma) */
    Prefactor=exp(lgamma(a+b)-lgamma(a)-lgamma(b)+a*log(x)+b*log(1.0-x));
    Ibeta=(x<((a+1.0)/(a+b+2.0)))? 
	    Prefactor*ibeta_contfrac(a, b, x, Err)/a:
	    1.0-Prefactor*ibeta_contfrac(b, a, 1.0-x, Err)/b;
    return Ibeta;
}
/* END of ibeta_regul(a,b,x,Err) */

/* ibeta_contfrac: continued fraction approximation of the beta function. */
static double ibeta_contfrac(double a, double b, double x, int *Err)
{
    register unsigned int i;
    register double Ni, A2, Frac, Oldfrac, Factor;
    double Nr, Dr;
    
    if (Err!=NULL) *Err=STATPROB_OK;
    /* setting up the continued fraction, d0==0.0 */
    Oldfrac=init_contfrac(&Nr, &Dr, 0.0);
    
    for (i=0, A2=a; i<STATPROB_MAXITER; ++i, A2+=2.0)
    {
	/* even coeff */
	Ni=!i? 1.0: i*(b-i)*x/((A2-1.0)*A2);
	Factor=eval_contfrac(&Nr, &Dr, Ni, 1.0);
	Frac=Factor*Oldfrac;
	if (fabs(Factor-1.0)<STATPROB_EPSILON)
	    break;  	/* converged */
	Oldfrac=Frac;
	
	/* odd coeff */
	Ni=-(a+i)*(a+b+i)*x/(A2*(A2+1.0));
	Factor=eval_contfrac(&Nr, &Dr, Ni, 1.0);
	Frac=Factor*Oldfrac;
	if (fabs(Factor-1.0)<STATPROB_EPSILON)
	    break;  	/* converged */
	Oldfrac=Frac;
    }
    
    /* no convergence :-( */
    if (i>=STATPROB_MAXITER)
    {
	fprintf(stderr, "\n? ibeta_contfrac(): Cannot converge in %d steps\n", STATPROB_MAXITER);
	if (Err!=NULL) *Err=STATPROB_NOCONV;
    }
    return Frac;
}
/* END of ibeta_contfrac */

/* ---- Continued fractions ---- */

/*
 * Inits the continued fraction evaluation via Lentz's algorithm.
 * Nr and Dr point to two temporary variables that have to be preserved
 * within calls [see eval_contfrac(_)].
 * Returns the first approximation (d0, or STATPROB_SMALL if d0==0.0).
 */
static double init_contfrac(double *Nr, double *Dr, double d0)
{
    if (d0==0.0) d0=STATPROB_SMALL;
    *Nr=d0; *Dr=0.0;
    return d0;
}
/* END of init_contfrac(Nr, Dr, d0) */

/*
 * Given a continued fraction of the form
 * 
 * \verbatim
 *      n1  n2	   nk
 * d0 + --- --- ...---
 *      d1+ d2+    dk
 * \endverbatim
 * 
 * it gets evaluated here to the k:th term, given nk and dk,
 * provided the evaluation up to the k-1:th term has already been done.
 * Uses the modified Lentz algorithm: *Nr and *Dr must be preserved
 * between calls. 
 * Returns the factor with which the old approximation must be multiplied
 * to get the new approximation. If the factor is very close to 1.0 then
 * the iteration has converged. (Not tested here.)
 */
static double eval_contfrac(double *Nr, double *Dr, double nk, double dk)
{
    register double Newn, Newd, Factor;
    
    Newn=dk+nk/(*Nr);
    if (fabs(Newn)<STATPROB_SMALL) Newn=STATPROB_SMALL;
    *Nr=Newn;
    
    Newd=dk+nk*(*Dr);
    if (fabs(Newd)<STATPROB_SMALL) Newd=STATPROB_SMALL;
    *Dr=Newd=1.0/Newd;
    
    Factor=Newn*Newd;
    return Factor;
}
/* END of eval_contfrac(Nr, Dr, nk, dk) */
    
/* ==== END OF statprob.c ==== */
