/* ==== FUNCTIONS student.c ==== */

/*
 * t-test and F-test for comparing two means.
 * \author Andras Aszodi
 * \date 22-Nov-2001.
 */

/* ---- MODULE HEADERS ---- */

#include "student.h"
#include "statprob.h"

/* ---- PROTOTYPES ---- */

static int var_check(const char* Funcnm, double Var1, double Var2, int n1, int n2);

/* ==== FUNCTIONS ==== */

/*
 * t_two(t, Df): returns the two-sided probability
 * that t is equal to 0 with Df degrees of freedom.
 * Returns -1.0 on error.
 */
double t_two(double t, int Df)
{
    static const double b=0.5;
    double a, x;
    
    a=(double)Df/2.0;
    x=Df/(Df+t*t);
    return ibeta_regul(a, b, x, NULL);
}
/* END of t_two(t, Df) */

/* t_samevar: Student's t-test for two distributions that have the same
 * "true" variance. Ave1, 2 are the averages, Var1, 2 are the variances.
 * (These should pass an F-test for the Var1==Var2 hypothesis.)
 * n1, n2 are the number of data points. 
 * Return values: returns the probability (significance) level 
 * and also returns the t-stat value in *T if T!=NULL. Returns -1.0 if
 * something was fishy plus prints a warning on stderr.
 */
double t_samevar(double Ave1, double Ave2, double Var1, double Var2, 
	int n1, int n2, double *T)
{
    int Df;
    double Svar, Tval, Prob;
    
    if (var_check("t-test", Var1, Var2, n1, n2)<0)
	return -1.0;
    
    Df=n1+n2-2;	    /* degrees of freedom */
    Svar=((n1-1)*Var1+(n2-1)*Var2)/Df;	/* pooled variance */
    Tval=fabs(Ave1-Ave2)/sqrt(fabs(Svar*(1.0/n1+1.0/n2)));
    Prob=ibeta_regul(0.5*Df,0.5,(double)Df/(Df+Tval*Tval), NULL);
    if (T!=NULL) *T=Tval;
    return(Prob);
}
/* END of t_samevar */

/* t_diffvar: Student's t-test for two distributions with different
 * variances. Ave1, 2 are the averages, Var1, 2 are the variances.
 * (These should pass an F-test for the Var1!=Var2 hypothesis.)
 * n1, n2 are the number of data points. 
 * Return values: returns the probability (significance) level 
 * and also returns the t-stat value in *T if T!=NULL. Returns -1.0 if
 * something was fishy plus prints a warning on stderr.
 */
double t_diffvar(double Ave1, double Ave2, double Var1, double Var2, 
	int n1, int n2, double *T)
{
    double Tval, Df, Prob;
    
    if (var_check("t_diffvar", Var1, Var2, n1, n2)<0)
	return -1.0;
    
    Var1/=n1; Var2/=n2;
    Tval=fabs(Ave1-Ave2)/sqrt(fabs(Var1+Var2));   /* t-value */
    /* degrees of freedom */
    
    Df=(Var1+Var2)*(Var1+Var2)/
	(Var1*Var1/(n1-1)+Var2*Var2/(n2-1));
    Prob=ibeta_regul(0.5*Df,0.5,Df/(Df+Tval*Tval), NULL);
    if (T!=NULL) *T=Tval;
    return(Prob);
}
/* END of t_diffvar */

/* f_test: Fisher's F-test to decide whether Var1, 2 are different
 * variances. n1, n2 are the no. of data points.
 * Return values: returns the F-statistics probability (significance) 
 * level. Also returns the F-value in *Fval if Fval!=NULL.
 * Returns -1.0 if something silly happens.
 */
double f_test(double Var1, double Var2, int n1, int n2, double *Fval)
{
    int Df1, Df2;
    double F, Prob;
    
    /* check */
    if (var_check("f_test", Var1, Var2, n1, n2)<0)
	return -1.0;
    
    /* swap if Var2<Var1 */
    if (Var1>Var2)
    {
	F=Var1/Var2; Df1=n1-1; Df2=n2-1;
    }
    else
    {
	F=Var2/Var1; Df1=n2-1; Df2=n1-1;
    }
    Prob=2.0*ibeta_regul(0.5*Df2,0.5*Df1,Df2/(Df2+Df1*F), NULL);
    if (Prob>1.0) Prob=2.0-Prob;
    if (Fval!=NULL) *Fval=F;
    return(Prob);
}
/* END of f_test */

/* ---- AUXILIARIES ---- */

/*
 * var_check(Var1,Var2,n1,n2):
 * The variances Var1,2 must be greater than 0.0, -1 or -2 is returned if not.
 * The number of data points n1,2 must be greater than 1, -3 or -4 returned if not.
 * If everything is OK, then 0 is returned.
 */
static int var_check(const char *Funcnm, double Var1, double Var2, int n1, int n2)
{
    if (Var1<=0.0)
    {
	fprintf(stderr, "\n? %s(): Invalid variance Var1: %.2e",
		Funcnm, Var1);
	return -1;
    }
    if (Var2<=0.0)
    {
	fprintf(stderr, "\n? %s(): Invalid variance Var2: %.2e",
		Funcnm, Var2);
	return -2;
    }
    if (n1<=1)
    {
	fprintf(stderr, "\n? %s(): Too few points n1: %d",
		Funcnm, n1);
	return -3;
    }
    if (n2<=1)
    {
	fprintf(stderr, "\n? %s(): Too few points n2: %d",
		Funcnm, n2);
	return -4;
    }
    
    return 0;	/* was OK */
}
/* END of var_check(Var1,Var2,n1,n2) */

/* ==== END OF FUNCTIONS student.c ==== */

