#ifndef ALIGNMENT_HEADER
#define ALIGNMENT_HEADER

// ==== HEADER Alignment.hh ====

/**
 * \file Alignment.hh
 * Class for representing multiple sequence alignments.
 */

// 18-Aug-1999, Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// ANSI C++ standard only.

// ==== STANDARD HEADERS ====

#include <stdlib.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <vector>
#else
#include <iostream.h>
#include <vector.h>
#endif

#include <string>
using namespace std;
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== MODULE HEADERS ====

#include"Monomer.hh"
#include"Gapseq.hh"
#include "Rangedimexc.hh"
#include"Bioseqio.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Alignment_: the multiple alignment class.
 * Implements some simple manipulations and I/O.
 */
class Alignment_: protected Bioseqio_
{
    // data
    private:
	    
    Monomer_ Monomer;	    	///< contains the monomer alphabet
    vector<Gapseq_> Sequences;	///< the aligned sequences
    mutable size_t Maxlen;  	///< the maximal (gapped) alignment length
    mutable bool Dirty;     ///< set to true after non-const access
    
    // methods
    public:
	    
    /// Creates an empty alignment
    Alignment_(): Monomer(Monomer_::UNKNOWN), Sequences(), Maxlen(0) {}
    
    /// Creates an alignment to hold N sequences
    explicit Alignment_(size_t N): Sequences(N), Maxlen(0) {}
    
    /// monomer(): provides const access to the monomer type.
    const Monomer_& monomer() const { return Monomer; }
    
    /// seq_no(): returns the number of sequences in the alignment.
    size_t seq_no() const { return Sequences.size(); }
    
    /// max_len(): returns the maximal gapped length in the alignment.
    size_t max_len() const { set_maxlen(); return Maxlen; }
    
    /// [Idx]: unchecked const access to the Idx:th sequence.
    const Gapseq_& operator[](size_t Idx) const { return Sequences[Idx]; }
    
    /**
    * at(Idx): checked const access to the Idx:th sequence.
    * Throws Indexrangexc_ if Idx is illegal.
    */
    const Gapseq_& at(size_t Idx) const;
    
    /**
    * find(ID): finds the sequence with ID string ID in the alignment.
    * Returns a number in [ 0..seq_no() ) if found, -1 if not found.
    */
    int find(const string& ID) const;
    
    /**
    * vertical_string(Apos):
    * creates and returns a string that contains the Apos:th
    * vertical position. E.g. for this alignment:
    * \verbatim
    * -ALI-GN-
    * --MIGGN-
    * SAVL-GNP
    * \endverbatim
    * the vertical string at position 2 is "LMV".
    * Apos must be in the range [0..max_len()). If the sequences
    * are unaligned, then the right margin of the alignment is "ragged",
    * in this case space characters will be used as padding.
    * Throws Indexrangexc_ if Apos>=max_len().
    */
    string vertical_string(size_t Apos) const;
    
    	/// Sequence manipulations
    
    /**
    * mmer_compat(Mseq, [Mali]):
    * returns true if the monomer type in Mseq for a sequence
    * is compatible with the monomer type in Mali for an alignment
    * (if omitted, the calling object's current monomer() object
    * is used). DNA, RNA and NUCL monomer types are compatible
    * with each other (basically we allow any nucleotide sequence
    * to be aligned to any other nucleotide sequence).
    * All other monomer types are compatible only with themselves,
    * UNKNOWN is incompatible with all definite types.
    */
    bool mmer_compat(const Monomer_& Mseq, const Monomer_& Mali) const;
    bool mmer_compat(const Monomer_& Mseq) const
    { return mmer_compat(Mseq, monomer()); }
    
    /**
    * add_seq(Seq, [Idx]): adds the sequence Seq before position Idx.
    * If the position is omitted then Seq will be added to the end.
    * If Idx is illegal, Indexrangexc_ will be thrown.
    * However, if the calling object was empty, then Idx is ignored
    * and the sequence will be added anyway.
    * The type of this first sequence will set the type of the 
    * whole alignment.
    * Seq must have compatible monomers as the alignment itself
    * as defined by the mmer_compat() method
    * otherwise it will not be added and a warning is issued.
    */
    void add_seq(const Gapseq_& Seq);
    void add_seq(const Gapseq_& Seq, size_t Idx);
    
    /**
    * remove_seq([Idx]): removes the Idx:th sequence from the alignment.
    * If Idx is omitted then all sequences are deleted (complete reset).
    * Nothing is done if the calling object was already empty.
    * Throws Indexrangexc_ otherwise if Idx is illegal.
    */
    void remove_seq(size_t Idx);
    void remove_seq();
    
    	/// Gap manipulations
    
    /**
    * insert_gaps(Seqidx, Netpos, Gaplen):
    * invokes Gapseq_::insert_gaps(Netpos, Gaplen) for the Seqidx:th
    * sequence in the alignment.
    * Throws Indexrangexc_ if Seqidx is illegal.
    */
    void insert_gaps(size_t Seqidx, size_t Netpos, size_t Gaplen);
    
    /**
    * remove_gaps([Seqidx],[Netpos]):
    * When both parameters are specified, then Gapseq_::remove_gaps(Netpos)
    * is invoked for the Seqidx:th sequence in the alignment.
    * If Netpos is omitted, then all gaps are removed from the
    * Seqidx:th sequence.
    * Throws Indexrangexc_ if Seqidx is illegal.
    * Without any parameters, all gaps are deleted from all sequences.
    */
    void remove_gaps(size_t Seqidx, size_t Netpos);
    void remove_gaps(size_t Seqidx);
    void remove_gaps();
    
    	// I/O
    
    /// available I/O formats
    enum Formatflags_ {UNKNOWN=0, FASTA=1, GCG=2, PIR=4, ANYFORMAT=7};

    /**
    * read_align(Inf, Type, Format):
    * tries to read a multiple alignment from the input stream Inf.
    * If Type is set to Monomer_::PROT, then only protein
    * sequences will be read, if it is set to Monomer_::[DEOXY[RIBO]]NUCL,
    * then only the corresponding nucleic acid sequences will be read,
    * if Type is Monomer_::USER (default) or an OR-ed combination of PROT
    * and *NUCL, then all types are considered and a polymer
    * type autodetection is attempted. This is not foolproof.
    * Format can be set to ANYFORMAT (default), in this case automatic
    * file format detection is performed, or to PIR (proteins only),
    * FASTA, GCG (protein or nucleic acid).
    * No read is performed and FORMATERR is returned when
    * incompatible type and format are specified.
    * Return value: OK or an error enum if something went wrong.
    */
    Bioseqio_::Errtype_ read_align(istream& Inf, int Type=Monomer_::USER, int Format=ANYFORMAT);

    /**
    * write_align(Outf, Format): writes the contents of the calling object
    * to Outf according to the format specification in Format.
    * Returns OK or an error enum if something went wrong.
    */
    Bioseqio_::Errtype_ write_align(ostream& Outf, int Format) const;
    
    /// hidden methods
    protected:
	    
    void set_monomer_type(const Monomer_& Mseq, Monomer_& Mali);
    void set_monomer_type(const Monomer_& Mseq)
    { set_monomer_type(Mseq, Monomer); }
    
    size_t set_maxlen() const;

    Bioseqio_::Errtype_ read_concat(istream& Inf, int Type, int Format);
    Bioseqio_::Errtype_ read_msf(istream& Inf);
    Bioseqio_::Errtype_ write_concat(ostream& Outf, int Format) const;
    Bioseqio_::Errtype_ write_msf(ostream& Outf) const;
};
// END OF CLASS Alignment_

}   // RazorBack

// ==== END OF HEADER Alignment.hh ====

#endif	    // ALIGNMENT_HEADER
