#ifndef BIOSEQ_HEADER
#define BIOSEQ_HEADER

// ==== HEADER Bioseq.hh ====

/**
 * \file Bioseq.hh
 * Base class for storing biological sequences.
 */

// 9-Aug-1999. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// ANSI C++ standard only.

// ==== STANDARD HEADERS ====

#include <stdlib.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#else
#include <iostream.h>
#endif

#include <string>
using namespace std;
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== MODULE HEADERS ====

#include"Monomer.hh"
#include "Rangedimexc.hh"
#include"Bioseqio.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Bioseq_: the base class for biological sequences.
 * Stores the monomer alphabet, the sequence,
 * its ID, title and some annotation.
 * Uses the I/O methods inherited from Bioseqio_.
 */
class Bioseq_: protected Bioseqio_
{
    // data
    private:
    
    Monomer_ Monomer;	///< the monomer alphabet
    bool Nocase;    ///< allows case-independent monomer match when true
    string Seq;     ///< the sequence (1-char abbreviations)
    string ID;	    ///< the short identification tag of the sequence
    string Title;    ///< the "title" of the sequence (max. one line)
    string Annot;   ///< annotation (free format)
    
    // methods
    public:
    
    /**
    * Create a sequence object with the prescribed monomer type
    * (amino acid codes by default). Consult "Monomer.hh"
    * for available monomer types.
    */
    explicit Bioseq_(Monomer_::Type_ Mt=Monomer_::PROT):
	    Monomer(Mt), Seq(""), ID(""), Title(""), Annot(""), 
	    Nocase(false), Bioseqio_() {}

    /**
    * Create a sequence object with the prescribed monomer alphabet
    * given in Codes. Consult "Monomer.hh" for code string syntax.
    */
    explicit Bioseq_(const char* Codes):
	    Monomer(Codes), Seq(""), ID(""), Title(""), Annot(""), 
	    Nocase(false), Bioseqio_() {}

    /// copy ctor
    Bioseq_(const Bioseq_& B):
	    Monomer(B.Monomer), Seq(B.Seq), ID(B.ID), Title(B.Title), Annot(B.Annot),
	    Nocase(B.Nocase), Bioseqio_(B) {}
    
    /**
    * id(), id(I), title(), title(T), annot(), annot(A), nocase(), nocase(N):
    * methods for getting and setting the corresponding members.
    */
    const string& id() const { return ID; }
    string& id() { return ID; }
    void id(const string& I) { ID=I; }
    const string& title() const { return Title; }
    string& title() { return Title; }
    void title(const string& T) { Title=T; }
    const string& annot() const { return Annot; }
    string& annot() { return Annot; }
    void annot(const string& A) { Annot=A; }
    bool nocase() const { return Nocase; }
    bool nocase(bool N) { bool On=Nocase; Nocase=N; return On; }
    
    /**
    * monomer(): const access to the monomer alphabet.
    * The monomer type of a sequence object can be changed only
    * through assignment or input.
    */
    const Monomer_& monomer() const { return Monomer; }
    
    /**
    * seq(): returns the sequence contained in the object.
    * See sequence() for hidden non-const access.
    */
    const string& seq() const { return Seq; }
    
    /**
    * seq(S): stores the sequence S in the calling object.
    * Refuses storage if S does not correspond to calling obj type.
    */
    virtual
    void seq(const string& S)
    {
	if (!is_legal(S))
	    cerr<<"\n? Bioseq_::seq(...): Illegal char(s) in sequence\n";
	else Seq=S;
    }
    
    /// Returns the length of the sequence in the calling object.
    size_t length() const { return seq().length(); }
    
    /// []: unsafe index operator for the Idx:th char in the sequence.
    char operator[](size_t Idx) const { return Seq[Idx]; }
    char& operator[](size_t Idx) { return Seq[Idx]; }
    
    /**
    * at_1(Idx), at_1(Idx, C):
    * safe index operator for the Idx:th char in the sequence.
    * The non-const version refuses to store characters not corresponding
    * to the current type of the calling object.
    * Throws Indexrangexc_ if Idx is illegal.
    * Returns true on success, false when C was illegal.
    */
    virtual
    char at_1(size_t Idx) const;
    virtual
    bool at_1(size_t Idx, char C);
   
    /**
    * at_3(Idx), at_3(Idx, Ccc):
    * safe index operator for getting and setting
    * a monomer using the 3-letter abbreviations.
    * The non-const version refuses to store abbreviations
    * that are illegal.
    * Throws Indexrangexc_ if Idx is illegal.
    * Returns true on success, false when Ccc was illegal.
    */
    const char* at_3(size_t Idx) const;
    bool at_3(size_t Idx, const char* Ccc);
   
    /**
    * uppercase(), lowercase():
    * convert the sequence string to all-uppercase or all-lowercase.
    * Return the number of characters converted.
    */
    int uppercase();
    int lowercase();
    
    	/// Legal issues
    
    /**
    * is_legal(C): returns true if C is a legal character
    * in the calling object under the current monomer alphabet
    * and case-sensitivity setting.
    * is_legal(Ccc): does the same for 3-letter abbreviations.
    * is_legal(S): does the same for the complete sequence string S.
    */
    virtual
    bool is_legal(char C) const { return (Monomer.is_legal(C, Nocase)); }
    virtual
    bool is_legal(const char* Ccc) const { return (Monomer.is_legal(Ccc, Nocase)); }
    bool is_legal(const string& S) const;
    
    	// Input-output
    
    /**
    * read_seq(Inf, Type, Format):
    * tries to read a sequence from the input stream Inf.
    * If Type is set to Monomer_::PROT, then only protein
    * sequences will be read, if it is set to Monomer_::[DEOXY[RIBO]]NUCL,
    * then only the corresponding nucleic acide sequences will be read,
    * if Type is Monomer_::USER (default) or an OR-ed combination of PROT
    * and *NUCL, then all types are considered and a polymer
    * type autodetection is attempted. This is not foolproof.
    * Format can be set to ANYFORMAT (default), in this case automatic
    * file format detection is performed, or to a concrete
    * file format. Type and Format may be incompatible:-
    * PIR, SWISS are for proteins only, EMBL is for nucleic acids only.
    * No read is performed and FORMATERR is returned when
    * incompatible type and format are specified.
    * Return value: OK or an error enum if something went wrong.
    */
    virtual
    Errtype_ read_seq(istream& Inf, int Type=Monomer_::USER, int Format=ANYFORMAT);

    /**
    * write_seq(Outf, Format): writes the contents of the calling object
    * to Outf according to the format specification in Format.
    * Returns OK or an error enum if something went wrong.
    */
    virtual
    Errtype_ write_seq(ostream& Outf, int Format) const;
    
    // hidden methods
    protected:
	    
    Monomer_& chg_monomer() { return Monomer; } 	// non-const access protected
    virtual
    bool is_legal(int Mt, char C) const;
    string& sequence() { return Seq; }
    
};
// END OF CLASS Bioseq_

}   // RazorBack

// ==== END OF HEADER Bioseq.hh ====

#endif      // BIOSEQ_HEADER
