#ifndef MONOMER_HEADER
#define MONOMER_HEADER

// ==== HEADER Monomer.hh ====

/**
 * \file Monomer.hh
 * Class for representing monomer alphabets for biomolecules.
 */

// 6-Aug-1999. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <string.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <vector>
using namespace std;
#else
#include <vector.h>
#endif

// ==== CLASSES ====

namespace RazorBack {

/**
 * Monomer_ : stores the legal 1- and 3-letter abbreviations
 * for biomolecule monomers. Amino acids and nucleotides are
 * conveniently predefined.
 */
class Monomer_
{
    // subclass
    // even more hidden methods
    public:
	    
    /**
    * Subclass Abbrev_: stores a 1-letter and a 3-letter abbreviation
    * for a monomer together.
    */
    class Abbrev_
    {
	friend class Monomer_;
	
	private:
		
	char Code1; ///< one-letter abbreviation
	char Code3[4];	///< three-letter abbreviation
	
	// methods
	public:
		
    	/**
	* Init to store 'C' and "Ccc" as abbreviations.
	* Defaults are 'X' and "Xxx".
    	*/
	explicit Abbrev_(char C='X', const char* Ccc="Xxx")
	{
	    set_codes(C, Ccc);
	}
	
	/// code_1(), code_3(): get the abbreviations
	char code_1() const { return Code1; }
	const char* code_3() const { return Code3; }
	
	// hidden methods
	protected:
		
    	/**
	* set_codes(C, Ccc): the abbreviations may only be set together.
	* Non-const access only for Monomer_
    	*/
	void set_codes(char C, const char* Ccc)
	{
	    Code1=C;
	    memset(Code3, 0, 4);    // paranoid...
	    strncpy(Code3, Ccc, 3); // only first 3 chars from Ccc
	}
    };
    // END OF SUBCLASS Abbrev_
    
    // typedefs
    public:
	    
    /**
    * Type_: symbolic constants for a few popular monomer types.
    * Here is the list of the accepted 1-letter symbols,
    * first the unambiguous codes, then the allowed ambiguous codes:-
    *
    * PROT=ACDEFGHIKLMNPQRSTVWY, B=(D|N), Z=(E|Q), X=any, *=end
    * DNA=ACGT, R=(A|G), Y=(C|T), (N|X)=any
    * RNA=ACGU, R=(A|G), Y=(C|U), (N|X)=any
    * NUCL=ACGTU, R=(A|G), Y=(C|T|U), (N|X)=any
    *
    * USER is user-defined, NUCL==DNA|RNA.
    * Some routines may depend on UNKNOWN==0: do not change!
    */
    enum Type_ { UNKNOWN=0, USER=1, PROT=2, DNA=4, RNA=8, NUCL=12 };
    
    private:
	    
    vector<Abbrev_> Alphabet;	///< all legal monomer types
    Type_ Mtype;     	///< symbolic const for monomer types
    
    // data
    public:
	    
    /// iterator shorthand
    typedef vector<Abbrev_>::const_iterator Aciter_;
    
    	// ctors
    
    /**
    * Inits to hold the alphabet for some predefined type.
    * USER is illegal here, PROT (the default) will be used instead.
    */
    explicit Monomer_(int Mt=PROT)
    {
	if (Mt==UNKNOWN || Mt==USER)    // Mt was illegal
	    Mt=PROT;
	set_alphabet(Mt);
    }
    
    /**
    * Inits to hold the user-defined alphabet Codes.
    * For the Codes string format, see set_alphabet(Codes).
    */
    explicit Monomer_(const char* Codes)
    {
	set_alphabet(Codes);
    }
    
    	/// access
    
    /**
    * set_alphabet(Mt):
    * sets the alphabet to one of the predefined monomer types as indicated by Mt.
    * If Mt==USER or UNKNOWN, no action is taken and a warning is printed.
    */
    void set_alphabet(int Mt);
    
    /** 
    * set_alphabet(Codes):
    * sets the alphabet to contain the abbreviations listed in Codes.
    * The format of Codes: 4 chars per monomer type, first the 1-letter 
    * code, then the 3-letter code like "AAlaBAsxCCys...".
    * Nothing is done and warnings are printed if Codes is NULL, empty string
    * or its length is not a multiple of 4.
    * On success, the Mtype of the calling object is set to USER.
    */
    void set_alphabet(const char* Codes);
    
    /// alphabet(): only const manipulations with the monomer list
    const vector<Abbrev_>& alphabet() const { return Alphabet; }
    
    /// type(): only const access
    Type_ type() const { return Mtype; }
    
    /**
    * ==: two monomer objects are considered equal either if both
    * are of type USER and their alphabets are the same,
    * or if they have the same type. In particular,
    * NUCL and RNA or NUCL and DNA are not equal.
    */
    bool operator==(const Monomer_& M) const;
    bool operator!=(const Monomer_& M) const { return !(*this==M); }
    
    /**
    * contains(M, Nocase):
    * returns true if all the monomers in M can be found in
    * the calling object, false otherwise.
    */
    bool contains(const Monomer_& M, bool Nocase) const;
    
    /**
    * find(C, Nocase), find(Ccc, Nocase):
    * tries to find the 1-letter (C) or 3-letter (Ccc) abbreviation
    * in the calling object.
    * If Nocase==true, then a case-insensitive match is enough.
    * Returns a const iterator to the abbreviation if found,
    * or Alphabet.end() if not found.
    */
    Aciter_ find(char C, bool Nocase=false) const;
    Aciter_ find(const char* Ccc, bool Nocase=false) const;
    
    /**
    * is_legal(C, Nocase), is_legal(Ccc, Nocase):
    * return true if C or Ccc are "legal" abbreviations of a monomer.
    * If Nocase==true, then a case-insensitive match is enough.
    */
    bool is_legal(char C, bool Nocase=false) const
    {
	return (Alphabet.end()!=find(C, Nocase));
    }
    bool is_legal(const char* Ccc, bool Nocase=false) const
    {
	return (Alphabet.end()!=find(Ccc, Nocase));
    }
    
    /**
    * one_2_three(C, Nocase):
    * returns the 3-letter abbrev. corresponding to C if C is legal
    * depending on Nocase (when true, then case-insensitive match is OK),
    * or NULL if C is illegal.
    */
    const char* one_2_three(char C, bool Nocase=false) const
    {
	Aciter_ Ai=find(C, Nocase);
	return (Ai!=Alphabet.end())? Ai->code_3(): NULL;
    }
    
    /** 
    * three_2_one(Ccc, Nocase):
    * the inverse of one_2_three(C, Nocase).
    * Returns '\0' if Ccc is illegal.
    */
    char three_2_one(const char* Ccc, bool Nocase=false) const
    {
	Aciter_ Ai=find(Ccc, Nocase);
	return (Ai!=Alphabet.end())? Ai->code_1(): '\0';
    }
    
};
// END OF CLASS Monomer_

} // RazorBack

// ==== END OF HEADER Monomer.hh ====

#endif	    // MONOMER_HEADER
