// ==== METHODS Alignment.cc ====

// Class for representing multiple sequence alignments.

// 18-Aug-1999, Andras Aszodi

// ==== STANDARD HEADERS ====

#include <string.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iomanip>
#include <sstream>
#include <algorithm>
using namespace std;
#else
#include <iomanip.h>
#include <strstream.h>
#define istringstream istrstream
#include <algo.h>
#endif

// ==== MODULE HEADERS ====

#include"Alignment.hh"
using namespace RazorBack;

// ==== Alignment_ METHODS ====

// ---- Access ----

// at(Idx): checked const access to the Idx:th sequence.
// Throws Indexrangexc_ if Idx is illegal.
const Gapseq_& Alignment_::at(size_t Idx) const
{
    if (Idx>=seq_no())
	throw Indexrangexc_("Alignment_::at(Idx) const", Idx, seq_no()-1);
    return Sequences[Idx];
}
// END of at(Idx)

// find(ID): finds the sequence with ID string ID in the alignment.
// Returns a number in [ 0..seq_no() ) if found, -1 if not found.
int Alignment_::find(const string& ID) const
{
    vector<Gapseq_>::const_iterator Sptr;
    for (Sptr=Sequences.begin(); Sptr!=Sequences.end(); ++Sptr)
	if (Sptr->id()==ID)
	    return Sptr-Sequences.begin();
    return -1;
}
// END of find(ID)

// vertical_string(Apos):
// creates and returns a string that contains the Apos:th
// vertical position. E.g. for this alignment:
// -ALI-GN-
// --MIGGN-
// SAVL-GNP
// the vertical string at position 2 is "LMV".
// Apos must be in the range [0..max_len()). If the sequences
// are unaligned, then the right margin of the alignment is "ragged",
// in this case space characters will be used as padding.
// Throws Indexrangexc_ if Apos>=max_len().
string Alignment_::vertical_string(size_t Apos) const
{
    if (Apos>=Maxlen)
	throw Indexrangexc_("Alignment_::vertical_string()", Apos, Maxlen-1);
    
    string Vstr;
    char C;
    vector<Gapseq_>::const_iterator Sptr;
    for (Sptr=Sequences.begin(); Sptr!=Sequences.end(); ++Sptr)
    {
	if (Apos>=Sptr->gross_length()) // over ragged end
	    C=' ';
	else C=Sptr->at_1(Sptr->net_pos(Apos));
	Vstr+=C;
    }
    return Vstr;
}
// END of vertical_string(Apos)

// ---- Sequence manipulation ----

// mmer_compat(Mseq, [Mali]):
// returns true if the monomer type in Mseq for a sequence
// is compatible with the monomer type in Mali for an alignment
// (if omitted, the calling object's current monomer() object
// is used). DNA, RNA and NUCL monomer types are compatible
// with each other (basically we allow any nucleotide sequence
// to be aligned to any other nucleotide sequence).
// All other monomer types are compatible only with themselves,
// UNKNOWN is incompatible with all definite types.
bool Alignment_::mmer_compat(const Monomer_& Mseq, const Monomer_& Mali) const
{
    // UNKNOWN is incompatible even with itself
    if (Mali.type()==Monomer_::UNKNOWN
	    || Mseq.type()==Monomer_::UNKNOWN)
	return false;
    
    // NUCL, DNA, RNA are compatible with each other
    if ((Mali.type() & Monomer_::NUCL) && (Mseq.type() & Monomer_::NUCL))
	return true;
    else return (Mali==Mseq);	// otherwise equality is needed
}
// END of mmer_compat(Mseq, [Mali])

// set_monomer_type(Mseq, [Mali]):
// this little method takes care of the special case of
// nucleic acids that are allowed to coexist within an Alignment_
// object (by default, all sequences must share the same alphabet).
// Mseq will be OR-ed to Mali (own monomer member if not specified)
// if either of them was a NUCL, DNA, RNA. Otherwise Mali will simply
// bet set to Mseq.
// It is assumed that the monomer compatibility was checked
// before the invocation of this method.
// Hidden
void Alignment_::set_monomer_type(const Monomer_& Mseq, Monomer_& Mali)
{
    if (Monomer==Mseq) return;	// no need to set
    
    // all kinds of nucleic acids are allowed,
    // Monomer will have the types all OR-ed together
    if ((Mali.type() & Monomer_::NUCL)
	    && (Mseq.type() & Monomer_::NUCL))
	Mali.set_alphabet(Mali.type() | Mseq.type());
    else Mali=Mseq;	// simple assignment for PROT, etc.
}
// END of set_monomer_type(Mseq, [Mali])

// add_seq(Seq, [Idx]): adds the sequence Seq before position Idx.
// If the position is omitted then Seq will be added to the end.
// If Idx is illegal, Indexrangexc_ will be thrown.
// However, if the calling object was empty, then Idx is ignored
// and the sequence will be added anyway.
// The type of this first sequence will set the type of the 
// whole alignment.
// Seq must have compatible monomers as the alignment itself
// as defined by the mmer_compat() method
// otherwise it will not be added and a warning is issued.
void Alignment_::add_seq(const Gapseq_& Seq)
{
    if (seq_no()>=1 && !mmer_compat(Seq.monomer()))
    {
	cerr<<"\n? Alignment_::add_seq(Seq): Monomer type mismatch, not added\n";
	return;
    }
    if (seq_no()==1) Monomer=Seq.monomer(); // first sets overall monomer type
    set_monomer_type(Seq.monomer());
    Sequences.push_back(Seq);
    Dirty=true;     // non-const access
}
void Alignment_::add_seq(const Gapseq_& Seq, size_t Idx)
{
    if (!seq_no())  // was empty
	add_seq(Seq);	// just disregard Idx, add anyway
    
    if (!mmer_compat(Seq.monomer()))
    {
	cerr<<"\n? Alignment_::add_seq(Seq): Monomer type mismatch, not added\n";
	return;
    }
    if (Idx>=seq_no())
	throw Indexrangexc_("Alignment_::add_seq()", Idx, seq_no()-1);
    set_monomer_type(Seq.monomer());
    Sequences.insert(Sequences.begin()+Idx, Seq);
    Dirty=true;
}
// END of add_seq(Seq, [Idx])

// remove_seq([Idx]): removes the Idx:th sequence from the alignment.
// If Idx is omitted then all sequences are deleted (complete reset).
// Nothing is done if the calling object was already empty.
// Throws Indexrangexc_ otherwise if Idx is illegal.
void Alignment_::remove_seq(size_t Idx)
{
    if (!seq_no()) return;
    if (Idx>=seq_no())
	throw Indexrangexc_("Alignment_::remove_seq()", Idx, seq_no()-1);
    Sequences.erase(Sequences.begin()+Idx);
    Dirty=true;
}
void Alignment_::remove_seq()
{
    Monomer.set_alphabet(Monomer_::UNKNOWN);	// no alphabet
    Sequences.clear();
    Maxlen=0; Dirty=false;
}
// END of remove_seq([Idx])

// ---- Gap manipulations ----

// insert_gaps(Seqidx, Netpos, Gaplen):
// invokes Gapseq_::insert_gaps(Netpos, Gaplen) for the Seqidx:th
// sequence in the alignment.
// Throws Indexrangexc_ if Seqidx is illegal.
void Alignment_::insert_gaps(size_t Seqidx, size_t Netpos, size_t Gaplen)
{
    if (Seqidx>=seq_no())
	throw Indexrangexc_("Alignment_::insert_gaps(Seqidx)", Seqidx, seq_no()-1);
    Sequences[Seqidx].insert_gaps(Netpos, Gaplen);
    Dirty=true;     // maximal gross length may change
}
// END of insert_gaps(Seqidx, Netpos, Gaplen)

// remove_gaps([Seqidx],[Netpos]):
// When both parameters are specified, then Gapseq_::remove_gaps(Netpos)
// is invoked for the Seqidx:th sequence in the alignment.
// If Netpos is omitted, then all gaps are removed from the
// Seqidx:th sequence.
// Throws Indexrangexc_ if Seqidx is illegal.
// Without any parameters, all gaps are deleted from all sequences.
void Alignment_::remove_gaps(size_t Seqidx, size_t Netpos)
{
    if (Seqidx>=seq_no())
	throw Indexrangexc_("Alignment_::remove_gaps(Seqidx)", Seqidx, seq_no()-1);
    Sequences[Seqidx].remove_gaps(Netpos);
    Dirty=true;
}
void Alignment_::remove_gaps(size_t Seqidx)
{
    if (Seqidx>=seq_no())
	throw Indexrangexc_("Alignment_::remove_gaps(Seqidx)", Seqidx, seq_no()-1);
    Sequences[Seqidx].remove_gaps();
    Dirty=true;
}
void Alignment_::remove_gaps()
{
    vector<Gapseq_>::iterator Sptr;
    for (Sptr=Sequences.begin(); Sptr!=Sequences.end(); ++Sptr)
    	Sptr->remove_gaps();
    Dirty=true;
}
// END of remove_gaps([Seqidx],[Netpos])

// ---- I/O ----

// read_align(Inf, Type, Format):
// tries to read a multiple alignment from the input stream Inf.
// If Type is set to Monomer_::PROT, then only protein
// sequences will be read, if it is set to Monomer_::[DEOXY[RIBO]]NUCL,
// then only the corresponding nucleic acid sequences will be read,
// if Type is Monomer_::USER (default) or an OR-ed combination of PROT
// and *NUCL, then all types are considered and a polymer
// type autodetection is attempted. This is not foolproof.
// Format can be set to ANYFORMAT (default), in this case automatic
// file format detection is performed, or to PIR (proteins only),
// FASTA, GCG (protein or nucleic acid).
// No read is performed and FORMATERR is returned when
// incompatible type and format are specified.
// Return value: OK or an error enum if something went wrong.
Bioseqio_::Errtype_ Alignment_::read_align(istream& Inf, int Type, int Format)
{
// NOTE: the "new" ISO C++ standard iostream library has different types
// and class structure than the "old" iostream. Until all the compilers
// migrate to this (notable exception is the buggy SGI "new" iostream),
// we provide a workaround as follows.
#ifdef __USE_STD_IOSTREAM
#define IOS_IOSTATE ios_base::iostate
#else
#define IOS_IOSTATE int
#endif
    static const char* WARN="\n? Alignment_::read_align(): ";
    
    if (!Inf.good()) return Bioseqio_::FILEERR;	// no read from bad files
    
    if (Type==Monomer_::USER)
	Type=Monomer_::PROT|Monomer_::NUCL;
    if (Format==UNKNOWN) Format=ANYFORMAT;
    if (Format==PIR && !(Type & Monomer_::PROT))
    {
	cerr<<WARN<<"Incompatible Type and Format\n";
	return Bioseqio_::FORMATERR;
    }
    
    streampos Filepos=Inf.tellg();
    IOS_IOSTATE Infstate=Inf.rdstate(); // save old input state
    Bioseqio_::Errtype_ Err;

    // try PIR (proteins only)
    if ((Type & Monomer_::PROT) && (Format & PIR))
    {
	Err=read_concat(Inf, Monomer_::PROT, Bioseqio_::PIR);
	if (Err==Bioseqio_::OK)	// success, do not try others
	    return Err;
    }
        
    // try FASTA (protein or nucleic acid)
    if (Format & FASTA)
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	Err=read_concat(Inf, Type, Bioseqio_::FASTA);
	if (Err==Bioseqio_::OK)	// success, do not try others
	return Err;
    }

    // try GCG's MSF format (protein or nucleic acid)
    if (Format & GCG)
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	Err=read_msf(Inf);
	if (Err==Bioseqio_::OK)	// success, do not try others
	return Err;
    }

    // ...try other formats...
    
    // no success
    Inf.seekg(Filepos);
    return Err;
}
// END of read_align(Inf, Type, Format)

// read_concat(Inf, Type, Format):
// reads the "concatenated" alignment formats. In this format,
// the gapped sequences making up the alignment are simply listed
// one after the other. Can be done with FASTA and PIR.
// Inf is assumed to be good(), not tested.
// Return value: OK or one of the Bioseq_ error codes on error.
// Hidden
Bioseqio_::Errtype_ Alignment_::read_concat(istream& Inf, int Type, int Format)
{
    Monomer_ Mtemp; Mtemp.set_alphabet(Monomer_::UNKNOWN);
    Gapseq_ Tempseq;
    vector<Gapseq_> Tempseqs;
    Bioseqio_::Errtype_ Error=Bioseqio_::OK;
    
    // process all sequences in stream
    while (Inf.good())
    {
	// read one gapped sequence
	Error=Tempseq.read_seq(Inf, Type, Format);
	if (Error!=Bioseqio_::OK) continue;	// just skip
	
	// check monomer type
	if (Mtemp.type()==Monomer_::UNKNOWN)	// first sequence seen
	    Mtemp=Tempseq.monomer();	    // determines alignment type
	else if (!mmer_compat(Tempseq.monomer(), Mtemp)) continue;	// just skip
	
	Tempseqs.push_back(Tempseq);	// OK, worth storing
    }
    
    if (!Tempseqs.empty())  // success, at least one sequence is there
    {
	Monomer=Mtemp;	// copy into calling object
	Sequences=Tempseqs;
	Dirty=true; set_maxlen();
    }
    return Error;
}
// END of read_concat(Inf, Type, Format)

// Seqrec_: little local struct to hold temp seq data for read_msf(Inf)
struct Seqrec_
{
    string Name;	// the ID of the sequence
    int Check;  	// the GCG checksum of the sequence
    string Seq; 	// the gapped raw sequence itself

    Seqrec_(): Name(""), Check(0), Seq("") {}
};

// read_msf(Inf): reads the MSF format of GCG.
Bioseqio_::Errtype_ Alignment_::read_msf(istream& Inf)
{
    static const char* WARN="\n? Alignment_::read_msf(): ";
    make_buffer();
    
    // make "!!(N|A)A_MULTIPLE_ALIGNMENT" as first line mandatory
    bool Bangbang=false;
    while (fetch_buffer(Inf))
    {
	if (Bangbang = (NULL!=strstr(buffer(), "!!NA_MULTIPLE_ALIGNMENT")
		 || NULL!=strstr(buffer(), "!!AA_MULTIPLE_ALIGNMENT")))
	    break;
    }
    if (!Bangbang) return Bioseqio_::TITLEERR;
    
    // Look for the "dividing line".
    // This must contain "MSF: xxx" (the total length),
    // "Check: xxx" (the overall checksum) and ".."
    size_t Totallen=0, Overallcheck=0;
    char Typechar='\0';
    bool Dotdot=false;
    char *Readptr=NULL, *Tptr=NULL;
    while (!Dotdot && fetch_buffer(Inf))
    {
    	if (NULL==(Readptr=strstr(buffer(), "MSF:"))) continue;
	Readptr+=4;
	istringstream Msf(Readptr);
	Msf>>Totallen;
	Tptr=strstr(Readptr, "Type:");
	if (Tptr!=NULL) Typechar=*(Tptr+6);
	if (NULL==(Readptr=strstr(Readptr, "Check:"))) continue;
	Readptr+=6;
	istringstream Check(Readptr);
	Check>>Overallcheck;
	Dotdot=(NULL!=strstr(Readptr, ".."));
    }
    if (!Dotdot) return Bioseqio_::TITLEERR;
    
    // Read the names, lengths and checksums of the individual sequences
    istringstream Inbuffer(buffer());
    bool Separ=false;
    Seqrec_ Tempseqrec;
    vector<Seqrec_> Seqrecs;
    string Varname, Tempname;
    size_t Templen, Tempcheck, Sumcheck=0;
    while (fetch_buffer(Inf))
    {
	if (Separ=(NULL!=strstr(buffer(), "//"))) break;
	
	Inbuffer.clear(); Inbuffer.seekg(0);  // rewind input buffer
	Varname=Tempname="";
	Inbuffer>>Varname>>Tempname;
	if (Varname=="Name:" && Tempname.length())
	    Tempseqrec.Name=Tempname;
	else continue;
	Templen=0;
	Inbuffer>>Varname>>Templen;
	if (Varname=="Len:" && Templen)
	{
	    if (Templen!=Totallen)
		cerr<<WARN<<"Sequence \""<<Tempseqrec.Name
			<<"\": Len="<<Totallen<<" expected, "
			<<Templen<<" found\n";
	}
	else continue;
	Tempcheck=0;
	Inbuffer>>Varname>>Tempcheck;
	if (Varname=="Check:" && Tempcheck)
	{
	    Tempseqrec.Check=Tempcheck;
	    Sumcheck+=Tempcheck;
	}
	else continue;
	Seqrecs.push_back(Tempseqrec);	// store sequence description
    }
    if (!Separ || Seqrecs.empty())
	return Bioseqio_::IDERR;
    if (Sumcheck % 10000 != Overallcheck)
    	cerr<<WARN<<"Individual checksums give overall checksum "
		<<(Sumcheck % 10000)<<", "<<Overallcheck<<" expected\n";
    
    // read the sequences
    size_t Linecnt=0, Seqrecidx;
    while (fetch_buffer(Inf))
    {
	// skip empty lines and position indicator lines
    	if (buffer(0)=='\0'
		|| strspn(buffer(), " \t\n0123456789")==strlen(buffer()))
	    continue;

	Inbuffer.clear(); Inbuffer.seekg(0);
	Seqrecidx=Linecnt % Seqrecs.size();
	Inbuffer>>Tempname;
	if (Tempname!=Seqrecs[Seqrecidx].Name)
	{
	    cerr<<WARN<<"Name \""<<Seqrecs[Seqrecidx].Name
		    <<"\" expected, \""<<Tempname<<"\" found at seqline "
		    <<Linecnt<<endl;
	    return Bioseqio_::SEQERR; // we are quite strict here
	}
	
	// ws will be skipped by reading by substrings
	size_t Nulltail;
	while (Inbuffer>>Tempname)
	{
	    // get rid of the incredible trailing '\0'-s
	    // this is a most bizarre feature of the std strings:
	    // '\0'-s are counted in the length
	    // and the >> operator may read in extra '\0'-s
	    // completely nonsensical
	    Nulltail=Tempname.find_first_of('\0');
	    if (Nulltail!=NPOS) Tempname.erase(Nulltail);
	    if (!Tempname.empty()) Seqrecs[Seqrecidx].Seq+=Tempname;
	}
	Linecnt++;
    }

    // a few more checks and the transfer
    vector<Seqrec_>::iterator Srptr;
    vector<Gapseq_> Tempseqs;
    Gapseq_ Tempseq;
    Monomer_ Mtemp(Monomer_::UNKNOWN);
    for (Srptr=Seqrecs.begin(); Srptr!=Seqrecs.end(); ++Srptr)
    {
	if (!Srptr->Seq.length())   // sequence was missing
	    return Bioseqio_::SEQERR;
	Tempcheck=gcg_checksum(Srptr->Seq);
	if (Srptr->Check!=Tempcheck)
	    cerr<<WARN<<"Seq \""<<Srptr->Name<<"\" checksum="
		    <<Tempcheck<<", expected "<<Srptr->Check<<endl;
	
	// convert to gapped sequence with type autodetection
	Tempseq.gapped_sequence(Srptr->Seq);
	if (Tempseqs.empty())	// first sequence
	    Mtemp=Tempseq.monomer();
	else	// not first, Mtemp already set
	{
	    if (!mmer_compat(Tempseq.monomer(), Mtemp))
	    {
		cerr<<WARN<<"Seq \""<<Srptr->Name<<"\": type mismatch\n";
		continue;
	    }
	    set_monomer_type(Tempseq.monomer(), Mtemp);
	}
	Tempseq.id(Srptr->Name);    // save seq name as ID in the seq itself
	Tempseqs.push_back(Tempseq);	// added
    }
    if (Mtemp.type()==Monomer_::PROT && Typechar!='P'
	    || (Mtemp.type() & Monomer_::NUCL) && Typechar!='N')
	cerr<<WARN<<"Type does not match typechar \'"<<Typechar<<"\'\n";
    
    // transfer at last!
    Monomer=Mtemp;
    Sequences=Tempseqs;
    Dirty=true; set_maxlen();
    return Bioseqio_::OK;
}
// END of read_msf(Inf)

// write_align(Outf, Format): writes the contents of the calling object
// to Outf according to the format specification in Format.
// Returns OK or an error enum if something went wrong.
Bioseqio_::Errtype_ Alignment_::write_align(ostream& Outf, int Format) const
{
    if (!Outf.good()) return Bioseqio_::FILEERR;
    switch (Format)
    {
	case FASTA:
	    return write_concat(Outf, FASTA);
	case PIR:
	    if (monomer().type()==Monomer_::PROT)
		return write_concat(Outf, PIR);
	    else return Bioseqio_::FORMATERR;
	case GCG:
	    return write_msf(Outf);
	default:
	    cerr<<"\n? Alignment_::write_align(): Format not implemented\n";
	    return Bioseqio_::FORMATERR;
    }
}
// END of write_align(Outf, Format)

// write_concat(Outf, Format):
// writes the "concatenated" alignment formats. In this format,
// the gapped sequences making up the alignment are simply listed
// one after the other. Can be done with FASTA and PIR.
// Outf is assumed to be good(), not tested.
// Return value: OK or one of the Bioseqio_ error codes on error.
// Hidden
Bioseqio_::Errtype_ Alignment_::write_concat(ostream& Outf, int Format) const
{
    if (!seq_no()) return Bioseqio_::OK;	// empty alignment, no write
    
    Bioseqio_::Errtype_ Err=Bioseqio_::OK;
    vector<Gapseq_>::const_iterator Seqptr;
    for (Seqptr=Sequences.begin(); Seqptr!=Sequences.end(); ++Seqptr)
    {
	Err=Seqptr->write_seq(Outf, Format);
	if (Err!=Bioseqio_::OK) break;
    }
    return Err;
}
// END of write_concat(Outf, Format)

// write_msf(Outf): writes the calling object to Outf
// in the GCG/MSF format. No comments/annotation lines are provided.
// Return value: OK or one of the Bioseqio_ error codes on error.
// Hidden
Bioseqio_::Errtype_ Alignment_::write_msf(ostream& Outf) const
{
    if (!Outf.good()) return FILEERR;
    if (!seq_no()) return SEQERR; // no sequences

    // print extra type information for GCG search programs
    bool Isprotein=(monomer().type()==Monomer_::PROT);
    if (Isprotein)
	Outf<<"!!AA_MULTIPLE_ALIGNMENT 1.0\n";
    else Outf<<"!!NA_MULTIPLE_ALIGNMENT 1.0\n";
    
    // transfer the gapped sequences to temp vector
    Seqrec_ Tempseqrec;
    vector<Seqrec_> Seqrecs;
    vector<Gapseq_>::const_iterator Seqptr;
    int Overallcheck=0;
    int Namewidth=0, Eidx;
    for (Seqptr=Sequences.begin(); Seqptr!=Sequences.end(); ++Seqptr)
    {
    	Tempseqrec.Name=Seqptr->id();
	Eidx=Tempseqrec.Name.find_first_of('\0');
	if (Eidx==NPOS) Eidx=Tempseqrec.Name.size();
	if (Eidx>Namewidth) Namewidth=Eidx; // get longest name length
	Tempseqrec.Seq=Seqptr->gapped_sequence(true);	// GCG gaps
	Tempseqrec.Check=gcg_checksum(Tempseqrec.Seq);
	Overallcheck+=Tempseqrec.Check;
	Seqrecs.push_back(Tempseqrec);
    }
    Overallcheck %= 10000;
    
    // write "dividing line"
    Outf<<endl<<"         MSF: "<<max_len()<<"  Type: "
	    <<(Isprotein? 'P': 'N')<<"  "<<current_datestr()
	    <<"  Check: "<<Overallcheck<<" ..\n\n";
    
    // write sequence names, checksums and weights
    vector<Seqrec_>::const_iterator Temptr;
    for (Temptr=Seqrecs.begin(); Temptr!=Seqrecs.end(); ++Temptr)
    {
	Outf<<"Name: ";
	Outf.setf(ios::left); // left justified
	Outf<<setw(Namewidth)<<Temptr->Name;
	Outf.setf(ios::right);    // back to default
	Outf<<"  Len: "<<setw(5)
		<<max_len()<<"  Check: "<<Temptr->Check<<"  Weight:  1.00\n";
    }
    Outf<<"\n//\n\n";

    // write the alignment with the decorative position numbering
    size_t Begpos, Endpos, Spaceno, i;
    for (Begpos=0; Begpos<max_len(); Begpos=Endpos)
    {
	Endpos=Begpos+50;
	if (Endpos>=max_len()) Endpos=max_len();
	Outf<<setw(Namewidth+2)<<"  ";
	Outf.setf(ios::left, ios::adjustfield);
	Outf<<setw(5)<<(Begpos+1);
	Outf.setf(ios::right, ios::adjustfield);
	Spaceno=Endpos-Begpos-5;
	Spaceno+=Spaceno/10;
	Outf<<setw(Spaceno)<<Endpos<<endl;
	
	for (Temptr=Seqrecs.begin(); Temptr!=Seqrecs.end(); ++Temptr)
	{
	    Outf<<setw(Namewidth)<<Temptr->Name<<" ";
	    for (i=Begpos; i<Endpos; ++i)
	    {
		if (i%10 == 0) Outf<<' ';
		Outf<<Temptr->Seq[i];
	    }
	    Outf<<endl;
	}
	Outf<<endl;
    }
    return Bioseqio_::OK;
}
// END of write_msf(Inf)

// ---- Auxiliaries ----

// set_maxlen(): recalculates the maximum of the gross (gapped)
// length of the sequences in the alignment.
// Sets and returns Maxlen.
// Hidden
size_t Alignment_::set_maxlen() const
{
    if (!Dirty) return Maxlen;
    
    Maxlen=0;
    vector<Gapseq_>::const_iterator Sptr;
    for (Sptr=Sequences.begin(); Sptr!=Sequences.end(); ++Sptr)
	Maxlen=max(Maxlen, Sptr->gross_length());
    Dirty=false;
    return Maxlen;
}
// END of set_maxlen()

// ==== END OF METHODS Alignment.cc ====
