// ==== METHODS Bioseq.cc ====

// Base class for storing biological sequences.

// 9-Aug-1999. Andris Aszodi

// ---- COMPATIBILITY NOTES ----

// Uses the ANSI C++ standard string class.

// ---- STANDARD HEADERS ----

#include <ctype.h>
#include <time.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iomanip>
using namespace std;
#else
#include <iomanip.h>
#endif

// ---- MODULE HEADER ----

#include"Bioseq.hh"
using namespace RazorBack;

// ==== Bioseq_ METHODS ====

// ---- Access ----

// at_1(Idx), at_1(Idx, C):
// safe index operator for the Idx:th char in the sequence.
// The non-const version refuses to store characters not corresponding
// to the current type of the calling object.
// Throws Indexrangexc_ if Idx is illegal.
// Returns true on success, false when C was illegal.
char Bioseq_::at_1(size_t Idx) const
{
    if (Idx>=length())
	throw Indexrangexc_("Bioseq_::at_1() const", Idx, length()-1);
    return Seq[Idx];
}
bool Bioseq_::at_1(size_t Idx, char C)
{
    if (is_legal(C))
    {
	if (Idx>=length())
	    throw Indexrangexc_("Bioseq_::at_1()", Idx, length()-1);
	Seq[Idx]=C;
	return true;
    }
    else
    {
	cerr<<"\n? Bioseq_::at_1("<<Idx<<", "<<C<<"): Illegal char\n";
	return false;
    }
}

// at_3(Idx), at_3(Idx, Ccc):
// safe index operator for getting and setting
// a monomer using the 3-letter abbreviations.
// The non-const version refuses to store abbreviations
// that are illegal.
// Throws Indexrangexc_ if Idx is illegal.
// Returns true on success, false when Ccc was illegal.
const char* Bioseq_::at_3(size_t Idx) const
{
    if (Idx>=length())
	throw Indexrangexc_("Bioseq_::at_3() const", Idx, length()-1);
    return (monomer().one_2_three(Seq[Idx], Nocase));
}
bool Bioseq_::at_3(size_t Idx, const char* Ccc)
{
    if (monomer().is_legal(Ccc, Nocase))
    {
	if (Idx>=length())
	    throw Indexrangexc_("Bioseq_::at_3()", Idx, length()-1);
	Seq[Idx]=monomer().three_2_one(Ccc, Nocase);
	return true;
    }
    else
    {
	cerr<<"\n? Bioseq_::at_3("<<Idx<<", \""<<Ccc<<"\"): Illegal 3-letter abbrev\n";
	return false;
    }
}
// END of at_3(...)

// ---- Legal issues ----

// is_legal(S): tests legality for the complete sequence string S.
bool Bioseq_::is_legal(const string& S) const
{
    string::const_iterator Sptr;
    bool Legal=true;
    
    for (Sptr=S.begin(); Legal && Sptr!=S.end(); Sptr++)
	Legal=Legal && is_legal(*Sptr);
    return Legal;
}
// END of is_legal(S)

// is_legal(Mt, C): tests if C is legal if the monomertype
// of the calling object were Mt. This is a virtual fn so
// derived classes may add extra logic (such as gap characters)
// to the standard Monomer_ alphabet test.
// Hidden
bool Bioseq_::is_legal(int Mt, char C) const
{
    Monomer_ M(Mt);
    return M.is_legal(C, nocase());
}

// uppercase(), lowercase():
// convert the sequence string to all-uppercase or all-lowercase.
// Return the number of characters converted.
int Bioseq_::uppercase()
{
    string::iterator Sptr;
    int Converted=0;
    char C;
    for (Sptr=sequence().begin(); Sptr!=sequence().end(); ++Sptr)
    {
	C=toupper(*Sptr);
	if (C!=*Sptr)
	{
	    *Sptr=C; ++Converted;
	}
    }
    return Converted;
}
int Bioseq_::lowercase()
{
    string::iterator Sptr;
    int Converted=0;
    char C;
    for (Sptr=sequence().begin(); Sptr!=sequence().end(); ++Sptr)
    {
	C=tolower(*Sptr);
	if (C!=*Sptr)
	{
	    *Sptr=C; ++Converted;
	}
    }
    return Converted;
}
// END of toupper(). tolower()

// ---- Input ----

// NOTE: the gory details will be taken care of by Bioseqio_.

// read_seq(Inf, Type, Format):
// tries to read a sequence from the input stream Inf.
// If Type is set to Monomer_::PROT, then only protein
// sequences will be read, if it is set to Monomer_::[DEOXY[RIBO]]NUCL,
// then only the corresponding nucleic acide sequences will be read,
// if Type is Monomer_::USER (default) or an OR-ed combination of PROT
// and *NUCL, then all types are considered and a polymer
// type autodetection is attempted. This is not foolproof.
// Format can be set to ANYFORMAT (default), in this case automatic
// file format detection is performed, or to a concrete
// file format. Type and Format may be incompatible:-
// PIR, SWISS are for proteins only, EMBL is for nucleic acids only.
// No read is performed and FORMATERR is returned when
// incompatible type and format are specified.
// Return value: OK or an error enum if something went wrong.
Bioseqio_::Errtype_ Bioseq_::read_seq(istream& Inf, int Type, int Format)
{
// NOTE: the "new" ISO C++ standard iostream library has different types
// and class structure than the "old" iostream. Until all the compilers
// migrate to this (notable exception is the buggy SGI "new" iostream),
// we provide a workaround as follows.
#ifdef __USE_STD_IOSTREAM
#define IOS_IOSTATE ios_base::iostate
#else
#define IOS_IOSTATE int
#endif
    static const char* WARN="\n? Bioseq_::read_seq(): ";
    
    if (!Inf.good()) return FILEERR;	// no read from bad files
    if (Type==Monomer_::USER)
	Type=Monomer_::PROT|Monomer_::NUCL;
    if (Type==Monomer_::PROT && Format==EMBL
	    || (!(Format & (PIR|SWISS)) && !(Type & Monomer_::PROT)))
    {
	cerr<<WARN<<"Incompatible Type and Format\n";
	return FORMATERR;
    }
    
    static string Tempid, Temptitle, Tempannot, Tempseq;  // input buffers
    streampos Filepos=Inf.tellg();
    IOS_IOSTATE Infstate=Inf.rdstate(); // save old input state
    Errtype_ Err=OK;

    // try PIR (proteins only)
    if ((Type & Monomer_::PROT) && (Format & PIR))
    {
	Err=read_pir(Inf, Tempid, Temptitle, Tempannot, Tempseq);
	if (Err==OK)    // successful
	{
	    if (clean_seq(Tempseq, Monomer_::PROT, nocase()))	// check chars, remove spaces etc.
	    {
		id(Tempid); title(Temptitle);
		annot()=Tempannot;
		sequence()=Tempseq;  // save sequence w/o further checks
		chg_monomer().set_alphabet(Monomer_::PROT);
		return OK;
	    }
	    else return SEQERR;
	}
    }
        
    // try FASTA (protein or nucleic acid)
    if (Format & FASTA)
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	Err=read_fasta(Inf, Tempid, Temptitle, Tempseq);
	if (Err==OK)    // successful
	{
	    // Autotype will contain the narrowest typing
	    int Autotype=Type & autodetect_type(Tempseq);
	    if (Autotype==Monomer_::UNKNOWN)
		return TYPEERR;
	    if (clean_seq(Tempseq, Autotype, nocase()))	// check chars, remove spaces etc.
	    {
		id(Tempid); title(Temptitle);
		annot().erase();   // FASTA has no annotation
		sequence()=Tempseq;  // save sequence w/o further checks
	    	chg_monomer().set_alphabet(Autotype);
		return OK;
	    }
	    else return SEQERR;
	}
    }
    
    // try GCG (protein or nucleic acid)
    if (Format & GCG)
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	int Gcglen, Gcgcheck;   // sequence length and checksum in GCG file
	char Typechar='\0';  	    // sequence type in GCG file: P or N
	Err=read_gcg(Inf, Tempid, Tempannot, Tempseq, Gcglen, Gcgcheck, Typechar);
	if (Err==OK)    // successful
	{
	    // GCG knows if protein or nucl,
	    // do autotyping only if Typechar was missing for some reason
	    int Autotype=(Typechar=='P')?
		    Monomer_::PROT: Type & autodetect_type(Tempseq);
	    if (Autotype==Monomer_::UNKNOWN)
		return TYPEERR;
		
	    if (clean_seq(Tempseq, Autotype, nocase()))	// check chars, remove spaces etc.
	    {
		// check & warn against various GCG hints (if present)
		if (Gcglen>-1 && Tempseq.size()!=Gcglen)
		    cerr<<WARN<<"Length mismatch: GCG="
			    <<Gcglen<<", found "<<Tempseq.size()<<endl;
		int Ourcheck=gcg_checksum(Tempseq);
		if (Gcgcheck>-1 && Ourcheck!=Gcgcheck)
		    cerr<<WARN<<"Checksum mismatch: GCG="
			    <<Gcgcheck<<", found "<<Ourcheck<<endl;

		id(Tempid);
		title().erase();	// GCG has no title
		annot(Tempannot);
		sequence()=Tempseq;  // save sequence w/o further checks
		chg_monomer().set_alphabet(Autotype);
		return OK;
	    }
	    else return SEQERR;
	}
    }

    // try SwissProt (protein only)
    if ((Type & Monomer_::PROT) && (Format & SWISS))
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	int Swlen, Mw;
	unsigned int Crc32;
	Err=read_swiss(Inf, Tempid, Tempannot, Tempseq, Swlen, Mw, Crc32);
	if (Err==OK)    	// succeeded
	{
	    
	    if (clean_seq(Tempseq, Monomer_::PROT, nocase())) // check chars, remove ws...
	    {
		// check against SwissProt hints (none of them critical)
		if (Swlen!=Tempseq.size())
		    cerr<<WARN<<"Length mismatch: SwissProt="
			    <<Swlen<<", found "<<Tempseq.size()<<endl;
		int Ourmw=mol_weight(Tempseq);
		if (Mw!=Ourmw)
		    cerr<<WARN<<"MW mismatch: SwissProt="
			    <<Mw<<", found "<<Ourmw<<endl;
		unsigned long Ourcrc32=crc_32(Tempseq);
		if (Crc32!=Ourcrc32)
		    cerr<<WARN<<"CRC32 mismatch: SwissProt="
			    <<hex<<Crc32<<", found "<<hex<<Ourcrc32<<endl;
		id(Tempid);
		title().erase();	// SwissProt has no title
		annot(Tempannot);
		sequence()=Tempseq;  // save sequence w/o further checks
		chg_monomer().set_alphabet(Monomer_::PROT);
		return OK;
	    }
	    else return SEQERR;
	}
    }

    // try EMBL (nucleic acid only)
    if ((Type & Monomer_::NUCL) && (Format & EMBL))
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	int Embllen;
	int Acgtx[5];
	Err=read_embl(Inf, Tempid, Tempannot, Tempseq, Embllen, Acgtx);
	if (Err==OK)    	// succeeded
	{
	    // Autotype will contain the narrowest typing
	    int Autotype=Type & autodetect_type(Tempseq);
	    if (Autotype==Monomer_::UNKNOWN)
		return TYPEERR;
	    if (clean_seq(Tempseq, Autotype, nocase())) // check chars, remove ws...
	    {
		// check against EMBL hints (none of them critical)
		if (Embllen!=Tempseq.size())
		    cerr<<WARN<<"Length mismatch: EMBL="
			    <<Embllen<<", found "<<Tempseq.size()<<endl;
		if (!compare_basecount(Tempseq, Acgtx, true))
		    cerr<<WARN<<"Base count mismatch\n";
		id(Tempid);
		title().erase();	// EMBL has no title
		annot(Tempannot);
		sequence()=Tempseq;  // save sequence w/o further checks
		chg_monomer().set_alphabet(Autotype);
		return OK;
	    }
	    else return SEQERR;
	}
    }

    // ... try other formats ...
    
    // no success
    Inf.seekg(Filepos);
    return Err;
}
// END of read_seq(Inf, Type, Format)

// ---- Output ----

// write_seq(Outf, Format): writes the contents of the calling object
// to Outf according to the format specification in Format.
// Returns OK or an error enum if something went wrong.
Bioseqio_::Errtype_ Bioseq_::write_seq(ostream& Outf, int Format) const
{
    if (!Outf.good()) return FILEERR;
    switch (Format)
    {
	case FASTA: return write_fasta(Outf, id(), title(), seq());
	case PIR:
	    if (Monomer.type()==Monomer_::PROT)
		return write_pir(Outf, id(), title(), annot(), seq());
	    else return FORMATERR;
	case GCG: return write_gcg(Outf, id(), title(), annot(), 
		seq(), Monomer.type());
	case SWISS:
	    if (Monomer.type()==Monomer_::PROT)
		return write_swiss(Outf, id(), title(), annot(), seq());
	    else return FORMATERR;
	case EMBL:
	    if (Monomer.type() & Monomer_::NUCL)
		return write_embl(Outf, id(), title(), annot(), 
			seq(), Monomer.type());
	    else return FORMATERR;
	default:
	    cerr<<"\n? Bioseq_::write_seq(): Format not implemented\n";
	    return FORMATERR;
    }
}
// END of write_seq(Outf, Format)

// ==== END OF METHODS Bioseq.cc ====
