// ==== METHODS Bioseqio.cc ====

// Base class for biological sequence I/O.

// 10-Aug-1999. Andris Aszodi

// ---- COMPATIBILITY NOTES ----

// Uses the ANSI C++ standard string class.

// ---- STANDARD HEADERS ----

#include <ctype.h>
#include <time.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iomanip>
#include <sstream>
using namespace std;
#else
#include <iomanip.h>
#include <strstream.h>
#define istringstream istrstream
#endif

// ---- MODULE HEADER ----

#include"Bioseqio.hh"
using namespace RazorBack;

// ==== Bioseqio_ METHODS ====

// ---- Static init ----

const int Bioseqio_::BUFFERSIZE=256;

// clean_seq(S, Type, Nocase): cleans up the sequence S.
// Removes white space characters silently
// and illegal characters with a warning.
// Illegality is defined as the is_legal(Char,Nocase)
// method of a monomer object initialised with Type.
// The alphabet used corresponds to Type,
// UNKNOWN or USER are illegal here.
// If GapsOK==true (default false), then gap characters are skipped.
// Return value: the length of the processed sequence.
// Hidden
int Bioseqio_::clean_seq(string& S, int Type, bool Nocase, bool GapsOK) const
{
    if (Type==Monomer_::UNKNOWN || Type==Monomer_::USER)
    {
	cerr<<"\n? Bioseqio_::clean_seq(): UNKNOWN or USER monomer type is illegal here\n";
	return 0;
    }
    int Badno=0;
    string::iterator Pto;
    string::const_iterator Pfrom;
    char Ch;
    Monomer_ Mtype(Type);
    
    for (Pfrom=Pto=S.begin(); Pfrom!=S.end(); ++Pfrom)
    {
	Ch=*Pfrom;
	if (isspace(Ch))  // always skip & filter out whitespace
	    continue;
	if (!(NULL!=strchr("-.~", Ch) && GapsOK)	// keep gaps if GapsOK==true
		&& !Mtype.is_legal(Ch, Nocase))
	{
	    cerr<<"\n? Bioseqio_::clean_seq(): Removing illegal char \'"
		<<Ch<<"\'\n";
	    Badno++;	// count no. of bad chars
	    continue;   // skip bad char
	}
    	*Pto++=Ch;	// Ch is accepted
    }
    if (Pfrom>Pto)
	S.erase(Pto, S.end());	// remove tail
    
    if (Badno)
	cerr<<"\n? Bioseqio_::clean_seq(): # of illegal chars removed="
		<<Badno<<endl;
    return S.size();
}
// END of clean_seq(S)

// autodetect_type(S):
// tries to figure out the monomer type of the sequence string S.
// Returns Monomer_::UNKNOWN if the type is unknown,
// or something more informative on success.
// Legality is checked generously using case-insensitive matching,
// only letters [A-Za-z] are processed, anything else is skipped,
// in particular whitespace, numbers, punctuation ('-' etc).
// [DEOXY]RNA has priority over PROT, so the sequence "AACAAT"
// will be diagnosed as DNA, the sequence "CCGGGC" as NUCL.
// Hidden static
int Bioseqio_::autodetect_type(const string& S)
{
    static const Monomer_ PROT(Monomer_::PROT),
	    DNA(Monomer_::DNA), RNA(Monomer_::RNA);
    
    int Type=Monomer_::PROT|Monomer_::NUCL;
    string::const_iterator Sptr;
    
    // exit criterion uses Monomer_::UNKNOWN==0
    for (Sptr=S.begin(); Type!=Monomer_::UNKNOWN && Sptr!=S.end(); ++Sptr)
    {
    	if (!isalpha(*Sptr)) continue;	// skip whitespace, numbers, '-' etc.
	
	if ((Type & Monomer_::PROT) && !PROT.is_legal(*Sptr, true))
	    Type&=~Monomer_::PROT;
	if ((Type & Monomer_::DNA) && !DNA.is_legal(*Sptr, true))
	    Type&=~Monomer_::DNA;
	if ((Type & Monomer_::RNA) && !RNA.is_legal(*Sptr, true))
	    Type&=~Monomer_::RNA;
    }
    if (Type & Monomer_::NUCL) Type&=~Monomer_::PROT;
    return Type;
}
// END of autodetect_type(S)

// ---- Input ----

// fetch_buffer(Inf): fills up the cleared input buffer
// with a line from Inf.
// Returns true if Inf is still "good" after input.
// Line-by-line file read can be accomplished by
// while(fetch_buffer(Inf)) { /*... process line ...*/ }
bool Bioseqio_::fetch_buffer(istream& Inf)
{
    memset(Buffer, '\0', BUFFERSIZE);
    Inf.getline(Buffer, BUFFERSIZE);
    return Inf.good();
}
// END of fetch_buffer(Inf)

// read_xxx(Inf, Id, Title, Annot, Seq):
// These methods read one sequence from the input stream Inf.
// Id is usually a word containing a label for the sequence.
// Title is usually a line containing a brief description
// (not found in GCG format).
// Annot is the annotation, may span multiple lines
// (not found in FASTA format).
// Seq is the raw sequence information, completely unchecked.
// Return value: OK or an error enum if something went wrong.
Bioseqio_::Errtype_ Bioseqio_::read_fasta(istream& Inf, string& Id,
	    string& Title, string& Seq)
{
    if (!Inf.good()) return FILEERR;
    make_buffer();
    
    // locate next ID line (begins with '>' but not with ">P1;")
    bool Isid=false;
    while (fetch_buffer(Inf)
	    && strncmp(buffer(), ">P1;", 4)
	    && !(Isid=(buffer(0)=='>'))); 
    if (Isid)	// found ID line
    {	
	string Tok1, Rest;
	parse_title(buffer()+1, Tok1, Rest);
	if (Tok1.empty()) return IDERR;	// ID (first token) missing
	Id=Tok1;
	Title=Rest;	// may be empty but who cares
    }
    else return TITLEERR;   // no title line found
    
    // read the sequence: all lines up to EOF or new title line
    Errtype_ Err=SEQERR;
    streampos Offs;
    Isid=false;
    Seq.erase(); // clear Seq
    while (fetch_buffer(Inf) && !(Isid=(buffer(0)=='>')))
    {
	Err=OK; // could read this line
	Offs=Inf.tellg();   // save current input position
	Seq+=buffer();	// append contents of Buffer to Seq
    }
    if (Isid)
	Inf.seekg(Offs);    // skip before new title line
    return Err;
}
// END of read_fasta(Inf, Id, Title, Seq)

Bioseqio_::Errtype_ Bioseqio_::read_pir(istream& Inf, string& Id,
	    string& Title, string& Annot, string& Seq)
{
    if (!Inf.good()) return FILEERR;
    make_buffer();
    
    // locate next ID line (begins with ">P1;")
    bool Isid=false;
    while (fetch_buffer(Inf)
	    && !(Isid=(0==strncmp(buffer(),">P1;",4))));
    if (Isid)	// found ID line
    {	
	string Tok1, Rest;
	parse_title(buffer()+4, Tok1, Rest);
	if (Tok1.empty()) return IDERR;	// ID (first token) missing
	Id=Tok1;
	Title=Rest;	// may be empty but who cares
    }
    else return TITLEERR;   // no title line found
    
    // the next line is the annotation
    if (fetch_buffer(Inf))
	Annot=buffer();
    else return ANNOTERR;
    
    // read the sequence: all lines up to '*' or new title line
    Errtype_ Err=SEQERR;
    char *Starpos;
    Seq.erase(); // clear Seq
    while (fetch_buffer(Inf))
    {
	Err=OK; // could read this line
	Starpos=strrchr(buffer(), '*');	// is this the last line?
	if (Starpos!=NULL) *Starpos='\0';   // end the buffer there
	Seq+=buffer();	// append contents of Buffer to Seq
	if (Starpos!=NULL) break;   // end-of-seq has been seen
    }
    return Err;
}
// END of read_pir(Inf, Id, Title, Annot, Seq)

Bioseqio_::Errtype_ Bioseqio_::read_gcg(istream& Inf, string& Id,
	    string& Annot, string& Seq, 
	    int &Gcglen, int &Gcgcheck, char &Typechar)
{
    if (!Inf.good()) return FILEERR;
    make_buffer();
    
    // make "!![NA]A_SEQUENCE" as first line mandatory
    bool Bangbang=false;
    while (fetch_buffer(Inf))
    {
	if (Bangbang = (NULL!=strstr(buffer(), "!!NA_SEQUENCE")
		 || NULL!=strstr(buffer(), "!!AA_SEQUENCE")))
	    break;
    }
    if (!Bangbang) return TITLEERR;
    
    // read annotation: everything up to separator line containing ".."
    bool Dotdot=false;
    Annot.erase();
    while (fetch_buffer(Inf))
    {
	 if (Dotdot = (NULL!=strstr(buffer(),"..")))
	     break; 	// title line ("separator") found
	 Annot+=buffer(); // add to annotation
	 Annot+='\n';	// this was bitten off by getline(), add 
    }
    if (!Dotdot) return TITLEERR;  // no separator line found
    
    // now Buffer contains the separator line.
    // This usually looks like:-
    //
    // "ID_string Length: xxx <Date,time> Type: X Check: xxxx .. "
    //
    // Some of the components are easier to parse using C functions
    
    string Tok1, Rest;
    parse_title(buffer(), Tok1, Rest);
    if (Tok1=="..") Tok1.erase();
    Id=Tok1;	// empty ID is allowed
    
    // we get a C pointer to Rest and parse it
    const char *Restptr, *Temptr;
    
    Restptr=Rest.c_str();
    
    // look for length
    Temptr=strstr(Restptr, "Length:");
    Gcglen=-1;
    if (Temptr!=NULL)
    {
	istringstream Lenstr(Temptr+7);
	Lenstr>>Gcglen;
    }
    
    // look for type
    Temptr=strstr(Restptr, "Type:");
    Typechar='\0';
    if (Temptr!=NULL)
    {
	istringstream Typestr(Temptr+5);
	Typestr>>Typechar;
    }

    // look for checksum
    Temptr=strstr(Restptr, "Check:");
    Gcgcheck=-1;
    if (Temptr!=NULL)
    {
	istringstream Checkstr(Temptr+6);
	Checkstr>>Gcgcheck;
    }
    
    // read the sequence
    Errtype_ Err=SEQERR;
    Seq.erase(); // clear Seq
    while (fetch_buffer(Inf))
    {
	if (!buffer(0)) continue;    // skip decorative empty GCG lines
	Err=OK; // could read this line
	Temptr=buffer()+strspn(buffer(), " \t0123456789");	// skip monomer nos.
	Seq+=Temptr;	// append contents of Buffer to Seq
    }
    return Err;
}
// END of read_gcg(Inf, Id, Annot, Seq, Gcglen, Gcgcheck, Typechar)

Bioseqio_::Errtype_ Bioseqio_::read_swiss(istream& Inf, string& Id,
	    string& Annot, string& Seq, 
	    int& Swlen, int& Mw, unsigned int& Crc32)
{
    if (!Inf.good()) return FILEERR;
    make_buffer();
    
    // locate the ID line
    bool Isid=false;
    while (fetch_buffer(Inf)
	    && !(Isid=!strncmp(buffer(), "ID   ", 5)));
    Swlen=-1;
    if (Isid)
    {
	string Tok1, Rest;
	parse_title(buffer()+5, Tok1, Rest);
	if (Tok1.empty() || Rest.empty()) return IDERR; // no ID: mandatory
	Id=Tok1;
	
	const char *Restptr, *Temptr;
	Restptr=Rest.c_str();
	Temptr=strstr(Restptr, "PRT;");
	if (Temptr!=NULL)
	{
	    istringstream Idstr(Temptr+4);
	    Idstr>>Swlen;
	}
	else return IDERR;  // malformed ID line
    }
    else return IDERR;	    // missing ID line
    
    // read all annotation lines w/o further parsing
    Annot.erase();
    bool Seqid=false;
    while (fetch_buffer(Inf)
	    && !(Seqid=!strncmp(buffer(), "SQ   ", 5)))
    {
	Annot+=buffer();
	Annot+='\n';	// add bitten-off newline
    }
    if (!Seqid) return SEQERR;	// SQ line missing

    // parse SQ line
    istringstream Sqline(buffer()+2);
    char Dummy[80];
    int Swlen2;
    Sqline>>Dummy>>Swlen2>>Dummy>>Mw>>Dummy>>hex>>Crc32;
    if (!Sqline.good())
	return SEQERR;	// malformed SQ line
    if (Swlen!=Swlen2)
	cerr<<"\n? Bioseqio_::read_swiss(): Internal length mismatch: "
		<<Swlen<<"!="<<Swlen2<<endl;
    
    // read sequence
    Seq.erase();
    Errtype_ Err=SEQERR;
    while (fetch_buffer(Inf)
	    && strncmp(buffer(), "//", 2))
    {
	Err=OK;
	Seq+=buffer();
    }
    return Err;
}
// END of read_swiss(Inf, ID, Annot, Seq, Swlen, Mw, Crc32)

Bioseqio_::Errtype_ Bioseqio_::read_embl(istream& Inf, string& Id,
	    string& Annot, string& Seq, 
	    int& Embllen, int Acgtx[5])
{
    if (!Inf.good()) return FILEERR;
    make_buffer();
    
    // locate the ID line
    bool Isid=false;
    char Dummy[80]; // for skipping strings :-)
    
    while (fetch_buffer(Inf)
	    && !(Isid=!strncmp(buffer(), "ID   ", 5)));
    Embllen=-1;
    if (Isid)
    {
	string Tok1, Rest;
	parse_title(buffer()+5, Tok1, Rest);
	if (Tok1.empty() || Rest.empty()) return IDERR; // no ID: mandatory
	Id=Tok1;
	
	const char *Restptr, *Temptr;
	Restptr=Rest.c_str();
	Temptr=strstr(Restptr, "NA;");	// actually "[DR]NA;"
	if (Temptr!=NULL && (Temptr[-1]=='D' || Temptr[-1]=='R'))
	{
	    istringstream Idstr(Temptr+3);
	    Idstr>>Dummy>>Embllen;
	}
	else return IDERR;	// hopeless
    }
    else return IDERR;	// missing ID line
    
    // read all annotation lines w/o further parsing
    Annot.erase();
    bool Seqid=false;
    while (fetch_buffer(Inf)
	    && !(Seqid=!strncmp(buffer(), "SQ   ", 5)))
    {
	Annot+=buffer();
	Annot+='\n';	// add bitten-off newline
    }
    if (!Seqid) return SEQERR;	// SQ line missing

    // parse SQ line
    istringstream Sqline(buffer()+2);
    int Embllen2;
    Sqline>>Dummy>>Embllen2>>Dummy>>Acgtx[0]>>Dummy>>Acgtx[1]
	    >>Dummy>>Acgtx[2]>>Dummy>>Acgtx[3]>>Dummy>>Acgtx[4];
    if (!Sqline.good())
	return SEQERR;	// malformed SQ line
    if (Embllen!=Embllen2)
	cerr<<"\n? Bioseqio_::read_embl(): Internal length mismatch: "
		<<Embllen<<"!="<<Embllen2<<endl;
    
    // read sequence
    Seq.erase();
    Errtype_ Err=SEQERR;
    while (fetch_buffer(Inf)
	    && strncmp(buffer(), "//", 2))
    {
	Err=OK;
	buffer(71)='\0';    // chop off nucleotide number
	Seq+=buffer();
    }
    return Err;
}
// END of read_embl(Inf, ID, Annot, Seq, Embllen, Acgtx)

// parse_title(Strc, Tok1, Rest): cuts up the input C string Strc
// according to the following scheme (<ws> is whitespace,
//[x] means 0 or more x):-
//
// "[<ws>]firsttoken<ws>[<ws>]restofthestring..."
//  ^     ^        ^          ^
//  |     |--------|          |-----------...
//  Strc     Tok1                Rest
// 
// and returns the substrings Tok1 and Rest.
// Hidden static
void Bioseqio_::parse_title(const char *Strc, string& Tok1, string& Rest)
{
    static const char *Ws=" \t\n";  // whitespace chars
    const string& Str(Strc);
    
    Tok1.erase(); Rest.erase(); 	// clear both
    
    int Tok1beg=Str.find_first_not_of(Ws);
    if (Tok1beg==NPOS) return;
    
    int Tok1end=Str.find_first_of(Ws, Tok1beg);
    Tok1=Str.substr(Tok1beg, Tok1end-Tok1beg);	// first token
    
    int Tok2beg=Str.find_first_not_of(Ws, Tok1end);
    if (Tok2beg==NPOS) return;
    
    Rest=Str.substr(Tok2beg);	// rest of the string
}
// END of parse_title(Strc, Tok1, Rest)

// ---- Output ----

// write_xxx(Inf, Id, Title, Annot, Seq):
// These methods write one sequence to the output stream Outf.
// Id is usually a word containing a label for the sequence.
// Title is usually a line containing a brief description.
// Annot is the annotation, may span multiple lines. 
// Note that there is no extra annotation in the FASTA format
// so this parameter is missing for read_fasta().
// Seq is the sequence.
// Return value: OK or an error enum if something went wrong.

Bioseqio_::Errtype_ Bioseqio_::write_fasta(ostream& Outf, const string& ID,
	const string& Title, const string& Seq)
{
    if (!Outf.good()) return FILEERR;
    if (ID.empty() && Title.empty()) return TITLEERR;	// no title at all
    if (Seq.empty()) return SEQERR; // no sequence
    
    Outf<<'>'<<ID<<' '<<Title<<endl;	// print title line
    
    int i, Len=Seq.size();
    for (i=0; i<Len; i++)
    {
	Outf<<Seq[i];
	if ((i+1) % 60 ==0) Outf<<endl;
    }
    Outf<<endl;
    return OK;
}
// END of write_fasta(Outf, ID, Title, Seq)

Bioseqio_::Errtype_ Bioseqio_::write_pir(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq)
{
    if (!Outf.good()) return FILEERR;
    if (Seq.empty()) return SEQERR; // no sequence
    if (ID.empty() && Title.empty()) return TITLEERR;	// no title line at all
    
    Outf<<">P1;"<<ID<<' '<<Title<<endl;	// print title line
    
    if (Annot.empty()) Outf<<"(no annotation)\n";
    else 
    {
	// PIR has only one annotation line
	// so print the first
	size_t Firsteoln=Annot.find('\n');
	Outf<<Annot.substr(0, Firsteoln)<<endl;
    }
    
    int i, Len=Seq.size();
    for (i=0; i<Len; i++)
    {
	Outf<<Seq[i];
	if ((i+1) % 60 ==0) Outf<<endl;
    }
    Outf<<"*\n";
    return OK;
}
// END of write_pir(Outf, ID, Title, Annot, Seq)

Bioseqio_::Errtype_ Bioseqio_::write_gcg(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq,
	Monomer_::Type_ Type)
{
    if (!Outf.good()) return FILEERR;
    if (Seq.empty()) return SEQERR; // no sequence

    // extra type information for GCG search programs
    bool Isprotein=(Type==Monomer_::PROT);
    if (Isprotein)
	Outf<<"!!AA_SEQUENCE 1.0\n";
    else Outf<<"!!NA_SEQUENCE 1.0\n";
    
    // title and annotation (both may be empty)
    if (!Title.empty()) Outf<<Title<<endl;
    if (!Annot.empty()) Outf<<Annot<<endl;
    
    // print the separator line
    if (!ID.empty()) Outf<<ID<<' '; // ID may be missing
    Outf<<"Length: "<<Seq.size()<<' '<<current_datestr()
	    <<" Type: "<<(Isprotein? 'P': 'N')
	    <<" Check: "<<gcg_checksum(Seq)
	    <<" ..\n\n     1 ";
    
    // print the sequence
    for (int i=0; i<Seq.size(); i++)
    {
	Outf<<Seq[i];
	if ((i+1) % 10 == 0) Outf<<' ';
	if ((i+1) % 50 == 0)
	    Outf<<"\n\n"<<setw(6)<<(i+2)<<" ";
    }
    Outf<<endl;
    return OK;
}
// END of write_gcg(Outf, ID, Title, Annot, Seq, Isprotein)   

// NOTE: the "new" ISO C++ standard iostream library has different types
// and class structure than the "old" iostream. Until all the compilers
// migrate to this (notable exception is the buggy SGI "new" iostream),
// we provide a workaround as follows.
#ifdef __USE_STD_IOSTREAM
#define IOS_FLAGTYPE ios_base::fmtflags
#define IOS_BASE ios_base
#else
#define IOS_FLAGTYPE long
#define IOS_BASE ios
#endif

Bioseqio_::Errtype_ Bioseqio_::write_swiss(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq)
{
    if (!Outf.good()) return FILEERR;
    if (Seq.empty()) return SEQERR;
    if (ID.empty()) return IDERR;
    
    // print the ID line
    int Len=Seq.size();
    Outf<<"ID   "<<ID<<"  standard; PRT; "<<Len<<" AA.\n";

    // The SwissProt/EMBL annotation is very rich and if the sequence
    // was read from another format then it cannot be reproduced easily.
    // We look for the "AC" and "DE" lines in the annotation and if
    // they are found then we dump the annotation as it is, and ignore
    // the Title field. Otherwise, it is assumed that the sequence
    // comes from another format and we copy the title and the
    // annotation as "CC" records to the output.
    if (Title.empty() && Annot.empty())
	Outf<<"CC   (no annotation)\n";
    else if (NPOS!=Annot.find("AC   ") && NPOS!=Annot.find("DE   "))
	    Outf<<Annot;    // looks like original SwissProt annotation
    else
    {
	swissembl_comment(Outf, Title); // convert non-SW title and annot
	swissembl_comment(Outf, Annot); // by prepending "CC" to each line
    }
    
    // print the SQ line
    Outf<<"SQ   SEQUENCE  "<<Len<<" AA;  "
	    <<mol_weight(Seq)<<" MW;  ";
    IOS_FLAGTYPE Oldflags=Outf.flags();	    	    // save old output flags
    Outf.flags(Oldflags|IOS_BASE::uppercase);    // print hex value in uppercase
    Outf<<hex<<crc_32(Seq)<<" CRC32;\n";
    Outf.flags(Oldflags);   	    	    // reset old flags
    
    // print the sequence
    register int i;
    for (i=0; i<Len; i++)
    {
	if (i % 60 == 0) Outf<<"     ";
	Outf<<Seq[i];
	if ((i+1) % 60 == 0) Outf<<endl;
	else if ((i+1) % 10 == 0) Outf<<' ';
    }
    if (i % 60 != 0) Outf<<endl;
    Outf<<"//\n";
    
    return OK;
}
// END of write_swiss(Outf, ID, Title, Annot, Seq)
	     
Bioseqio_::Errtype_ Bioseqio_::write_embl(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq,
	Monomer_::Type_ Type)
{
    if (!Outf.good()) return FILEERR;
    if (Seq.empty()) return SEQERR;
    if (ID.empty()) return IDERR;
    
    // print the ID line
    bool Isdna=(Type==Monomer_::DNA);
    int Len=Seq.size();
    Outf<<"ID   "<<ID<<"  standard; "
    	<<(Isdna? "DNA": "RNA")<<"; XXX; "<<Len<<" BP.\n";

    // The SwissProt/EMBL annotation is very rich and if the sequence
    // was read from another format then it cannot be reproduced easily.
    // We look for the "AC" and "DE" lines in the annotation and if
    // they are found then we dump the annotation as it is, and ignore
    // the Title field. Otherwise, it is assumed that the sequence
    // comes from another format and we copy the title and the
    // annotation as "CC" records to the output.
    if (Title.empty() && Annot.empty())
	Outf<<"CC   (no annotation)\n";
    else if (NPOS!=Annot.find("AC   ") && NPOS!=Annot.find("DE   "))
	    Outf<<Annot;    // looks like original SwissProt annotation
    else
    {
	swissembl_comment(Outf, Title); // convert non-SW title and annot
	swissembl_comment(Outf, Annot); // by prepending "CC" to each line
    }
    
    // print the SQ line
    Outf<<"SQ   Sequence  "<<Len<<" BP; ";
    int Acgtx[5];
    count_bases(Seq, Acgtx, Isdna);
    Outf<<Acgtx[0]<<" A; "<<Acgtx[1]<<" C; "<<Acgtx[2]<<" G; "
	    <<Acgtx[3]<<(Isdna? " T; ": " U; ")<<Acgtx[4]<<" other;\n";
    
    // print the sequence
    register int i;
    for (i=0; i<Len; i++)
    {
	if (i % 60 == 0) Outf<<"     ";
	Outf<<Seq[i];
	if ((i+1) % 60 == 0) Outf<<setw(10)<<(i+1)<<endl;
	else if ((i+1) % 10 == 0) Outf<<' ';
    }
    int Lastlinelen=i % 60, Lastspaces=(Lastlinelen-1)/10;
    if (i % 60 != 0)
	Outf<<setw(75-Lastlinelen-Lastspaces)<<Len<<endl;
    Outf<<"//\n";
    
    return OK;
}
// END of write_embl(Outf, ID, Title, Annot, Seq, Isdna)
	     
// ---- Checksums ----

// gcg_checksum(S): calculates the GCG sequence checksum.
int Bioseqio_::gcg_checksum(const string& S)
{
    int i, Cnt, Sum;
    char C;
    if (S.empty()) return(0);
    
    for (i=Sum=0, Cnt=1; i<S.size(); ++i, ++Cnt)
    {
	C=toupper(S[i]);
	// if (C=='~') C='.';
	Sum+=Cnt*(int)C; // counter*ASCII value of residue code
	if (Cnt==57) Cnt=0;
    }
    Sum %= 10000;
    return(Sum);
}
// END gcg_checksum(S)

// SwissProt/EMBL CRC-32 checksum calculation: the routines were
// kindly provided by Dr Elisabeth Gasteiger,
// Swiss Institute of Bioinformatics. Slightly after-edited
void Bioseqio_::BuildCRCTable(unsigned int CRCTable[256])
{
    static const unsigned long CRC32_POLYNOMIAL = 0xEDB88320L;
    
    register int i, j;
    register unsigned long crc;

    for ( i = 0; i <= 255 ; i++ )
    {
	crc = i;
	for ( j = 8 ; j > 0; j-- )
        {
            if ( crc & 1 )
        	crc = ( crc >> 1 ) ^ CRC32_POLYNOMIAL;
            else
        	crc >>= 1;
        }
	CRCTable[ i ] = crc;
    }
}

unsigned int Bioseqio_::crc_32 (const string& S)
{
    static unsigned int CRCTable[256];
    static bool Tablemade=false;
    
    register unsigned int crc = 0xFFFFFFFFL;
    const char *buffer=S.c_str();
    register unsigned int temp1, temp2, i=S.size();
    
    if (!Tablemade) 	// build table only once
    {
	BuildCRCTable(CRCTable);
	Tablemade=true;
    }
    while ( i-- )
    {
    	temp1 = ( crc >> 8 ) & 0x00FFFFFFL;
    	temp2 = CRCTable[ ( (int) crc ^ toupper(*buffer++) ) & 0xff ];
    	crc = temp1 ^ temp2;
    }
    return( crc );
}
// END of Swissprot/EMBL CRC-32 checksum

// ---- Auxiliaries ----

// error_str(), format_str(): return a const char ptr to
// a human-readable short description of the various enums. Static
const char *Bioseqio_::error_str(Errtype_ Err)
{
    switch (Err)
    {
	case OK: return "OK";
	case FILEERR: return "File error";
	case FORMATERR: return "Bad format";
	case TITLEERR: return "Title error";
	case IDERR: return "ID error";
	case ANNOTERR: return "Annotation error";
	case SEQERR: return "Sequence error";
	case TYPEERR: return "Seq type error";
	default: return "Error";
    }
}
// END of error_str(Err);

const char *Bioseqio_::format_str(Formatflags_ F)
{
    switch (F)
    {
	case ANYFORMAT: return "Any format (automatic)";
	case FASTA: return "FASTA";
	case PIR: return "PIR";
	case GCG: return "GCG";
	case SWISS: return "SwissProt";
	case EMBL: return "EMBL";
	default: return "Unknown format";
    }
}
// END of format_str(F)

// current_datestr(): returns the current date and time
// in a format required by GCG. Static hidden
const char *Bioseqio_::current_datestr()
{
    #define TSTAMPLEN 60
    static char Timestamp[TSTAMPLEN+1];
    time_t Tt;
    struct tm *Tm;
    
    // get the current date in a fancy format
    Tt=time(NULL);  // current system time
    Tm=localtime(&Tt);
    strftime(Timestamp, TSTAMPLEN, "%B 0.000000e+00, %Y %H:%M", Tm);
    return((const char *)Timestamp);
    #undef TSTAMPLEN
}
// END of current_datestr()

// mol_weight(S): calculates the molecular weight of the
// amino acid sequence S. Rounds to the nearest integer.
// To be used with SwissProt. Static hidden
int Bioseqio_::mol_weight(const string& S)
{
    // mol. wt. of amino acid monomers
    static const double AAWGTS[26]=
    {
	89.09, 132.61, 121.15, 133.10, 147.13,	// ABCDE
	165.19, 75.07, 155.16, 131.17, 0.00,	// FGHIJ
	146.19, 131.17, 149.21, 132.12, 0.00,	// KLMNO
	115.13, 146.15, 174.20, 105.09, 119.12, // PQRST
	0.00, 117.15, 204.23, 128.16, 181.19, 146.64	// UVWXYZ
    };
    static const double WATER=18.02;	// MW of water
    
    double Sum;
    int i, Len, Isum;
    char C;
    
    if (S.empty()) return(0);
    
    Len=S.size(); Sum=0.0;
    for (i=0; i<Len; i++)
    {
	C=toupper(S[i]);
	if (!isalpha(C)) C='X';
	Sum+=AAWGTS[C-'A'];
    }
    Sum-=(Len-1)*WATER;
    Isum=(int)Sum;
    if (Sum-Isum>=0.5) Isum++;	// round up
    return(Isum);
}
// END of mol_weight(S)

// count_bases(S, Acgtx, Isdna): counts the nucleotides in
// sequence S and puts the result in Acgtx
// [0]==A, [1]==C, [2]==G, [3]==T|U, [4]==X.
// Looks for T if Isdna==true, for U otherwise.
// Static
void Bioseqio_::count_bases(const string& S, int Acgtx[5], bool Isdna)
{
    int i;
    char C;
    string::const_iterator Sptr;
    
    for (i=0; i<5; i++) Acgtx[i]=0;
    
    for (Sptr=S.begin(); Sptr!=S.end(); Sptr++)
    {
	C=toupper(*Sptr);
	switch(C)
	{
	    case 'A': Acgtx[0]++; break;
	    case 'C': Acgtx[1]++; break;
	    case 'G': Acgtx[2]++; break;
	    case 'T': if (Isdna) Acgtx[3]++; else Acgtx[4]++; break;
	    case 'U': if (!Isdna) Acgtx[3]++; else Acgtx[4]++; break;
	    default: Acgtx[4]++; break;
	}
    }
}
// END of count_bases(S, Acgtx, Isdna)

// compare_basecount(S, Acgtx, Isdna): counts the bases in S
// (cf. count_bases(...)), and compares the result to what is
// contained in Acgtx. Returns true if the two match, false otherwise.
// Static
bool Bioseqio_::compare_basecount(const string& S, const int Acgtx[5], bool Isdna)
{
    static int Bases[5];
    
    count_bases(S, Bases, Isdna);
    for (int i=0; i<5; i++)
	if (Bases[i]!=Acgtx[i]) return false;
    return true;
}
// END of compare_basecount(S, Acgtx, Isdna)

// swissembl_comment(Outf, S): breaks up the (multi-line) string S
// in individual lines and prepends each line with the
// SwissProt/EMBL comment tag "CC   ". Prints the whole thing to Outf.
// Static hidden
void Bioseqio_::swissembl_comment(ostream& Outf, const string& S)
{
    static const int MAXWIDTH=75;
    
    int i, k, Len;
    char C;
    
    if (S.empty()) return;
    Len=S.size();
    
    for (i=0, k=1; i<Len; i++, k++)
    {
	if (k==1) Outf<<"CC   ";
	Outf<<(C=S[i]);
	if (C=='\n') k=0;    // reset at EOLN
	else if (k==MAXWIDTH)
	{
	    Outf<<endl;
	    k=0;    // new line comes
	}
    }
    if (k) Outf<<endl;
}
// END of swissembl_comment(Outf, S)

// ==== END OF METHODS Bioseqio.cc ====
