// ==== METHODS Gapseq.cc ====

// Class for storing gapped biological sequences.

// 10-Aug-1999. Andris Aszodi

// ---- COMPATIBILITY NOTES ----

// Uses the ANSI C++ standard string class.

// ==== MODULE HEADERS ====

#include"Gapseq.hh"
using namespace RazorBack;

// ==== Gapseq_ METHODS ====

// ---- Static init ----

const char Gapseq_::GAP='-';	    // usual gap
const char* Gapseq_::GAP3="---";
const char Gapseq_::GCG_GAP='.';    // GCG gap
const char* Gapseq_::GCG_GAP3="...";
const char Gapseq_::GCG_ENDGAP='~'; // GCG padding gap at the ends
const char* Gapseq_::GCG_ENDGAP3="~~~";

// ---- Access ----

// at_1(Idx), at_1(Idx, C):
// safe index operator for the Idx:th char in the sequence.
// at_1(NPOS) returns a gap character.
// The non-const version refuses to store characters not corresponding
// to the current type of the calling object.
// Throws Indexrangexc_ if Idx is NPOS or illegal.
// Returns true on success, false when C was illegal.
char Gapseq_::at_1(size_t Idx) const
{
    if (Idx==NPOS) return GAP;
    if (Idx>=length())
	throw Indexrangexc_("Gapseq_::at_1() const", Idx, length()-1);
    return seq()[Idx];
}
bool Gapseq_::at_1(size_t Idx, char C)
{
    if (is_legal(C))
    {
	if (Idx==NPOS || Idx>=length())
	    throw Indexrangexc_("Gapseq_::at_1()", Idx, length()-1);
	sequence()[Idx]=C;
	return true;
    }
    else
    {
	cerr<<"\n? Gapseq_::at_1("<<Idx<<", "<<C<<"): Illegal char\n";
	return false;
    }
}

// at_3(Idx), at_3(Idx, Ccc):
// safe index operator for getting and setting
// a monomer using the 3-letter abbreviations.
// The non-const version refuses to store abbreviations
// that are illegal.
// Returns true on success, false when Ccc was illegal.
const char* Gapseq_::at_3(size_t Idx) const
{
    if (Idx==NPOS) return GAP3;
    if (Idx>=length())
	throw Indexrangexc_("Gapseq_::at_3() const", Idx, length()-1);
    char C=seq()[Idx];
    return monomer().one_2_three(C, nocase());
}
bool Gapseq_::at_3(size_t Idx, const char* Ccc)
{
    if (!is_legal(Ccc))
    {
	cerr<<"\n? Gapseq_::at_3("<<Idx<<", \""<<Ccc<<"\"): Illegal 3-letter abbrev\n";
	return false;
    }
    else
    {
	if (Idx==NPOS || Idx>=length())
	    throw Indexrangexc_("Gapseq_::at_3()", Idx, length()-1);
	sequence()[Idx]=monomer().three_2_one(Ccc, nocase());
	return true;
    }
}
// END of at_3(...)

// gapped_sequence([Gcg]):
// create and return a string with the gap characters actually inserted.
// If Gcg==true (default false), then the funny GCG gap characters
// are used, ie. '~' for trailing gaps and '.' for internal gaps.
// gapped_sequence(Gseq, [Mt]):
// set the calling object to contain the gapped sequence Gseq where
// the gap positions are actually indicated by the gap characters.
// Here all gap characters are accepted, no matter where they are.
// The new monomer type can be forced by specifying Mt, if it is
// Monomer_::UNKNOWN (the default) then autodetection will be done.
string Gapseq_::gapped_sequence(bool Gcg) const 
{
    if (Gaplist.empty()) return seq();
    
    string Gseq;
    list<Gap_>::const_iterator Gptr;
    size_t Netidx=0, Gpos;
    char Gapsymbol;
    for (Gptr=Gaplist.begin(); Gptr!=Gaplist.end(); ++Gptr)
    {
	Gpos=Gptr->pos();
	if (Gcg)
	    Gapsymbol=(!Gpos || Gpos==NPOS)? GCG_ENDGAP: GCG_GAP;
	else Gapsymbol=GAP;
	
	if (Gpos==NPOS) Gpos=net_length();
	if (Gpos>0)
	    Gseq.append(seq(), Netidx, Gpos-Netidx);	// ungapped part
	Gseq.append(Gptr->len(), Gapsymbol);	// gap itself
	Netidx=Gpos;
    }
    if (Netidx<=net_length()-1)
	Gseq.append(seq(), Netidx, net_length()-Netidx);    // last ungapped
    return Gseq;
}

void Gapseq_::gapped_sequence(const string& Gseq, int Mtype)
{
    static const string GAPCHARS="-.~";    // normal plus GCG
    
    sequence()="";  // erase completely
    remove_gaps();
    
    size_t Aftergap, Beforegap, Ungappedlen, Gappedlen;
    size_t Grosspos=0, Netpos=0;
    while (Grosspos<Gseq.length())
    {
	Aftergap=Gseq.find_first_not_of(GAPCHARS, Grosspos);
	if (Aftergap==NPOS) Aftergap=Gseq.length();
	if (Aftergap==Grosspos)     // no gap at Grosspos
	{
	    // [Grosspos..Beforegap) is ungapped
	    Beforegap=Gseq.find_first_of(GAPCHARS, Grosspos);
	    if (Beforegap==NPOS) Beforegap=Gseq.length();
	    Ungappedlen=Beforegap-Grosspos;
	    sequence().append(Gseq, Grosspos, Ungappedlen);
	    Grosspos=Beforegap;
	    Netpos+=Ungappedlen;
	}
	else	// Aftergap>Grosspos
	{
	    // [Grosspos..Aftergap) is a gapped region
	    Gappedlen=Aftergap-Grosspos;
	    if (Aftergap==Gseq.length()) Netpos=NPOS;	// tail gap
	    Gaplist.push_back(Gap_(Netpos, Gappedlen)); // store gap info
	    Gaplen+=Gappedlen;	    // update total gap count
	    Grosspos=Aftergap;
	}
    }
    if (Mtype==Monomer_::UNKNOWN)
	Mtype=autodetect_type(seq());
    chg_monomer().set_alphabet(Mtype);
}
// END of gapped_sequence([Gseq])   

// ---- Legal issues ----

// is_legal(Mt, C): tests if C is legal if the monomertype
// of the calling object were Mt. This is a virtual fn so
// derived classes may add extra logic (such as gap characters)
// to the standard Monomer_ alphabet test.
// Hidden
bool Gapseq_::is_legal(int Mt, char C) const
{
    Monomer_ M(Mt);
    return (C==GAP || M.is_legal(C, nocase()));
}
// END of is_legal(Mt, C)

// ---- Gap manipulation ----

// insert_gaps(Idx, N):
// insert N gap characters before position [Idx].
// If Idx==NPOS, the gap will be placed after the sequence.
// If Idx is out-of-range, then Indexrangexc_ will be thrown.
// If there is already a gap before Idx, then the
// length of this existing gap will be changed to N.
// N must be non-0, if N==0 then nothing is done.
void Gapseq_::insert_gaps(size_t Idx, size_t N)
{
    if (Idx!=NPOS && Idx>=length())
	throw Indexrangexc_("Gapseq_::insert_gaps()", Idx, length()-1);
    if (!N)
    {
	cerr<<"\n? Gapseq_::insert_gaps("<<Idx<<", Len=0)\n";
	return;
    }
    
    // first gap?
    if (Gaplist.empty())
    {
	Gaplist.push_back(Gap_(Idx, N));
	Gaplen+=N;
	return;
    }
    
    list<Gap_>::iterator Gpos;

    // locate gap object before which new gap is inserted
    for (Gpos=Gaplist.begin(); Gpos!=Gaplist.end(); ++Gpos)
	if (Gpos->pos()>=Idx) break;
    
    // add new gap
    if (Gpos->pos()==Idx)	// there was already a gap before [Idx]
    {
	Gaplen-=Gpos->len();
	Gpos->len(N);   	// just change its length
    }
    else
	Gaplist.insert(Gpos, Gap_(Idx, N));	// add new gap
    Gaplen+=N;
}
// END of insert_gaps(Idx, N)

// remove_gaps(Idx):
// removes the gap characters before net position [Idx].
// Nothing is done if there were no gaps there.
// If Idx is out of range, Indexrangexc_ will be thrown.
void Gapseq_::remove_gaps(size_t Idx)
{
    if (Gaplist.empty()) return;    // ungapped
    
    if (Idx!=NPOS && Idx>=length())
	throw Indexrangexc_("Gapseq_::remove_gaps()", Idx, length()-1);

    // locate gap object
    list<Gap_>::iterator Gpos;
    for (Gpos=Gaplist.begin(); Gpos!=Gaplist.end(); ++Gpos)
	if (Gpos->pos()==Idx) break;
    if (Gpos!=Gaplist.end())	    // found
    {
	Gaplen-=Gpos->len();	// gap number goes down
	Gaplist.erase(Gpos);	// gap removed
    }
}
// END of remove_gaps(Idx)

// Net/gross position conversion:
// The "net" position of a monomer in the gapped sequence
// is the position it would have if there were no gaps.
// The "gross" position is its position with the gaps included.
// Both positions start with 0 in the best C tradition :-)
// For example in the gapped sequence "--AC-F-D" 
// the net position of 'F' is 2, the gross position is 5.

// net_pos(Gross): given the "gross" position Gross, the ungapped
// net position is returned. 
// If there is a gap at Gross then NPOS is returned.
// If Gross is illegal (>=gross_length()), then Indexrangexc_ is thrown.
size_t Gapseq_::net_pos(size_t Gross) const
{
    if (Gross>=gross_length())
	throw Indexrangexc_("Gapseq_::net_pos()", Gross, gross_length()-1);
    if (Gaplist.empty()) return Gross;
    
    list<Gap_>::const_iterator Gptr;
    size_t Net=0;
    for (Gptr=Gaplist.begin(); Gptr!=Gaplist.end(); ++Gptr)
    {
    	Net=Gptr->pos();
	if (Net==NPOS)	// tail gap
	{
	    if (Gross>=net_length()) Gross=NPOS;
	    break;
	}
	
	if (Gross<Net) break;
	else if (Gross<Net+Gptr->len())
	{
	    Gross=NPOS;
	    break;
	}
	else Gross-=Gptr->len();
    }
    return Gross;
}
// END of net_pos(Gross)

// gross_pos(Net): given the "net" position Net, the gapped
// gross position is returned (inverse of net_pos(Gross) above).
// If Net is illegal (>=net_length()), then Indexrangexc_ is thrown.
size_t Gapseq_::gross_pos(size_t Net) const
{
    if (Net>=net_length())
	throw Indexrangexc_("Gapseq_::gross_pos()", Net, net_length()-1);
    if (Gaplist.empty()) return Net;
    
    list<Gap_>::const_iterator Gptr;
    size_t Gapsum=0, Netpos;
    for (Gptr=Gaplist.begin(); Gptr!=Gaplist.end(); ++Gptr)
    {
	Netpos=Gptr->pos();
    	if (Netpos==NPOS || Netpos>Net) break;
	Gapsum+=Gptr->len();
    }
    return Net+Gapsum;
}
// END of gross_pos(Net)

// ---- I/O ----

// NOTE: the "new" ISO C++ standard iostream library has different types
// and class structure than the "old" iostream. Until all the compilers
// migrate to this (notable exception is the buggy SGI "new" iostream),
// we provide a workaround as follows.
#ifdef __USE_STD_IOSTREAM
#define IOS_IOSTATE ios_base::iostate
#else
#define IOS_IOSTATE int
#endif

// read_seq(Inf, Type, Format):
// tries to read a sequence from the input stream Inf.
// If Type is set to Monomer_::PROT, then only protein
// sequences will be read, if it is set to Monomer_::[DEOXY[RIBO]]NUCL,
// then only the corresponding nucleic acid sequences will be read,
// if Type is Monomer_::USER (default) or an OR-ed combination of PROT
// and *NUCL, then all types are considered and a polymer
// type autodetection is attempted. This is not foolproof.
// Format can be set to ANYFORMAT (default), in this case automatic
// file format detection is performed, or to 
// FASTA (protein or nucleic acid) or to PIR (proteins only).
// The other formats provided by the Bioseq_ class are not accepted
// because those cannot be used in multiple alignments.
// No read is performed and FORMATERR is returned when
// incompatible type and format are specified.
// Return value: OK or an error enum if something went wrong.
Bioseq_::Errtype_ Gapseq_::read_seq(istream& Inf, int Type, int Format)
{
    static const char* WARN="\n? Gapseq_::read_seq(): ";
    
    if (!Inf.good()) return FILEERR;	// no read from bad files
    if (Type==Monomer_::USER)
	Type=Monomer_::PROT|Monomer_::NUCL;
    Format=Format & (FASTA|PIR);    // only these are acceptable
    if (Format==UNKNOWN)    // was not FASTA or PIR
    {
	cerr<<WARN<<"Only FASTA and PIR formats are supported\n";
	return FORMATERR;
    }
    if (Format == PIR && !(Type & Monomer_::PROT))
    {
	cerr<<WARN<<"Incompatible Type and Format\n";
	return FORMATERR;
    }
    
    static string Tempid, Temptitle, Tempannot, Tempseq;  // input buffers
    streampos Filepos=Inf.tellg();
    IOS_IOSTATE Infstate=Inf.rdstate(); // save old input state
    Errtype_ Err=OK;

    // try PIR (proteins only)
    if ((Type & Monomer_::PROT) && (Format & PIR))
    {
	Err=read_pir(Inf, Tempid, Temptitle, Tempannot, Tempseq);
	if (Err==OK)    // successful
	{
	    if (clean_seq(Tempseq, Monomer_::PROT, nocase(), true))	// check chars, remove spaces etc.
	    {
		id(Tempid); title(Temptitle);
		annot()=Tempannot;
		chg_monomer().set_alphabet(Monomer_::PROT);
		gapped_sequence(Tempseq);   // make internal gap representation
		return OK;
	    }
	    else return SEQERR;
	}
    }
        
    // try FASTA (protein or nucleic acid)
    if (Format & FASTA)
    {
	Inf.clear(Infstate);    // reset stream state
	Inf.seekg(Filepos);     // go back to original position
	Err=read_fasta(Inf, Tempid, Temptitle, Tempseq);
	if (Err==OK)    // successful
	{
	    // Autotype will contain the narrowest typing
	    int Autotype=Type & autodetect_type(Tempseq);
	    if (Autotype==Monomer_::UNKNOWN)
		return TYPEERR;
	    if (clean_seq(Tempseq, Autotype, nocase(), true))	// check chars, remove spaces etc.
	    {
		id(Tempid); title(Temptitle);
		annot().erase();   // FASTA has no annotation
		chg_monomer().set_alphabet(Autotype);
		gapped_sequence(Tempseq);   // make internal gap representation
		return OK;
	    }
	    else return SEQERR;
	}
    }
    // no success
    Inf.seekg(Filepos);
    return Err;
}
// END of read_seq(Inf, Type, Format)

// write_seq(Outf, Format): writes the contents of the calling object
// to Outf according to the format specification in Format.
// Only FASTA and PIR formats are accepted, see comments to read_seq(..).
// Returns OK or an error enum if something went wrong.
Bioseq_::Errtype_ Gapseq_::write_seq(ostream& Outf, int Format) const
{
    if (!Outf.good()) return FILEERR;
    string Gapped=gapped_sequence();
    switch (Format)
    {
	case FASTA:
	    return write_fasta(Outf, id(), title(), Gapped);
	case PIR:
	    if (monomer().type()==Monomer_::PROT)
		return write_pir(Outf, id(), title(), annot(), Gapped);
	    else return FORMATERR;
	default:
	    cerr<<"\n? Gapseq_::write_seq(): Format not implemented\n";
	    return FORMATERR;
    }
}
// END of write_seq(Outf, Format)

// ==== END OF METHODS Gapseq.cc ====
