// ==== METHODS Monomer.cc ====

// Class for representing monomer alphabets for biomolecules.

// 10-Aug-1999. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <ctype.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ==== MODULE HEADERS ====

#include"Monomer.hh"
using namespace RazorBack;

// ==== PROTOTYPES ====

static int str_ncasecmp(const char *S1, const char *S2, size_t N);

// ==== METHODS ====

// ---- Access ----

// set_alphabet(Mt):
// sets the alphabet to one of the predefined monomer types as indicated by Mt.
// If Mt==USER, no action is taken and a warning is printed.
// If Mt==UNKNOWN, a complete reset is done.
void Monomer_::set_alphabet(int Mt)
{
    switch(Mt)
    {
	case PROT:
	    set_alphabet("AAlaBAsxCCysDAspEGluFPheGGlyHHisIIle\
KLysLLeuMMetNAsnPProQGlnRArgSSerTThrVValWTrpXXxxYTyrZGlx*End");
	break;
	case DNA:
	    set_alphabet("AAdeCCytGGuaTThyRPurYPyrXXxxNAny");
	break;
	case RNA:
	    set_alphabet("AAdeCCytGGuaUUraRPurYPyrXXxxNAny");
	break;
	case NUCL:
	    set_alphabet("AAdeCCytGGuaTThyUUraRPurYPyrXXxxNAny");
	break;
	case UNKNOWN:	// reset
	    Alphabet.clear();
	break;
	case USER:
	default:
	    cerr<<"\n? Monomer_::set_alphabet(Mt): Not predefined, \
		please use method set_alphabet(Codes)\n";
	return;
    }
    Mtype=(Type_)Mt;	// set_alphabet(Codes) has set this to USER, so correct here
}
// END of set_alphabet(Mt)

// set_alphabet(Codes):
// sets the alphabet to contain the abbreviations listed in Codes.
// The format of Codes: 4 chars per monomer type, first the 1-letter 
// code, then the 3-letter code like "AAlaBAsxCCys...".
// Nothing is done and warnings are printed if Codes is NULL, empty string
// or its length is not a multiple of 4.
// On success, the Mtype of the calling object is set to USER.
void Monomer_::set_alphabet(const char* Codes)
{
    // paranoia
    if (Codes==NULL || !*Codes)
    {
	cerr<<"\n? Monomer_::set_alphabet(\""<<Codes<<"\"): Empty argument\n";
	return;
    }
    size_t Codeslen=strlen(Codes);
    if ( Codeslen % 4 )
    {
	cerr<<"\n? Monomer_::set_alphabet(\""<<Codes<<"\"): Length="
		<<Codeslen<<" is not multiple of 4\n";
	return;
    }
    Codeslen/=4;
    
    // throw away old alphabet
    if (!Alphabet.empty())
	Alphabet.clear();
    Alphabet.reserve(Codeslen);
    
    // build up new alphabet
    const char* Codeptr;
    register size_t i;
    Abbrev_ Elem;
    for (i=0, Codeptr=Codes; i<Codeslen; ++i, Codeptr+=4)
    {
	Elem.set_codes(*Codeptr, Codeptr+1);
	Alphabet.push_back(Elem);
    }
    Mtype=USER;
}
// END of set_alphabet(Codes)

// ---- Alphabet comparisons ----

// ==: two monomer objects are considered equal either if both
// are of type USER and their alphabets are the same
// (case-insensitive match is enough),
// or if they have the same type. In particular,
// NUCL and RNA or NUCL and DNA are not equal.
bool Monomer_::operator==(const Monomer_& M) const
{
    if (this==&M) return true;
    if (type()==USER && M.type()==USER)
	return (contains(M, true) && M.contains(*this, true));
    else
	return (type()==M.type());
}
// END of ==

// contains(M, Nocase):
// returns true if all the monomers in M can be found in
// the calling object, false otherwise.
bool Monomer_::contains(const Monomer_& M, bool Nocase) const
{
    Aciter_ Ai;
    for (Ai=M.Alphabet.begin(); Ai!=M.Alphabet.end(); ++Ai)
    {
    	if (find(Ai->code_1(), Nocase)!=Alphabet.end() // M's 1-letter not in this
		|| find(Ai->code_3(), Nocase)!=Alphabet.end()) // M's 3-letter not in this
	    return false;
    }
    return true;
}
// END of contains(M, Nocase)

// ---- find ----

// find(C, Nocase), find(Ccc, Nocase):
// tries to find the 1-letter (C) or 3-letter (Ccc) abbreviation
// in the calling object.
// If Nocase==true, then a case-insensitive match is enough.
// Returns a const iterator to the abbreviation if found,
// or Alphabet.end() if not found.
Monomer_::Aciter_ Monomer_::find(char C, bool Nocase) const
{
    Aciter_ Ai;
    for (Ai=Alphabet.begin(); Ai!=Alphabet.end(); ++Ai)
	if (!Nocase && Ai->code_1()==C
	    || Nocase && toupper(Ai->code_1())==toupper(C))
	    	return Ai;
    return Alphabet.end();  // not found
}
Monomer_::Aciter_ Monomer_::find(const char* Ccc, bool Nocase) const
{
    Aciter_ Ai;
    for (Ai=Alphabet.begin(); Ai!=Alphabet.end(); ++Ai)
	if (!Nocase && !strncmp(Ai->code_3(), Ccc, 3)
	    || Nocase && !str_ncasecmp(Ai->code_3(), Ccc, 3))
	    	return Ai;
    return Alphabet.end();  // not found
}
// END of find(...)

// str_ncasecmp(S1, S2, N):
// implements the BSD function strncasecmp that gets switched off
// on SGI-s by the _POSIX_C_SOURCE macro that is needed for
// thread-safe compilation. :-(
int str_ncasecmp(const char *S1, const char *S2, size_t N)
{
    register size_t i;
    register char X1, X2;
    register const char *P1, *P2;
    
    if (!N) return 0;
    if (S1==NULL && S2==NULL) return 0;
    if (S1==NULL) return 1;
    if (S2==NULL) return -1;
    
    for (i=0, P1=S1, P2=S2; i<N; ++i, ++P1, ++P2)
    {
	if (!*P1 && !*P2) return 0; /* both were shorter than N and == */
	if (!*P1) return 1; /* S2 is longer */
	if (!*P2) return -1;	/* S1 is longer */
	X1=toupper(*P1);
	X2=toupper(*P2);
	if (X1==X2) continue;
	return (X1<X2? 1: -1);	/* mismatch */
    }
    return 0;
}
// END of str_ncasecmp(S1, S2, N)

// ==== END OF METHODS Monomer.cc ====   
