#ifndef LU_HEADER
#define LU_HEADER

// ==== HEADER Lu.hh ====

/**
 * \file Lu.hh
 * LU-decomposition and linear equation solver
 * routines for square matrices.
 */

// 23-Nov-1999. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stdlib.h>

// ==== MODULE HEADERS ====

#include"Vector.hh"
#include"Rectmat.hh"
#include "Rangedimexc.hh"
#include "Squarexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Lu_: a class for doing LU-decomposition. Can be asked to
 * calculate determinants and solve linear equation systems.
 */
class Lu_
{
    // data
    private:
    
    static const double LU_EPSILON; ///< a very small number
    
    Rmatdbl_ Lu;	///< holds the decomposition
    vector<unsigned int> Perm;	///< permutation vector
    int Psign;	///< permutation sign
    
    // methods
    public:
    
    /// Inits to an N x N problem (default 3).
    explicit Lu_(unsigned int N=3): Lu(N), Perm(N, 0), Psign(0) {}
    
    /// Inits with the decomposition of the square matrix A.
    explicit Lu_(const Rmatdbl_& A): Lu(A.rno()), Perm(A.rno(), 0), Psign(0)
    {
	decomp(A);
    }
    
    /**
    * decomp(A): performs an LU-decomposition on the square matrix A.
    * Return value: the permutation sign (0 if A was singular).
    * Throws Squarexc_ if A was not NxN.
    */
    int decomp(const Rmatdbl_& A);
    
    /**
    * det(): calculates the determinant from the decomposition results
    * stored inside. It is assumed that decomp() has been called beforehand.
    */
    double det() const;
    
    /**
    * solve(B): solves the linear equation A*x=B and returns x.
    * The calling object is supposed to have been primed with
    * decomp(A) before this call. B is the "right-hand-side"
    * vector which is preserved. 
    * Throws a Dimexc_ exception if B's dimension does not match A's.
    */
    Vecdbl_ solve(const Vecdbl_& B) const;
    
    /**
    * lineq(A, B, X, Maxit): solves the nxn linear equation A*X=B and returns X.
    * Iterative improvement is done a la Num. Recipes (max. Maxit times, 
    * default 0).
    * This routine should be used if only one right-hand side vector B
    * is present. For several right-hand sides, use the decomposition
    * and solver routines separately. 
    * Return value: 1 means OK, 0 means A was singular.
    * Throws Squarexc_ if A was not NxN, Dimexc_ on dim mismatches.
    */
    int lineq(const Rmatdbl_& A, const Vecdbl_& B, Vecdbl_& X, unsigned int Maxit=0);
    
};
// END OF CLASS Lu_

} // RazorBack

// ==== END OF HEADER Lu.hh ====

#endif	// LU_HEADER
