#ifndef RSYMDIAG_HEADER
#define RSYMDIAG_HEADER

// ==== HEADER Rsymdiag.hh ====

/**
 * \file Rsymdiag.hh
 * Real symmetric matrix diagonalisation class (double precision).
 * Handles special cases such as eigenvalues only, all eigenvalues and eigenvectors, 
 * all eigenvalues/some eigenvectors, and selected eigenvalue/eigenvector pairs.
 * Implementation is based on the Matrix Computations book by Golub and Van Loan.
 */

// 19-Nov-2001, Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <math.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <list>
using namespace std;
#else
#include <list.h>
#endif

// ==== MODULE HEADERS ====

#include"Vector.hh"
#include"Rectmat.hh"
#include"Symmat.hh"
#include "Fparch.hh"

// ==== CLASSES ====

namespace RazorBack { 

/**
 * Class Rsymdiag_: 
 * an object from this class can be initialised with a real symmetric matrix
 * and its methods allow various diagonalisation tasks to be carried out,
 * such as eigenvalues only, all eigenvalues and eigenvectors, 
 * all eigenvalues/some eigenvectors, and selected eigenvalue/eigenvector pairs.
 */
class Rsymdiag_
{
    // constants
    private:
	    
    static const Fparch_<double> Fpdbl; ///< double-prec arithmetics mach constants
    static const Fparch_<float> Fpflo;	///< single-prec arithmetics mach constants
    static const double EPSILON; ///< error tolerance
    
    // helper classlets
    private:
	    
    /**
     * Eigidx_: object for eigenvalue sorting where the eigenvalue
     * remembers where it came from.
     */
    struct Eigidx_
    {
	double Eig;	///< an eigenvalue along with its
	size_t Idx;	///< original position

	friend bool operator<(const Eigidx_& E1, const Eigidx_& E2)
	{ return (E1.Eig>E2.Eig); } // use > to get descending order!
    };

    struct Evaloc_;
    friend struct Evaloc_;
    /**
     * Evaloc_: object for keeping track of intervals
     * containing eigenvalues during the bisection process.
     * This classlet is a friend of the enclosing class because
     * it needs to access Rsymdiag_::EPSILON.
     */
    struct Evaloc_
    {
	double Low, Up; ///< interval boundaries
	double Eval;	///< eigenvalue estimate
	size_t Lowno, Upno, No;  ///< #eigenvalues below Low, Up, and in interval
	
	/// init
	Evaloc_(double L, size_t Ln, double U, size_t Un):
		Low(L), Lowno(Ln), Up(U), Upno(Un), No(Un-Ln), Eval(0.0) {}
	
	/**
	 * bracket_loose: returns true if the relative width
	 * of the interval is larger than machine precision.
	 */
	bool bracket_loose() const
	{
	    return fabs(Up-Low)>EPSILON*(fabs(Up)+fabs(Low));
	}
    };

    // data
    private:
	    
    size_t Size;    ///< size of the matrices (problem dimension)
    Rmatdbl_ Q; ///< Householder transformation matrix
    Vecdbl_ Diag, Subdiag;  ///< tridiagonalised matrix
    Vecdbl_ Beta, Gamma;    ///< decomposition vectors for inverse iteration
    bool Tridiagdone,	///< flag to indicate if tridiagonalisation is done
	    Qdone,  	///< transformation matrix for eigenvectors is done
    	    QLdone;     ///< QL done (destructive!)
    
    // methods
    public:
	    
    /// Inits to an empty object.
    Rsymdiag_(): Size(0), Tridiagdone(false), Qdone(false), QLdone(false) {}
    
    /// Inits to diagonalise Mat.
    explicit Rsymdiag_(const Smatdbl_& Mat) { re_init(Mat); }
    
    /**
    * re_init(Mat):
    * reinitialises the calling object with the real symmetric matrix Mat.
    * Returns the size.
    */
    size_t re_init(const Smatdbl_& Mat);
    
    /// Returns the size of the matrix being diagonalised.
    size_t size() const { return Size; }
    
    // All-eigenvalue methods
    // NOTE: these methods are to be used unless you need
    // approx. less than one-fourth of the eigenvalues.
    
    /**
    * all_evalevec(Eval, Evec, Sortevals):
    * Produces all eigenvalues in Eval and the eigenvectors in Evec (columns).
    * The sizes of these are adjusted silently if necessary.
    * if Sortevals is true (the default) then eigenvalues (and the
    * corresponding eigenvectors) are sorted in decreasing order.
    * Return value: no. of dimensions.
    * Throws Rsymdiagexc_ if the calling object was not re-initialised
    * or if the iteration no. was exceeded in the implicit QL step.
    */
    int all_evalevec(Vecdbl_& Eval, Rmatdbl_& Evec, bool Sortevals=true);
    
    /**
    * all_eval(Eval, Sortevals):
    * Produces all eigenvalues in Eval (size adjusted silently).
    * if Sortevals is true (the default) then the eigenvalues 
    * are sorted in decreasing order.
    * Return value: no. of dimensions.
    * Throws Rsymdiagexc_ if the calling object was not re-initialised
    * or if the iteration no. was exceeded in the implicit QL step.
    */
    int all_eval(Vecdbl_& Eval, bool Sortevals=true);
    
    // Selected eigenvalue methods
    // NOTE: these methods provide a speedup only if you need
    // approx. less than one-fourth of the eigenvalues.
    
    /**
    * range_evalevec(Low, Up, Eval, Evec):
    * Finds all eigenvalues within the range [Low, Up],
    * and puts them into Eval in ascending order by default,
    * or in descending order if Descend==true.
    * The corresponding eigenvectors are put into the columns of Evec.
    * Return value: the number of eigenvalues found.
    * Throws Rsymdiagexc_ if a reinit is needed.
    */
    int range_evalevec(double Low, double Up,
	    Vecdbl_& Eval, Rmatdbl_& Evec, bool Descend=false);
    
    /**
    * range_eval(Low, Up, Eval, Descent):
    * Finds all eigenvalues within the range [Low, Up],
    * and puts them into Eval in ascending order by default
    * or in descending order if Descend==true.
    * Return value: the number of eigenvalues found,
    * or -1 if reinit is needed.
    */
    int range_eval(double Low, double Up,
	    Vecdbl_& Eval, bool Descend=false);
    
    /**
    * largest_evalevec(N, Eval, Evec, Descend):
    * calculates the N largest eigenvalues and the corresponding eigenvectors.
    * Returns the number of eigenvalues found.
    * Throws Argexc_ if N is not in the range [1..size()].
    * Throws Rsymdiagexc_ if a reinit is needed.
    * See range_evalevec(..) for a description of the other variables.
    */
    int largest_evalevec(size_t N, Vecdbl_& Eval,
	    Rmatdbl_& Evec, bool Descend=false);

    /**
    * largest_eval(N, Eval, Descend):
    * calculates the N largest eigenvalues.
    * Returns the number of eigenvalues found.
    * Throws Argexc_ if N is not in the range [1..size()].
    * Throws Rsymdiagexc_ if a reinit is needed.
    * See range_eval(..) for a description of the other variables.
    */
    int largest_eval(size_t N, Vecdbl_& Eval, bool Descend=false);

    /**
    * evalevec_above(Thr, Eval, Evec, Descend):
    * calculates the largest eigenvalues and the corresponding eigenvectors
    * above the threshold Thr. Use for getting all positive eigenvalues for example.
    * Returns the number of eigenvalues found.
    * This can be 0 if Thr is set too high.
    * Throws Rsymdiagexc_ if a reinit is needed.
    * See range_evalevec(..) for a description of the other variables.
    */
    int evalevec_above(double Thr, Vecdbl_& Eval,
	    Rmatdbl_& Evec, bool Descend=false);

    /**
    * eval_above(Thr, Eval, Descend):
    * calculates the largest eigenvalues above the threshold Thr.
    * Use for getting all positive eigenvalues for example.
    * Returns the number of eigenvalues found.
    * This can be 0 if Thr is set too high.
    * Throws Rsymdiagexc_ if a reinit is needed.
    * See range_eval(..) for a description of the other variables.
    */
    int eval_above(double Thr, Vecdbl_& Eval, bool Descend=false);

    /**
    * gershgorin_limits(Low, Up):
    * calculates the limits of the range [Low,Up] that contains
    * all eigenvalues of the input matrix.
    * Note that this method performs a Housholder tridiagonalisation
    * but the transform matrix is not accumulated, you would not get
    * eigenvectors without calling re_init(..) after calling this.
    */
    void gershgorin_limits(double& Low, double& Up);
    
    // hidden methods
    protected:
	    
    void housholder_tridiag(bool Wantevec);
    int implicit_ql(int Itno, bool Wantevec);
    
    void largest_range(size_t N, double& Low, double& Up);
    size_t sturm_seq(double X) const;
    double sturm_newton(double X) const;
    void bisect_evals(list<Evaloc_>& Evalist,
	    list<Evaloc_>::iterator Iv);
    int inverse_iter(double Eval, Vecdbl_& Evec, double& Tol);
    void inverse_iter(double Eval, size_t Multiple, Rmatdbl_& Evecs);
    static void gram_schmidt(Rmatdbl_& Mat);
    static double subspace_diff(const Rmatdbl_& V, const Rmatdbl_& W);

    int tridiag_decomp(const Vecdbl_& Diagshift);
    void tridiag_solve(const Vecdbl_& Rhs, Vecdbl_& X) const;

    static void random_vector(Vecdbl_& Vec);
    static void random_matrix(Rmatdbl_& Mat);
    static double round_tozero(double X)
    { return ((fabs(X)<EPSILON)? 0.0: X); }
};
// END OF CLASS Rsymdiag_

/**
 * An Rsymdiagexc_ exception can be thrown if something went wrong.
 * Stores an error description string only.
 */
class Rsymdiagexc_: public Utilsexc_
{
    public:
    explicit Rsymdiagexc_(const char *Thnm, const char *Msg, bool F=false):
	    Utilsexc_(Thnm, F, Msg) {}
    private:
    Rsymdiagexc_();	// no default ctor
};
// END OF CLASS Rsymdiagexc_

} // RazorBack

// ==== END OF HEADER Rsymdiag.hh ====

#endif	    // RSYMDIAG_HEADER
