// ==== FUNCTIONS Lu.ccc ====

// LU-decomposition and linear equation solver
// routines for square matrices.

// 8-Oct-2001. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <math.h>
#include <limits.h>
#include <float.h>

// ==== MODULE HEADER ==== 

#include"Lu.hh"
using namespace RazorBack;

// ==== STATIC INIT ====

#ifdef FLT_MIN
const double Lu_::LU_EPSILON=10.0*FLT_MIN;
#else
const double Lu_::LU_EPSILON=1.0e-30;
#endif

// ==== Lu_ METHODS ====

// ---- Decomposition ----

// decomp(A): performs an LU-decomposition on the square matrix A.
// Return value: the permutation sign (0 if A was singular).
// Throws Squarexc_ if A was not NxN.
int Lu_::decomp(const Rmatdbl_& A)
{
    if (A.rno()!=A.cno())
	throw Squarexc_("Lu_::decomp(A)", A.rno(),A.cno());  // must be NxN
    
    register unsigned int i, j, k, imax, n=A.rno();
    register double Large, Pivot, Tmp, Tmp2;
        
    // get implicit scaling: if a row contains 0-s only,
    // then the matrix is singular which will be indicated
    // by setting Psign=0. Precision is controlled by
    // the constant LU_EPSILON
    vector<double> Scal(n, 0.0);
    Psign=1;
    Lu=A;
    for (i=0; Psign && i<n; i++)
    {
	Large=0.0;
	for (j=0; j<n; j++)
	    if ((Tmp=fabs(Lu[i][j]))>Large) Large=Tmp;
	if (Large<LU_EPSILON)	// (almost) singular
	{ Psign=0; break; }
	Scal[i]=1.0/Large;
    }
    
    // loop over columns
    vector<double> Tmpa(n);
    Perm=vector<unsigned int>(n,0);
    for (j=0; Psign && j<n; j++)
    {
	for (i=0; i<j; i++)
	{
	    Tmp=Lu[i][j];
	    for (k=0; k<i; k++) Tmp-=Lu[i][k]*Lu[k][j];
	    Lu[i][j]=Tmp;
	}
	
	// find largest pivot
	Large=0.0;
	for (i=j; i<n; i++)
	{
	    Tmp=Lu[i][j];
	    for (k=0; k<j; k++) Tmp-=Lu[i][k]*Lu[k][j];
	    Lu[i][j]=Tmp;
	    
	    if ((Tmp2=Scal[i]*fabs(Tmp))>=Large)    // best so far
	    { Large=Tmp2; imax=i; }
	}
	
	// interchange rows?
	if (j!=imax)
	{
	    copy(Lu[imax], Lu[imax]+n, Tmpa.begin());	// row interchange with STL
	    copy(Lu[j], Lu[j]+n, Lu[imax]);
	    copy(Tmpa.begin(), Tmpa.end(), Lu[j]);
	    
	    Psign*=(-1);    // parity change
	    Scal[imax]=Scal[j];
	}
	Perm[j]=imax;
	
	// get the pivot
	Pivot=Lu[j][j];
	if (fabs(Pivot)<LU_EPSILON)	// singularity
	{ Psign=0; break; }
	
	// divide by the pivot
	if (j<n-1)
	    for (i=j+1; i<n; i++) Lu[i][j]/=Pivot;
    }	// for j
    
    return(Psign);
}
// END of decomp(A)

// det(): calculates the determinant from the decomposition results
// stored inside. It is assumed that decomp() has been called beforehand.
double Lu_::det() const
{
    if (!Psign) return(0.0);	// matrix is singular
    
    register unsigned int i;
    register double Det=Psign;
    
    // the "sum-of-logarithms" trick is NOT used here
    for (i=0; i<Lu.rno(); i++) Det*=Lu[i][i];
    return(Det);
}
// END of det()

// ---- Solution ----

// solve(B): solves the linear equation A*x=B and returns x.
// The calling object is supposed to have been primed with
// decomp(A) before this call. B is the "right-hand-side"
// vector which is preserved. 
// Throws a Dimexc_ exception if B's dimension does not match A's.
Vecdbl_ Lu_::solve(const Vecdbl_& B) const
{
    register unsigned int n=Lu.rno();

    if (n!=B.dim())
    	throw Dimexc_("Lu_::solve()", n, B.dim());
    
    register int i;
    register unsigned int j, ip;
    register double Tmp;
    Vecdbl_ X(B);

    // permute forward
    for (i=0; i<n-1; i++)
	if ((ip=Perm[i])!=i)
	{ Tmp=X[ip]; X[ip]=X[i]; X[i]=Tmp; }
    
    // forward substitution
    for (i=0; i<n; i++)
    {
	Tmp=X[i];
	for (j=0; j<i; j++) Tmp-=Lu[i][j]*X[j];
	X[i]=Tmp;
    }
    
    // back substitution
    for (i=n-1; i>=0; i--)
    {
	Tmp=X[i];
	for (j=i+1; j<n; j++) Tmp-=Lu[i][j]*X[j];
	X[i]=Tmp/Lu[i][i];
    }
    return(X);
}
// END of solve(B)

// lineq(A, B, X, Maxit): solves the nxn linear equation A*X=B and returns X.
// Iterative improvement is done a la Num. Recipes (max. Maxit times, 
// default 0).
// This routine should be used if only one right-hand side vector B
// is present. For several right-hand sides, use the decomposition
// and solver routines separately. 
// Return value: 1 means OK, 0 means A was singular.
// Throws Squarexc_ if A was not NxN, Dimexc_ on dim mismatches.
int Lu_::lineq(const Rmatdbl_& A, const Vecdbl_& B, Vecdbl_& X, unsigned int Maxit)
{
    if (A.rno()!=A.cno())
	throw Squarexc_("Lu_::lineq(A,..,..,..)", A.rno(),A.cno());  // must be NxN
    
    // preparation
    unsigned int n=A.rno();
    if (n!=B.dim())
    	throw Dimexc_("Lu_::lineq(A,B,..,..)", n, B.dim());
    
    // LU-decomposition
    decomp(A);
    if (!Psign)
    {
	cerr<<"?\n Lu_::lineq(): matrix is singular\n";
	return(0);
    }
    
    // solution (sets the dimension of X)
    X=solve(B);
    
    // iterative improvement
    if (Maxit)
    {
	Vecdbl_ Dx(n);
	for (unsigned int i=0; i<Maxit; i++)
	{
	    Dx=solve(A*X-B);
	    if (Dx.vec_len()<LU_EPSILON) break;
	    else X-=Dx;
	}
    }
    
    // OK, result has been calculated at last
    return(1);
}
// END of lineq(A, B, X, Maxit)

// ==== END OF FUNCTIONS Lu.cc ====

