// ==== METHODS Rsymdiag.cc ====

// Real symmetric matrix diagonalisation class (double precision).

// 19-Nov-2001, Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
#include <algorithm>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#include <algo.h>
#endif

// ==== MODULE HEADERS ====

#include"Rsymdiag.hh"
using namespace RazorBack;

// ==== STATIC INIT ====

const Fparch_<double> Rsymdiag_::Fpdbl;
const Fparch_<float> Rsymdiag_::Fpflo;
const double Rsymdiag_::EPSILON=10.0*Rsymdiag_::Fpdbl.eps();

// ==== METHODS ====

// Implementation notes:-
// The full matrix diagonalisation is performed by
// Housholder tridiagonalisation followed by QL-iteration
// with implicit shifts. 
// These algorithms are available in all FORTRAN
// linear algebra packages.

// The selected eigenvalues/eigenvectors methods
// also use the Housholder tridiagonalisation step.
// Isolation of the eigenvalues of the symmetric
// tridiagonal matrix is done via bisection and Newton steps
// using my recursive algorithm based on the theory
// presented by Golub and van Loan. Eigenvectors are found
// by inverse iteration with separate decomposition and solution steps.
// Inverse iteration for multiple eigenvalues is my invention:
// a set of orthonormal vectors are subjected to inverse iteration
// and the stop criterion is when the new vector set is completely
// embedded into the old set (so that they live in the same
// invariant subspace).

// ---- Setup ----

// re_init(Mat):
// reinitialises the calling object with the real symmetric matrix Mat.
// Returns the size.
size_t Rsymdiag_::re_init(const Smatdbl_& Mat)
{
    Size=Mat.rno();
    Q=Mat;
    Diag.dim(Size); Subdiag.dim(Size);
    Beta.dim(Size); Gamma.dim(Size);
    Tridiagdone=false; Qdone=false; QLdone=false;
    return Size;
}
// END of re_init(Mat)

// ---- All eigenvalues ----

// all_evalevec(Eval, Evec, Sortevals):
// Produces all eigenvalues in Eval and the eigenvectors in Evec (columns).
// The sizes of these are adjusted silently if necessary.
// if Sortevals is true (the default) then eigenvalues (and the
// corresponding eigenvectors) are sorted in decreasing order.
// Return value: no. of dimensions.
// Throws Rsymdiagexc_ if the calling object was not re-initialised
// or if the iteration no. was exceeded in the implicit QL step.
int Rsymdiag_::all_evalevec(Vecdbl_& Eval, Rmatdbl_& Evec, bool Sortevals)
{
    static const int ITERNO=100;    // max. number of iterations
    
    if (!size() || QLdone || Tridiagdone && !Qdone)
	throw Rsymdiagexc_("Rsymdiag_::all_evalevec()","Please (re)init first");
    
    // Size checks. Eval and Evec may not have the right dimensions,
    // if this happens then the size is adjusted.
    Eval.dim(size()); Evec.size(size(), size());
    
    // Housholder tridiagonalisation
    // prepare for eigenvectors
    housholder_tridiag(true);

    // apply shifted QL-transforms: get eigenvalues in Diag,
    // eigenvectors in Q's columns
    if (implicit_ql(ITERNO, true))
	throw Rsymdiagexc_("Rsymdiag_::all_evalevec()","Iter limit exceeded");

    // Sort eigenvalues in decreasing order. The unsorted
    // eigenvalues are copied from Diag into a vector
    // of Eigidx_ structures, sorting is STL
    register size_t i;
    if (Sortevals)
    {
	vector<Eigidx_> Evs(Size);
	
	for (i=0; i<Size; i++)
	{
	    Evs[i].Eig=round_tozero(Diag[i]);  // small eigenvalues are rounded to 0.0
	    Evs[i].Idx=i;	// store index as well
	}
	sort(Evs.begin(), Evs.end());	// STL quicksort

	// permute eigenvectors according to the new order
	// of the eigenvalues and copy them into the columns
	// of the Evec matrix. Also copy eigenvalues to Eval
	register size_t k;
	for (i=0; i<Size; i++)
	{
	    Eval[i]=Evs[i].Eig;	// copy back eigenvalues
	    k=Evs[i].Idx;
	    Evec.col(Q.col(k), i);	// copy eigenvectors
	}
    }
    else    // no sorting
    {
	for (i=0; i<Size; i++)
	    Eval[i]=round_tozero(Diag[i]);
	Evec=Q;
    }
    return Size;
}
// END of all_evalevec(Eval, Evec, Sortevals)

// all_eval(Eval, Sortevals):
// Produces all eigenvalues in Eval (size adjusted silently).
// if Sortevals is true (the default) then the eigenvalues 
// are sorted in decreasing order.
// Return value: -2 if iteration limit was exceeded,
// -1 if reinit is needed, no. of dimensions otherwise.
int Rsymdiag_::all_eval(Vecdbl_& Eval, bool Sortevals)
{
    static const int ITERNO=100;    // max. number of iterations
    
    if (!size() || QLdone)
	throw Rsymdiagexc_("Rsymdiag_::all_eval()","Please (re)init first");
    
    // Size check
    Eval.dim(size());
    
    // Housholder tridiagonalisation
    // don't prepare for eigenvectors
    housholder_tridiag(false);

    // apply shifted QL-transforms: get eigenvalues in Diag
    if (implicit_ql(ITERNO, false))
	throw Rsymdiagexc_("Rsymdiag_::all_eval()","Iter limit exceeded");

    // Sort eigenvalues in decreasing order in Diag
    register size_t i;
    if (Sortevals)
	sort(Diag.begin(), Diag.end(), greater<double>());
    for (i=0; i<Size; i++)
	Eval[i]=round_tozero(Diag[i]);
    return Size;
}
// END of all_eval(Eval, Sortevals)

// housholder_tridiag(Wantevec):
// Housholder tridiagonalisation of the symmetric matrix put in Q.
// The main diagonal of the tridiagonal output is put into Diag,
// the subdiagonal into Subdiag with Subdiag[0]==0.0. On return, Q contains
// the cumulated Housholder rotation matrix.
// If Wantevec==true, then preparations are made for eigenvector
// calculation; if false, then runtime goes down by a factor of 2
// for large matrices. Price: you get eigenvalues only.
// Hidden
void Rsymdiag_::housholder_tridiag(bool Wantevec)
{
    if (Tridiagdone)
	return;
    
    register int l, k, j, i, n=size();	// get size
    register double Scale, InvScale, hh, h, Invh, g, f;

    for (i=n-1; i>=1; --i)
    {
	l=i-1;
	h=Scale=0.0;
	if (l >= 1)
	{
	    for (k=0; k<=l; ++k)
		Scale += fabs(Q[i][k]);
	    if (Scale < EPSILON)
		Subdiag[i]=Q[i][l];
	    else
	    {
		InvScale=1.0/Scale;
		for (k=0; k<=l; ++k)
		{
		    Q[i][k] *= InvScale;
		    h += Q[i][k]*Q[i][k];
		}
		
		f=Q[i][l];
		g = (round_tozero(f)>0.0) ? -sqrt(h) : sqrt(h);
		Subdiag[i]=Scale*g;
		h -= f*g;
		Q[i][l]=f-g;
		f=0.0;
		
		Invh=1.0/h;
		for (j=0; j<=l; ++j)
		{
		    if (Wantevec)
		    	Q[j][i]=Q[i][j]*Invh;
		    g=0.0;
		    for (k=0; k<=j; ++k)
			g += Q[j][k]*Q[i][k];
		    for (k=j+1; k<=l; ++k)
			g += Q[k][j]*Q[i][k];
		    Subdiag[j]=g*Invh;
		    f += Subdiag[j]*Q[i][j];
		}
		
		hh=f/(2*h);
		for (j=0; j<=l; ++j)
		{
		    f=Q[i][j];
		    Subdiag[j]=g=Subdiag[j]-hh*f;
		    for (k=0; k<=j; ++k)
			Q[j][k] -= (f*Subdiag[k]+g*Q[i][k]);
		}
	    }
	}
	else
	    Subdiag[i]=Q[i][l];
	Diag[i]=h;
    }
    
    if (Wantevec)
    	Diag[0]=Subdiag[0]=0.0;
    
    for (i=0; i<n; ++i)
    {
	if (Wantevec)
	{
	    l=i-1;
	    if (round_tozero(Diag[i])!=0.0) // skips i==0, l==-1
	    {	
		for (j=0; j<=l; ++j)
		{
		    g=0.0;
		    for (k=0; k<=l; ++k)
			g += Q[i][k]*Q[k][j];
		    for (k=0; k<=l; ++k)
			Q[k][j] -= g*Q[k][i];
		}
	    }
	}
	
	Diag[i]=(fabs(Q[i][i])<EPSILON)? 0.0: Q[i][i];
	
	if (Wantevec)
	{
	    Q[i][i]=1.0;
	    for (j=0; j<=l; ++j)
		Q[j][i]=Q[i][j]=0.0;
	}
    }
    
    // tell the other methods what has been done
    // the eval-only methods do not need the Q matrix
    Tridiagdone=true; Qdone=Wantevec;
}
// END of housholder_tridiag(Wantevec)

// implicit_ql(Itno, Wantevec):
// QL algorithm with implicit shifts on tridiagonal matrices
// to be invoked after housholder_tridiag(..).
// Takes the tridiagonal form from Diag and Subdiag,
// together with the Housholder transformation in Q.
// On output, Diag contains the eigenvalues and Q the
// eigenvectors, with the k-th _column_ corresponding to the k-th eigenvalue.
// If Wantevec==false, then no eigenvectors are calculated,
// Q is left untouched, and the runtime will be 30N^2 instead of 3N^3.
// Itno supplies the maximum allowable no. of iterations. 
// Return value: 0 if OK, >=1 if iteration limit has been exceeded.
// Hidden
int Rsymdiag_::implicit_ql(int Itno, bool Wantevec)
{
    // set right here to indicate that implicit_ql(..)
    // always screws up Diag, and Q as well if Wantevec==true
    QLdone=true;
    
    register int m,l,iter,i,k, n=size();
    register double s,r,ra,p,g,f,c,b,dd;

    for (i=1; i<n; ++i) Subdiag[i-1]=Subdiag[i];
    Subdiag[n-1]=0.0;
    for (l=0; l<n; ++l)
    {
	iter=0;
	do
	{
	    for (m=l; m<n-1; ++m)
	    {
		dd=(fabs(Diag[m])+fabs(Diag[m+1]));
		if (fabs(Subdiag[m])+dd == dd) break; 
	    }
	    if (m != l)
	    {
		if (iter++ >= Itno)	// too many iters
		    return Itno;
		
		g=(Diag[l+1]-Diag[l])/(2.0*Subdiag[l]);
		r=sqrt((g*g)+1.0);
		ra=(round_tozero(g)<0.0) ? -fabs(r) : fabs(r);
		g=Diag[m]-Diag[l]+Subdiag[l]/(g+ra);
		s=c=1.0;
		p=0.0;
		for (i=m-1; i>=l; --i)
		{
		    f=s*Subdiag[i];
		    b=c*Subdiag[i];
		    if (fabs(f) >= fabs(g))
		    {
			c=g/f;
			r=sqrt((c*c)+1.0);
			Subdiag[i+1]=f*r;
			c *= (s=1.0/r);
		    }
		    else
		    {
			s=f/g;
			r=sqrt((s*s)+1.0);
			Subdiag[i+1]=g*r;
			s *= (c=1.0/r);
		    }
		    g=Diag[i+1]-p;
		    r=(Diag[i]-g)*s+2.0*c*b;
		    p=s*r;
		    Diag[i+1]=g+p;
		    g=c*r-b;
		    
		    if (Wantevec)
		    {
			for (k=0; k<n; ++k)
			{
			    f=Q[k][i+1];
			    Q[k][i+1]=s*Q[k][i]+c*f;
			    Q[k][i]=c*Q[k][i]-s*f;
			}
		    }
		}
		Diag[l]=Diag[l]-p;
		Subdiag[l]=g;
		Subdiag[m]=0.0;
	    }
	}
	while (m != l);
    }
    return 0;
}
// END of implicit_ql(Itno, Wantevec)

// ---- Selected eigenvalues ----

// range_evalevec(Low, Up, Eval, Evec, Descend):
// Finds all eigenvalues within the range [Low, Up],
// and puts them into Eval in ascending order by default
// or in descending order if Descend==true.
// The corresponding eigenvectors are put into the columns of Evec.
// Return value: the number of eigenvalues found.
// Throws Rsymdiagexc_ if a reinit is needed.
int Rsymdiag_::range_evalevec(double Low, double Up,
	Vecdbl_& Eval, Rmatdbl_& Evec, bool Descend)
{
    if (!size() || QLdone  || Tridiagdone && !Qdone)
	throw Rsymdiagexc_("Rsymdiag_::range_evalevec()","Please (re)init first");

    if (Low==Up) return 0;  // no eigenvalues in zero-length range
    if (Low>Up) swap(Low, Up);
    
    // Housholder tridiagonalisation
    // prepare for eigenvectors
    housholder_tridiag(true);
    
    // check if there are eigenvalues in the given range at all
    size_t Lowno=sturm_seq(Low), Upno=sturm_seq(Up);
    size_t Evalno=Upno-Lowno;
    if (!Evalno)
	return 0;	// don't bother any further
    Eval.dim(Evalno); Evec.size(size(), Evalno);    // resize
    
    Evaloc_ Fullrange(Low, Lowno, Up, Upno);	// init search

    // locate eigenvalues in ascending order
    list<Evaloc_> Evalist(1, Fullrange);
    bisect_evals(Evalist, Evalist.begin());
    
    // you may get the eigenvalues in descending order, too
    if (Descend) Evalist.reverse();
    
    // get corresponding eigenvectors
    Vecdbl_ Tmpvec(size());
    size_t Idx;
 
    // process eigenvalue list
    list<Evaloc_>::const_iterator Evp;
    for (Idx=0, Evp=Evalist.begin(); Evp!=Evalist.end(); ++Evp, ++Idx)
    {
	double Ev=Evp->Eval;

	if (Evp->No==1)
	{
	    int Err=1, Iterno;
	    bool Bestseen=false;
	    static const size_t MAXITERNO=10;
	    double Tol, Besttol=Fpflo.largest(), Tmplen;
	    Vecdbl_ Besttmp(size());

	    // multiplicity is 1
	    for (Iterno=0; Err && Iterno<MAXITERNO; ++Iterno)
	    {
		random_vector(Tmpvec);
		Tmplen=Tmpvec.vec_norm();
		if (Tmplen<1e-8)
		{
		    --Iterno; continue; // backstep on dodgy random vectors
		}

    		Err=inverse_iter(Ev, Tmpvec, Tol);
		if (Err==-1)
		{
		    Ev*=1.000000123;   // perturb last few digits of the eigenvalue
		    Besttol=Fpflo.largest();   	// and start again
		    Bestseen=false;
		    continue;
		}
		if (Tol<Besttol)
		{
		    Bestseen=true;
		    Besttol=Tol;    // save eigenvector with best tolerance so far
		    Besttmp=Tmpvec;
		}
	    }
	    if (!Bestseen)
	    {
		cerr<<"\n? Rsymdiag_::range_evalevec(): No eigenvector found for eval["
			<<Idx<<"]="<<Ev<<": nullvector used\n";
		Besttmp.set_values();
	    }
	    else if (Iterno>=MAXITERNO)
	    {
		cerr<<"\n? Rsymdiag_::range_evalevec(): Evec convergence problem @ eval["
			<<Idx<<"]="<<Ev<<": best tol="<<Besttol<<endl;
	    }
	    Eval[Idx]=Ev;
	    Evec.col(Q*Besttmp, Idx);
	}
	else
	{
	    // multiple eigenvectors
	    Rmatdbl_ Evecs(size(), Evp->No);
	    inverse_iter(Ev, Evp->No, Evecs);
	    for (size_t m=0; m<Evp->No; ++m)
	    {
		Eval[Idx+m]=Ev;
		Tmpvec=Q*Evecs.col(m);
		Evec.col(Tmpvec, Idx+m);
	    }
	    Idx+=Evp->No-1;
	}
    }
	    
    return Idx;
}
// END of range_evalevec(Low, Up, Eval, Evec, Descend)

// range_eval(Low, Up, Eval, Descent):
// Finds all eigenvalues within the range [Low, Up],
// and puts them into Eval in ascending order by default
// or in descending order if Descend==true.
// Return value: the number of eigenvalues found.
// Throws Rsymdiagexc_ if a reinit is needed.
int Rsymdiag_::range_eval(double Low, double Up,
	Vecdbl_& Eval, bool Descend)
{
    if (!size() || QLdone)
	throw Rsymdiagexc_("Rsymdiag_::range_eval()","Please (re)init first");

    if (Low==Up) return 0;  // no eigenvalues in zero-length range
    if (Low>Up) swap(Low, Up);
    
    // Housholder tridiagonalisation
    housholder_tridiag(false);

    // check if there are eigenvalues in the given range at all
    size_t Lowno=sturm_seq(Low), Upno=sturm_seq(Up);
    size_t Evalno=Upno-Lowno;
    if (!Evalno)
	return 0;	// don't bother any further
    Eval.dim(Evalno);   // resize
    
    Evaloc_ Fullrange(Low, Lowno, Up, Upno);	// init search
    
    // locate eigenvalues in ascending order
    list<Evaloc_> Evalist(1, Fullrange);
    bisect_evals(Evalist, Evalist.begin());
    
    // you may get the eigenvalues in descending order, too
    if (Descend) Evalist.reverse();
    
    // process eigenvalue list
    list<Evaloc_>::const_iterator Evp;
    int Idx;
    for (Idx=0, Evp=Evalist.begin(); Evp!=Evalist.end(); ++Evp, Idx+=Evp->No)
	for (size_t m=0; m<Evp->No; ++m)
	    Eval[Idx+m]=Evp->Eval;
	    
    return Idx;
}
// END of range_eval(Low, Up, Eval, Descend)

// largest_evalevec(N, Eval, Evec, Descend):
// calculates the N largest eigenvalues and the corresponding eigenvectors.
// Returns the number of eigenvalues found.
// Throws Argexc_ if N is not in the range [1..size()].
// Throws Rsymdiagexc_ if a reinit is needed.
// See range_evalevec(..) for a description of the other variables.
int Rsymdiag_::largest_evalevec(size_t N, Vecdbl_& Eval,
	Rmatdbl_& Evec, bool Descend)
{
    if (!N || N>size())
    	throw Argexc_("Rsymdiag_::largest_evalevec()", N, 1, size());
    if (!size() || QLdone  || Tridiagdone && !Qdone)
	throw Rsymdiagexc_("Rsymdiag_::largest_evalevec()","Please (re)init first");
    
    // we need the tridiagonal matrix and the Housholder transform matrix
    // largest_range(..) invokes gershgorin_limits(..)
    // but the tridiag step will not be done twice
    housholder_tridiag(true);
    
    double Low, Up;
    largest_range(N, Low, Up);	// get range holding the evals
    return range_evalevec(Low, Up, Eval, Evec, Descend);
}
// END of largest_evalevec(N, Eval, Evec, Descend)

// largest_eval(N, Eval, Descend):
// calculates the N largest eigenvalues.
// Returns the number of eigenvalues found.
// Throws Argexc_ if N is not in the range [1..size()].
// Throws Rsymdiagexc_ if a reinit is needed.
// See range_eval(..) for a description of the other variables.
int Rsymdiag_::largest_eval(size_t N, Vecdbl_& Eval, bool Descend)
{
    if (!N || N>size())
    	throw Argexc_("Rsymdiag_::largest_evalevec()", N, 1, size());
    if (!size() || QLdone)
	throw Rsymdiagexc_("Rsymdiag_::largest_eval()","Please (re)init first");
    
    // we need the tridiagonal matrix only
    housholder_tridiag(false);
    
    double Low, Up;
    largest_range(N, Low, Up);	// get range holding the evals
    return range_eval(Low, Up, Eval, Descend);
}
// END of largest_eval(N, Eval, Descend)

// evalevec_above(Thr, Eval, Evec, Descend):
// calculates the largest eigenvalues and the corresponding eigenvectors
// above the threshold Thr. Use for getting all positive eigenvalues for example.
// Returns the number of eigenvalues found.
// This can be 0 if Thr is set too high.
// Throws Rsymdiagexc_ if a reinit is needed.
// See range_evalevec(..) for a description of the other variables.
int Rsymdiag_::evalevec_above(double Thr, Vecdbl_& Eval,
	Rmatdbl_& Evec, bool Descend)
{
    if (!size() || QLdone  || Tridiagdone && !Qdone)
	throw Rsymdiagexc_("Rsymdiag_::evalevec_above()","Please (re)init first");
    
    // we need the tridiagonal matrix 
    // and the Housholder transform matrix
    housholder_tridiag(true);
    if (sturm_seq(Thr)==size())	// no eigenvalues above Thr
	return 0;
    
    double Low, Up;
    gershgorin_limits(Low, Up);	// get range holding all evals
    if (Thr<=Low)   // all above
    {
	cerr<<"\n? Rsymdiag_::evalevec_above("
		<<Thr<<",...): all eigenvalues above threshold, Descend="
		<<Descend<<" ignored, all_evalevec(..) invoked\n";
	return all_evalevec(Eval, Evec);
    }
    return range_evalevec(Thr, Up, Eval, Evec, Descend);
}
// END of evalevec_above(N, Eval, Evec, Descend)

// eval_above(Thr, Eval, Descend):
// calculates the largest eigenvalues above the threshold Thr.
// Use for getting all positive eigenvalues for example.
// Returns the number of eigenvalues found.
// This can be 0 if Thr is set too high.
// Throws Rsymdiagexc_ if a reinit is needed.
// See range_eval(..) for a description of the other variables.
int Rsymdiag_::eval_above(double Thr, Vecdbl_& Eval, bool Descend)
{
    if (!size() || QLdone)
	throw Rsymdiagexc_("Rsymdiag_::eval_above()","Please (re)init first");
    
    // we need the tridiagonal matrix only
    housholder_tridiag(false);
    
    if (sturm_seq(Thr)==size())	// no eigenvalues above Thr
	return 0;
    
    double Low, Up;
    gershgorin_limits(Low, Up);	// get range holding all evals
    if (Thr<=Low)   // all above
    {
	cerr<<"\n? Rsymdiag_::eval_above("
		<<Thr<<",...): all eigenvalues above threshold, Descend="
		<<Descend<<" ignored, all_eval(..) invoked\n";
	return all_eval(Eval);
    }
    return range_eval(Thr, Up, Eval, Descend);
}
// END of eval_above(N, Eval, Descend)

// gershgorin_limits(Low, Up):
// calculates the limits of the range [Low,Up] that contains
// all eigenvalues of the input matrix.
// Note that this method performs a Housholder tridiagonalisation
// but the transform matrix is not accumulated, you would not get
// eigenvectors without calling re_init(..) after calling this.
void Rsymdiag_::gershgorin_limits(double& Low, double& Up)
{
    if (!size() || QLdone)
	throw Rsymdiagexc_("Rsymdiag_::gershgorin_limits()","Please (re)init first");

    housholder_tridiag(false);
    
    register double Ai, Bi, Bi1, L, U;
    register size_t i, N=size();
    
    // begin: special
    Ai=Diag[0]; Bi=fabs(Subdiag[1]);
    Low=Ai-Bi; Up=Ai+Bi;

    // get min, max
    for (i=1; i<N-1; ++i)
    {
	Bi1=Bi; // i.e. Bi1=fabs(Subdiag[i-1])
	Ai=Diag[i]; Bi=fabs(Subdiag[i+1]);
	L=Ai-Bi-Bi1;
	if (Low>L) Low=L;
	U=Ai+Bi+Bi1;
	if (Up<U) Up=U;
    }
    
    // last: special again
    Ai=Diag[N-1]; Bi1=fabs(Subdiag[N-1]);
    L=Ai-Bi1;
    if (Low>L) Low=L;
    U=Ai+Bi1;
    if (Up<U) Up=U;
    
    // make limits a bit loose
    // this hack was needed for a spectrum with triple 2.0 and quadruple 0.0
    // the upper limit was exactly 2.0
    // and that screwed up sturm_seq()
    if (Low<0.0) Low*=1.001; else Low*=0.999;
    if (Up>0.0) Up*=1.001; else Up*=0.999;
}
// END of gershgorin_limits(Low, Up)

// largest_range(N, Low, Up):
// returns the range Low..Up that brackets the N largest eigenvalues.
// Hidden
void Rsymdiag_::largest_range(size_t N, double& Low, double& Up)
{
    double L, U, X;
    size_t Xno=size()-N, Xbelow;
    
    gershgorin_limits(L, Up);     // initial bracketing
    U=Up; Xbelow=0;
    while (Xbelow!=Xno)
    {
	X=(U-L)/2.0;     // bisection
	Xbelow=sturm_seq(X);
	if (Xbelow<Xno)
	    L=X;
	else if (Xbelow>Xno)
	    U=X;
    }
    Low=X;
}
// END of largest_range(N, Low, Up)

// sturm_seq(X):
// calculates the characteristic polynomial of the tridiagonalised matrix
// at X and counts the sign changes. This gives the number of
// eigenvalues that are less than X.
// Hidden
size_t Rsymdiag_::sturm_seq(double X) const
{
    size_t Signchange=0;
    register size_t i;
    register double b, P, P1, P2;
    
    P1=1.0; P=Diag[0]-X;
    if (P<0.0) ++Signchange;
    
    for (i=1; i<size(); ++i)
    {
	b=Subdiag[i];
	P2=P1; P1=P;
	P=(Diag[i]-X)*P1-b*b*P2; // next polynomial
	if (P1>=0.0 && P<0.0 || P1<=0.0 && P>0.0)
	    ++Signchange;
    }
    
    return Signchange;
}
// END of sturm_seq(X)

// sturm_newton(X):
// calculates the negative ratio of the characteristic polynomial
// and its first derivative at X. This is needed for 
// the Sturm-Newton iteration that looks for isolated eigenvalues.
// Hidden
double Rsymdiag_::sturm_newton(double X) const
{
    register size_t i;
    register double b, bb, Dix, P, P1, P2, Pd, Pd1, Pd2;
    
    P1=1.0; P=Diag[0]-X;
    Pd1=0.0; Pd=-1.0;
    
    for (i=1; i<size(); ++i)
    {
	b=Subdiag[i]; bb=b*b;
	Dix=Diag[i]-X;
	P2=P1; P1=P;
	P=Dix*P1-bb*P2; // next polynomial
	Pd2=Pd1; Pd1=Pd;
	Pd=-P1+Dix*Pd1-bb*Pd2; // next derivative
    }
    
    return -(P/Pd);
}
// END of sturm_newton(X)

// bisect_evals(Evalist, Iv):
// locates the eigenvalues by bisection.
// On invocation Evalist is a list containing one element,
// the range in which we look for eigenvalues.
// Iv shall point to this element (use Evalist.begin()).
// On return, Evalist contains the list of ranges
// bracketing the eigenvalues in ascending order.
// The eigenvalues are already bracketed very tightly,
// relative error is about EPSILON.
// Hidden
void Rsymdiag_::bisect_evals(list<Evaloc_>& Evalist,
	list<Evaloc_>::iterator Iv)
{
    double X;
    size_t Ebelowx;
    
    if (Iv->No>1)
    {
	// try to separate: "drive a wedge between them"
	while (Iv->bracket_loose())
	{
	    X=Iv->Eval=(Iv->Up+Iv->Low)/2.0; // bisection
	    Ebelowx=sturm_seq(X);
	    if (Ebelowx==Iv->Lowno)
		Iv->Low=X;  // shrink interval from below
	    else if (Ebelowx==Iv->Upno)
		Iv->Up=X;   // shrink interval from above
	    else
	    {
		// X separates!
		// make new interval below old
		Evaloc_ New(Iv->Low, Iv->Lowno, X, Ebelowx);

    		// adjust old interval: lower limit is X now
		Iv->Low=X; Iv->Lowno=Ebelowx;
		Iv->No=Iv->Upno-Iv->Lowno;
		
		// add new interval before current position
		// in the C version this is the other way round
		// because it is easier to insert after current pos
		// in a singly-linked list :-)
		list<Evaloc_>::iterator Newp=Evalist.insert(Iv, New);
		
		// recursive calls to process the two new intervals
		bisect_evals(Evalist, Newp);
		bisect_evals(Evalist, Iv);
		break;
	    }
	}
    }
    else    // only one eigenvalue in interval, find it
    {
	double DeltaX, Xnew;
	X=(Iv->Up+Iv->Low)/2.0; // bisection to start Sturm-Newton iter

	// max. no of iterations is limited
	static const double Ln2=0.69314718;
	int Nmax=int(log((Iv->Up-Iv->Low)/EPSILON)/Ln2);
	register int i;
	for (i=0; i<Nmax; ++i)
	{
	    DeltaX=sturm_newton(X);
	    if (fabs(DeltaX)<fabs(X)*EPSILON)
	    {
		Iv->Eval=X; // eigenvalue found
		break;	    // end of iteration
	    }
	    Xnew=X+DeltaX;

	    // If Xnew is outside interval, use bisection
	    if (Xnew<Iv->Low || Xnew>Iv->Up)
	    {
		X=(Iv->Up+Iv->Low)/2.0; // bisect (throw away "bad" X)
		Ebelowx=sturm_seq(X);
		if (Ebelowx==Iv->Lowno)
		{
		    Iv->Low=X;  // shrink interval from below
		    Iv->Lowno=Ebelowx;
		}
		else
		{
		    Iv->Up=X;   // shrink interval from above
		    Iv->Upno=Ebelowx;
		}
		X=(Iv->Up+Iv->Low)/2.0;
		
		// start again!
		i=0; Nmax=static_cast<int>(log((Iv->Up-Iv->Low)/EPSILON)/Ln2);
	    }
	    else
		X=Xnew; // Newton step accepted
	}
	Iv->Eval=X;
    }
}
// END of bisect_evals(Evalist, Iv)

// inverse_iter(Eval, Evec, Tol):
// given a good approximate eigenvalue Eval and an initial eigenvector Evec
// that is not orthogonal to Eval's subspace, this method
// produces a normalised eigenvector by inverse iteration.
// Evec is assumed to be of unit length.
// Return value: 0 if OK, -1 if the decomposition for the
// linear equation solving did not work (try to perturb Eval),
// -2 if the iteration did not converge in MAXSTEPNO steps
// (try a different initial Evec).
// Tol contains on return the tolerance value (relative deviation
// of the scalar product of the old and new eigenvector from +/-1.0).
// Hidden
int Rsymdiag_::inverse_iter(double Eval, Vecdbl_& Evec, double& Tol)
{
    Vecdbl_ Dshf(Diag), Oldevec(Evec);
    double Scalprod=0.0;
    size_t Stepno;
    static const size_t MAXSTEPNO=30;	// usually 2..3 iters are enough
    static const double TOL=1e-6;
    
    // decompose matrix: do only once as eigenvalues are OK already
    // the subdiagonal comes from Subdiag
    Dshf-=Eval;	// A-lambda*I
    if (tridiag_decomp(Dshf))
	return -1;  // could not decompose, eval problematic?
    
    // iterate
    Tol=10.0*TOL;
    for (Stepno=0; Stepno<MAXSTEPNO &&  Tol>TOL; ++Stepno)
    {
	tridiag_solve(Oldevec, Evec);
	Evec.vec_norm();
	Scalprod=Oldevec*Evec; // must be +/- 1 for convergence
	Tol=fabs(fabs(Scalprod)-1.0);
	Oldevec=Evec;
    }
    return (Stepno>=MAXSTEPNO)? -2: 0;	// no convergence or OK
}
// END of inverse_iter(Eval, Evec, Tol)

// inverse_iter(Eval, Multiple, Evecs):
// finds all eigenvectors belonging to the multiple eigenvalue Eval
// (multiplicity is Multiple>=2), and puts them into the matrix Evecs
// (size()xMultiple, the eigenvectors are the columns).
// Hidden
void Rsymdiag_::inverse_iter(double Eval, size_t Multiple, Rmatdbl_& Evecs)
{
    Evecs.size(size(), Multiple);
    
    // decompose tridiag matrix first (same for all evecs here) (no check)
    Vecdbl_ Dshf=Diag; 
    Dshf-=Eval;	// A-lambda*I
    tridiag_decomp(Dshf);
    
    Rmatdbl_ Oldevecs(size(), Multiple);
    Vecdbl_ Oldevec(size()), Evec(size());
    double Subspdiff=1.0;
    size_t Stepno, m;
    static const size_t MAXSTEPNO=10;
    
    random_matrix(Oldevecs);
    gram_schmidt(Oldevecs);
	
    // inverse iteration on all eigenvectors belonging to Eval
    for (Stepno=0; Stepno<MAXSTEPNO && Subspdiff>EPSILON; ++Stepno)
    {
	// one inverse iteration step for all old vectors in subspace
	for (m=0; m<Multiple; ++m)
	{
	    Oldevec=Oldevecs.col(m);
	    tridiag_solve(Oldevec, Evec);
	    Evecs.col(Evec, m);
	}
	
	// orthonormalisation and subspace difference
	gram_schmidt(Evecs);
	Subspdiff=subspace_diff(Oldevecs, Evecs);
    }
}
// END of inverse_iter(Eval, Multiple, Evecs)

// gram_schmidt(Mat):
// Makes an orthonormal set from the columns of Mat using
// the modified Gram-Schmidt algorithm. See Golub & Van Loan for details.
// Hidden static
void Rsymdiag_::gram_schmidt(Rmatdbl_& Mat)
{
    register size_t k, i, j, m=Mat.rno(), n=Mat.cno();
    register double Rkk, Aik, Rkj;
    
    for (k=0; k<n; ++k)
    {
	// normalise k:th column that is already orthogonal
	// to all columns [0..k-1]
	Rkk=0.0;
	for (i=0; i<m; ++i)
	{
	    Aik=Mat[i][k];
	    Rkk+=Aik*Aik;
	}
	Rkk=1.0/sqrt(Rkk);
	for (i=0; i<m; ++i)
	    Mat[i][k]*=Rkk;
	
	// orthogonalisation
	for (j=k+1; j<n; ++j)
	{
	    // get scalarproduct of k:th and j:th column
	    Rkj=0.0;
	    for (i=0; i<m; ++i)
		Rkj+=Mat[i][k]*Mat[i][j];
	    
	    // subtract projection of j:th column on k:th from j:th
	    for (i=0; i<m; ++i)
		Mat[i][j]-=Mat[i][k]*Rkj;
	}
    }
}
// END of gram_schmidt(Mat)

// subspace_diff(V, W):
// for all column vectors in W, it subtracts the components that
// lie in span(V), and gets the maximal length of what remains.
// If this length is very small, then span(V)==span(W) approximately.
// This is the stop criterion for the inverse iteration for
// multiple eigenvalues (only the subspace of the eigenvectors
// is invariant, and we need just a good orthonormal base in it).
// The columns of both V and W are assumed to be orthonormal
// and the dims must match of course.
// Hidden static
double Rsymdiag_::subspace_diff(const Rmatdbl_& V, const Rmatdbl_& W)
{
    register size_t j, p, m=V.rno(), n=V.cno();
    register double Sp, Len, Diff=0.0;
    
    // IMPLEMENTATION NOTE:-
    // this is experimental code, it seems to work,
    // but it certainly can be improved. One day perhaps...
    
    // check if the p:th col of W is in span(V)
    Vecdbl_ Wp(m), Vj(m), X(m);
    for (p=0; p<n; ++p)
    {
	Wp=W.col(p);
	X=Wp;
	
	// subtract the components of Wp in direction of Vj from Wp (in X)
	for (j=0; j<n; ++j)
	{
	    Vj=V.col(j);    // the j:th V vector
	    Sp=Vj*Wp;	    // Wp's component to Vj
	    Vj*=Sp;
	    X-=Vj;
	}
	
	// X now contains the component orthogonal to span(V)
	// store its length, the shorter the better
	Len=X.vec_len();
	if (Len>Diff) Diff=Len; // get maximum
    }
    return Diff;
}
// END of subspace_diff(V, W)

// tridiag_decomp(Diagshift):
// Decomposition for solving the inverse iteration equation
// (A-Eval*I)Xnew=X
// where A is the symmetric tridiagonal matrix produced by
// housholder_tridiag(..), Eval is an (approximate) eigenvalue,
// I is the unit matrix, X and Xnew are the old and new
// approximate eigenvectors, respectively.
// Diagshift is the main diagonal of (A-Eval*I),
// the subdiagonal is the member Subdiag.
// Return value: 0 if OK, >=1 on decomposition error.
// Hidden
int Rsymdiag_::tridiag_decomp(const Vecdbl_& Diagshift)
{
    // Implementation note:-
    // this routine and the associated method tridiag_solve() below
    // are is adapted to the special needs of Rsymdiag_.
    // For a general-purpose implementation, see the C version
    // in "tridiag.c"
    
    register size_t j;

    if (Diagshift[0] == 0.0) return 1;
    Beta[0]=Diagshift[0];
    
    // decomposition
    for (j=1; j<size(); ++j)
    {
	Gamma[j]=Subdiag[j]/Beta[j-1];    // Superdiag==Subdiag here
	Beta[j]=Diagshift[j]-Subdiag[j]*Gamma[j];
	if (Beta[j] == 0.0) return j+1;
    }
    return 0;	// OK
}
// END of tridiag_decomp(Diagshift)

// tridiag_solve(Rhs, X):
// solves the linear equation system for X with right-hand-side Rhs
// that was decomposed by tridiag_decomp(..).
// Hidden
void Rsymdiag_::tridiag_solve(const Vecdbl_& Rhs, Vecdbl_& X) const
{
    register int j;
    
    // forward substitution
    X[0]=Rhs[0]/Beta[0];
    for (j=1; j<size(); ++j)
	X[j]=(Rhs[j]-Subdiag[j]*X[j-1])/Beta[j];
    
    // backward substitution
    for (j=size()-2; j>=0; --j)
    	X[j] -= Gamma[j+1]*X[j+1];
}
// END of tridiag_solve(Rhs, X)

// ---- Auxiliaries ----

// random_vector(Vec): generates a vector with random
// coordinates. Uses the dodgy built-in RNG.
// Hidden static
void Rsymdiag_::random_vector(Vecdbl_& Vec)
{
    register size_t i;
    for (i=0; i<Vec.dim(); ++i)
	Vec[i]=rand();
}
// END of random_vector(Vec)

// random_matrix(Vec): generates a matrix with random
// elements.
// Hidden static
void Rsymdiag_::random_matrix(Rmatdbl_& Mat)
{
    register size_t i, j;
    for (i=0; i<Mat.rno(); ++i)
	for (j=0; j<Mat.cno(); ++j)
	    Mat[i][j]=rand();
}
// END of random_matrix(Mat)

// ==== END OF Rsymdiag.cc ====



    
