// ==== FUNCTIONS Svd.cc ====

// Singular value decomposition.

// 18-Aug-1998. Andras Aszodi

// ----	HEADER ----

#include"Svd.hh"
using namespace RazorBack;

// ---- PROTOTYPES ----

extern "C"
{
    static double sign(double a, double b)
    {
	return((b>=0.0)? fabs(a): -fabs(a));
    }
}

// ==== Static initialisation ====

const Safety_ Svd_::Safe; 	// see Safety.hh for defaults
#ifdef FLT_EPSILON
const double Svd_::SVD_EPSILON=10.0*FLT_EPSILON;
#else
const double Svd_::SVD_EPSILON=1e-7;
#endif

// ==== Svd_ MEMBER FUNCTIONS ====

// ---- AUXILIARIES ----

// utb(): the vector B is premultiplied by the transpose of the
// U matrix in the decomposition and the result is placed into
// Utb. Dim checks are not carried out. Private
inline
void Svd_::utb(const Vecdbl_& B, Vecdbl_& Utb) const
{
    register unsigned int i, j;
    register double Temp;
    
    for (j=0; j<C; j++)
    {
	Temp=0.0;
	for (i=0; i<R; i++)
	    Temp+=U[i][j]*B[i];
	Utb[j]=Temp;
    }
}
// END of utb()

// ---- Constructors ----

// Set up SVD for a Row x Col matrix where Row>=Col. If Row<Col,
// then the rows of U will be padded to make it ColxCol
// with a warning. If either Row or Col is 0, will be changed
// to 3 (cf matrix/vector ctors) and a warning printed.
// The default size is also 3x3.
// The members will be initialised to 0.0 (all elements).
Svd_::Svd_(unsigned int Row, unsigned int Col) :
    U(((Row<Col)? Col: Row), Col), // row padding
    W(Col), 
    V(Col, Col)
{
    // size check warnings (actually the members are init'd already)
    if (!Row) { cerr<<"\n? Svd_: Row==0, was set to 3\n"; Row=3; }
    if (!Col) { cerr<<"\n? Svd_: Col==0, was set to 3\n"; Col=3; }
    C=Col; Rorig=Row;
    if (Row<Col)
    {
	cerr<<"\n? Svd_: Row="<<Row<<", padded to "<<Col<<endl;
	Row=Col;
    }
    R=Row;
}

// ---- Size ----

// set_size(): modifies the sizes of the SVD components to
// accommodate a Row x Col matrix. If Row < Col, then the
// rows of U will be padded to give a Col x Col matrix.
// If Row or Col is 0, then 3 will be used instead and a warning printed.
// Without arguments, the size is set to 3x3.
void Svd_::set_size(unsigned int Row, unsigned int Col)
{
    if (R==Row && C==Col) return;	// no change
    
    if (!Row) { cerr<<"\n? Svd_::set_size(): Row==0, was set to 3\n"; Row=3; }
    if (!Col) { cerr<<"\n? Svd_::set_size(): Col==0, was set to 3\n"; Col=3; }

    C=Col; Rorig=Row;
    if (Row<Col)
    {
	cerr<<"\n? Svd_::set_size(): Row="<<Row<<", padded to "<<Col<<endl;
	Row=Col;
    }
    R=Row;
    U.size(R, C); W.dim(C); V.size(C, C);
}
// END of set_size()

// ---- SVD routines ----

// make_decomp(): the SVD done
// on the general matrix A (which is preserved). The
// three member objects U, W, V will be set in the calling
// object. Throws Svdexc_ if there was a convergence problem
// or a div-by-zero attempt in the core routine.
void Svd_::make_decomp(const Rmatdbl_& A)
{
    // adjust size if necessary
    set_size(A.rno(), A.cno());
    
    // copy A to U for decomposition
    if (A.rno()>=A.cno()) U=A;
    else    // copy rows only
    {
	for (unsigned int i=0; i<A.rno(); i++)
	    U.row(A.row(i), i);
    }

    // actual decomposition: may throw Svdexc_
    svd_core();
}
// END of make_decomp()

// rank_cond(): checks the N singular values W[] of a matrix 
// after SVD. The condition number in *Cond
// (ratio of the largest and smallest singular value) is also
// calculated if Cond!=NULL. The singular values which are smaller than
// Eps times the largest are set to 0.0.
// Return value: the rank of the matrix.
unsigned int Svd_::rank_cond(double Eps, double *Cond)
{
    register double Wmax=-HUGE_VAL, Wmin=HUGE_VAL;
    register unsigned int i;
    
    // get the largest and smallest singular value
    for (i=0; i<C; i++)
    {
	if (W[i]>Wmax) Wmax=W[i];
	if (W[i]<Wmin) Wmin=W[i];
    }
    
    // calc the condition number: set to HUGE_VAL if Wmin==0.0
    if (Cond!=NULL)
	*Cond=(Wmin==0.0)? HUGE_VAL: Wmax/Wmin;
    
    // set all singular values which are smaller than Eps*Wmax
    // to zero: this is the conditioning. Calc the rank
    Wmax*=fabs(Eps);
    unsigned int Rank, Maxrank;
    Rank=Maxrank=(Rorig>=C)? C: Rorig;
    for (i=0; i<Maxrank; i++)
	if (W[i]<Wmax) 
	{
	    W[i]=0.0; Rank--;
	}
    return(Rank);
}
// END of rank_cond()

// lin_solve(): back-substitution routine for solving linear equations
// AX=B. A should be SV-decomposed into U, W and V' by make_decomp()
// and the weight vector should be "conditioned" (Safe.small() entries
// zeroed) by rank_cond() prior to the call to this routine.
// The right-hand-side vector B must have a dim equal to the original
// no. of rows in A otherwise a Dimexc_ exception will be thrown.
// Return value: the X vector.
Vecdbl_ Svd_::lin_solve(const Vecdbl_& B) const
{
    // dim mismatches
    unsigned int Bdim=B.dim();
    if (Bdim!=Rorig)
	throw Dimexc_("Svd_::lin_solve", Bdim, Rorig);
    
    Vecdbl_ Wub(C);	// this will be W*U'*B
    // the "fewer eqns than unknowns" case, warn only
    if (R>Rorig && Bdim==Rorig)
    {
	cerr<<"\n? Svd_::lin_solve: fewer eqns ("<<Rorig
	    <<") than unknowns ("<<R<<"), B zero-padded\n";
	Vecdbl_ Bpad(B); Bpad.dim(R);   // pad B w/ zeroes
	utb(Bpad, Wub);
    }
    else	// normal cases
	utb(B, Wub);
    
    for (unsigned int j=0; j<C; j++)
	if (W[j]==0.0) Wub[j]=0.0; else Wub[j]/=W[j];
    
    // get solution vector and return
    Vecdbl_ X=V*Wub;
    return(X);
}
// END of lin_solve()

// ---- SVD CORE ----

// svd_core(): Performs the SVD of the member matrix U (not preserved).
// (It is assumed to have been initialised before the call.)
// Based on the Golub & Van Loan algorithm: the FORTRAN
// indexing convention is used within and the matrix ptrs are
// restored on exit. Throws Svdexc_ if iter limit is exceeded or
// there was an attempt to div by zero.
void Svd_::svd_core()
{
    const int SVD_ITMAX=100, m=R, n=C;	// max. itno, row and column size
    register int i,its=1,j,jj,k,l,nm;
    bool Flg;
    register double An=0.0,c,f,g=0.0,h,s,Scal=0.0,Iscal, x,y,z;
    Vecdbl_ Tempvec(n);

    // FORTRAN indexing: the matrices and vectors will have
    // [1..N] style indexing.
    // If the matrices were indexed in C style like "Mat[i][j]"
    // then here it is how to convert these into FORTRAN-style
    // "Mat(i,j)" expressions (this is needed by my matrix library
    // to do FORTRAN indexing and optional range checking):-
    // FIND the regular expression "\[([^]]+)\]\[([^]]+)\]"
    // and REPLACE with "\(\1,\2\)"
    // Converting vectors from "V[i]" to "V(i)" is easier,
    // this is left as an exercise to the reader :-)
    U.ftn_idx(true); V.ftn_idx(true);
    Tempvec.ftn_idx(true); W.ftn_idx(true);
    
    for (i=1;i<=n;i++)
    {
	l=i+1;
	Tempvec(i)=Scal*g;
	g=s=Scal=0.0;
	if (i <= m)
	{
	    for (k=i;k<=m;k++) Scal += fabs(U(k,i));
	    if (Scal>Safe.small())
	    {
		Iscal=1.0/Scal;
		for (k=i;k<=m;k++)
		{
		    U(k,i) *= Iscal;
		    s += U(k,i)*U(k,i);
		}
		f=U(i,i);
		g = -sign(sqrt(s),f);
		h=Safe.safe_div(1.0,f*g-s,__LINE__);
		U(i,i)=f-g;
		for (j=l;j<=n;j++)
		{
		    for (s=0.0,k=i;k<=m;k++) s += U(k,i)*U(k,j);
		    f=s*h;
		    for (k=i;k<=m;k++) U(k,j) += f*U(k,i);
		}
		for (k=i;k<=m;k++) U(k,i) *= Scal;
	    }
	}
	W(i)=Scal *g;
	g=s=Scal=0.0;
	if (i <= m && i != n)
	{
	    for (k=l;k<=n;k++) Scal += fabs(U(i,k));
	    if (Scal>Safe.small())
	    {
		Iscal=1.0/Scal;
		for (k=l;k<=n;k++)
		{
			U(i,k) *= Iscal;
			s += U(i,k)*U(i,k);
		}
		f=U(i,l);
		g = -sign(sqrt(s),f);
		h=Safe.safe_div(1.0,f*g-s,__LINE__);
		U(i,l)=f-g;
		for (k=l;k<=n;k++) Tempvec(k)=U(i,k)*h;
		for (j=l;j<=m;j++)
		{
			for (s=0.0,k=l;k<=n;k++) s += U(j,k)*U(i,k);
			for (k=l;k<=n;k++) U(j,k) += s*Tempvec(k);
		}
		for (k=l;k<=n;k++) U(i,k) *= Scal;
	    }
	}
	An=max(An,(fabs(W(i))+fabs(Tempvec(i))));
    }
    for (i=n;i>=1;i--)
    {
	if (i < n)
	{
	    if (fabs(g)>Safe.small())
	    {
		g=1.0/g;
		for (j=l;j<=n;j++)
		    V(j,i)=(U(i,j)/U(i,l))*g;
		for (j=l;j<=n;j++) {
		    for (s=0.0,k=l;k<=n;k++) s += U(i,k)*V(k,j);
		    for (k=l;k<=n;k++) V(k,j) += s*V(k,i);
		}
	    }
	    for (j=l;j<=n;j++) V(i,j)=V(j,i)=0.0;
	}
	V(i,i)=1.0;
	g=Tempvec(i);
	l=i;
    }
    for (i=(m<n)? m: n;i>=1;i--)
    {
	l=i+1;
	g=W(i);
	for (j=l;j<=n;j++) U(i,j)=0.0;
	if (fabs(g)>Safe.small())
	{
	    g=1.0/g;
	    for (j=l;j<=n;j++)
	    {
		for (s=0.0,k=l;k<=m;k++) s += U(k,i)*U(k,j);
		f=Safe.safe_div(s,U(i,i),__LINE__);
		f*=g;
		for (k=i;k<=m;k++) U(k,j) += f*U(k,i);
	    }
	    for (j=i;j<=m;j++) U(j,i) *= g;
	} else for (j=i;j<=m;j++) U(j,i)=0.0;
	(U(i,i))++;
    }
    for (k=n;k>=1;k--)
    {
	for (its=1;its<=SVD_ITMAX;its++)
	{
	    Flg=true;
	    for (l=k;l>=1;l--)
	    {
		nm=l-1;
		if ((fabs(Tempvec(l))+An) == An)
		{
		    Flg=false;
		    break;
		}
		if ((fabs(W(nm))+An) == An)
		    break;
	    }
	    if (Flg)
	    {
		c=0.0; s=1.0;
		for (i=l;i<=k;i++)
		{
		    f=s*Tempvec(i);
		    Tempvec(i)=c*Tempvec(i);
		    if ((fabs(f)+An) == An) break;
		    g=W(i);
		    h=Safe.pythag(f,g);
		    W(i)=h;
		    h=Safe.safe_div(1.0,h,__LINE__);
		    c=g*h;
		    s = -f*h;
		    for (j=1;j<=m;j++)
		    {
			y=U(j,nm);
			z=U(j,i);
			U(j,nm)=y*c+z*s;
			U(j,i)=z*c-y*s;
		    }
		}
	    }
	    z=W(k);
	    if (l == k)
	    {
		if (z < 0.0)
		{
		    W(k) = -z;
		    for (j=1;j<=n;j++) V(j,k) = -V(j,k);
		}
		break;
	    }

	    // serious trouble, leave...
	    if (its == SVD_ITMAX)
		throw Svdexc_("Svd_::svd_core()", "No convergence", true);

	    x=W(l);
	    nm=k-1;
	    y=W(nm);
	    g=Tempvec(nm);
	    h=Tempvec(k);
	    f=Safe.safe_div((y-z)*(y+z)+(g-h)*(g+h),(2.0*h*y),__LINE__);
	    g=Safe.pythag(f,1.0);
	    if (fabs(x)<Safe.small())
		throw Svdexc_("Svd_::svd_core()", "Impending div by zero (x)", true);

    	    f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
	    c=s=1.0;
	    for (j=l;j<=nm;j++)
	    {
		i=j+1;
		g=Tempvec(i);
		y=W(i);
		h=s*g;
		g=c*g;
		z=Safe.pythag(f,h);
		if (z<Safe.small())
		    throw Svdexc_("Svd_::svd_core()", "Impending div by zero (z)", true);
		Tempvec(j)=z;
		c=f/z;
		s=h/z;
		f=x*c+g*s;
		g = g*c-x*s;
		h=y*s;
		y *= c;
		for (jj=1;jj<=n;jj++)
		{
		    x=V(jj,j);
		    z=V(jj,i);
		    V(jj,j)=x*c+z*s;
		    V(jj,i)=z*c-x*s;
		}
		z=Safe.pythag(f,h);
		W(j)=z;
		if (z>Safe.small())
		{
		    z=1.0/z;
		    c=f*z;
		    s=h*z;
		}
		f=c*g+s*y;
		x=c*y-s*g;
		for (jj=1;jj<=m;jj++)
		{
		    y=U(jj,j);
		    z=U(jj,i);
		    U(jj,j)=y*c+z*s;
		    U(jj,i)=z*c-y*s;
		}
	    }
	    Tempvec(l)=0.0;
	    Tempvec(k)=f;
	    W(k)=x;
	}
    }

    // get back to C indexing and clean up
    U.ftn_idx(false); V.ftn_idx(false);
    W.ftn_idx(false); Tempvec.ftn_idx(false);
}
// END of svd_core()

// ---- GLOBAL FUNCTIONS ----

// <<: the overloaded output operator. Prints a neat listing to Out */
namespace RazorBack {
ostream& operator<<(ostream& Out, const Svd_& Svd)
{
    Out<<Svd.Rorig<<"x"<<Svd.C<<" singular decomposition";
    unsigned int Ex=Svd.R-Svd.Rorig;
    if (Ex) Out<<" ("<<Ex<<" row"<<((Ex==1)? " ":"s ")<<"added)\n";
    else Out<<'\n';
    Out<<"Singular values:\n"<<Svd.W;
    Out<<"The U matrix:\n"<<Svd.U;
    Out<<"The V matrix:\n"<<Svd.V;
    return(Out);
}
}
// END of <<

// ==== END OF FUNCTIONS Svd.cc ====
