#ifndef RECTMAT_TMPL_METHODS
#define RECTMAT_TMPL_METHODS

// ==== TEMPLATE METHODS Rectmat.cc ====

// 28-Jul-1999. Andras Aszodi

namespace RazorBack {

// ---- Access ----

// The two-argument function call operator provides optional checked access
// when Check is true, and FORTRAN-style indexing when Ftnidx is true.
template <class T_>
const T_& Rectmat_<T_>::operator()(size_t Ridx, size_t Cidx) const
{
    if (ftn_idx()) { Ridx--; Cidx--; }
    if (range_check())
    {
	if (Ridx>=rno()) throw(Indexrangexc_("Rectmat_[ROW][](const)", Ridx, rno()-1));
	if (Cidx>=cno()) throw(Indexrangexc_("Rectmat_[][COL](const)", Cidx, cno()-1));
    }
    return(elems(Ridx*cno()+Cidx));
}
template <class T_>
T_& Rectmat_<T_>::operator()(size_t Ridx, size_t Cidx)
{
    if (ftn_idx()) { Ridx--; Cidx--; }
    if (range_check())
    {
	if (Ridx>=rno()) throw(Indexrangexc_("Rectmat_[ROW][](lval)", Ridx, rno()-1));
	if (Cidx>=cno()) throw(Indexrangexc_("Rectmat_[][COL](lval)", Cidx, cno()-1));
    }
    clone();	// impending modification
    return(elems(Ridx*cno()+Cidx));
}
// END of operator(Ridx,Cidx)

// size(Rno, Cno): changes the size of the calling object to Rno x Cno.
// Nothing is done if the size parameters are the same as the current
// size or if either of them is 0. When the matrix "shrinks", then
// the upper left corner is preserved, when it grows, then the new
// rows and/or columns are padded with "zeroes" T_(0).
template <class T_>
void Rectmat_<T_>::size(size_t Rno, size_t Cno)
{
    if (!Rno || !Cno || rno()==Rno && cno()==Cno) return;   // no change
    
    Matstorage_<T_> *Newmat=new Matstorage_<T_>(Rno*Cno);
    unsigned int i,
	Rmin=min(rno(), Rno),	// STL
	Cmin=min(cno(), Cno);
    vector<T_>::iterator Newrows;
    vector<T_>::const_iterator Rowsi;
    bool Padrows=(Cno>cno());
    
    // copy rows, pad with "0" if necessary
    for (i=0, Newrows=Newmat->elems().begin(), Rowsi=elems().begin(); 
	    i<Rmin; 
	    i++, Newrows+=Cno, Rowsi+=cno())
    {
	copy(Rowsi, Rowsi+Cmin, Newrows);	// STL copy
	if (Padrows)
	    fill_n(Newrows+cno(), Cno-cno(), T_(0));    // pad rows with "zeroes"
    }
    // pad the columns if necessary
    if (Rno>rno())
	fill(Newrows, Newmat->elems().end(), T_(0));
    
    // replace old with new
    R=Rno; C=Cno;
    replace_storage(Newmat);
}
// END of size(Rno, Cno)

// row(Idx): returns the Idx-th row as a Vector_ object safely.
// row(Vec,Idx): sets the elements in the Idx-th row to the elements
// of the vector Vec, provided Idx is legal and Vec has the right
// number of elements.
template <class T_>
Vector_<T_> Rectmat_<T_>::row(size_t Idx) const
{
    if (Idx>=rno()) throw(Indexrangexc_("Rectmat_<T_>::row(Idx)", Idx, rno()-1));
    Vector_<T_> Rowvec(cno());
    vector<T_>::const_iterator Rowidx=elems().begin()+Idx*cno();
    
    copy(Rowidx, Rowidx+cno(), Rowvec.begin());	// STL
    return(Rowvec);
}
// END of row(Idx)

template <class T_>
void Rectmat_<T_>::row(const Vector_<T_>& Vec, size_t Idx)
{
    if (Vec.dim()!=cno()) throw(Dimexc_("Rectmat_<T_>::row(Vec, Idx)", Vec.dim(), cno()-1));
    if (Idx>=rno()) throw(Indexrangexc_("Rectmat_<T_>::row(Vec, Idx)", Idx, rno()-1));

    clone();	// modified
    vector<T_>::iterator Rowidx=elems().begin()+Idx*cno();
    
    copy(Vec.begin(), Vec.end(), Rowidx);   // STL
}
// END of row(Vec,Idx)

// col(Idx): returns the Idx-th column as a Vector_ object safely.
// col(Vec,Idx): sets the elements in the Idx-th column to the elements
// of the vector Vec, provided Idx is legal and Vec has the right
// number of elements.
template <class T_>
Vector_<T_> Rectmat_<T_>::col(size_t Idx) const
{
    if (Idx>=cno()) throw(Indexrangexc_("Rectmat_<T_>::col(Idx)", Idx, cno()-1));
    Vector_<T_> Colvec(rno());
    
    for (register unsigned int i=0; i<rno(); i++)
	Colvec[i]=elems(i*cno()+Idx);
    return(Colvec);
}
// END of col(Idx)

template <class T_>
void Rectmat_<T_>::col(const Vector_<T_>& Vec, size_t Idx)
{
    if (Vec.dim()!=rno()) throw(Dimexc_("Rectmat_<T_>::col(Vec, Idx)", Vec.dim(), rno()-1));
    if (Idx>=cno())  throw(Indexrangexc_("Rectmat_<T_>::col(Vec, Idx)", Idx, cno()-1));
    
    clone();	// modified
    for (register unsigned int i=0; i<rno(); i++)
	elems(i*cno()+Idx)=Vec[i];
}
// END of col(Vec,Idx)

// ---- Simple scalar arithmetics ----

template<class T_>
Rectmat_<T_> Rectmat_<T_>::operator*(const T_& Factor) const
{
    Rectmat_<T_> Temp=(*this);
    Temp*=Factor;
    return(Temp);
}

template<class T_>
Rectmat_<T_> operator*(const T_& Factor, const Rectmat_<T_>& Bm)
{
    Rectmat_<T_> Temp=Bm;
    Temp*=Factor;
    return(Temp);
}

template<class T_>
Rectmat_<T_> Rectmat_<T_>::operator/(const T_& Div) const
{
    Rectmat_<T_> Temp=(*this);
    Temp/=Div;
    return(Temp);
}

// ---- Simple linear algebra ----

// Matrix addition and subtraction.
// Dimexc_ is thrown on dimension mismatches.
template<class T_>
Rectmat_<T_>& Rectmat_<T_>::operator+=(const Rectmat_<T_>& Mat)
{
    if (rno()!=Mat.rno()) throw(Dimexc_("Mat1+=Mat2 (row)",rno(), Mat.rno()));
    if (cno()!=Mat.cno()) throw(Dimexc_("Mat1+=Mat2 (col)",cno(), Mat.cno()));
    clone();
    std::transform(elems().begin(), elems().end(),
	    Mat.elems().begin(), elems().begin(), 
	    plus<T_>());    // STL vector sum in place
    return(*this);
}
template<class T_>
Rectmat_<T_> Rectmat_<T_>::operator+(const Rectmat_<T_>& Mat) const
{
    Rectmat_<T_> Sum=(*this);
    Sum+=Mat;
    return Sum;
}
// END of +=, +

template<class T_>
Rectmat_<T_>& Rectmat_<T_>::operator-=(const Rectmat_<T_>& Mat)
{
    if (rno()!=Mat.rno()) throw(Dimexc_("Mat1-=Mat2 (row)", rno(), Mat.rno()));
    if (cno()!=Mat.cno()) throw(Dimexc_("Mat1-=Mat2 (col)", cno(), Mat.cno()));
    clone();
    std::transform(elems().begin(), elems().end(),
	    Mat.elems().begin(), elems().begin(), 
	    minus<T_>());    // STL vector diff in place
    return(*this);
}
template<class T_>
Rectmat_<T_> Rectmat_<T_>::operator-(const Rectmat_<T_>& Mat) const
{
    Rectmat_<T_> Dif=(*this);
    Dif-=Mat;
    return Dif;
}
// END of -=, -

#undef STD_

// Matrix*matrix and matrix*vector multiplication.
// Dimexc_ is thrown on dimension mismatches.
template<class T_>
Rectmat_<T_> Rectmat_<T_>::operator*(const Rectmat_<T_>& Mat) const
{
    if (Mat.rno()!=cno())
	throw(Dimexc_("Rectmat_*Rectmat_", Mat.rno(), cno()));

    Rectmat_<T_> Prod(rno(), Mat.cno());
    register unsigned int i, j, k;
    vector<T_>::const_iterator Rowi;	// our i:th row
    vector<T_>::iterator Prodi;     	// i:th row of the product matrix
    T_ Temp;
    
    for (i=0, Rowi=elems().begin(), Prodi=Prod.elems().begin(); 
	    i<rno();
	    i++, Rowi+=cno(), Prodi+=Prod.cno())
    {
	for (j=0; j<Mat.cno(); j++)
	{
	    Temp=0;
	    for (k=0; k<cno(); k++)
		Temp+=Rowi[k]*Mat[k][j];
	    Prodi[j]=Temp;
	}
    }
    return(Prod);
}
// END of Mat*Mat

template<class T_>
Vector_<T_> Rectmat_<T_>::operator*(const Vector_<T_>& Vec) const
{
    if (cno()!=Vec.dim()) throw(Dimexc_("Mat*Vec", cno(), Vec.dim()));
    Vector_<T_> Prod(rno());
    register unsigned int i;
    vector<T_>::const_iterator Rowi;
    T_ Temp;
    
    for (i=0, Rowi=elems().begin(); i<rno(); i++, Rowi+=cno())
    {
	Temp=inner_product(Rowi, Rowi+cno(), Vec.begin(), T_(0));
	Prod[i]=Temp;
    }
    return(Prod);
}
// END of Mat*Vec

// ---- Transpose ----

// get_transpose(): returns the transpose of the calling object.
template<class T_>
Rectmat_<T_> Rectmat_<T_>::get_transpose() const
{
    Rectmat_<T_> Tr(cno(), rno());
    register unsigned int i, j;
    vector<T_>::const_iterator Rowi;
    
    for (i=0, Rowi=elems().begin(); i<rno(); i++, Rowi+=cno())
    {
	for (j=0; j<cno(); j++)
	    Tr[j][i]=Rowi[j];
    }
    return(Tr);
}
// END of get_transpose()

// transpose_inplace(): replaces the calling object with its transpose.
// Returns a reference to the calling object.
template <class T_>
Rectmat_<T_>& Rectmat_<T_>::transpose_inplace()
{
    register unsigned int i, j;
    
    if (rno()==cno())	// simple swap for square matrices
    {
	clone();
	for (i=1; i<rno(); i++)
	    for (j=0; j<i; j++)
		swap((*this)[i][j], (*this)[j][i]); // STL
    }
    else    // layout changes for general rectangular matrices
    {
	Matstorage_<T_> *Newmat=new Matstorage_<T_>(rno()*cno());
	vector<T_>::const_iterator Rowi;

	for (i=0, Rowi=elems().begin(); i<rno(); i++, Rowi+=cno())
	{
	    for (j=0; j<cno(); j++)
		Newmat->elems()[j*rno()+i]=Rowi[j];
	}
	replace_storage(Newmat);
	swap(R, C);     // STL
    }
    return(*this);
}
// END of transpose_inplace()

// ---- Diagonal ----

// diag(): copies the diagonal into a Vector_ object.
// diag(Vec): sets the diagonal to the values in the Vector_ object Vec.
// If the dimensions don't match, a Dimexc_ exception is thrown.
// If the calling object is not a square matrix, then Squarexc_ is thrown.
template <class T_>
Vector_<T_> Rectmat_<T_>::diag() const
{
    if (rno()!=cno()) throw(Squarexc_("diag()", rno(), cno()));
    
    Vector_<T_> Diag(rno());
    for (register unsigned int i=0; i<rno(); i++)
	Diag[i]=elems((cno()+1)*i);
    return(Diag);
}
// END of diag()

template <class T_>
void Rectmat_<T_>::diag(const Vector_<T_>& Vec)
{
    if (rno()!=cno()) throw(Squarexc_("diag(V)", rno(), cno()));
    if (rno()!=Vec.dim()) throw(Dimexc_("diag(V)", rno(), Vec.dim()));

    clone();
    for (register unsigned int i=0; i<rno(); i++)
	elems((cno()+1)*i)=Vec[i];
    return;
}
// END of diag(Vec)

// diag_matrix(Dval): turns the calling object into a diagonal matrix 
// with the value Dval in all diagonal positions. Dval=="1" by default
// so diag_matrix() produces the unit matrix w/o a parameter.
// If the matrix is not square, a Squarexc_ exception is thrown.
template <class T_>
void Rectmat_<T_>::diag_matrix(const T_& Dval)
{
    if (rno()!=cno()) throw(Squarexc_("diag_matrix()", rno(), cno()));

    set_values();   // zeroes (invokes clone() as well)
    for (register unsigned int i=0; i<rno(); i++)
	elems((cno()+1)*i)=Dval;
}
// END of diag_matrix(Dval)

// trace(): returns the sum of the diagonal elements.
// If the matrix is not square, a Squarexc_ exception is thrown.
template <class T_>
T_ Rectmat_<T_>::trace() const
{
    if (rno()!=cno()) throw(Squarexc_("diag_matrix()", rno(), cno()));
    T_ Trace(0);
    for (register unsigned int i=0; i<rno(); i++)
	Trace+=elems((cno()+1)*i);
    return(Trace);
}
// END of trace()

// ---- I/O ----

// read_matrix(In): reads into the calling object from In.
// Invoked by >>. Protected
template <class T_>
void Rectmat_<T_>::read_matrix(istream& In)
{
#define LINELEN 120
    static char Line[LINELEN+1];
    
    Rectmat_<T_> Temp;
    size_t Lineno=0, Dashlen, Cidx=0, Newcidx, Cxtemp, 
	    Ridx=0, Rxtemp, Rexpect;
    char *Eolnptr;
    bool Colidx=false, Topdelim, Chunkdelim, Botdelim=false;
    char Pipe;
    T_ Tempcoord;
    
    // read input line by line
    while (In.getline(Line, LINELEN, '\n').good())
    {
	++Lineno;
	// skip empty lines or comments
	if (!Line[0] || Line[0]=='\n' || Line[0]=='#')
	    continue;
	
	// chop off terminating \n if present
	Eolnptr=strrchr(Line, '\n');
	if (Eolnptr!=NULL) *Eolnptr='\0';
	
	// look for column index line (on top)
	// the next chunk will be between Cidx..Newcidx
	if (!Colidx)
	{
	    istringstream Cidxline(Line);
	    for (Newcidx=Cidx; Cidxline>>Cxtemp && Cxtemp==Newcidx; ++Newcidx);
	    if (Cidxline.eof() && Newcidx>Cidx)
	    {
		Temp.size(Ridx, Newcidx);
		Colidx=true;
		Topdelim=false; 
	    }
	    continue;
	}
	
	// col idx line seen, check for top delimiter "----"
	if (!Topdelim)
	{
	    Dashlen=strspn(Line, "-");
	    if (Dashlen==strlen(Line))
	    {
		Topdelim=true;	// found top delimiter
		Chunkdelim=false;
		Rexpect=0;  // expected row number
	    }
	    continue;
	}
	
	// check for bottom delimiter "===" if rows have been seen
	if (Ridx && Topdelim && !Botdelim)
	{
	    Dashlen=strspn(Line, "=");
	    if (Dashlen==strlen(Line))
	    {
		Botdelim=true;    // found bottom delimiter
		break;	// we're done
	    }
	}
	
	// check for chunk delimiter "---" if rows have been seen
	if (Ridx && Topdelim && !Chunkdelim)
	{
	    Dashlen=strspn(Line, "-");
	    if (Dashlen==strlen(Line))
	    {
		Chunkdelim=true;    // found chunk delimiter
		Colidx=false;	// look for next column index line
		Cidx=Newcidx;
	    	continue;
	    }
	}
	
	// get the rows
	// look for data line "x | yyyy zzzz ..."
	istringstream Instr(Line); 
	Instr>>Rxtemp;
	if (!Instr.good() || Rxtemp!=Rexpect)
	{
	    cerr<<"\n? >>Rectmat_: Row index "<<Rexpect<<" expected (line "
		    <<Lineno<<")\n";
	    break;
	}
	++Rexpect;  // next row index to be expected
	if (Rexpect>Ridx)   // first chunk
	{
	    Ridx=Rexpect;   // count rows
	    Temp.size(Ridx, Newcidx);	// make room for new row
	}
	Instr>>Pipe;
	if (!Instr.good() || Pipe!='|')
	{
	    cerr<<"\n? >>Rectmat_: Separator \'|\' expected, \'"
		    <<Pipe<<"\' found (line "<<Lineno<<")\n";
	    break;
	}
	
	// read the matrix elements in the [Rxtemp] row chunk
	for (Cxtemp=Cidx; Cxtemp<Newcidx && Instr>>Tempcoord; ++Cxtemp)
	    Temp[Rxtemp][Cxtemp]=Tempcoord;
	if (!(Instr.good() || Instr.eof()))
	{
	    cerr<<"\n? >>Rectmat_: Data field problem (line "<<Lineno<<")\n";
	    break;
	}
    }	    // while
    
    // input was OK if bottom delimiter has been seen
    if (Botdelim) *this=Temp;
    else In.clear(ios::failbit);    // set error condition
#undef LINELEN
}
// END of read_matrix(In)

// print_rows(): print the rows of a rectangular matrix so that
// if the matrix is big then it is cut up into chunks nicely.
// Protected
template <class T_>
void Rectmat_<T_>::print_rows(ostream& Out, size_t Rw, size_t Jbeg, 
	size_t Items, size_t Width, size_t Prec) const
{
    vector<T_>::const_iterator Rowi;
    unsigned int i;
    
    for (i=0, Rowi=elems().begin(); i<rno(); i++, Rowi+=cno())
    {
	Out.unsetf(ios::left);
	Out << setw(Rw) << i << setw(0) << " | " ;   // row idx
	Out.setf(ios::left);
	
	// list a row: either all items or Items which fit on one line
	for (unsigned int j=Jbeg; j<cno() && j<Jbeg+Items; j++)
	    Out << setw(Width) << setprecision(Prec) << Rowi[j];
	Out << endl;
    }
}
// END of print_rows()

} // RazorBack

// ==== END OF TEMPLATE METHODS Rectmat.cc ====

#endif      // RECTMAT_TMPL_METHODS
