#ifndef SYMMAT_TMPL_METHODS
#define SYMMAT_TMPL_METHODS

// ==== TEMPLATE METHODS Symmat.cc ====

// 20-Apr-1999. Andras Aszodi

namespace RazorBack {

// ---- Conversions to/from Rectmat ----

// Symmat->Rectmat conversion
template <class T_>
Symmat_<T_>::operator Rectmat_<T_>() const
{
    Rectmat_<T_> Rect(rno(), rno());	// make it symmetric
    for (register unsigned int i=0; i<rno(); i++)
    {
	Rect[i][i]=elems(tri_size(i)+i);
	for (register unsigned int j=0; j<i; j++)
	    Rect[i][j]=Rect[j][i]=elems(tri_size(i)+j);
    }
    return(Rect);
}
// END of operator Rectmat_<T_>

// triangle(Rect, Low): copies the lower triangle of Rect
// to the calling object if Low is true (default) or the upper triangle
// if Low is false. Rectmat must be square, otherwise Squarexc_ is thrown.
// The calling object is resized silently.
// Returns reference to calling object.
template <class T_>
Symmat_<T_>& Symmat_<T_>::triangle(const Rectmat_<T_>& Rect, bool Low)
{
    if (Rect.rno()!=Rect.cno())
	throw(Squarexc_("triangle(Rectmat)", Rect.rno(), Rect.cno()));

    size(Rect.rno());
    clone();	// needed if size did not change anything
    for (register unsigned int i=0; i<rno(); i++)
    {
	elems(tri_size(i)+i)=Rect[i][i];
	for (register unsigned int j=0; j<i; j++)
	    elems(tri_size(i)+j)=Low? Rect[i][j]: Rect[j][i];
    }
    return(*this);
}
// END of triangle(Rect, Low)
    
// ---- Access ----

// The two-argument function call operator provides optional checked access
// when Check is true, and FORTRAN-style indexing when Ftnidx is true.
// Indices are always swapped silently so that Sym(i,j) is Sym[j][i]
// when i<j.
template <class T_>
const T_& Symmat_<T_>::operator()(size_t Ridx, size_t Cidx) const
{
    if (ftn_idx()) { Ridx--; Cidx--; }
    if (range_check())
    {
	if (Ridx>=rno()) throw(Indexrangexc_("Symmat_[ROW][](const)", Ridx, rno()-1));
	if (Cidx>=cno()) throw(Indexrangexc_("Symmat_[][COL](const)", Cidx, cno()-1));
    }
    return at(Ridx,Cidx);
}
template <class T_>
T_& Symmat_<T_>::operator()(size_t Ridx, size_t Cidx)
{
    if (ftn_idx()) { Ridx--; Cidx--; }
    if (range_check())
    {
	if (Ridx>=rno()) throw(Indexrangexc_("Symmat_[ROW][](lval)", Ridx, rno()-1));
	if (Cidx>=cno()) throw(Indexrangexc_("Symmat_[][COL](lval)", Cidx, cno()-1));
    }
    return at(Ridx, Cidx);  // non-const at() does the cloning
}
// END of operator(Ridx,Cidx)

// size(S): changes the size of the calling object to S x S.
// Nothing is done if the size parameter is the same as the current
// size or if it is 0. When the matrix "shrinks", then
// the upper left corner is preserved, when it grows, then the new
// rows and columns are padded with "zeroes" T_(0).
template <class T_>
void Symmat_<T_>::size(size_t S)
{
    if (!S || rno()==S) return;   // no change
    
    clone();
    eno(tri_size(S));	// appropriate resize
    Size=S;
}
// END of size(S)

// row(Idx): returns the Idx-th row as a Vector_ object safely.
// row(Vec,Idx): sets the elements in the Idx-th row to the elements
// of the vector Vec, provided Idx is legal and Vec has the right
// number of elements.
// Throws Indexrangexc_ on error.
template <class T_>
Vector_<T_> Symmat_<T_>::row(size_t Idx) const
{
    if (Idx>=rno()) throw(Indexrangexc_("Symmat_<T_>::row/col(Idx)", Idx, rno()-1));
    Vector_<T_> Rowvec(cno());
    vector<T_>::const_iterator Rowidx=elems().begin()+tri_size(Idx);
    
    // STL copy of row up to and including main diagonal
    copy(Rowidx, Rowidx+Idx+1, Rowvec.begin());
    
    // traditional elem-by-elem copy for the column under the main diag
    for (register unsigned int i=Idx+1; i<rno(); i++)
	Rowvec[i]=elems(tri_size(i)+Idx); // copy col

    return(Rowvec);
}
// END of row(Idx)

template <class T_>
void Symmat_<T_>::row(const Vector_<T_>& Vec, size_t Idx)
{
    if (Vec.dim()!=cno()) throw(Dimexc_("Symmat_<T_>::row/col(Vec, Idx)", Vec.dim(), cno()));
    if (Idx>=rno()) throw(Indexrangexc_("Symmat_<T_>::row/col(Vec, Idx)", Idx, rno()-1));

    clone();	// modified
    vector<T_>::iterator Rowidx=elems().begin()+tri_size(Idx);
    
    // STL copy of row up to and including main diagonal
    copy(Vec.begin(), Vec.begin()+Idx+1, Rowidx);
    
    // traditional elem-by-elem copy for the column under the main diag
    for (register unsigned int i=Idx+1; i<rno(); i++)
	elems(tri_size(i)+Idx)=Vec[i]; // copy col

}
// END of row(Vec,Idx)

// ---- Simple scalar arithmetics ----

template<class T_>
Symmat_<T_> Symmat_<T_>::operator*(const T_& Factor) const
{
    Symmat_<T_> Temp=(*this);
    Temp*=Factor;
    return(Temp);
}

template<class T_>
Symmat_<T_> operator*(const T_& Factor, const Symmat_<T_>& Bm)
{
    Symmat_<T_> Temp=Bm;
    Temp*=Factor;
    return(Temp);
}

template<class T_>
Symmat_<T_> Symmat_<T_>::operator/(const T_& Div) const
{
    Symmat_<T_> Temp=(*this);
    Temp/=Div;
    return(Temp);
}

// ---- Simple linear algebra ----

// Matrix addition and subtraction.
// Dimexc_ is thrown on dimension mismatches.
template<class T_>
Symmat_<T_>& Symmat_<T_>::operator+=(const Symmat_<T_>& Mat)
{
    if (rno()!=Mat.rno()) throw(Dimexc_("Symmat1+=Symmat2", rno(), Mat.rno()));
    clone();
    std::transform(elems().begin(), elems().end(),
	    Mat.elems().begin(), elems().begin(), 
	    plus<T_>());    // STL vector sum in place
    return(*this);
}
template<class T_>
Symmat_<T_> Symmat_<T_>::operator+(const Symmat_<T_>& Mat) const
{
    Symmat_<T_> Sum=(*this);
    Sum+=Mat;
    return Sum;
}
template<class T_>
Rectmat_<T_> Symmat_<T_>::operator+(const Rectmat_<T_>& Mat) const
{
    if (rno()!=Mat.rno())
	throw(Dimexc_("Symmat+Rectmat (row)", rno(), Mat.rno()));
    if (cno()!=Mat.cno())
	throw(Dimexc_("Symmat+Rectmat (col)", cno(), Mat.cno()));
    
    Rectmat_<T_> Sum=Mat;
    T_ Temp;
    register unsigned int i, j;
    
    for (i=0; i<rno(); i++)
    {
	Sum[i][i]+=elems(tri_size(i)+i);
	for (j=0; j<i; j++)
	{
	    Temp=elems(tri_size(i)+j);
	    Sum[i][j]+=Temp; Sum[j][i]+=Temp;
	}
    }
    return Sum;
}
// END of +=, +

template<class T_>
Symmat_<T_>& Symmat_<T_>::operator-=(const Symmat_<T_>& Mat)
{
    if (rno()!=Mat.rno()) throw(Dimexc_("Symmat1-=Symmat2", rno(), Mat.rno()));
    clone();
    std::transform(elems().begin(), elems().end(),
	    Mat.elems().begin(), elems().begin(), 
	    minus<T_>());    // STL vector diff in place
    return(*this);
}
template<class T_>
Symmat_<T_> Symmat_<T_>::operator-(const Symmat_<T_>& Mat) const
{
    Symmat_<T_> Dif=(*this);
    Dif-=Mat;
    return Dif;
}
template<class T_>
Rectmat_<T_> operator-(const Rectmat_<T_>& Mat, const Symmat_<T_>& Sym)
{
    if (rno()!=Mat.rno())
	throw(Dimexc_("Symmat+Rectmat (row)", rno(), Mat.rno()));
    if (cno()!=Mat.cno())
	throw(Dimexc_("Symmat+Rectmat (col)", cno(), Mat.cno()));
    
    Rectmat_<T_> Dif=Mat;
    T_ Temp;
    register unsigned int i, j;
    
    for (i=0; i<rno(); i++)
    {
	Dif[i][i]-=Sym[i][i];
	for (j=0; j<i; j++)
	{
	    Temp=Sym[i][j];
	    Dif[i][j]-=Temp; Dif[j][i]-=Temp;
	}
    }
    return Dif;
}
// END of -=, -

// Matrix*matrix and matrix*vector multiplication.
// Mixed Symmat/Rectmat operators are provided to avoid conversions.
// Dimexc_ is thrown on dimension mismatches.
template<class T_>
Rectmat_<T_> Symmat_<T_>::operator*(const Symmat_<T_>& Mat) const
{
    if (Mat.rno()!=cno())
	throw(Dimexc_("Symmat_*Symmat_", Mat.rno(), cno()));

    Rectmat_<T_> Prod(rno(), Mat.cno());
    register unsigned int i, j, k;
    T_ Temp;
    
    for (i=0; i<rno(); i++)
	for (j=0; j<Mat.cno(); j++)
	{
	    Temp=0;
	    for (k=0; k<cno(); k++)
		Temp+=at(i,k)*Mat.at(k,j);
	    Prod[i][j]=Temp;
	}
    return(Prod);
}

template<class T_>
Rectmat_<T_> Symmat_<T_>::operator*(const Rectmat_<T_>& Mat) const
{
    if (Mat.rno()!=cno())
	throw(Dimexc_("Symmat_*Rectmat_", Mat.rno(), cno()));

    Rectmat_<T_> Prod(rno(), Mat.cno());
    register unsigned int i, j, k;
    T_ Temp;
    
    for (i=0; i<rno(); i++)
	for (j=0; j<Mat.cno(); j++)
	{
	    Temp=0;
	    for (k=0; k<cno(); k++)
		Temp+=at(i,k)*Mat[k][j];
	    Prod[i][j]=Temp;
	}
    return(Prod);
}

template<class T_>
Rectmat_<T_> operator*(const Rectmat_<T_>& Rect, const Symmat_<T_>& Sym)
{
    if (Sym.rno()!=Rect.cno())
	throw(Dimexc_("Rectmat_*Symmat_", Sym.rno(), Rect.cno()));

    Rectmat_<T_> Prod(Rect.rno(), Sym.cno());
    register unsigned int i, j, k;
    T_ Temp;
    
    for (i=0; i<Rect.rno(); i++)
	for (j=0; j<Sym.cno(); j++)
	{
	    Temp=0;
	    for (k=0; k<Rect.cno(); k++)
		Temp+=Rect[i][k]*Sym.at(k,j);
	    Prod[i][j]=Temp;
	}
    return(Prod);
}
// END of Mat*Mat

template<class T_>
Vector_<T_> Symmat_<T_>::operator*(const Vector_<T_>& Vec) const
{
    if (cno()!=Vec.dim()) throw(Dimexc_("Symmat*Vec", cno(), Vec.dim()));
    Vector_<T_> Prod(rno());
    register unsigned int i, j;
    T_ Temp;
    
    for (i=0; i<rno(); i++)
    {
	Temp=0;
	for (j=0; j<cno(); j++)
	    Temp+=at(i, j)*Vec[j];
	Prod[i]=Temp;
    }
    return(Prod);
}
// END of Mat*Vec

// ---- Diagonal ----

// diag(): copies the diagonal into a Vector_ object.
// diag(Vec): sets the diagonal to the values in the Vector_ object Vec.
// If the dimensions don't match, a Dimexc_ exception is thrown.
template <class T_>
Vector_<T_> Symmat_<T_>::diag() const
{
    Vector_<T_> Diag(rno());
    for (register unsigned int i=0; i<rno(); i++)
	Diag[i]=elems(tri_size(i)+i);
    return(Diag);
}
// END of diag()

template <class T_>
void Symmat_<T_>::diag(const Vector_<T_>& Vec)
{
    if (rno()!=Vec.dim()) throw(Dimexc_("Symmat_<T_>::diag(V)", rno(), Vec.dim()));

    clone();
    for (register unsigned int i=0; i<rno(); i++)
	elems(tri_size(i)+i)=Vec[i];
    return;
}
// END of diag(Vec)

// diag_matrix(Dval): turns the calling object into a diagonal matrix 
// with the value Dval in all diagonal positions. Dval=="1" by default
// so diag_matrix() produces the unit matrix w/o a parameter.
template <class T_>
void Symmat_<T_>::diag_matrix(const T_& Dval)
{
    set_values();   // zeroes (invokes clone())
    for (register unsigned int i=0; i<rno(); i++)
	elems(tri_size(i)+i)=Dval;
}
// END of diag_matrix(Dval)

// trace(): returns the sum of the diagonal elements.
template <class T_>
T_ Symmat_<T_>::trace() const
{
    T_ Trace(0);
    for (register unsigned int i=0; i<rno(); i++)
	Trace+=elems(tri_size(i)+i);
    return(Trace);
}
// END of trace()

// ---- I/O ----

// read_matrix(In): reads into the calling object from In.
// Invoked by >>. Protected
template <class T_>
void Symmat_<T_>::read_matrix(istream& In)
{
#define LINELEN 120
    static char Line[LINELEN+1];
    
    Symmat_<T_> Temp;
    size_t Lineno=0, Dashlen, Cidx=0, Newcidx, Cxtemp, 
	    Ridx=0, Rxtemp, Rexpect;
    char *Eolnptr;
    bool Colidx=false, Topdelim, Chunkdelim, Botdelim=false;
    char Pipe;
    T_ Tempcoord;
    
    // read input line by line
    while (In.getline(Line, LINELEN, '\n').good())
    {
	++Lineno;
	// skip empty lines or comments
	if (!Line[0] || Line[0]=='\n' || Line[0]=='#')
	    continue;
	
	// chop off terminating \n if present
	Eolnptr=strrchr(Line, '\n');
	if (Eolnptr!=NULL) *Eolnptr='\0';
	
	// look for column index line (on top)
	// the next chunk will be between Cidx..Newcidx
	if (!Colidx)
	{
	    istringstream Cidxline(Line);
	    for (Newcidx=Cidx; Cidxline>>Cxtemp && Cxtemp==Newcidx; ++Newcidx);
	    if (Cidxline.eof() && Newcidx>Cidx)
	    {
		Temp.size(max(Ridx, Newcidx));
		Colidx=true;
		Topdelim=false; 
	    }
	    continue;
	}
	
	// col idx line seen, check for top delimiter "----"
	if (!Topdelim)
	{
	    Dashlen=strspn(Line, "-");
	    if (Dashlen==strlen(Line))
	    {
		Topdelim=true;	// found top delimiter
		Chunkdelim=false;
		Rexpect=0;  // expected row number
	    }
	    continue;
	}
	
	// check for bottom delimiter "===" if rows have been seen
	if (Ridx && Topdelim && !Botdelim)
	{
	    Dashlen=strspn(Line, "=");
	    if (Dashlen==strlen(Line))
	    {
		Botdelim=true;    // found bottom delimiter
		break;	// we're done
	    }
	}
	
	// check for chunk delimiter "---" if rows have been seen
	if (Ridx && Topdelim && !Chunkdelim)
	{
	    Dashlen=strspn(Line, "-");
	    if (Dashlen==strlen(Line))
	    {
		Chunkdelim=true;    // found chunk delimiter
		Colidx=false;	// look for next column index line
		Cidx=Newcidx;
	    	continue;
	    }
	}
	
	// get the rows
	// look for data line "x | yyyy zzzz ..."
	istringstream Instr(Line); 
	Instr>>Rxtemp;
	if (!Instr.good() || Rxtemp!=Rexpect)
	{
	    cerr<<"\n? >>Rectmat_: Row index "<<Rexpect<<" expected (line "
		    <<Lineno<<")\n";
	    break;
	}
	++Rexpect;  // next row index to be expected
	if (Rexpect>Ridx)   // first chunk
	{
	    Ridx=Rexpect;   // count rows
	    Temp.size(max(Ridx, Newcidx));	// make room for new row
	}
	Instr>>Pipe;
	if (!Instr.good() || Pipe!='|')
	{
	    cerr<<"\n? >>Rectmat_: Separator \'|\' expected, \'"
		    <<Pipe<<"\' found (line "<<Lineno<<")\n";
	    break;
	}
	
	// read the matrix elements in the [Rxtemp] row chunk
	for (Cxtemp=Cidx; 
		Cxtemp<Newcidx && Cxtemp<=Rxtemp && Instr>>Tempcoord; ++Cxtemp)
	    Temp[Rxtemp][Cxtemp]=Tempcoord;
	if (!(Instr.good() || Instr.eof()))
	{
	    cerr<<"\n? >>Rectmat_: Data field problem (line "<<Lineno<<")\n";
	    break;
	}
    }	    // while
    
    // input was OK if bottom delimiter has been seen
    if (Botdelim) *this=Temp;
    else In.clear(ios::failbit);    // set error condition
#undef LINELEN
}
// END of read_matrix(In)

// print_rows(): print the rows of a symmetric matrix so that
// if the matrix is big then it is cut up into chunks nicely.
// Protected
template <class T_>
void Symmat_<T_>::print_rows(ostream& Out, size_t Rw, size_t Jbeg, 
	size_t Items, size_t Width, size_t Prec) const
{
    unsigned int i;
    
    for (i=0; i<rno(); i++)
    {
	Out.unsetf(ios::left);
	Out << setw(Rw) << i << setw(0) << " | " ;   // row idx
	Out.setf(ios::left);
	
	// list a row: either all items or Items which fit on one line
	for (unsigned int j=Jbeg; j<=i && j<Jbeg+Items; j++)
	    Out << setw(Width) << setprecision(Prec) << at(i, j);
	Out << endl;
    }
}
// END of print_rows()

} // RazorBack

// ==== END OF TEMPLATE METHODS Symmat.cc ====

#endif      // SYMMAT_TMPL_METHODS
