#ifndef VECTOR_TMPL_METHODS
#define VECTOR_TMPL_METHODS

// ==== TEMPLATE METHODS Vector.cc ====

// Linear algebra vector template class based on STL. */

// 12-Jul-1999. Andris Aszodi

// ---- COMPATIBILITY NOTES ----

// Exception support is assumed.
// This module depends on the Standard Template Library.

// ---- COMPILER COMMAND LINE MACROS ----

// Define the macro INCLUDE_TMPL_DEFS if your compiler
// does not support automatic template source inclusion.

// ---- STANDARD HEADERS ----

#include <string.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <sstream>
using namespace std;
#else
#include <strstream.h>
#define istringstream istrstream 
#endif

// ==== METHODS ====
namespace RazorBack {

// (): allows range-checked and FORTRAN-style indexing.
// Throws Indexrangexc_ on range errors.
template <class T_>
inline
const T_& Vector_<T_>::operator()(size_t Idx) const
{
    if (ftn_idx()) Idx--;
    if (range_check() && Idx>=dim()) throw Indexrangexc_("Vector_(Idx) (const)", Idx, dim()-1);
    return(Vec[Idx]);
}
template <class T_>
inline
T_& Vector_<T_>::operator()(size_t Idx)
{
    if (ftn_idx()) Idx--;
    if (range_check() && Idx>=dim()) throw Indexrangexc_("Vector_(Idx)", Idx, dim()-1);
    return(Vec[Idx]);
}

// set_values(): sets all coordinates to X (0 by default).
template <class T_>
Vector_<T_>& Vector_<T_>::set_values(const T_& X)
{
    fill(Vec.begin(), Vec.end(), X);
    return(*this);
}

// dim(): returns the dimension (size) of the vector.
// dim(D): adjusts the dimension to D or to 3 if D==0.
// Returns previous dimension.
template <class T_>
size_t Vector_<T_>::dim(size_t D)
{
    size_t Oldim=dim();
    D=!D? 3: D;
    if (D!=Oldim)
    {
	if (D>Oldim)
	    Vec.insert(Vec.end(), D-Oldim, T_(0));    // storage grows, pad w/ "0"
	else
	    Vec.erase(Vec.begin()+D, Vec.end());	// remove high-dim. coords
	Dim=D;  // store new dimension
    }
    return(Oldim);
}

// Scalar arithmetics: add, subtract, multiply or divide 
// each coordinate the T_ type object X "in place". The modified
// calling object is returned.
template <class T_>
Vector_<T_>& Vector_<T_>::operator+=(const T_& X)
{
    for (register unsigned int i=0; i<dim(); i++) Vec[i]+=X;
    return(*this);
}
template <class T_>
Vector_<T_>& Vector_<T_>::operator-=(const T_& X)
{
    for (register unsigned int i=0; i<dim(); i++) Vec[i]-=X;
    return(*this);
}
template <class T_>
Vector_<T_>& Vector_<T_>::operator*=(const T_& X)
{
    for (register unsigned int i=0; i<dim(); i++) Vec[i]*=X;
    return(*this);
}
template <class T_>
Vector_<T_> Vector_<T_>::operator*(const T_& X) const
{
    Vector_<T_> Tmp=*this;
    Tmp*=X;
    return(Tmp);
}
template <class T_>
Vector_<T_> operator*(const T_& X, const Vector_<T_>& V) // X*Vec
{
    Vector_<T_> Tmp=V;
    Tmp*=X;
    return(Tmp);
}
template <class T_>
Vector_<T_>& Vector_<T_>::operator/=(const T_& X)
{
    register unsigned int i;
    for (i=0; i<dim(); i++) Vec[i]/=X;
    return(*this);
}
template <class T_>
Vector_<T_> Vector_<T_>::operator/(const T_& X) const
{
    Vector_<T_> Tmp=*this;
    Tmp/=X;
    return(Tmp);
}

// Vector arithmetics: addition, subtraction, scalar product.
// The operands must have equal dimensions otherwise a Dimexc_ is thrown.

template <class T_>
Vector_<T_>& Vector_<T_>::operator+=(const Vector_<T_>& V)
{
    if (dim()!=V.dim()) throw Dimexc_("Vector_+=Vector_", dim(), V.dim());
    // uses STL
    std::transform(Vec.begin(), Vec.end(), V.Vec.begin(), 
	Vec.begin(), plus<T_>());
    return(*this);
}
template <class T_>
Vector_<T_> Vector_<T_>::operator+(const Vector_<T_>& V) const
{
    Vector_<T_> Tmp=*this;
    Tmp+=V;
    return(Tmp);
}
template <class T_>
Vector_<T_>& Vector_<T_>::operator-=(const Vector_<T_>& V)
{
    if (dim()!=V.dim()) throw Dimexc_("Vector_-=Vector_", dim(), V.dim());
    // uses STL
    std::transform(Vec.begin(), Vec.end(), V.Vec.begin(), 
	Vec.begin(), minus<T_>());
    return(*this);
}
template <class T_>
Vector_<T_> Vector_<T_>::operator-(const Vector_<T_>& V) const
{
    Vector_<T_> Tmp=*this;
    Tmp-=V;
    return(Tmp);
}
template <class T_>
T_ Vector_<T_>::operator*(const Vector_<T_>& V) const
{
    if (dim()!=V.dim())
	throw Dimexc_("Vector_*Vector_ (scalar product)", dim(), V.dim());
    // uses STL
    return(inner_product(Vec.begin(), Vec.end(), V.Vec.begin(), T_(0)));
}

// cross_prod(): returns the cross-product of two 3D vectors.
// If either of the arguments are not 3D, then a Dimexc_ exception
// is thrown.
template <class T_>
Vector_<T_> cross_prod(const Vector_<T_>& V1, const Vector_<T_>& V2)
{
    // exception stores the first "wrong" dimension and 3
    if (V1.dim()!=3)
	throw Dimexc_("Vector_ x Vector_ (cross product)", V1.dim(), 3);
    if (V2.dim()!=3)
	throw Dimexc_("Vector_ x Vector_ (cross product)", 3, V2.dim());
    
    Vector_<T_> Temp(3);
    Temp[0]=V1[1]*V2[2]-V1[2]*V2[1];
    Temp[1]=V1[2]*V2[0]-V1[0]*V2[2];
    Temp[2]=V1[0]*V2[1]-V1[1]*V2[0];
    return(Temp);
}
// END of cross_prod()

// Modulus
// vec_norm(): normalises the calling object to a unit vector "in place".
// Returns the original length. The type T_ must support
// a sqrt(T_) operation. "Null"-vectors should ideally
// throw a "division by zero" exception later.
template <class T_>
T_ Vector_<T_>::vec_norm()
{
    T_ L=vec_len();
    (*this)/=L;		// normalise
    return(L);
}
// END of vec_norm()

// diff_len2(): returns the squared length of V1-V2. On dim
// mismatch a Dimexc_ exception is thrown.
template <class T_>
T_ diff_len2(const Vector_<T_>& V1, const Vector_<T_>& V2)
{
    register unsigned int Dim=V1.dim();
    if (Dim!=V2.dim())
    {
	throw Dimexc_("diff_len2(V1, V2)", Dim, V2.dim());
    }
    register unsigned int i;
    T_ D, L=0;
    for (i=0; i<Dim; i++)
    {
	D=V1[i]-V2[i];
	L+=D*D;
    }
    return(L);
}
// END of diff_len2()

// ---- I/O ----

#define LINEWIDTH 80

// read(In): reads from In into the calling object. 
template <class T_>
void Vector_<T_>::read(istream& In)
{
    static char Line[LINEWIDTH+1];
    vector<T_> Temp;
    bool Topdelim=false, Botdelim=false;
    char *Eolnptr;
    size_t Tempsize=0, Idx, Dashlen, Lineno=0;
    T_ Tempcoord;
    char Pipe;
    
    // read input line by line
    while (In.getline(Line, LINEWIDTH, '\n').good())
    {
	++Lineno;
	// skip empty lines or comments
	if (!Line[0] || Line[0]=='\n' || Line[0]=='#')
	    continue;
	
	// chop off terminating \n if present
	Eolnptr=strrchr(Line, '\n');
	if (Eolnptr!=NULL) *Eolnptr='\0';
	
	if (!Topdelim)	    // look for top delimiter '----'
	{
	    Dashlen=strspn(Line, "-");
	    if (Dashlen==strlen(Line))
		Topdelim=true;	// found top delimiter
	    continue;
	}
	
	if (Tempsize && !Botdelim)  // check for bottom delimiter "==="
	{
	    Dashlen=strspn(Line, "=");
	    if (Dashlen==strlen(Line))
	    {
		Botdelim=true;
		break;
	    }
	}
	
	// look for data line "x | yyyy"
	istringstream Instr(Line);
	Instr>>Idx;
	if (!Instr.good() || Idx!=Tempsize)
	{
	    cerr<<"\n? >>Vector_: Index "<<Tempsize<<" expected (line "
		    <<Lineno<<")\n";
	    break;
	}
	Instr>>Pipe;
	if (!Instr.good() || Pipe!='|')
	{
	    cerr<<"\n? >>Vector_: Separator \'|\' expected, \'"
		    <<Pipe<<"\' found (line "<<Lineno<<")\n";
	    break;
	}
	Instr>>Tempcoord;
	if (!(Instr.good() || Instr.eof()))
	{
	    cerr<<"\n? >>Vector_: Data field problem (line "<<Lineno<<")\n";
	    break;
	}
	
	// may save temp coord
	Temp.push_back(Tempcoord);
	++Tempsize;
    }	    // while
    
    // input was OK if both delimiters have been found
    if (Topdelim && Botdelim)
    {
	Vec=Temp;	// store temp data
	dim(Tempsize);    // adjust dimension
    }
    else In.clear(ios::failbit);    // set error condition
}
// END of read(In)

// output
template <class T_>
ostream& operator<<(ostream& Out, const Vector_<T_>& V)
{
// NOTE: the "new" ISO C++ standard iostream library has different types
// and class structure than the "old" iostream. Until all the compilers
// migrate to this (notable exception is the buggy SGI "new" iostream),
// we provide a workaround as follows.
#ifdef __USE_STD_IOSTREAM
#define IOS_FLAGTYPE ios_base::fmtflags
#define IOS_BASE ios_base
#else
#define IOS_FLAGTYPE long
#define IOS_BASE ios
#endif
    static const size_t WIDTH=13, PREC=6;   // for float/double
    
    // figure out how many chars are needed for the index
    size_t Nwidth=int(ceil(log10(double(V.dim()))));

    // make the top and bottom delimiters
    char Topdelim[LINEWIDTH+1], Botdelim[LINEWIDTH+1];
    size_t Delimwidth=Nwidth+3+WIDTH;
    memset(Topdelim, '-', Delimwidth); Topdelim[Delimwidth]='\0';
    memset(Botdelim, '=', Delimwidth); Botdelim[Delimwidth]='\0';
    
    // output at last
    Out<<Topdelim<<endl;
    IOS_FLAGTYPE Oldflags=Out.flags();	// store previous settings here
    int Oldprec=Out.precision(PREC);
    
    Out.unsetf(IOS_BASE::right|IOS_BASE::internal);   // right-justified by default
    Out.setf(IOS_BASE::showpoint|IOS_BASE::scientific);	// 0.000000e+00 in C
    
    for (size_t i=0; i<V.dim(); ++i)
	Out<<setw(Nwidth)<<i<<setw(0)<<" | "
		<<V[i]<<endl;
    Out<<Botdelim<<endl;
    
    Out.width(0);	    // each item uses its length as width
    Out.precision(Oldprec);	    // old precision
    Out.flags(Oldflags);    // reset flags to state before call
    return(Out);
}
// END of <<

#undef LINEWIDTH

} // RazorBack

// ==== END OF TEMPLATE METHODS Vector.cc ====

#endif	// VECTOR_TMPL_METHODS



