#ifndef BITS_HEADER
#define BITS_HEADER

// ==== HEADER Bits.hh ====

/**
 * \file Bits.hh
 * A compressed bit-array class for storing Boolean data.
 */

// 12-Jul-1999. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stddef.h>
#include <string.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
#include <climits>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#include <limits.h>
#endif

// ==== MODULE HEADERS ====

#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Bits_ : A bit-array class. Bits are stored in 
 * an array of unsigned ints. Individual bits can be set and tested
 * and global bitwise Boolean operations can be performed.
 */
class Bits_
{
    // data
    private:
    
    static const size_t BITS_PER_CHAR;     ///< bits per char
    static const size_t BITS_PER_UINT;	    ///< bits per unsigned int
  
    unsigned int *B;	///< the bit array
    size_t Cs, Bs;	///< no. of uints and bits
    
    // methods
    public:
    
    /**
    * Inits to hold N bits (default 0). Set all bits to 
    * Val (false by default).
    */
    explicit Bits_(unsigned int N=0, bool Val=false);
    
    /// copy constructor
    Bits_(const Bits_& Bits);
	
    /**
    * Inits by a string Str which should have the format "110101001",
    * 0 for "false", 1 for "true", 1st char is MSB, last is LSB.
    * Any char other than '0' or '1' will elicit a warning
    * and will be interpreted as "true". if Str==NULL or if it is the
    * empty string "" then an empty object will be created. 
    */
    explicit Bits_(const char *Str);
    
    /// destructor
    ~Bits_() { delete [] B; }
	
    /// assignment
    Bits_& operator=(const Bits_& Rhs);
    
    /**
    * Two Bits_ arrays are equal if they have the same number of bits
    * and all bits are set in the same way.
    */
    bool operator==(const Bits_& Other) const;
    
    /// Non-equality.
    bool operator!=(const Bits_& Other) const;
    
    /// The number of uints used in storing the bit array.
    unsigned int cno() const { return(Cs); }
    
    /**
    * get_bit(): returns the Boolean value of the Idx-th bit.
    * Throws Indexrangexc_ if Idx is out of range.
    * Throws Emptyexc_ if the calling object is empty.
    */
    bool get_bit(unsigned int Idx) const;
    
    /**
    * set_bit(): sets the Idx-th bit to Value (default is true).
    * Returns old value.
    * Throws Indexrangexc_ if Idx is out of range.
    * Throws Emptyexc_ if the calling object is empty.
    */
    bool set_bit(unsigned int Idx, bool Value=true);
    
    /// set_values(): sets all bits to Value (default is true).
    void set_values(bool Value=true);
    
    /// Returns the number of bits in the array.
    unsigned int len() const { return(Bs); }
    /**
    * len(Len): adjusts the size of the bit array to Len.
    * If Len==Bs,  no action is taken, if Len<Bs then the
    * tail of the array is lost, if Len>Bs then the tail of the
    * new array will be filled with false values (0).
    * Returns the old length.
    */
    unsigned int len(unsigned int Len);
    
    /// Returns the number of bits that are set to true.
    unsigned int on_no() const;
    
    /// Returns the number of bits that are set to false.
    unsigned int off_no() const;
    
	/// bitwise operations
    
    /**
    * The following overlaid operators (~,&,|,^) perform exactly the same
    * bitwise operations as their standard counterparts. 
    * The bit array sizes must be the same otherwise 
    * a Dimexc_ exception is thrown.
    */
    Bits_& operator~();
    Bits_& operator&=(const Bits_& Bits);
    Bits_& operator|=(const Bits_& Bits);
    Bits_& operator^=(const Bits_& Bits);
    Bits_ operator&(const Bits_& Bits) const;
    Bits_ operator|(const Bits_& Bits) const;
    Bits_ operator^(const Bits_& Bits) const;

	// shifts
    
    /**
    * The following operators (<<,<<=,>>,>>=) do exactly the same things as
    * their original C counterparts. For << and <<=, 0 bits are shifted 
    * in at the right; for >> and >>=, 0 bits are shifted in from the left.
    * Empty bit-vectors will be filled up with Shift 0-s.
    */
    Bits_& operator<<=(unsigned int Shift);
    Bits_& operator>>=(unsigned int Shift);
    Bits_ operator<<(unsigned int Shift) const;
    Bits_ operator>>(unsigned int Shift) const;

	// I/O

    /**
    * Read a bit array from stream In.
    * The ASCII format of a Bits_ object is:-
    * \verbatim
      ------------------------------------
                           00000 00100110 [76..64]
      00000000 00000000 00000000 00000000 [63..32]
      00000011 11111111 11111111 11111111 [31..0]
      ====================================
      \endverbatim
    */
    friend istream& operator>>(istream& In, Bits_& Bits);
    
    /// Writes a bit array to stream Out.
    friend ostream& operator<<(ostream& Out, const Bits_& Bits);
    
    /// hidden methods
    protected:
	    
    static unsigned int uints_needed(unsigned int Bitno);
    static unsigned int get_mask(unsigned int Idx, unsigned int& Offs);
    static unsigned int* alloc_arr(unsigned int Size, const unsigned int *Arr=NULL);
    unsigned int zero_mask() const;
    
    private:
    unsigned int read_dataline(istream& In, char *Buffer);

};
// END OF CLASS Bits_

} // RazorBack

// ==== END OF HEADER Bits.hh ====

#endif  // BITS_HEADER
