// ==== METHODS Bits.cc ====

// A simple bit-array class for storing Boolean data.

// 12-Jul-1999. Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef _STANDARD_C_PLUS_PLUS
#include <sstream>
using namespace std;
#else
#include <strstream.h>
#define istringstream istrstream 
#endif

// ==== MODULE HEADER ====

#include"Bits.hh"
using namespace RazorBack;

// ==== STATIC INIT ====

const size_t Bits_::BITS_PER_CHAR=CHAR_BIT;
const size_t Bits_::BITS_PER_UINT=CHAR_BIT*sizeof(unsigned int);

// ==== METHODS ====

// ---- Memory handling ----

// uints_needed(Bitno): calculates how many uints are needed
// to store Bitno bits.
// Hidden static
inline
unsigned int Bits_::uints_needed(unsigned int Bitno)
{ return(Bitno? (Bitno-1)/BITS_PER_UINT+1: 0); }

// alloc_arr(): creates an uint array containing Size items
// and returns a ptr to it (or NULL if allocation failed).
// If Arr!=NULL, then its contents (up to Size uints) are
// copied into the new array. Arr==NULL by default. Hidden static
inline
unsigned int* Bits_::alloc_arr(unsigned int Size, const unsigned int *Arr)
{
    if (!Size) return(NULL);	// do nothing
    
    unsigned int *New=new unsigned int [Size];
    if (New==NULL)
    {	// should throw exception here
	cerr<<"\n! Bits_::alloc_arr(): Out of memory\n";
	return(NULL);
    }
    if (Arr!=NULL) memcpy(New, Arr, Size*sizeof(unsigned int));
    return(New);
}
// END of alloc_arr()

// zero_mask(): the last uint in the B array may be incompletely filled.
// This method returns an uint which has 1-s for the valid positions
// and 0-s for the unused tail. For zeroing that tail, just & to the
// last uint. Hidden
inline
unsigned int Bits_::zero_mask() const
{
    register unsigned int Lastbits=Bs % BITS_PER_UINT,  // no. of bits in last active uint
	Zmask=~0;  // all 111...1
    if (Lastbits)
    {
	// make 00..011..11 when not full
	if (Lastbits<=BITS_PER_UINT/2) { Zmask<<=Lastbits; Zmask=~Zmask; }
	else Zmask>>=BITS_PER_UINT-Lastbits;
    }
    return(Zmask);  // all 1 if Lastbits==0
}
// END of zero_mask()

// ---- Constructors ----

// Init to hold N bits (default 0). Set all bits to 
// Val (false by default).
Bits_::Bits_(unsigned int N, bool Val):
    Bs(N), Cs(uints_needed(N))
{
    if (!N) { B=NULL; Cs=0; return; }    // empty
    
    B=alloc_arr(Cs);
    
    // this sets the extra bits at the upper end of the last uint, too!
    if (B!=NULL)
	memset(B, (!Val)? '\0': ~'\0', Cs*sizeof(unsigned int));
}

// copy constructor
Bits_::Bits_(const Bits_& Bits):
    B(alloc_arr(Bits.Cs, Bits.B)), Bs(Bits.Bs), Cs(Bits.Cs)
{
    if (!Bits.Bs || B==NULL) return;    // empty
}

// Init by a string Str which should have the format "110101001",
// 0 for "false", 1 for "true", 1st char is MSB, last is LSB.
// Any char other than '0' or '1' will elicit a warning
// and will be interpreted as "true". if Str==NULL or if it is the
// empty string "" then an empty object will be created. 
Bits_::Bits_(const char *Str):
    B(NULL), Bs(0), Cs(0)
{
    if (Str==NULL || *Str=='\0')
	return;	// empty
    
    Bs=strlen(Str);
    Cs=uints_needed(Bs);
    B=alloc_arr(Cs);
    if (B!=NULL)
    {
	memset(B, 0, Cs*sizeof(unsigned int));	// zero all
	for (register int i=0; i<Bs; i++)
	{
	    char Ch=Str[Bs-i-1];
	    if (Ch=='0') continue;
	    if (Ch!='1')
		cerr<<"\n? Bits_(str): Invalid char \'"<<Ch<<"\', \'1\' used\n";
	    set_bit(i);
	}
    }
}

// Assignment
Bits_& Bits_::operator=(const Bits_& Rhs)
{
    if (this==&Rhs) return(*this);	// x=x
    
    if (Cs<Rhs.Cs)	// realloc is necessary
    {
	unsigned int *Bnew=alloc_arr(Rhs.Cs, Rhs.B);
	if (Bnew==NULL) return(*this);
	delete [] B; B=Bnew; Cs=Rhs.Cs;
    }
    else memcpy(B, Rhs.B, Rhs.Cs*sizeof(unsigned int));
    
    Bs=Rhs.Bs;
    return(*this);
}
// END of =

// ---- Equality ----

// ==,!=: return an appropriate Boolean value. Two Bits_ arrays are
// equal if they have the same number of bits and all bits are set
// in the same way.
bool Bits_::operator==(const Bits_& Other) const
{
    if (this==&Other) return(true);	// x==x
    if (Bs!=Other.Bs) return(false);  // different sizes
    
    // check the "full" uint's in the array first
    register unsigned int i, Uino=uints_needed(Bs);
    for (i=0; i<Uino-1; i++)
	if (B[i]!=Other.B[i]) return(false);
    
    // check the bits in the last uint
    for (i=(Uino-1)*sizeof(unsigned int); i<Bs; i++)
	if (get_bit(i)!=Other.get_bit(i)) return(false);
    return(true);
}
// END of ==

bool Bits_::operator!=(const Bits_& Other) const
{
    if (this==&Other) return(false);	// x!=x
    if (Bs!=Other.Bs) return(true);  // different sizes
    
    // check the "full" uint's in the array first
    register unsigned int i, Uino=uints_needed(Bs);
    for (i=0; i<Uino-1; i++)
	if (B[i]!=Other.B[i]) return(true);
    
    // check the bits in the last uint
    for (i=(Uino-1)*sizeof(unsigned int); i<Bs; i++)
	if (get_bit(i)!=Other.get_bit(i)) return(true);
    return(false);
}
// END of !=

// ---- Access and size ----

// get_mask(): auxiliary function for "locating" individual bits.
// The Idx-th bit is accessed so that the offset of the unsigned int
// array is returned in Offs and an unsigned int bitmask is generated
// which has 1 at the corresponding position and 0 bits otherwise.
// NOTE: Idx is not checked here (assumed to be correct!).
// Return value: the bitmask or 0 if Idx was out of range.
// Hidden static
inline
unsigned int Bits_::get_mask(unsigned int Idx, unsigned int& Offs)
{
    Offs=Idx/BITS_PER_UINT;
    register unsigned int Mask=1; Mask<<=(Idx % BITS_PER_UINT);
    return(Mask);
}
// END of get_mask()

// get_bit(Idx): returns the Boolean value of the Idx-th bit.
// Throws Indexrangexc_ if Idx is out of range.
// Throws Emptyexc_ if the calling object is empty.
bool Bits_::get_bit(unsigned int Idx) const
{
    // exceptions if Idx is out of range or calling object is empty
    if (!Bs)
	throw Emptyexc_("Bits_::get_bit()");
    if (Idx>=Bs)
	throw Indexrangexc_("Bits_::get_bit()", Idx, Bs-1);
    
    register unsigned int Offs, Mask=get_mask(Idx, Offs);
    return(bool((Mask & B[Offs])!=0));    // to convert to 0|1
}
// END of get_bit(Idx)

// set_bit(Idx, Value): sets the Idx-th bit to Value (default is true).
// Returns old value.
// Throws Indexrangexc_ if Idx is out of range.
// Throws Emptyexc_ if the calling object is empty.
bool Bits_::set_bit(unsigned int Idx, bool Value)
{
    // exceptions if Idx is out of range or calling object is empty
    if (!Bs)
	throw Emptyexc_("Bits_::get_bit()");
    if (Idx>=Bs)
	throw Indexrangexc_("Bits_::get_bit()", Idx, Bs-1);
    
    register unsigned int Offs, Mask=get_mask(Idx, Offs);
    bool Oldbit=bool((Mask & B[Offs])!=0);
    if (Value) B[Offs]|=Mask; else B[Offs]&=~Mask;
    return(Oldbit);
}
// END of set_bit(Idx, Value)

// len(Len): adjusts the size of the bit array to Len.
// If Len==Bs,  no action is taken, if Len<Bs then the
// tail of the array is lost, if Len>Bs then the tail of the
// new array will be filled with false values (0).
// Returns the old length.
unsigned int Bits_::len(unsigned int Len)
{
    if (Len==Bs) return(Bs);    // no change
    
    unsigned int Oldlen=Bs;
    
    if (Len>Bs)	    // bit size grows (may fit into original Cs-long array)
    {
	register unsigned int Csnew=uints_needed(Len),  // new uint size
	    Csmin=uints_needed(Bs);	// minimal uint size for old
	
	if (Cs<Csnew)  // realloc is necessary (uint size grows as well)
	{
	    unsigned int *Bnew=alloc_arr(Csnew);	// create new array
	    if (Bnew==NULL) return(Oldlen);
	    
	    if (B!=NULL)	// there was a previous array (i.e. Bs>0,Cs>0)
	    {
		memcpy(Bnew, B, Csmin*sizeof(unsigned int));	// copy old
		delete [] B;
	    }
	    B=Bnew;	// replace old array w/ new
	}
	memset(B+Csmin, 0, (Csnew-Csmin)*sizeof(unsigned int));   // zero new uints
	
	// zero the last bits in the last (old) uint if necessary
	if (Csmin) B[Csmin-1]&=zero_mask();
	
	if (Cs<Csnew) Cs=Csnew;	// Cs had to be kept until now, cf zeroing above
    }
    
    Bs=Len;
    return(Oldlen);
}
// END of len(Len)

// set_values(Value): sets all bits to Value (default is true).
void Bits_::set_values(bool Value)
{
    if (!Bs) return;
    unsigned char Vals=(Value)? ~0: 0;
    memset(B, Vals, uints_needed(Bs)*sizeof(unsigned int));
}
// END of set_values(Value)

// on_no(), off_no(): return the number of bits that are ON or OFF,
// respectively.
unsigned int Bits_::on_no() const
{
    unsigned int No, i;
    for (No=0, i=0; i<Bs; i++)
	if (get_bit(i)) No++;
    return(No);
}
// END of on_no()

unsigned int Bits_::off_no() const
{
    unsigned int No, i;
    for (No=0, i=0; i<Bs; i++)
	if (!get_bit(i)) No++;
    return(No);
}
// END of off_no()

// ---- Bitwise operations ----

// The following overlaid operators (~,&,|,^) perform exactly the same
// bitwise operations as their standard counterparts. 
// The bit array sizes must be the same otherwise 
// a Dimexc_ exception is thrown.

Bits_& Bits_::operator~()
{
    for (register unsigned int i=0; i<uints_needed(Bs); i++) B[i]=~B[i];
    return (*this);
}

Bits_& Bits_::operator&=(const Bits_& Bits)
{
    if (Bs!=Bits.Bs) 
	throw Dimexc_("Bits_::x &= y", Bs, Bits.Bs);
    for (register unsigned int i=0; i<uints_needed(Bs); i++) B[i]&=Bits.B[i];
    return(*this);
}

Bits_& Bits_::operator|=(const Bits_& Bits)
{
    if (Bs!=Bits.Bs) 
	throw Dimexc_("Bits_::x |= y", Bs, Bits.Bs);
    else for (register unsigned int i=0; i<uints_needed(Bs); i++) B[i]|=Bits.B[i];
    return(*this);
}

Bits_& Bits_::operator^=(const Bits_& Bits)
{
    if (Bs!=Bits.Bs) 
	throw Dimexc_("Bits_::x ^= y", Bs, Bits.Bs);
    else for (register unsigned int i=0; i<uints_needed(Bs); i++) B[i]^=Bits.B[i];
    return(*this);
}

Bits_ Bits_::operator&(const Bits_& Bits) const
{
    if (Bs!=Bits.Bs) 
	throw Dimexc_("Bits_::x & y", Bs, Bits.Bs);
    Bits_ Temp=(*this); Temp&=Bits; return(Temp);
}

Bits_ Bits_::operator|(const Bits_& Bits) const
{
    if (Bs!=Bits.Bs) 
	throw Dimexc_("Bits_::x | y", Bs, Bits.Bs);
    Bits_ Temp=(*this); Temp|=Bits; return(Temp);
}

Bits_ Bits_::operator^(const Bits_& Bits) const
{
    if (Bs!=Bits.Bs) 
	throw Dimexc_("Bits_::x ^ y", Bs, Bits.Bs);
    Bits_ Temp=(*this); Temp^=Bits; return(Temp);
}

// ---- Shifts ----

// The following operators (<<,<<=,>>,>>=) do exactly the same things as
// their original C counterparts. For << and <<=, 0 bits are shifted 
// in at the right; for >> and >>=, 0 bits are shifted in from the left.
// Empty bit-vectors will be filled up with Shift 0-s.

Bits_& Bits_::operator<<=(unsigned int Shift)
{
    if (!Shift) return (*this);
    if (!Bs)	// empty, resize
    {
    	len(Shift);
	return (*this);
    }
    
    // if Shift>=BITS_PER_UINT, full uint's must be shifted first
    register unsigned int i, Csmin=uints_needed(Bs), Ush=Shift/BITS_PER_UINT;
    
    if (Ush)	// do uint-left-shift
    {
	for (i=Csmin-1; i>=Ush; i--) B[i]=B[i-Ush];
	memset(B, 0, Ush*sizeof(unsigned int));	    // zero the Ush rightmost uints
    }

    // do bitwise shifts, carrying over bits at each left side
    register unsigned int Bsh=Shift % BITS_PER_UINT;
    if (Bsh)
    {
	register unsigned int Carry, Prevcarry=0, Cmask=~0;	// all 11..1
	Cmask>>=Bsh; Cmask=~Cmask;  // 11..10000..0
	for (i=Ush; i<Csmin-Ush; i++)
	{
	    Carry=B[i] & Cmask;	// save leftmost Bsh bits
	    B[i] <<= Bsh;   // do the shift, rightmost Bsh bits are 0
	    B[i] |= Prevcarry;	// put there saved bits from previous shift
	    Prevcarry=Carry;	// save for next
	    Prevcarry>>=(BITS_PER_UINT-Bsh);    // move from left to right
	}
    }
    return(*this);
}
// END of <<=

Bits_& Bits_::operator>>=(unsigned int Shift)
{
    if (!Shift) return (*this);
    if (!Bs)	// empty, resize
    {
    	len(Shift);
	return (*this);
    }
    
    // if Shift>=BITS_PER_UINT, full uint's must be shifted first
    register unsigned int Csmin=uints_needed(Bs), Ush;
    register int i;
    
    // subtract the "tail" Csmin % Bs bits from Shift and do
    // uint-right-shifts only if Shift is larger than this
    Ush=(Csmin % Bs >= Shift)? 0: (Shift-Csmin % Bs)/BITS_PER_UINT;
    if (Ush)	// do uint-right-shift
    {
	for (i=0; i<Csmin-Ush; i++) B[i]=B[i+Ush];
	memset(B+Csmin-Ush, 0, Ush*sizeof(unsigned int));	    // zero the Ush leftmost uints
    }

    // do bitwise shifts, carrying over bits at each right side
    register unsigned int Bsh=Shift % BITS_PER_UINT;
    if (Bsh)
    {
	register unsigned int Carry, Prevcarry=0, Cmask=~0;	// all 11..1
	Cmask<<=Bsh; Cmask=~Cmask;  // 00..0011..1111
	B[Csmin-1]&=zero_mask();    // zero tail
	for (i=Csmin-Ush-1; i>=0; i--)    // proper -1 check needed, i not uint
	{
	    Carry=B[i] & Cmask;	// save rightmost Bsh bits
	    B[i] >>= Bsh;   // do the shift, leftmost Bsh bits are 0
	    B[i] |= Prevcarry;	// put there saved bits from previous shift
	    Prevcarry=Carry;	// save for next
	    Prevcarry<<=(BITS_PER_UINT-Bsh);    // move from right to left
	}
    }
    return(*this);
}
// END of >>=

Bits_ Bits_::operator<<(unsigned int Shift) const
{
    Bits_ Temp(*this); Temp<<=Shift;
    return(Temp);
}
// END of <<

Bits_ Bits_::operator>>(unsigned int Shift) const
{
    Bits_ Temp(*this); Temp>>=Shift;
    return(Temp);
}
// END of >>

// ---- I/O ----

// The ASCII format of a Bits_ object is:-
// 
// ------------------------------------
//                      00000 00100110 [76..64]
// 00000000 00000000 00000000 00000000 [63..32]
// 00000011 11111111 11111111 11111111 [31..0]
// ====================================
// 

#define LINELEN 120

// friend fn defs must be in namespace
namespace RazorBack {
istream& operator>>(istream& In, Bits_& Bits)
{
    if (!In) return In; // no good
    
    static char Buffer[LINELEN];
    streampos Oldpos=In.tellg();
    
    // read --- on top: need not be full length
    while (In>>Buffer)
    {
	if (strspn(Buffer, "-")==strlen(Buffer))
	    break;
    }
    if (!In)
    {
	In.clear(ios::failbit); // no Bits_ object
	In.seekg(Oldpos);   	// go back to old position
	return In;
    }
    
    // read data lines and "===" at bottom
    Bits_ Temp;     // empty temporary
    while (In && Temp.read_dataline(In, Buffer));
	
    if (In) 	// success
	Bits=Temp;
    else
    {
	In.clear(ios::failbit); // no Bits_ object
	In.seekg(Oldpos);   	// go back to old position
    }
    return In;
}
// END of >>
}   // RazorBack

// read_dataline(In, Buffer):
// reads from In a data line "00000 00100110 [76..64]"
// or the terminating "===...==" line into Buffer
// that is assumed to be LINELEN chars long.
// The calling object must be empty upon first invocation,
// the bitvector size will be adjusted according to what
// is found as highest significant position in the first
// data line.
// On error, 0 is returned and In's failbit is set;
// On having seen the last "===" 0 is returned but In stays "good";
// otherwise, the highest significant pos ("76" in the
// above example) is returned.
// Hidden
unsigned int Bits_::read_dataline(istream& In, char *Buffer)
{
    static const char *BINCHARS="01", *SIZECHARS="[0123456789.]";
    
    memset(Buffer, '\0', LINELEN);  // zero buffer
    char *Bufptr=NULL;
    unsigned int Bitlen;
    
    // Collect the binary strings for a uint in Buffer.
    // Bufptr points to the next position to be read,
    // the substrings are read one after the other
    // so that spaces are automagically skipped.
    // The last substring is the terminal string
    // that contains the highest and lowest bit position
    // in this line. Only the highest pos is read into Bitlen.
    for (Bufptr=Buffer, Bitlen=0; In>>Bufptr; Bufptr+=strlen(Bufptr))
    {
	// is it the end-of-Bits "==="?
	if (strspn(Bufptr, "=")==strlen(Bufptr))
	    return 0;	// OK
	
	// is it the terminal "[76..64]" string?
	if (*Bufptr=='[' && strspn(Bufptr, SIZECHARS)==strlen(Bufptr))
	{
	    istringstream Blenstr(Bufptr+1);
	    Blenstr>>Bitlen;	// read length
	    *Bufptr='\0';   // terminate binary string (from Buffer to Bufptr)
	    break;
	}
	
	// if not binary "0100101" then we have a problem
	if (strspn(Bufptr, BINCHARS)!=strlen(Bufptr))
	{
	    In.clear(ios::failbit); // not OK this time!
	    return 0;
	}
	
	// otherwise OK, proceed
    }
    
    // If the calling object was empty, this means that
    // we have just read the first (most significant) data line.
    // Set storage accordingly
    if (!len() && Bitlen) len(Bitlen+1);
    
    // "shift in" the binary string
    for (Bufptr=Buffer; *Bufptr!='\0'; ++Bufptr)
    {
	(*this)<<=1;
	if (*Bufptr=='1') set_bit(0);
    }
    
    return Bitlen;
}
// END of read_dataline(In, Buffer)
#undef LINELEN

// friend fn defs must be in namespace
namespace RazorBack {
ostream& operator<<(ostream& Out, const Bits_& Bits)
{
    unsigned int j, Mask, 
	    Lstb=Bits.len() % Bits_::BITS_PER_UINT, 
	    Uino=Bits.uints_needed(Bits.len());
    int i;
    const unsigned int Dw=sizeof(unsigned int)*(Bits_::BITS_PER_CHAR+1);
    
    // print --- on top
    for (j=0; j<Dw; j++) Out.put('-'); Out<<endl;
    
    // print last uint (might be incomplete)
    if (Lstb)
    {
	Mask=1<<(Lstb-1);	// 1000...000 (shorter)
	for (j=0; j<Bits_::BITS_PER_UINT; j++)
	{
	    if (j<Bits_::BITS_PER_UINT-Lstb) Out.put(' ');	    // just left pad
	    else
	    {
		Out.put((Bits.B[Uino-1] & Mask)? '1': '0');
		Mask>>=1;
	    }
	    if (!((j+1) % Bits_::BITS_PER_CHAR)) Out.put(' ');
	}
	Out<<'['<<Bits.Bs-1<<".."<<(Uino-1)*Bits_::BITS_PER_UINT<<"]\n";
    }
    
    // print "full" uints
    for (i=((Lstb)? Uino-2: Uino-1); i>=0; --i)
    {
	Mask=1U<<(Bits_::BITS_PER_UINT-1);	// 1000...000
	for (j=0; j<Bits_::BITS_PER_UINT; j++)
	{
	    Out.put((Bits.B[i] & Mask)? '1': '0');
	    if (!((j+1) % Bits_::BITS_PER_CHAR)) Out.put(' ');  // separate bytes
	    Mask>>=1;
	}
	Out<<'['<<(i+1)*Bits_::BITS_PER_UINT-1<<".."<<i*Bits_::BITS_PER_UINT<<"]\n";
    }
    
    // print === at bottom
    for (j=0; j<Dw; j++) Out.put('='); Out<<endl;
    return Out;
}
// END of <<
}   // RazorBack

// ==== END OF METHODS Bits.cc ====
