// ==== METHODS Symvar.cc ====

/**
 * Class to represent "symbolic variables" that have
 * a string value taken from a fixed list of possible strings.
 */

// 5-Oct-2001. Andras Aszodi

// ==== MODULE HEADERS ====

#include "Symvar.hh"
using namespace RazorBack;

// ==== METHODS ====

/**
 * Sets the allowed tokens to Tokenstr. This string shall contain the tokens
 * separated by one of the allowed separator characters (returned by the
 * static method legal_separators()). For example, set_tokens("tok1|tok2,tok3")
 * specifies the three acceptable tokens "tok1", "tok2", "tok3".
 * The current token iterator will be set to the first one.
 * Returns the number of tokens seen.
 */
unsigned int Symvar_::set_tokens(const string& Tokenstr)
{
    Toklist.clear();
    unsigned int Separpos, Newseparpos=0, Tokenno=0;
    
    for (Separpos=0; Newseparpos<Tokenstr.length() && Separpos<Tokenstr.length()-1; Separpos=Newseparpos+1)
    {
	Newseparpos=Tokenstr.find_first_of(separators(), Separpos);
	if (Separpos==Newseparpos)
	{
	    // skip duplicate separator char
	    continue;
	}
	
	// Newseparpos==NPOS indicates the last token (no separator found)
	// this special case is taken care of by the write-only ?: expression below
	string Newtok(Tokenstr.substr(Separpos, 
		((Newseparpos==NPOS)? Tokenstr.length(): Newseparpos)-Separpos));
    	Toklist.push_back(Newtok); ++Tokenno;
    }
    Curtokptr=Toklist.begin();
    return Tokenno;
}
// END of set_tokens(Tokenstr)

/**
 * Removes the token Tok from the list of allowed tokens.
 * If the token to be removed was the current one, then
 * the new current token will be the first in the internal list.
 */
void Symvar_::remove_token(const string& Tok)
{
    if (Tok==(*Curtokptr))
    {
	Toklist.erase(Curtokptr);
	Curtokptr=Toklist.begin();
    }
    else
    {
	// "a reasonable way of removing elements", the STL Guide says :-)
	Toklist.erase(remove(Toklist.begin(), Toklist.end(), Tok), Toklist.end());
    }
}
    
/**
 * Sets the current token to Tok if it is valid.
 * Returns the old current token.
 */
const string& Symvar_::token(const string& Tok)
{
    if (empty())
	throw Emptyexc_("Symvar_::token(Tok)");
    
    list<string>::iterator Oldptr=Curtokptr;
    Curtokptr=find(Toklist.begin(),Toklist.end(), Tok);
    if (Curtokptr==Toklist.end())	// not found
	Curtokptr=Oldptr;	    	// reset to previous
    return *Oldptr;
}
    
// friend fn defs must be in namespace
namespace RazorBack {
/**
 * Input: reads the next string into Sv.
 * If it is a valid token, then it will be stored in Sv,
 * otherwise Sv is not modified and In's failbit will be set.
 */
istream& operator>>(istream& In, Symvar_& Sv)
{
    if (!In) return In;
    
    string Tok;
    In>>Tok;
    if (Sv.valid_token(Tok))
	Sv.token(Tok);
    else
	In.clear(ios::failbit);
    return In;
}
// END of >>

/// Output: writes the token string in Sv to Out. No newline at the end.
ostream& operator<<(ostream& Out, const Symvar_& Sv)
{
    if (!Sv.empty()) Out<<Sv.token();
    return Out;
}
// END of <<
} // RazorBack
    
/**
 * Constructs and returns a string that contains all legal tokens,
 * separated by the first char in the string returned by separator().
 * If the calling object is empty, then "" is returned.
 */
string Symvar_::all_tokens() const
{
    if (empty()) return string("");
    
    string Tokstr=Toklist.front();
    list<string>::const_iterator Tp=Toklist.begin();
    ++Tp;
    for ( ; Tp!=Toklist.end(); ++Tp)
	Tokstr+=(Separators[0]+ *Tp);
    return Tokstr;
}
// END of all_tokens()

// ---- aux ----

// curtokptr_offset(): returns the offset of Curtokptr from Toklist.begin().
// curtokptr_offset(Offs): moves Curtokptr so that it points to Offs elements
// from Toklist.begin().
// These are needed because one cannot do full pointer arithmetic on
// STL list iterators (why??).
// Hidden
unsigned int Symvar_::curtokptr_offset() const
{
    list<string>::const_iterator It;
    unsigned int Offs;
    
    for (Offs=0, It=Curtokptr; It!=Toklist.begin(); --It, ++Offs);
    return Offs;
}
void Symvar_::curtokptr_offset(unsigned int Offs)
{
    unsigned int Idx;
    for (Idx=0, Curtokptr=Toklist.begin(); Idx<Offs; ++Idx, ++Curtokptr);
}
// END of curtokptr_offset([Offs])

// ==== END OF METHODS Symvar.cc ====
