#ifndef REFSTORAGE_TMPL_HEADER
#define REFSTORAGE_TMPL_HEADER

// ==== TEMPLATE HEADER Refstorage.hh ====

/**
 * \file Refstorage.hh
 * Storage template based on reference counting and copy-on-write.
 * Useful for implementing storage for "fat" classes
 * that have lots of data and need efficient copying.
 */

// 15-Dec-2000. Andras Aszodi

// ---- MODULE HEADERS ----

#include "Utilsexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Refstorage_: implements a handler and a data container
 * that support copy-on-write and reference counting.
 * There are no public methods: this class is intended
 * as a (protected) base class for other classes that
 * have lots of data. The data objects should be contained
 * in a separate class that will be substituted as T_.
 */
template <class T_>
class Refstorage_
{
    // nested class
    private:
    
    /**
    * Storage_: this nested class holds one copy of the actual data object
    * plus a reference count. Implements deep and shallow copy
    */
    struct Storage_
    {
	// data
	T_* Dataptr;    //< pointer to the data to be stored
	int Refcount;	//< reference count
	
    	/// Default ctor, inits to empty.
	Storage_(): Dataptr(NULL), Refcount(0) {}
	
	/**
	 * Init with an already available data object.
	 * Ideally, this data object has to be allocated dynamically
	 * like 'Storage_(new T_(params...))'
	 * and ownership is taken by Refstorage_<T_>::Storage_
	 * so that de-allocation will be performed by it.
	 */
	explicit Storage_(const T_* Dptr): Dataptr((T_*)Dptr), Refcount(0) {}
	
	/// Copy ctor, implements deep copy.
	Storage_(const Storage_& S)
		: Dataptr(S.Dataptr==NULL? NULL: new T_(*(S.Dataptr))), 
		Refcount(0) {}
	
	/// destructor: deletes stored object, too
	~Storage_() { delete Dataptr; }
	
    };	    // END of nested class Storage_
    
    // data
    private:
	    
    Storage_ *Storptr;	    //< points to the storage object
    
    // methods
    protected:
    
    /// default ctor, inits to empty
    Refstorage_(): Storptr(NULL) {}
    
    /**
     * Inits with a freshly created data object pointed to by Dp.
     * Since ownership is taken by Refstorage_,
     * de-allocation will be performed by it.
     * For this reason, the only wise invocation of this ctor is
     * Refstorage_<T_>(new T_(params...))
     */
    explicit Refstorage_(const T_* Dp):
	    Storptr(new Storage_(Dp)) 
    {
	Storptr->Refcount=1;
    }
    
    /// copy ctor, implements shallow copy
    Refstorage_(const Refstorage_<T_>& Rf):
	    Storptr(Rf.Storptr)
    {
	if (Storptr!=NULL) Storptr->Refcount++;
    }
    
    /// destructor
    virtual ~Refstorage_()
    {
	unlink_old();	// reduce refcount or delete if no more refs
    }
    
    /// assignment
    Refstorage_<T_>& operator=(const Refstorage_<T_>& Rhs)
    {
	if (this==&Rhs || Storptr==Rhs.Storptr)
	    return(*this);	// x=x or same data, do nothing
	unlink_old();	    // get rid of lhs data
	Storptr=Rhs.Storptr;	// shallow copy rhs data
	if (Storptr!=NULL) Storptr->Refcount++;
	return(*this);
    }
	
    /**
    * unlink_old(): unlinks old storage (decr Refcount and delete
    * if there are no more links).
    */
    void unlink_old()
    {
	if (Storptr!=NULL && --Storptr->Refcount==0)
	    delete Storptr;
    }

    /// clone(): clones Storptr if it is referenced by at least another object.
    void clone()
    {
	if (Storptr!=NULL && Storptr->Refcount>1)
	{
	    Storage_* Newsptr=new Storage_(*Storptr);	// deep copy *Sptr
	    unlink_old();
	    Storptr=Newsptr; Storptr->Refcount=1;	// link new storage
    	}
    }
    
    /**
    * replace_storage(Dp): replaces old storage with new storage
    * so that the data Dp will be stored.
    * The same rule applies for Dp as for the ctor with the same arg:
    * Dp must point to a freshly created, dynamically allocated
    * object of type T_.
    */
    void replace_storage(const T_* Dp)
    {
	unlink_old();   // get rid of old storage
	Storptr=new Storage_(Dp);    // and link in new
	Storptr->Refcount=1;
    }

    /// Returns true if there is no stored data.
    bool empty() const { return (Storptr==NULL || Storptr->Dataptr==NULL); }
    
    /**
     * data(): (Const) access to the stored data itself.
     * Throws Emptyexc if the calling object was empty.
     */
    const T_& data() const
    {
	if (empty())
	    throw Emptyexc_("Refstorage_::data() (const)");
	return *(Storptr->Dataptr);
    }
    T_& data()
    {
	if (empty())
	    throw Emptyexc_("Refstorage_::data()");
	return *(Storptr->Dataptr);
    }
    
    /**
     * ptr(): Returns a (const) pointer to the stored data.
     * Returns NULL if the calling object was empty.
     */
    const T_* ptr() const { return (empty()? NULL: Storptr->Dataptr); }
    T_* ptr() { return (empty()? NULL: Storptr->Dataptr); }
};
// END OF CLASS Refstorage_

// NOTE: there is no associated template methods file.

} // RazorBack

// ==== END TEMPLATE HEADER Refstorage.hh ====
    
#endif  // REFSTORAGE_TMPL_HEADER
