#ifndef ANOVA1_HEADER
#define ANOVA1_HEADER

// ==== CLASS HEADER Anova1.hh ====

/**
 * \file
 * One-way analysis of variance.
 * Havlicek, L.L. and Crain, R.D.: Practical Statistics for the Physical Sciences.
 * ACS Professional Reference Book, Washington, 1988.
 * \author Andras Aszodi
 * \date 9-Oct-2001.
 */

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <math.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <vector>
#include <algorithm>
#include <iostream>
#include <iomanip>
#else
#include <vector.h>
#include <algo.h>
#include <iostream.h>
#include <iomanip.h>
#endif

#include <string>
using namespace std;
#define NPOS string::npos

// ==== MODULE HEADERS ====

#include "Statexc.hh"

// ==== CLASSES ====

namespace RazorBack {
    
class Anova1_
{
    // member types
    public:
	    
    /**
     * This class implements an observation group.
     * Contains the group size, the group mean and a description string.
     */
    class Group_
    {
	// data
	private:
		
	string Descr;
	double Sum;
	unsigned int Size;
	
	// methods
	public:
	
	/// Inits to contain the description D, otherwise empty.
	explicit Group_(const string& D=string()):
		Descr(D), Sum(0.0), Size(0) {}
	
	/// Inits to contain a data group with description D, sum S and size N.
	Group_(const string& D, double S, unsigned int N):
		Descr(D), Sum(S), Size(N) {}
	
	// access
	
	const string& descr() const { return Descr; }
	unsigned int size() const { return Size; }
	double mean() const { return Sum/Size; }
	
	// hidden methods
	protected:

    };
	    
    /// Indexes the various ANOVA sums.
    enum Sumidx_ {TOTAL=0, BETWEEN=1, WITHIN=2};
    
    // data
    private:
	    
    vector<Group_> Xgroups;
    vector<double> Sumsq, Meansq;
    unsigned int Groupno, Groupsize, Ntotal;
    vector<unsigned int> Degfree;
    bool Equalgroups;
    double Sumx, Sumx2, Sumgroup;
    
    // methods
    public:
	    
    /// Ctor: init to empty.
    Anova1_()
	    : Xgroups(), Sumsq(3), Meansq(3),
	    Groupno(0), Groupsize(0), Ntotal(0), Degfree(3),
	    Sumx(0.0), Sumx2(0.0), Sumgroup(0.0),
	    Equalgroups(true) {}
    
    /// Clears the calling object and prepares it for a new analysis.
    void clear();
    
    /**
     * Adds a group. The measurement data for the group are in the range
     * [First, Last) (can come from any STL sequence). The name of the 
     * group is Groupname. Throws Toofewexc_ if [First, Last) does not 
     * specify a valid range with at least 1 data point.
     */
    template <class Initer_>
    void add_group(const string& Groupname, Initer_ First, Initer_ Last)
    {
	// definition given in header because it is a template
	
	register double X, Sx=0.0, Sx2=0.0;
	register unsigned int n=0;
	
	// perform the summation over the data points in the group
	Initer_ Gp;
	for (Gp=First; Gp!=Last; ++Gp)
	{
	    X=*Gp;
	    Sx+=X; Sx2+=(X*X); ++n;
	}
	if (n<1)    // not enough data points
	    throw Toofewexc_("Anova1_::add_group()",
		    "Data points in group", n);
	
	// store the new group
	Group_ Newgroup(Groupname, Sx, n);
	Xgroups.push_back(Newgroup);
	
	// update sums of squares
	update_sumsq(Sx, Sx2, n);
    }
    
    // ---- Access ----
    
    /// Returns the vector of all groups (ie. all data points).
    const vector<Group_>& groups() const { return Xgroups; }
    
    /// Returns the sum of squares (Idx can be TOTAL, BETWEEN, WITHIN).
    double sum_squares(Sumidx_ Idx) const { return Sumsq[Idx]; }
    
    /// Returns the degrees of freedom.
    unsigned int degrees_freedom(Sumidx_ Idx) const { return Degfree[Idx]; }
    
    /// Returns the mean of squares.
    double mean_squares(Sumidx_ Idx) const { return Meansq[Idx]; }
    
    /// Returns the F-ratio (ratio of between/within mean squares).
    double f_ratio() const
    {
	return (Meansq[WITHIN]==0.0)?
		HUGE_VAL: Meansq[BETWEEN]/Meansq[WITHIN];
    }
    
    /*
     * Returns an iterator to the group called "Name" or groups().end()
     * if not found.
     */
    vector<Anova1_::Group_>::const_iterator
    find_group(const string& Name) const;
    
    /**
     * Returns the significance level for the test that the
     * between/within F ratio is larger than 1.0 (see f_ratio()).
     * If the value returned by f_test() is low, then the probability is large
     * that f_ratio() is high, ie. that some of the group means are significantly
     * different from some of the others ("omnibus F-test").
     */
    double omnibus_ftest() const
    {
	if (Groupno<2)
	    throw Toofewexc_("Anova1_::omnibus_ftest()", "Groupno", Groupno);
    
	return (Meansq[WITHIN]==0.0)?
		0.0: f_test(f_ratio());
    }
    
    /**
     * Returns the significance value that the two groups pointed to by G1, G2
     * have different means using the Scheffe test.
     * A small returned value means that the means are significantly different.
     */
    double scheffe_test(vector<Group_>::const_iterator G1,
    	    vector<Group_>::const_iterator G2) const;

    /**
     * Returns the significance value for the test that the two meta-groups of groups
     * between [F1, L1) and [F2, L2) have different means according to the
     * Scheffe procedure. [F1,L1) and [F2,L2) are supposed to be valid ranges
     * in the groups given to the Anova1_ object before. Currently, only
     * continuous ranges are supported (not any combination of groups).
     * A small returned value means that the means are significantly different.
     */
    double scheffe_test(vector<Group_>::const_iterator F1,
    	    	    		vector<Group_>::const_iterator L1,
    	    	    		vector<Group_>::const_iterator F2,
    	    	    		vector<Group_>::const_iterator L2) const;
				
    /**
     * Returns the significance value for the test that the two meta-groups of groups
     * defined in the vectors of vector<Group_> iterators have different means according to the
     * Scheffe procedure. The iterators in Metagroup1,2 are supposed to be valid ranges
     * in the groups given to the Anova1_ object before. 
     * A small returned value means that the means are significantly different.
     */
    double scheffe_test(const vector< vector<Group_>::const_iterator >& Metagroup1,
    	    	    		const vector< vector<Group_>::const_iterator >& Metagroup2) const;

    /**
     * Returns the significance value that the two pointed to by G1, G2
     * have different means using the Tukey A test.
     * A small returned value means that the means are significantly different.
     */
    double tukey_atest(vector<Group_>::const_iterator G1,
    	    vector<Group_>::const_iterator G2) const;

    // hidden methods
    protected:
    
    void update_sumsq(double Sx, double Sx2, unsigned int n);
    double f_test(double Fratio) const;
    double q_test(double q) const;
};
// END OF CLASS Anova1_

} // Razorback

// ==== END OF CLASS HEADER Anova1.hh ====

#endif	    // ANOVA1_HEADER
