#ifndef DISTRAPPROX_HEADER
#define DISTRAPPROX_HEADER

// ==== HEADER Distrapprox.hh ====

/**
 * \file Distrapprox.hh
 * Class for approximating univariate continuous distributions.
 */

// 15-Nov-2002. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <math.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <vector>
#include <list>
#include <algorithm>
#include <utility>
using namespace std;
#else
#include <vector.h>
#include <list.h>
#include <algo.h>
#include <pair.h>
#endif

// ---- MODULE HEADERS ---- 

#include "Random.hh"
#include "Distrbase.hh"
#include "Stat2.hh"
#include "Statexc.hh"
#include "Iovar.hh"

// ==== CLASSES ====

namespace RazorBack {

class Distrapprox_: public Distrbase_
{
    // member classes
    protected:
	    
    /**
     * A Bin_ object represents the i:th histogram bin.
     * All bins are Width wide (see enclosing class for details).
     */
    class Bin_
    {
	// data
	private:
		
	int Idx;	///< the bin index (can be negative)
	unsigned int Count; 	///< no. of samples falling into the bin
	
	// methods
	public:
		
	/// Constructs the i:th bin (default 0)
	explicit Bin_(int i=0): Idx(i), Count(0) {}
	
	/// Returns the index of the bin.
	int index() const { return Idx; }
	
	/// Returns the no. of samples in the bin.
	unsigned int count() const { return Count; }
	
	/// Comparison according to index
	bool operator<(const Bin_& B) { return index()<B.index(); }
	
	/// Prefix increment: adds 1 to the bin count.
	Bin_& operator++() { ++Count; return *this; }
	
	/// Output: the format is "<index> <count>"
	friend ostream& operator<<(ostream& Out, const Bin_& B)
	{
	    Out<<B.index()<<" "<<B.count();
	    return Out;
	}
	
	/// Input: the format is "<index> <count>"
	friend istream& operator>>(istream& In, Bin_& B)
	{
	    if (!In.good()) return In;
	    int i;
	    unsigned int Cnt;
	    In>>i>>Cnt;
	    if (In)
	    {
		B.Idx=i; B.Count=Cnt;
	    }
	    return In;
	}
    };
    // END of member class Bin_
    
    // shorthand
    typedef list<Bin_>::iterator Biniter_;
    
    // data
    private:
	    
    list<Bin_> Binlist;     ///< histogram bins sorted in increasing center order
    double Width;  ///< bin width
    vector<double> Pdf;     ///< approximation of PDF
    vector<double> Cdf;     ///< approximation of CDF
    unsigned int Binno;     ///< length of Pdf and Cdf
    double Low, High;	    ///< lowest and highest X value covered by the bins
    bool Dirty;     ///< true when a new value is added
    Randomuni_ Rand;	///< uniform random number generator
    Stat_ Stat;     ///< statistics for average and SD
    
    // methods
    public:
	    
    /**
     * Constructs an object so that it is ready to do an approximation
     * with a bin width W.
     * If W<0.0, then its absolute value is used.
     * If W==0.0, then a Badvalexc_ is thrown.
     */
    explicit Distrapprox_(double W) { reset(W); }

    /**
     * reset(W):
     * Resets the calling object so that it is ready to do an approximation
     * with a bin width W. All previous data are lost.
     * If W<0.0, then its absolute value is used.
     * If W==0.0, then a Badvalexc_ is thrown.
     */
    void reset(double W);
    
    /**
     * clear():
     * clears the calling object. All data are lost except the histogram bin width.
     */
    void clear();
    
    /// +=: adds the value X (a sample from the distribution to be approximated).
    Distrapprox_& operator+=(double X);
    
    /// evaluate(): normalise the counts and make the PDF and CDF.
    void evaluate();
    
    /**
     * Returns the lowest value covered by the approximation.
     * Throws Forgottenexc_ if evaluate() was not invoked beforehand.
     */
    double low() const
    {
    	if (Dirty)
	    throw Forgottenexc_("Distrapprox_::low()", "Distrapprox_::evaluate()");
	return Low;
    }
    
    /**
     * Returns the highest value covered by the approximation.
     * Throws Forgottenexc_ if evaluate() was not invoked beforehand.
     */
    double high() const
    {
    	if (Dirty)
	    throw Forgottenexc_("Distrapprox_::low()", "Distrapprox_::evaluate()");
	return High;
    }
    
    /**
     * Returns a linearly interpolated value for the PDF.
     * Throws Forgottenexc_ if evaluate() was not invoked beforehand.
     */
    virtual double pdf(double X) const { return lin_intp(X, Pdf, 0.0); }
    
    /**
     * Returns a linearly interpolated value for the CDF.
     * Throws Forgottenexc_ if evaluate() was not invoked beforehand.
     */
    virtual double cdf(double X) const { return lin_intp(X, Cdf, 1.0); }
    
    /** 
     * Returns an interpolated value for the inverse CDF, X where CDF(X)=Y.
     * Throws Forgottenexc_ if evaluate() was not invoked beforehand.
     */
    virtual double inv_cdf(double Y) const;
    
    /// Returns the estimated mean of the distribution.
    virtual double mean() const { return Stat.avg(); }
    
    /// Returns the estimated variance of the distribution.
    virtual double var() const { return Stat.sd()*Stat.sd(); }
    
    /// Returns the skewness of the distribution.
    virtual double skew() const { return Stat.skew(); }
    
    /// Returns the kurtosis of the distribution.
    virtual double kurt() const { return Stat.kurt(); }
    
    /**
     * entropy(): approximates the entropy of the distribution.
     * Throws Forgottenexc_ if evaluate() was not invoked beforehand.
     */
    virtual double entropy() const;
    
    /// random(): returns a random value from the approximated distribution.
    virtual double random(void) const { return inv_cdf(Rand()); }
    
    // I/O (serialization)
    
    /**
     * Writes (serializes) a Distrapprox_ object to Out. The format is:-
     * \verbatim
	DISTRAPPROX
	WIDTH = 0.1
	STAT
	N = 1000
	MIN = -4.7362
	MAX = 6.30655
	SUM = 957.413
	SUM2 = 4575.58
	SUM3 = 11291.8
	SUM4 = 57072.3
	ENDSTAT
	-47 1	    <-- bin offset index and bin count, see Bin_
	-45 3
	...
	61 2
	63 1
	ENDDISTRAPPROX
	\endverbatim
    */
    friend ostream& operator<<(ostream& Out, const Distrapprox_& D);
    
    /**
     * Reads (deserializes) a Distrapprox_ object from In.
     * See operator<< for a format description.
     */
    friend istream& operator>>(istream& In, Distrapprox_& D);
    
    // hidden methods
    protected:
	    
    int bin_index(double X) const
    {
	return static_cast<int>(floor(X/Width+0.5));
    }
    double bin_midpoint(unsigned int Idx) const { return (Width*(Idx+0.5)+Low); }
    int locate_index(double X) const;
    double lin_intp(double X, const vector<double>& YY, double YYmax) const;
};
// END OF CLASS Distrapprox_

} // RazorBack

// ==== END OF CLASS HEADER Distrapprox.hh ====

#endif	    // DISTRAPPROX_HEADER
