#ifndef DISTRFUNC_HEADER
#define DISTRFUNC_HEADER

// ==== HEADER Distrfunc.hh ====

/**
 * \file Distrfunc.hh
 * Classes for univariate continuous distributions.
 */

// 19-Sep-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <math.h>
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifdef _STANDARD_C_PLUS_PLUS
#include <vector>
    using namespace std;
#else
#include <vector.h>
#endif

// ---- MODULE HEADERS ---- 

#include"Distrbase.hh"
#include"Random.hh"
#include "Rangedimexc.hh"    // out-of-range exceptions

// ==== CLASSES ====

namespace RazorBack {

/**
 * Distrfunc_: function object abstract base class that implements 
 * a univariate PDF with an arbitrary number of parameters.
 * Derive your PDF from this base class (see examples below)
 * so that you overload the pdf() and/or cdf() calls appropriately.
 */
class Distrfunc_: public Distrbase_
{
    // data
    vector<double> Parameters;	///< parameters as STL vector
    
    // methods
    public:
    
    /// default ctor: init to hold N parameters.
    Distrfunc_(unsigned int N=0): Parameters(N) {}
    
    virtual ~Distrfunc_() {}
    
    /// param(N): returns the value of the N-th parameter.
    double param(unsigned int N) const;
    
    /**
    * param(N, X): set the N-th parameter to X, returns old value.
    * Throws Indexrangexc_ if N is out of range.
    */
    double param(unsigned int N, double X);
    
};
// END OF CLASS Distrfunc_

// ==== PDF CLASSES ====

// The classes below implement important distributions
// which are often used in practice.
// Other distributions may be implemented in the same way
// by deriving a new class from Distrfunc_.

/// Uniform_: implements the uniform distribution in a range.
class Uniform_: public Distrfunc_
{
    /// data
    Randomuni_ Rand;	///< uniform random number generator
    double Range;   ///< upper-lower limit
    
    // methods
    public:
	    
    /// init to range Low...Up
    Uniform_(double Low=0.0, double Up=1.0);
    
    /// Returns the lower limit of the range.
    double low() const { return param(0); }
    
    /// Sets the lower limit to L, returns old lower limit.
    double low(double L)
    {
	return ((L<param(1))? param(0, L): param(0));
    }
    
    /// Returns the upper limit of the range.
    double up() const { return param(1); }
    
    /// Sets the upper limit to U, returns old upper limit.
    double up(double U)
    {
	return ((U>param(0))? param(1, U): param(1));
    }
    
    /// Calculates the PDF at X.
    virtual double pdf(double X) const
    {
	if (X<low() || X>up()) return 0.0;
	else return(1.0/Range);
    }
    
    /// Calculates the CDF at X.
    virtual double cdf(double X) const
    {
	if (X<low()) return 0.0;
	if (X>up()) return 1.0;
	return (X-low())/Range;
    }
    
    /// Calculates the inverse CDF, returns X where CDF(X)=Y.
    virtual double inv_cdf(double Y) const
    {
	if (Y<0.0 || Y>1.0)
	    throw Argexc_("Uniform_::inv_cdf()", Y, 0.0, 1.0);
	return (Y*Range+low());
    }
    
    /// Returns the mean.
    virtual double mean() const { return (low()+up())/2.0; }
    
    /// Returns the variance.
    virtual double var() const { return (Range*Range)/12.0; }
    
    /// Returns the skewness of the distribution.
    virtual double skew() const { return 0.0; }
    
    /// Returns the kurtosis of the distribution.
    virtual double kurt() const { return 1.8; /* exact is 9/5 :-) */ }
    
    /// Returns the entropy.
    virtual double entropy() const { return (-log(1.0/Range)); }
    
    /// Returns a random number from this uniform distribution
    virtual double random() const { return(Rand.rnd_lim(low(), up())); }
};
// END OF CLASS Uniform_

/// Gaussian_: implements the Normal distribution.
class Gaussian_: public Distrfunc_
{
    // data
    static const double SQRT_2PI, SQRT2;
    Randomnorm_ Rand;	///< Gaussian random number generator
    
    // methods
    public:
	
    /**
     * default ctor: init to mean M, standard deviation S.
     * Default values are for the standard normal distribution.
     * Throws Badvalexc_ if S==0. If S<0.0 then the absolute value
     * will be used.
     */
    Gaussian_(double M=0.0, double S=1.0):
	    Distrfunc_(2), Rand()
    {
	S=fabs(S);
	if (S==0.0)
	    throw Badvalexc_("Gaussian_(M,S)", S);
	param(0, M); param(1, S);
    }
    
    // Returns the center (mu) parameter of the Gaussian.
    double mu() const { return param(0); }
    
    // Sets the center of the Gaussian, returns old value.
    double mu(double M) { return param(0, M); }
    
    // Returns the spread (sigma) parameter of the Gaussian.
    double sigma() const { return param(1); }
    
    /**
     * Sets the spread of the Gaussian, returns old value.
     * Throws Badvalexc_ if S==0. If S<0.0 then the absolute value
     * will be used.
     */
    double sigma(double S)
    {
	S=fabs(S);
	if (S==0.0)
	    throw Badvalexc_("Gaussian_::sigma(S)", S);
	return param(1, S);
    }
    
    /// Returns the PDF at X.
    virtual double pdf(double X) const;
    
    /// Returns the CDF at X.
    virtual double cdf(double X) const;
    
    /// Returns the inverse CDF, the value of X for which CDF(X)=Y.
    virtual double inv_cdf(double Y) const;
    
    /// Returns the mean (this is the mu parameter).
    virtual double mean() const { return mu(); }
    
    /// Returns the variance (this is the squared sigma parameter).
    virtual double var() const { return sigma()*sigma(); }
    
    /// Returns the skewness of the distribution.
    virtual double skew() const { return 0.0; }
    
    /// Returns the kurtosis of the distribution.
    virtual double kurt() const { return 3.0; }
    
    /// Returns the entropy.
    virtual double entropy() const
    {
	return((log(2.0*M_PI*var())+1)*0.5);
    }
    
    /// Returns a random number from this normal distribution.
    virtual double random() const { return Rand.rnd_msd(mu(), sigma()); }
    
    // hidden methods
    protected:
	    
    static double cdf_std(double Z);
    static double invcdf_ratio(double P);
};
// END OF CLASS Gaussian_

} // RazorBack

// ==== END OF HEADER Distrfunc.hh ====

#endif  // DISTRFUNC_HEADER





