// ==== METHODS Distrfunc.cc ====

// Classes for univariate continuous distributions.

// 23-June-1999. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <float.h>

// ---- MODULE HEADER ---- 

#include"Distrfunc.hh"
using namespace RazorBack;

// ==== Distrfunc_ METHODS ====

// param(N): returns the value of the N-th parameter.
// param(N, X): set the N-th parameter to X, returns old value.
// Throws Indexrangexc_ if N is out of range.
double Distrfunc_::param(unsigned int N) const
{
    if (N>=Parameters.size())   // throw range exception here
	throw Indexrangexc_("Distrfunc_::param() const", N, Parameters.size()-1);
    return Parameters[N];
}
double Distrfunc_::param(unsigned int N, double X)
{
    if (N>=Parameters.size())   // throw range exception here
	throw Indexrangexc_("Distrfunc_::param()", N, Parameters.size()-1);
    double Oldp=Parameters[N];
    Parameters[N]=X;
    return Oldp;
}

// ==== Uniform_ METHODS ====

// init to range Low...Up
Uniform_::Uniform_(double Low, double Up):
	Distrfunc_(2)
{
    if (Low>Up) swap(Low, Up);
    if (Low-Up==0.0)    // exception here... later
    {
	cerr<<"\n? Uniform_("<<Low<<","<<Up<<"): Range is zero\n";
	Low=0.0; Up=1.0;
    }
    param(0, Low); param(1, Up);
    Range=param(1)-param(0);
}

// ==== Gaussian_ METHODS ====

// ---- Static member init ----

// Under Linux, with G++ 2.95.2, static init with a std function
// does not work when this library is compiled to a shared object.
// So we give the constants as calc'd by Mathematica :-)
// 13-Jun-2000
const double Gaussian_::SQRT_2PI=2.506628274631; // sqrt(2.0*M_PI);
const double Gaussian_::SQRT2=1.414213562373;    // sqrt(2.0);

// ---- pdf, cdf ----

// pdf(X): implement the actual function call here.
double Gaussian_::pdf(double X) const
{
    register double Z;

    Z=(X-mu())/sigma();
    return(exp(-Z*Z/2)/(SQRT_2PI*sigma()));
}

// cdf_std(X): the CDF for the Standard Normal distribution.
// Hidden static
inline
double Gaussian_::cdf_std(double Z)
{
    register double Y;
    Y=(Z<0.0)? erfc(-Z/SQRT2): (1.0+erf(Z/SQRT2));
    return 0.5*Y;
}
    
// cdf(X): we can calculate it directly.
double Gaussian_::cdf(double X) const
{
    register double Z;

    Z=(X-mu())/sigma();
    return cdf_std(Z);
}

// inv_cdf(Y): this implementation is a modification of the
// dinvnr() routine from the DCDFLIB library (Version 1.1)
// by B.W. Brown, J. Lovato, K. Russell.
double Gaussian_::inv_cdf(double Y) const
{
    static const size_t MAXITER=100;	// Newton-Raphson iterations
    static const double EPS=1.0e-13,	// iteration tolerance
	    R2PI=0.3989422804014326;
    
    if (Y<0.0 || Y>1.0)    // arg range exception
    	throw Argexc_("Gaussian_::inv_cdf()", Y, 0.0, 1.0);
    
    if (Y<DBL_EPSILON) return -FLT_MAX; 	// around the tails
    if (Y>(1.0-DBL_EPSILON)) return FLT_MAX;
    
    // use complement arg if Y>0.5
    bool Ylesshalf=(Y<=0.5);
    if (!Ylesshalf) Y=1.0-Y;
    
    // init iteration
    register double Strtx=invcdf_ratio(Y);	// starting approximation
    register double Xcur=Strtx, Cum, Dx, Dinvnr;
    bool Success=false;
    
    // Newton/Raphson iterations
    for(register size_t i=1; !Success && i<=MAXITER; i++)
    {
        Cum=cdf_std(Xcur);
        Dx=(Cum-Y)/R2PI*exp(-0.5*Xcur*Xcur);
        Xcur-=Dx;
        Success=(fabs(Dx) < EPS*fabs(Xcur));
    }

    if (Success)
	Dinvnr=Ylesshalf? Xcur: -Xcur;
    else
    	Dinvnr=Ylesshalf? Strtx: -Strtx; // no convergence
    return (sigma()*Dinvnr+mu());	// scale
}
// END of inv_cdf(Y)

// invcdf_ratio(P): approximates the inverse of the std normal CDF
// with a 4:4 rational function.
// Return value: an X for which CDF(X)=P.
// Algorithm: page 95 of Kennedy and Gentle,
// Statistical Computing, Marcel Dekker, NY, 1980.
// This implementation is a modification of the stvaln()
// routine in the DCDFLIB library (see above).
// Hidden static
double Gaussian_::invcdf_ratio(double P)
{
    static const double Denomcoeff[5] = {
    	    9.93484626060e-2, 0.588581570495, 0.531103462366,
	    0.103537752850, 3.8560700634e-3
    };
    static const double Numcoeff[5] = {
    	    -0.322232431088, -1.0, -0.342242088547,
	    -2.04231210245e-2, -4.53642210148e-5
    };

    register double Sign, Y, Z, Retval;
    if (P<=0.5)
    {
	Z=P; Sign=-1.0;
    }
    else
    {
	Z=1.0-P; Sign=1.0;
    }
    Y=sqrt(-(2.0*log(Z)));
    
    // evaluate rational function of Y (two Horners in one cycle)
    register int i;
    register double Num=Numcoeff[4], Denom=Denomcoeff[4];
    for (i=3; i>=0; --i)
    {
	Num=Numcoeff[i]+Num*Y;
	Denom=Denomcoeff[i]+Denom*Y;
    }
    Retval=Sign*(Y+Num/Denom);
    return Retval;
}
// END of invcdf_ratio(P)

// ==== END OF METHODS Distrfunc.cc ====
