// ==== METHODS Linls.cc ====

// Linear least squares parameter estimation.
// Based on SVD as explained in Golub/Van Loan (Chapter 5)

// 14-Nov-2001. Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ==== MODULE HEADERS ====

#include "Linls.hh"
#include "Rangedimexc.hh"
using namespace RazorBack;

// ==== METHODS ====

// lin_reg(Xmeas, Ymeas, Eps):
// Carries out a linear least squares fit using SVD.
// Xmeas is the data matrix (Nm rows of observations for Np indep. vars,
// Nm>Np), Ymeas is the Nm observations of the dependent variable.
// Eps is a small tolerance value used for finding the best SVD rank.
// Should be a "small multiple of the machine zero". If in doubt,
// set it to 0.0 and it will be figured out for you.
// Return value: the chi-square statistics of the fit,
// or a negative value on error.
double Linls_::lin_reg(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas, double Eps)
{
    adjust_dimensions(Xmeas, Ymeas);
    Svd.make_decomp(Xmeas); // may throw Svdexc_

    make_utb(Ymeas);
    Rank=calc_rank(Ymeas, Eps);
    if (!Rank)
	throw Argexc_("Linls_::lin_reg", 0, 1, Np);

    // obtain parameters, SD, correl matrix, chi-square
    calc_params(Xmeas, Ymeas);

    // get Chi2
    return chi_square();
}
// END of lin_reg(Xmeas, Ymeas, Eps)

// adjust_dimensions(Xmeas, Ymeas): checks that everything has
// an appropriate size, and sets the internal data accordingly.
// Throws Dimexc_ exceptions where appropriate.
// Hidden
void Linls_::adjust_dimensions(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas)
{
    // Xmeas should be Nm x Nx, Nm>Nx, Ymeas should be Nm
    size_t M=Xmeas.rno(), X=Xmeas.cno(), Y=Ymeas.dim(), F=M-X;
    if (F<=0)
	throw Dimexc_("Linls_::adjust_dimensions(): Not enough measurement points", M, X);
    if (M!=Y)
	throw Dimexc_("Linls_::adjust_dimensions()", M, Y);
    
    // OK, set sizes
    Nm=M; Np=X; Nf=F;
    Svd.set_size(Nm, Np);
    Utb.dim(Np);
}
// END of adjust_dimensions(Xmeas, Ymeas)

// make_utb(B):
// calculates the scalar products of the columns of U and B,
// the right-hand side vector.
// Hidden
void Linls_::make_utb(const Vecdbl_& B)
{
    Vecdbl_ Uj(Svd.rno());
    size_t j;
    
    for (j=0; j<Svd.cno(); ++j)
    {
	Uj=Svd.u().col(j); // get j:th col
	Utb[j]=Uj*B;
    }
}
// END of make_utb(B)

// calc_rank(B, Eps):
// Golub & Van Loan give an upper bound for the parameter error (Eqn. 5.5.11)
// as a function of the rank: this is implemented here.
// B is the right-hand side, Eps is a small number,
// if set to <=0.0 (the default) then 10xDBL_EPSILON is used.
// Returns the rank minimising the error upper bound.
// Hidden
size_t Linls_::calc_rank(const Vecdbl_& B, double Eps) const
{
    // a very small number: "small multiple of machine precision"
    // as Golub & Van Loan put it
    if (Eps<=0.0) Eps=10.0*DBL_EPSILON;
    
    double Term1=2.0*(1.0+Eps)*Eps*B.vec_len();
    double Term2=0.0, Sq, Singval, 
	Bound, Lowestbound=HUGE_VAL;
    int r;
    size_t Temprank=0;
    
    // try to lower the rank, see if bound goes down, too
    for (r=Svd.cno(); r>0; --r)
    {
    	if ((Singval=Svd.w()[r-1])<Eps)  // accept as machine zero
	    continue;	// "zero" singular value
	
	Bound=r*Term1/Singval+sqrt(Term2);
	if (Bound<Lowestbound)	// bound can still be improved
	{
	    Lowestbound=Bound;
	    Temprank=r;
	}
	if (Bound>Lowestbound)	// goes up again
	    break;
	
	Sq=Utb[r-1]/Singval;	// update 2nd term
	Term2+=Sq*Sq;	    	// will be used in next round!
    }
    return Temprank;
}
// END of calc_rank(B, Eps)

// calc_params(Xmeas, Ymeas):
// Calculates the parameters, their standard deviation
// and correlation matrix.
// Hidden
void Linls_::calc_params(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas)
{
    // invert the singular values
    size_t r;
    Vecdbl_ Invsval(rank());
    for (r=0; r<rank(); ++r)
	Invsval[r]=1.0/Svd.w()[r];
    
    // get the parameters
    P.set_values();
    for (r=0; r<rank(); ++r)
	P+=Utb[r]*Invsval[r]*Svd.v().col(r);
    
    // get the chi-square
    Chi2=calc_chi2(Xmeas, Ymeas);
    
    // covariances
    // these are to be scaled by the residual deviation
    // do this through the inverse singular values
    Invsval*=residual_dev();
    size_t j, k;
    double Covtemp;
    for (j=0; j<Np; ++j)
    {
	for (k=0; k<=j; ++k)
	{
	    Covtemp=0.0;
	    for (r=0; r<rank(); ++r)
		Covtemp+=Svd.v()[j][r]*Invsval[r]*Svd.v()[k][r]*Invsval[r];
	    Correl[j][k]=Covtemp;
	}
    }
    
    // norm the covariance matrix to correlation
    // copy the variances to the standard deviation
    double Norm;
    for (j=0; j<Np; ++j)
    {
	Sdev[j]=sqrt(fabs(Correl[j][j]));
	Correl[j][j]=1.0;
	for (k=0; k<j; ++k)
	{
	    Norm=Sdev[j]*Sdev[k];
	    if (Norm<10.0*DBL_EPSILON)
	    	Correl[j][k]=0.0;
	    else Correl[j][k]/=Norm;
	}
    }
}
// END of calc_params(Xmeas, Ymeas)

// calc_chi2(Xmeas, Ymeas):
// Returns the Chi2 error measure.
// Hidden
double Linls_::calc_chi2(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas) const
{
    double D, Chisq=0.0;
    Vecdbl_ Yest=Xmeas*P;   // estimated Y values
    register size_t im;

    // sum of squares
    for (im=0; im<Nm; ++im)
    {
	D=Yest[im]-Ymeas[im];
	Chisq+=D*D;
    }
    return Chisq;
}
// END of calc_chi2(Xmeas, Ymeas)

// ==== END OF METHODS Linls.cc ====
