// ==== METHODS Linreg.cc ====

// Linear regression with normal equations.

// 19-Sep-2001, Andras Aszodi

// ---- MODULE HEADERS ----

#include "Linreg.hh"
#include "Rangedimexc.hh"
#include "Statexc.hh"
using namespace RazorBack;

// ==== Linreg_ METHODS ====

// lin_reg(Xmeas, Ymeas [,Userwgt]): multivariate linear regression
// calculating the parameters P in the equation
// Y=P[0]*X[0]+...+P[Np-1]*X[Np-1].
// Xmeas is an Nm x Np matrix holding Nm rows of Np independent vars,
// Ymeas is an Nm-long vector holding the dependent vars.
// Userwgt is an optional Nm-long vector of weights.
// Throws Dimexc_ and Toofewexc_.
// Return value: the residual deviation.
// Invoke param(), sdev(), correl() for the detailed results.
double Linreg_::lin_reg(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas, const Vecdbl_& Userwgt)
{
    // setup and checks
    adjust_dimensions(Xmeas, Ymeas, Userwgt);	// may throw exceptions
    
    setup_weight(Userwgt);
    setup_xtx(Xmeas, Ymeas);
    normalise();
    solve(P, 0.0, 0.0);
    double Q=quality(Xmeas, Ymeas);
    sdev_correl(Q);
    return Q;
}
double Linreg_::lin_reg(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas)
{
    if (Weighting==USER)
    {
	cerr<<"\n? Linreg_::lin_reg(Xmeas, Ymeas): No USER weights, EQUAL used\n";
	Weighting=EQUAL;
    }
    
    // Pass Ymeas as Userwgt. Will be used for RELATIVE, ignored for EQUAL
    return (lin_reg(Xmeas, Ymeas, Ymeas));
}
// END of lin_reg(Xmeas, Ymeas [,Userwgt])

// adjust_dimensions(Xmeas, Ymeas ,Userwgt]): checks that everything has
// an appropriate size, and sets the internal data accordingly.
// Throws Dimexc_ and Toofewexc_.
// Hidden
void Linreg_::adjust_dimensions(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas, const Vecdbl_& Userwgt)
{
    if (Weighting==USER && Ymeas.dim()!=Userwgt.dim())
	throw Dimexc_("Linreg_::adjust_dimensions(): Measurement/weight mismatch",
		Ymeas.dim(), Userwgt.dim(), true);

    // Xmeas should be Nm x Nx, Nm>Nx, Ymeas should be Nm
    size_t M=Xmeas.rno(), X=Xmeas.cno(), Y=Ymeas.dim(), F=M-X;
    if (F<=0)
	throw Toofewexc_("Linreg_::adjust_dimensions()", "Measurement points", M, true);
    if (M!=Y)
	throw Dimexc_("Linreg_::adjust_dimensions(): Measurement number mismatch",
		X, Y, true);
    
    // OK, set sizes (except weight vector)
    Nm=M; Np=X; Nf=F;
    resize();	// see parent class
}
// END of adjust_dimensions(Xmeas, Ymeas, Userwgt)

// setup_weight(YW): inits the weight vector according to Weighting:
// if EQUAL, then W={1, 1, ...,1}, YW ignored;
// if RELATIVE, then YW is interpreted as Ymeas;
// if USER, then YW is interpreted as a user-defined weight vector.
// Hidden
void Linreg_::setup_weight(const Vecdbl_& YW)
{
    static const double WMIN=1e-6;
    double Wtemp, Wsum=0;  // for relative and user weights
    register unsigned int im;
    
    W.dim(Nm);	// set weight vector size
    switch (Weighting)
    {
	case EQUAL:
	    W.set_values(1.0);
	break;
	case RELATIVE:
	    for (im=0; im<Nm; ++im)
	    {
		Wtemp=fabs(YW[im])+WMIN;
		W[im]=1.0/(Wtemp*Wtemp);
		Wsum+=W[im];
	    }
	    W*=(Nm/Wsum);
	break;
	case USER:
	    for (im=0; im<Nm; ++im)
	    {
		Wtemp=fabs(YW[im])+WMIN;
		W[im]=Wtemp;
		Wsum+=Wtemp;
	    }
	    W*=(Nm/Wsum);
	break;
    }
}
// END of setup_weight(YW)

// setup_xtx(Xmeas, Ymeas): makes the X'WX and X'WY matrices.
// Hidden
void Linreg_::setup_xtx(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas)
{
    register double Tempx, Tempy;
    register unsigned int im, ip, jp;
    
    // NOTE: this is a typical example of
    // "make it run first, then make it run fast" :-)
    for (ip=0; ip<Np; ++ip)
    {
	// make X'WX (triangular)
	for (jp=0; jp<=ip; ++jp)
	{
	    Tempx=0.0;
	    for (im=0; im<Nm; im++)
		Tempx+=Xmeas[im][ip]*W[im]*Xmeas[im][jp];
	    Xtx[ip][jp]=Tempx;
	}

	// make X'WY
    	Tempy=0.0;
	for (im=0; im<Nm; ++im)
	    Tempy+=Xmeas[im][ip]*W[im]*Ymeas[im];
	Xty[ip]=Tempy;
    }
}
// END of setup_xtx(Xmeas, Ymeas)

// quality(Xmeas, Ymeas): calculates the residual deviation,
// standard deviation and correlation matrix.
// Hidden
double Linreg_::quality(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas)
{
    double D, Q;
    Vecdbl_ Yest=Xmeas*P;   // estimated Y values
    register unsigned int im;

    // sum of squares
    for (Q=0.0, im=0; im<Nm; ++im)
    {
	D=Yest[im]-Ymeas[im];
	Q+=W[im]*D*D;
    }
    Q=sqrt(Q/Nf);   // residual deviation
    return Q;
}
// END of quality(Xmeas, Ymeas)

// ==== END OF METHODS Linreg.cc ====
