// ==== METHODS Mvn.cc ====

// Class for multivariate Normal distributions.

// 23-Feb-2000. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <float.h>
#include <string.h>

// ==== MODULE HEADERS ====

#include"Mvn.hh"
#include"Distrfunc.hh"
#include "Rsymdiag.hh"
using namespace RazorBack;

// ==== Mvn_ METHODS ====

// ---- Static init ----

#ifdef M_PI
    const double Mvn_::TWO_PI=2.0*M_PI;
#else
    const double Mvn_::TWO_PI=6.2831853;
#endif

// ---- Constructors ----

// Inits to store a D-dimensional MVN centered on the origin,
// unit covariance matrix (standard MVN).
Mvn_::Mvn_(size_t D)
{
    dim(D); // adjust dimensions
    Avgvec.set_values();	// nullvector
    Covarmat.diag_matrix(1.0); // unit matrix
    Covrank=D;	    // full rank of course
    Eval.set_values(1.0);   // all eigenvalues are 1.0
    Evec.diag_matrix(1.0);  // also unit matrices
    Stdnmap.diag_matrix(1.0);
    Normfactor=1.0/pow(TWO_PI, dim()/2.0);
}

// Inits to store a MVN distribution with average A and
// covariance matrix C. These should have the same dimension of course.
Mvn_::Mvn_(const Vecdbl_& A, const Smatdbl_& C)
{
    if (A.dim()!=C.rno())
	throw Dimexc_("Mvn_(A, C)", A.dim(), C.rno());
    dim(A.dim());
    avg(A); covar(C);
}

// ---- Access ----

// dim(D): adjusts the calling object to represent a D-dimensional
// MVN distribution. If D<2 then it will be set to 2 silently.
// Does nothing if D is the same as the previous dimension.
// Do not forget to call avg(Newavg) and covar(Newcov) afterwards!
// Return value: the old dimension.
size_t Mvn_::dim(size_t D)
{
    if (D==dim()) return D; // do nothing if no change
    
    size_t Olddim=dim();    // save old dim for return value
    if (D<2) D=2;   // only real MVN-s are allowed
    
    // resize members
    Avgvec.dim(D);
    Covarmat.size(D);
    Eval.dim(D); Evec.size(D, D);
    Stdnmap.size(D, D);
    Dim=D;
    return Olddim;
}
// END of dim(D)

// ---- Calc ----

// random(X): generates a dim()-dimensional random vector
// from the MVN distribution stored in the calling object.
// The dimension of X is adjusted silently, previous contents
// are overwritten.
void Mvn_::random(Vecdbl_& X) const
{
    static Vecdbl_ Stdnorms;    // storage for indep. std. normals
    static Gaussian_ Stdn;  // N(0,1) function object
    
    X.dim(dim()); Stdnorms.dim(covrank());  // set dimensions
    
    // generate independent standard normal variates
    for (register size_t d=0; d<covrank(); ++d)
	Stdnorms[d]=Stdn.random();
    
    // mix them and shift average
    X=Stdnmap*Stdnorms+avg();
}
// END of random(X)   
  
// maha_dist(X):
// calculates and returns the Mahalanobis distance
// between X and the average vector in the calling object.
double Mvn_::maha_dist(const Vecdbl_& X) const
{
    if (dim()!=X.dim())
	throw Dimexc_("Mvn_::maha_dist(X)", dim(), X.dim());
    
    Vecdbl_ Normx=X-avg();
    register double Mdist=0.0, Scalprod;
    register size_t i, j;
    
    for (i=0; i<covrank(); ++i)
    {
	Scalprod=0.0; 
	for (j=0; j<dim(); ++j)
	    Scalprod+=Evec[j][i]*Normx[j];
	Mdist+=Scalprod*Scalprod/Eval[i];
    }

    return Mdist;
}
// END of maha_dist(X)

// covar(Newcov): sets the covariance of the distribution to Newcov
// if Newcov has the same dimension as the calling object and
// if it has at least rank 2. Otherwise nothing is changed.
// Return value: the rank if OK or 0 on error.
int Mvn_::covar(const Smatdbl_& Newcov)
{
    if (dim()!=Newcov.rno())
	throw Dimexc_("Mvn_::covar(Newcov)", dim(), Newcov.rno());
    
    // first check if diagonalisation & rank is OK
    Vecdbl_ Tempeval(dim());
    Rmatdbl_ Tempevec(dim(), dim());
    
    // diagonalise the covariance matrix
    Rsymdiag_ Diag(Newcov);
    if (0>Diag.all_evalevec(Tempeval, Tempevec))
    {
	cerr<<"\n? Mvn_::covar(): Diagonalisation failed\n";
	return 0;
    }

    // determine rank by ignoring almost-zero eigenvalues
    size_t Tempcovrank;
    for (Tempcovrank=dim(); 
	    Tempcovrank>=1 && Tempeval[Tempcovrank-1]<FLT_EPSILON; --Tempcovrank)
	Tempeval[Tempcovrank-1]=0.0;
    if (Tempcovrank<dim())
    {
	cerr<<"\n? Mvn_::update_covar(): Covar matrix rank deficient ("
		<<Tempcovrank<<"<"<<dim()<<")\n";
	if (Tempcovrank<2) return 0;
    }
    
    // was OK
    Covarmat=Newcov; Covrank=Tempcovrank;
    Eval=Tempeval; Evec=Tempevec;

    // calc the normalisation factor
    Normfactor=1.0;
    for (register size_t d=0; d<covrank(); ++d)
	Normfactor*=TWO_PI*Eval[d];
    Normfactor=1.0/sqrt(Normfactor);
    
    // calc the N(0,1)-->MVN mapping matrix
    Stdnmap.size(dim(), covrank());
    Vecdbl_ Tempe(dim());
    register double Sqrteval;
    
    // the j:th column of Stdnmap is the j:th eigenvector,
    // scaled by the square root of the j:th eigenvalue
    for (register size_t j=0; j<covrank(); ++j)
    {
	Sqrteval=sqrt(Eval[j]);
	Tempe=Evec.col(j);
	Tempe*=Sqrteval;
	Stdnmap.col(Tempe, j);
    }

    return (covrank()); // OK
}
// END of covar(Newcov)

// marginal(Bitmask, Marginal):
// changes Marginal so that it contains a marginal distribution
// of the calling object. The variables corresponding to the
// bits set in Bitmask will be kept, the others will be
// "integrated out". E.g. if the calling object is a 4-dim MVN
// with PDF(x1,x2,x3,x4), and the bitmask is "1010", then
// Marginal will be a 2-dim MVN with PDF(x2,x4).
// The bitmask should contain exactly as many bits as dim(),
// otherwise a Dimexc_ exception will be thrown.
// At least two bits must be set to ON so that Marginal remains
// a valid Mvn_ object; otherwise a warning is printed
// and no action will be taken.
// To get a 1-D marginal distribution,
// simply obtain the corresponding average coordinate and
// variance and use a Gaussian_ object.
// Return value: the dimension of Marginal.
size_t Mvn_::marginal(const Bits_& Bitmask, Mvn_& Marginal) const
{
    if (dim()!=Bitmask.len())
	throw Dimexc_("Mvn_::marginal()", dim(), Bitmask.len());
    size_t Margdim=Bitmask.on_no();
    if (Margdim<2)
	throw Utilsexc_("Mvn_::marginal()", false, "At least 2 bits should be on in Bitmask");
    
    Marginal.dim(Margdim);
    Vecdbl_ Margavg(Margdim);
    Smatdbl_ Margcovar(Margdim);
    register size_t i, j, mi, mj;
    
    for (i=mi=0; i<dim(); ++i)
    {
	if (!Bitmask.get_bit(i)) continue;
	
	for (j=mj=0; j<=i; ++j)
	{
	    if (!Bitmask.get_bit(j)) continue;
	    Margcovar[mi][mj]=covar()[i][j];
	    ++mj;
	}
	Margavg[mi]=avg()[i];
	++mi;
    }
    Marginal.avg(Margavg);
    Marginal.covar(Margcovar);
    return Margdim;
}
// END of marginal(Bitmask, Marginal)    

// ---- I/O ----

// The ASCII format of an Mvn_ object is:-
//
// MVN
// AVERAGE
// <dbl vector>     (center of distribution)
// COVARIANCE
// <dbl lower triangular matrix>    (covariance matrix)
// ENDMVN
//
// For the formats of the vectors and matrices, see
// the corresponding linalg header files.
// Empty lines are allowed, lines starting with '#' are comments.

// friend fn defs must be in scope
namespace RazorBack {

istream& operator>>(istream& In, Mvn_& Mvn)
{
#define LINELEN 120
    enum {MVN, AVERAGE, COVARIANCE, ENDMVN, DONE} Token;
    const char WARN[]="\n? >>Mvn_: ";
    char Line[LINELEN], *Eolnptr;
    bool Error=false;
    Vecdbl_ Tempavg;
    Smatdbl_ Tempcovar;
    size_t D;
    
    Token=MVN;
    
    // read input line by line
    while (!Error && Token!=DONE && In.getline(Line, LINELEN, '\n').good())
    {
	// skip empty lines or comments
	if (!Line[0] || Line[0]=='\n' || Line[0]=='#')
	    continue;
	
	// chop off terminating \n if present
	Eolnptr=strrchr(Line, '\n');
	if (Eolnptr!=NULL) *Eolnptr='\0';
	
	// action depends on which token we are looking for
	// after an action was successful, Token is set to
	// the next symbol
	switch (Token)
	{
	    case MVN:
		if (!strncmp(Line, "MVN", 3))
		    Token=AVERAGE;
		else
		{
		    cerr<<WARN<<"MVN token error\n";
		    Error=true;
		}
	    break;
	    case AVERAGE:
		if (!strncmp(Line, "AVERAGE", 7))
		{
		    if (In>>Tempavg)
		    {
			D=Tempavg.dim();
			if (D<2)
			{
			    cerr<<WARN<<"AVERAGE dim="
				    <<D<<" (must be at least 2)\n";
			    Error=true;
			}
			else Token=COVARIANCE;
		    }
		    else
		    {
			cerr<<WARN<<"AVERAGE error\n";
			Error=true;
		    }
		}
	    break;
	    case COVARIANCE:
		if (!strncmp(Line, "COVARIANCE", 10))
		{
		    if (In>>Tempcovar)
		    {
			if (Tempcovar.rno()!=D)
			{
			    cerr<<WARN<<"COVARIANCE size mismatch, "
				    <<D<<" expected\n";
			    Error=true;
			}
			else Token=ENDMVN;
		    }
		    else
		    {
			cerr<<WARN<<"COVARIANCE error\n";
			Error=true;
		    }
		}
	    break;
	    case ENDMVN:
		if (!strncmp(Line, "ENDMVN", 6))
		    Token=DONE;
	    break;
	}   // switch
    }	    // while
    
    if (!Error)     // success
    {
	if (Token!=DONE)
	    cerr<<WARN<<"\"ENDMVN\" token missing\n"; // just a warning
	
	// storage
	if (!Mvn.covar(Tempcovar))  // rank-deficiency
	{
	    cerr<<WARN<<"Covar rank deficient\n";
	    In.clear(ios::failbit);    // set error condition
	}
	else Mvn.avg(Tempavg);	// OK, store average, too
    }
    else In.clear(ios::failbit);    // set error condition
    return In;
#undef LINELEN
}
// END of >>

ostream& operator<<(ostream& Out, const Mvn_& Mvn)
{
    Out<<"MVN\nAVERAGE\n"<<Mvn.avg()<<endl;
    Out<<"COVARIANCE\n"<<Mvn.covar();
    Out<<"ENDMVN\n";
    return Out;
}
// END of <<

} // RazorBack

// === END OF METHODS Mvn.cc ====
