#ifndef BIOSEQIO_HEADER
#define BIOSEQIO_HEADER

// ==== HEADER Bioseqio.hh ====

/**
 * \file Bioseqio.hh
 * Base class for biological sequence input/output.
 */

// 18-Aug-1999. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// ANSI C++ standard only.

// ==== STANDARD HEADERS ====

#include <stdlib.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#else
#include <iostream.h>
#endif

#include <string>
using namespace std;
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== MODULE HEADERS ====

#include"Monomer.hh"
#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Bioseqio_: the base class for biological sequence input/output.
 * Defines the I/O methods for some popular formats.
 */
class Bioseqio_
{
    // data
    private:
    char *Buffer;   ///< I/O buffer
    static const int BUFFERSIZE;    ///< and its size
    
    // methods
    protected:
    
    /**
    * Create an empty I/O object. The buffer will be allocated
    * only when needed by an input method
    */
    Bioseqio_(): Buffer(NULL) {}

    /// copy ctor: we do not copy the buffer (see above)
    Bioseqio_(const Bioseqio_& B):Buffer(NULL) {}
    
    /**
    * assignment: buffer not overwritten!
    * essentially, this is a null operation
    */
    Bioseqio_& operator=(const Bioseqio_& B) { return *this; }
    
    /// dtor
    virtual ~Bioseqio_() { delete [] Buffer; }
    
    char* buffer() { return Buffer; }
    char& buffer(size_t Idx) { return Buffer[Idx]; }
    
    	/// Input-output
    
    public:
    
    /// I/O error types
    enum Errtype_ {OK=0, FILEERR, FORMATERR, TITLEERR, IDERR, ANNOTERR, SEQERR, TYPEERR};
    /// available I/O formats
    enum Formatflags_ {UNKNOWN=0, FASTA=1, GCG=2, PIR=4, SWISS=8, EMBL=16, ANYFORMAT=31};
    
    /**
    * error_str(), format_str(): return a const char ptr to
    * a human-readable short description of the various enums.
    */
    static const char *error_str(Errtype_ Err);
    static const char *format_str(Formatflags_ F);

    /// the rest of the I/O methods
    protected:
	    
    void make_buffer() { if (Buffer==NULL) Buffer=new char [BUFFERSIZE]; }
    bool fetch_buffer(istream& Inf);
    int clean_seq(string& S, int Type, bool Nocase, bool GapsOK=false) const;
    static int autodetect_type(const string& S);
    
    Errtype_ read_fasta(istream& Inf, string& Id,
		string& Title, string& Seq);
    Errtype_ read_pir(istream& Inf, string& Id,
	    string& Title, string& Annot, string& Seq);
    Errtype_ read_gcg(istream& Inf, string& Id,
	    string& Annot, string& Seq, 
	    int &Gcglen, int &Gcgcheck, char &Typechar);
    Errtype_ read_swiss(istream& Inf, string& Id,
	    string& Annot, string& Seq,
	    int& Swlen, int& Mw, unsigned int& Crc32);
    Errtype_ read_embl(istream& Inf, string& Id,
	    string& Annot, string& Seq, 
	    int& Embllen, int Acgtx[5]);
    
    static Errtype_ write_fasta(ostream& Outf, const string& ID,
	    const string& Title, const string& Seq);
    static Errtype_ write_pir(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq);
    static Errtype_ write_gcg(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq,
	Monomer_::Type_ Type);
    static Errtype_ write_swiss(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq);
    static Errtype_ write_embl(ostream& Outf, const string& ID,
	const string& Title, const string& Annot, const string& Seq,
	Monomer_::Type_ Type);
	    
    static int gcg_checksum(const string& S);
    static const char *current_datestr();
    static int mol_weight(const string& S);
    static void BuildCRCTable(unsigned int CRCTable[256]);
    static unsigned int crc_32 (const string& buffer);
    static void count_bases(const string& S, int Acgtx[5], bool Isdna);
    static bool compare_basecount(const string& S, const int Acgtx[5], bool Isdna);
    
    private:
	    
    static void parse_title(const char *Strc, string& Tok1, string& Rest);
    static void swissembl_comment(ostream& Outf, const string& S);
};
// END OF CLASS Bioseqio_

} // RazorBack

// ==== END OF HEADER Bioseqio.hh ====

#endif      // BIOSEQIO_HEADER
