#ifndef FPARCH_TMPL_HEADER
#define FPARCH_TMPL_HEADER

// ==== TEMPLATE HEADER Fparch.hh ====

/**
 * \file
 * Machine floating-point architecture constants. 
 * Based on freeware from NETLIB
 * originally written by W. J. Cody, Argonne National Laboratory
 * based on an algorithm due to M. Malcolm, Comm. ACM 15 (1972), pp. 949-951,
 * plus some improvements suggested by M. Gentleman and S. Marovich,
 * Comm. ACM 17 (1974), pp. 276-277.
 * \author Andras Aszodi
 * \date 14-Nov-2001.
 */

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <sys/utsname.h>

// ==== CLASS ====

namespace RazorBack {

/**
 * Objects of the Fparch_ class figure out as much as possible about the
 * floating-point architecture of a machine. The template parameter REAL
 * makes sense only if it is one of "float", "double" or "long double".
 * Functionality is similar to that of <numeric_limits> in the standard C++
 * library. This class is provided so that you can get information about the
 * floating-point architecture even if an ANSI C++ implementation is not available.
 * The main difference is that the members are not static as I was too lazy
 * to do the silly init for all static members before the ctor can be run. :-)
 */
template <class REAL>
class Fparch_
{
    private:
	    
    static const REAL ZERO, ONE, TWO;
    
    struct utsname Utsname;
    
    int Radix, Digits, Rnd, Negep, Machep, Nguard, Expdigits, Minexp, Maxexp;
    REAL Epsneg, Eps, Smallest, Largest;
    
    public:
	    
    /// default ctor
    Fparch_();
    
    /// returns the system description in a UNIX utsname struct.
    const struct utsname* system_descr() const { return &Utsname; }
    
    /// returns the radix of the floating-point representation.
    int radix() const { return Radix; }
    
    /// returns the number of significant base-Radix digits in the mantissa.
    int mantissa_digits() const { return Digits; }
    
    /// returns the number of significant base-Radix digits in the exponent.
    int exponent_digits() const { return Expdigits; }
    
    /// returns the minimal base-Radix exponent
    int min_exp() const { return Minexp; }
    
    /// returns the smallest representable floating-point number
    REAL smallest() const { return Smallest; }
    
    /// returns the maximal base-Radix exponent
    int max_exp() const { return Maxexp; }
    
    /// returns the largest representable floating-point number
    REAL largest() const { return Largest; }
    
    /// floating-point addition rounding index (0...5)
    int fpadd_round() const { return Rnd; }
    
    /// returns true if rounding is IEEE-754 compliant
    bool round_ieee754() const { return (Rnd==2 || Rnd==5); }
    
    /// returns true if some rounding is done but is not IEEE
    bool round_non_ieee() const { return (Rnd==1 || Rnd==4); }
    
    /// returns true if floating-point addition truncates
    bool fp_truncate() const { return (Rnd==0 || Rnd==3); }
    
    /// returns true if IEEE "graceful underflow" is available
    bool graceful_underflow() const { return (Rnd>=3); }
    
    /**
     * returns the largest negative integer such that
     * 1.0-radix^negep() != 1.0, bounded by -(digits()+3) from below
     */
    int negep() const { return Negep; }
    
    /// returns a small real number such that 1.0-epsneg() != 1.0
    REAL epsneg() const { return Epsneg; }
    
    /**
     * returns the largest negative integer such that
     * 1.0+radix^negep() != 1.0, bounded by -(digits()+3) from below
     */
    int machep() const { return Machep; }
    
    /// returns the smallest real number such that 1.0+eps() != 1.0
    REAL eps() const { return Eps; }
    
    /// Returns the number of guard digits for multiplication
    int nguard() const { return Nguard; }
    
    protected:
	    
    static REAL absolute(REAL x) { return (x>=ZERO? x: -x); }
};
// END OF CLASS Fparch_

} // RazorBack

#ifdef INCLUDE_TMPL_DEFS
#include "Fparch.cc"
#endif

// ==== END OF TEMPLATE HEADER Fparch.hh ====

#endif	    // FPARCH_TMPL_HEADER
