#ifndef GAPSEQ_HEADER
#define GAPSEQ_HEADER

// ==== HEADER Gapseq.hh ====

/**
 * \file Gapseq.hh
 * Class for storing gapped biological sequences.
 */

// 6-Aug-1999. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// ANSI C++ standard only.

// ==== STANDARD HEADERS ====

#include <stdlib.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <list>
#else
#include <iostream.h>
#include <list.h>
#endif

#include <string>
using namespace std;
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== MODULE HEADERS ====

#include"Monomer.hh"
#include"Bioseq.hh"
#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Gapseq_: this class stores possibly gapped biological sequences.
 * Provides methods for inserting and deleting gaps.
 */
class Gapseq_: public Bioseq_
{
    // data
    public:
	    
    /// usual gap symbol
    static const char GAP;  ///< the gap character '-'
    static const char* GAP3;	///< three-letter version "---"
    
    /// the slightly odd GCG gap symbols
    static const char GCG_GAP;  ///< the gap character '.'
    static const char* GCG_GAP3;	///< three-letter version "..."
    static const char GCG_ENDGAP;  ///< the end gap character '~'
    static const char* GCG_ENDGAP3;	///< three-letter version "~~~"
    
    private:
    
    /**
     * subclass Gap_: a Gap_ object stores the position in the sequence
     * before it is inserted and the gap length (>0). A gapped sequence
     * then contains a list of Gap_-s and the insertion/deletion of gaps
     * does not change the underlying original sequence itself.
     * A placeholder gap before the sequence has position 0,
     * and position NPOS after the sequence.
     */
    class Gap_
    {
	friend class Gapseq_;
	
	// data
	private:
	size_t Pos; 	///< position before gap is inserted [0..NPOS]
	size_t Len; 	///< length of gap, >0
	
	// methods
	public:
	
    	/// Inits a gap before ungapped position P with length L
	explicit Gap_(size_t P=0, size_t L=1): Pos(P), Len(L) {}
	
	size_t pos() const { return Pos; }
	void pos(size_t P) { Pos=P; }
	size_t len() const { return Len; }
	void len(size_t L) { if (L) Len=L; }
    };
    // END OF SUBCLASS Gap_
    
    list<Gap_> Gaplist;     ///< list of gaps in gapped sequence
    size_t Gaplen;  	///< total no. of gap chars in gapped sequence
    
    // methods
    public:
	    
    /**
    * Create a gapped sequence object with the prescribed monomer type
    * (amino acid codes by default). Consult "Monomer.hh"
    * for available monomer types.
    */
    explicit Gapseq_(Monomer_::Type_ Mt=Monomer_::PROT):
	    Bioseq_(Mt), Gaplist(), Gaplen(0) {}

    /**
    * Create a gapped sequence object with the prescribed monomer alphabet
    * given in Codes. Consult "Monomer.hh" for code string syntax.
    */
    explicit Gapseq_(const char* Codes):
	    Bioseq_(Codes), Gaplist(), Gaplen(0) {}

    /// copy ctor & conversion from base class
    Gapseq_(const Bioseq_& B):
	    Bioseq_(B), Gaplist(), Gaplen(0) {}
    
    /// copy ctor
    Gapseq_(const Gapseq_& G):
	    Bioseq_(G), Gaplist(G.Gaplist), Gaplen(G.Gaplen) {}
    
    /**
    * The unsafe [] operators and the at_(1|3) family
    * all use the "net" indexing, ie. gaps are ignored.
    * This means that you can get or set a monomer, but
    * cannot convert a monomer to a gap or vice versa.
    * See the "Gap manipulations" section.
    */
    
    /**
    * at_1(Idx), at_1(Idx, C):
    * safe index operator for the Idx:th char in the sequence.
    * at_1(NPOS) returns a GAP character.
    * The non-const version refuses to store characters not corresponding
    * to the current type of the calling object.
    * Throws Indexrangexc_ if Idx is NPOS or illegal.
    * Returns true on success, false when C was illegal.
    */
    virtual
    char at_1(size_t Idx) const;
    virtual
    bool at_1(size_t Idx, char C);
    
    /**
    * at_3(Idx), at_3(Idx, Ccc):
    * safe index operator for getting and setting
    * a monomer using the 3-letter abbreviations.
    * at_3(NPOS) returns a GAP3 string.
    * The non-const version refuses to store abbreviations
    * that are illegal.
    * Returns true on success, false when Ccc was illegal.
    */
    virtual
    const char* at_3(size_t Idx) const;
    virtual
    bool at_3(size_t Idx, const char* Ccc);
   
    /**
    * seq(): inherited from Bioseq_, returns a const ref
    * to the ungapped sequence.
    * seq(S): stores the sequence S in the calling object.
    * Refuses storage if S does not correspond to calling obj type.
    * Note that changing the underlying sequence would invalidate
    * the gap records, therefore all gaps are removed by this method.
    */
    const string& seq() const { return Bioseq_::seq(); }
    virtual
    void seq(const string& S)
    {
	if (!is_legal(S))
	    cerr<<"\n? Gapseq_::seq(...): Illegal char(s) in sequence\n";
	else { sequence()=S; remove_gaps(); }
    }
    
    /**
    * gapped_sequence([Gcg]):
    * create and return a string with the gap characters actually inserted.
    * If Gcg==true (default false), then the funny GCG gap characters
    * are used, ie. '~' for trailing gaps and '.' for internal gaps.
    * gapped_sequence(Gseq,[Mtype]):
    * set the calling object to contain the gapped sequence Gseq where
    * the gap positions are actually indicated by the gap characters.
    * Here all gap characters are accepted, no matter where they are.
    * The new monomer type can be forced by specifying Mt, if it is
    * Monomer_::UNKNOWN (the default) then autodetection will be done.
    */
    string gapped_sequence(bool Gcg=false) const;
    void gapped_sequence(const string& Gseq, int Mtype=Monomer_::UNKNOWN);
    
    	/// Legal issues
    
    /**
    * is_legal(C): returns true if C is a legal character
    * in the calling object under the current monomer alphabet
    * and case-sensitivity setting.
    * is_legal(Ccc): does the same for 3-letter abbreviations.
    * is_legal(S): does the same for the complete sequence string S.
    */
    virtual
    bool is_legal(char C) const
    {
    	return (C==GAP || monomer().is_legal(C, nocase()));
    }
    virtual
    bool is_legal(const char* Ccc) const
    {
	return (!strncmp(Ccc, GAP3, 3) || monomer().is_legal(Ccc, nocase()));
    }
    bool is_legal(const string& S) const { return Bioseq_::is_legal(S); }
    
    	/// Gap manipulation
    
    /**
    * insert_gaps(Idx, N):
    * insert N gap characters before position [Idx].
    * If Idx==NPOS, the gap will be placed after the sequence.
    * If Idx is out-of-range, then Indexrangexc_ will be thrown.
    * If there is already a gap before Idx, then the
    * length of this existing gap will be changed to N.
    * N must be non-0, if N==0 then nothing is done.
    */
    void insert_gaps(size_t Idx, size_t N);

    /**
    * remove_gaps([Idx]):
    * removes the gap characters before net position [Idx].
    * Nothing is done if there were no gaps there.
    * If Idx is out of range, Indexrangexc_ will be thrown.
    * If Idx is omitted, then all gaps are removed.
    */
    void remove_gaps(size_t Idx);
    void remove_gaps()
    {
	Gaplen=0;
	Gaplist.erase(Gaplist.begin(), Gaplist.end());
    }
    
    /**
    * Net/gross position conversion:
    * The "net" position of a monomer in the gapped sequence
    * is the position it would have if there were no gaps.
    * The "gross" position is its position with the gaps included.
    * Both positions start with 0 in the best C tradition :-)
    * For example in the gapped sequence "--AC-F-D" 
    * the net position of 'F' is 2, the gross position is 5.
    */

    /**
    * net_pos(Gross): given the "gross" position Gross, the ungapped
    * net position is returned. 
    * If there is a gap at Gross then NPOS is returned.
    * If Gross is illegal (>=gross_length()), then Indexrangexc_ is thrown.
    */
    size_t net_pos(size_t Gross) const;
    
    /**
    * gross_pos(Net): given the "net" position Net, the gapped
    * gross position is returned (inverse of net_pos(Gross) above).
    * If Net is illegal (>=net_length()), then Indexrangexc_ is thrown.
    */
    size_t gross_pos(size_t Net) const;
    
    /// net_length(): the length of the sequence without the gaps.
    size_t net_length() const { return seq().length(); }
    
    /// gross_length(): the total length of the sequence with gaps.
    size_t gross_length() const { return seq().length()+Gaplen; }

    	// I/O
    
    /**
    * read_seq(Inf, Type, Format):
    * tries to read a sequence from the input stream Inf.
    * If Type is set to Monomer_::PROT, then only protein
    * sequences will be read, if it is set to Monomer_::[DEOXY[RIBO]]NUCL,
    * then only the corresponding nucleic acid sequences will be read,
    * if Type is Monomer_::USER (default) or an OR-ed combination of PROT
    * and *NUCL, then all types are considered and a polymer
    * type autodetection is attempted. This is not foolproof.
    * Format can be set to ANYFORMAT (default), in this case automatic
    * file format detection is performed, or to 
    * FASTA (protein or nucleic acid) or to PIR (proteins only).
    * The other formats provided by the Bioseq_ class are not accepted
    * because those cannot be used in multiple alignments.
    * No read is performed and FORMATERR is returned when
    * incompatible type and format are specified.
    * Return value: OK or an error enum if something went wrong.
    */
    virtual
    Errtype_ read_seq(istream& Inf, int Type=Monomer_::USER, int Format=ANYFORMAT);
    
    /**
    * write_seq(Outf, Format): writes the contents of the calling object
    * to Outf according to the format specification in Format.
    * Only FASTA and PIR formats are accepted, see comments to read_seq(..).
    * Returns OK or an error enum if something went wrong.
    */
    virtual
    Errtype_ write_seq(ostream& Outf, int Format) const;
    
    // hidden methods
    protected:
	    
    virtual
    bool is_legal(int Mt, char C) const;
    
};
// END OF CLASS Gapseq_

}   // RazorBack

// ==== END OF HEADER Gapseq.hh ====

#endif	    // GAPSEQ_HEADER
