#ifndef IOVAR_TMPL_HEADER
#define IOVAR_TMPL_HEADER

// ==== TEMPLATE HEADER Iovar.hh ====

/**
 * \file Iovar.hh
 * Template class for representing variables
 * for nicely formatted ASCII I/O.
 */

// 28-Sep-2001, Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stdlib.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#else
#include <iostream.h>
#endif

#include <string>
using namespace std;
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== MODULE HEADERS ====

#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

// silly predefs for friend templates
template <class T_> class Iovar_;
template <class T_>
istream& operator>>(istream& In, Iovar_<T_>& Iov);
template <class T_>
ostream& operator<<(ostream& Out, const Iovar_<T_>& Iov);

/**
 * Class Iovar_: a template class that represents
 * a variable in an ASCII I/O format. Useful for
 * reading and writing program parameters to/from data files.
 * The (regexp-like) format is:-
 * \verbatim
   (descr_token)(ws)*(ws|(eqchar)(ws)*)(value)
   
   where
   (descr_token) is a string w/o whitespaces, this is the "name"
    	of the variable;
   (ws)* denotes zero or more whitespace chars;
   (eqchar) denotes "equal-to" chars like '=' or ':';
   (value) is the value of the variable.
   \endverbatim
 * Examples:-
 * "Intvar 1", "Intvar:2", "Dblvar = 3.5", "Bool= 1"
 * 
 * NOTE: this class is intended for POD types
 * like int, float, bool etc. only!
 */
template <class T_>
class Iovar_
{
    // data
    private:
	    
    string Descr;    //< description string
    bool Nocase;    //< true if description is case-insensitive
    char Eqchar;    //< the "equals" char, '\0' when not prescribed
    T_ Value;	    //< the value of the variable
    
    // methods
    public:
	    
    /**
    * Inits to store a variable with description token Dtok.
    * Specification of Dtok is mandatory, the ctor throws
    * a "fatal" Emptyexc_ if Dtok is NULL or "".
    * Defval is the default value of the variable, T_(0) by default.
    * Noc=true if the description token is meant to be case-insensitive,
    * this is false by default (exact match required).
    * Eq (by default '=') is the "equal" character, if you want
    * to omit it, then specify '\0'.
    */
    explicit Iovar_(const char *Dtok, T_ Defval=T_(0), bool Noc=false, char Eq='=');
    
    /// descr(): returns the description string.
    const string& descr() const { return Descr; }
    
    /// dlen(): returns the length of the description string.
    size_t dlen() const { return Descr.size(); }
    
    /**
    * match_descr(Str, Nocase): returns true if Str begins with descr().
    * If nocase()==true, then a case-insensitive comparison is done,
    * otherwise (the default) an exact match is required.
    */
    bool match_descr(const string& Str) const;
    
    /**
    * nocase(): returns true if the description token is
    * case-insensitive.
    */
    bool nocase() const { return Nocase; }
    
    /// eqchar(): returns the "equals" char.
    char eqchar() const { return Eqchar; }
    
    /// Returns the value.
    const T_& value() const { return Value; }
    
    /// Sets the value to V, returns old value.
    T_ value(const T_& V) { T_ Ov=Value; Value=V; return Ov; }
    
    // Input: reads an Iov object from In accepting the formats
    // "Descr Value" or "Descr=Value" (where '=' is eqchar() and
    // it can be surrounded by >=0 whitespace chars on either side).
    // On error, no warnings are printed but In's failbit is set
    // and the stream is rewound to its original position.
    friend
    istream& operator>> <>(istream& In, Iovar_<T_>& Iov);

    // Output: uses the format "Descr Value" or "Descr = Value"
    // when Eqchar (here '=') is specified. No newline at the end.
    friend
    ostream& operator<< <>(ostream& Out, const Iovar_<T_>& Iov);
    
    // hidden methods
    protected:
	    
    istream& read_value(istream& In);
    ostream& write_value(ostream& Out) const;
};
// END OF CLASS Iovar_

#ifndef RAZORBACK_LIB_COMPILE
// I/O for bool: the case-insensitive tokens "true","on","yes" are accepted
// in addition to "1", and "false","off","no" in addition to 0.
// The output strings are always lowercase "true" and "false".
// The matches are case-independent.

template<> istream& Iovar_<bool>::read_value(istream& In);
template<> ostream& Iovar_<bool>::write_value(ostream& Out) const;
#endif

} // RazorBack

#ifdef INCLUDE_TMPL_DEFS
#include "Iovar.cc"
#endif

// ==== END OF TEMPLATE HEADER Iovar.hh ====

#endif	    // IOVAR_TMPL_HEADER
