#ifndef LINLS_CLASS_HEADER
#define LINLS_CLASS_HEADER

// ==== HEADER Linls.hh ====

/**
 * \file Linls.hh
 * Linear least squares parameter estimation.
 * Based on SVD as explained in Golub/Van Loan (Chapter 5).
 */

// 9-Aug-2000. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <float.h>
#include <math.h>

// ==== MODULE HEADERS ====

#include "Vector.hh"
#include "Rectmat.hh"
#include "Symmat.hh"
#include "Svd.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Linls_: Class for carrying out linear least-squares using the SVD.
 * Note that unlike Linreg_, this class is not derived from Paramest_.
 */
class Linls_
{
    // data
    private:
    
    size_t Nm, Np, Nf; ///< no. of measurements, no.of params, degrees of freedom
    Svd_ Svd;	    	///< an SVD object
    size_t Rank;    	///< the rank estimate
    Vecdbl_ Utb;    	///< scalprods of U cols w/ RHS
    Vecdbl_ P, Sdev;	///< parameter values and standard deviations
    Smatdbl_ Correl; 	///< parameter correlation matrix
    double Chi2;    	///< chi-square
    
    // methods
    public:
	    
    /// Init to empty.
    explicit Linls_(size_t NM=0, size_t NP=0):
	    Nm(NM), Np(NP), Svd(NM, NP), Rank(NP), Utb(NP),
	    P(NP), Sdev(NP), Correl(NP) {}
    
    /**
     * lin_reg(Xmeas, Ymeas, Eps):
     * Carries out a linear least squares fit using SVD.
     * Xmeas is the data matrix (Nm rows of observations for Np indep. vars,
     * Nm>Np), Ymeas is the Nm observations of the dependent variable.
     * Eps is a small tolerance value used for finding the best SVD rank.
     * Should be a "small multiple of the machine zero". If in doubt,
     * set it to 0.0 and it will be figured out for you.
     * Return value: the chi-square statistics of the fit,
     * or a negative value on error.
     */
    double lin_reg(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas, double Eps=1e-10);
    
    /// Returns the rank.
    size_t rank() const { return Rank; }
    
    /// Returns the condition number.
    double condition() const { return (!rank())? HUGE_VAL: Svd.w()[0]/rank(); }
    
    /// Returns the degrees of freedom.
    size_t deg_freedom() const { return Nf; }
    
    /// Returns the parameter estimate.
    const Vecdbl_& param() const { return P; }
    
    /// Returns the standard deviation of the parameters.
    const Vecdbl_& sdev() const { return Sdev; }
    
    /// Returns the correlation matrix of the parameters.
    const Smatdbl_& correl() const { return Correl; }
    
    /// Returns the chi-square statistic for the fit quality.
    double chi_square() const { return Chi2; }
    
    /// Returns the residual deviation for the fit.
    double residual_dev() const { return sqrt(Chi2/Nf); }
    
    // hidden methods
    protected:
    
    void adjust_dimensions(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas);
    void make_utb(const Vecdbl_& B);
    size_t calc_rank(const Vecdbl_& B, double Eps) const;
    void calc_params(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas);
    double calc_chi2(const Rmatdbl_& Xmeas, const Vecdbl_& Ymeas) const;

};
// END OF CLASS Linls_

} // RazorBack

// ==== END OF HEADER Linls.hh ====

#endif	    // LINLS_CLASS_HEADER
