#ifndef MVN_HEADER
#define MVN_HEADER

// ==== HEADER Mvn.hh ====

/**
 * \file Mvn.hh
 * Class representing multivariate Normal distributions.
 */

// 6-Jul-1999. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <math.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ==== MODULE HEADERS ====

#include "Vector.hh"
#include "Rectmat.hh"
#include "Symmat.hh"
#include "Rangedimexc.hh"
#include "Bits.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Mvn_: a class representing multivariate Normal distributions.
 * Can be asked to calculate the MVN density and to generate
 * MVN-distributed random vectors. I/O is provided.
 */
class Mvn_
{
    // static data
    private:
	    
    static const double TWO_PI;
    
    // data
    private:
	    
    size_t Dim, Covrank;     ///< dimension of random vector, covar mtx rank
    Vecdbl_ Avgvec; ///< average
    Smatdbl_ Covarmat;	///< covariance matrix
    Vecdbl_ Eval;   ///< eigenvalues of Covarmat
    Rmatdbl_ Evec;  ///< eigenvectors of Covarmat
    Rmatdbl_ Stdnmap;    ///< lin. comb. coeffs that map N(0,1)-s onto MVN
    double Normfactor;	///< the density normalisation factor
    
    // methods
    public:
	 
    /**
    * Inits to store a D-dimensional MVN centered on the origin,
    * unit covariance matrix (standard MVN).
    */
    explicit Mvn_(size_t D=2);
    
    /**
    * Inits to store a MVN distribution with average A and
    * covariance matrix C. These should have the same dimension of course.
    */
    Mvn_(const Vecdbl_& Avg, const Smatdbl_& Cov);
    
    /// dim(): returns the actual dimension.
    size_t dim() const { return Dim; }
    
    /**
    * dim(D): adjusts the calling object to represent a D-dimensional
    * MVN distribution. If D<2 then it will be set to 2 silently.
    * Does nothing if D is the same as the previous dimension.
    * Do not forget to call avg(Newavg) and covar(Newcov) afterwards!
    * Return value: the old dimension.
    */
    size_t dim(size_t D);
    
    /**
    * covrank(): returns the rank of the covariance matrix.
    * It is always true that covrank()<=dim().
    */
    size_t covrank() const { return Covrank; }
    
    /// avg(): returns a const ref. to the average of the distribution.
    const Vecdbl_& avg() const { return Avgvec; }
    
    /// avg(Newavg): sets the average of the distribution to Newavg.
    void avg(const Vecdbl_& Newavg)
    {
	if (dim()!=Newavg.dim())
	    throw Dimexc_("Mvn_::avg(Newavg)", dim(), Newavg.dim());
	Avgvec=Newavg;
    }
    
    /// covar(): returns a const ref. to the covariance of the distribution.
    const Smatdbl_& covar() const { return Covarmat; }
    
    /**
    * covar(Newcov): sets the covariance of the distribution to Newcov
    * if Newcov has the same dimension as the calling object and
    * if it has at least rank 2. Otherwise nothing is changed.
    * Return value: the rank if OK or 0 on error.
    */
    int covar(const Smatdbl_& Newcov);
    
    /**
    * pdf(X): calculates the probability density function
    * of the calling object at point X.
    */
    double pdf(const Vecdbl_& X) const
    {
	return (Normfactor*exp(-maha_dist(X)/2.0));
    }
    
    /**
    * random(X): generates a dim()-dimensional random vector
    * from the MVN distribution stored in the calling object.
    * The dimension of X is adjusted silently, previous contents
    * are overwritten.
    */
    void random(Vecdbl_& X) const;
    
    /**
    * maha_dist(X):
    * calculates and returns the Mahalanobis distance
    * between X and the average vector in the calling object.
    */
    double maha_dist(const Vecdbl_& X) const;
    
    /**
    * marginal(Bitmask, Marginal):
    * changes Marginal so that it contains a marginal distribution
    * of the calling object. The variables corresponding to the
    * bits set in Bitmask will be kept, the others will be
    * "integrated out". E.g. if the calling object is a 4-dim MVN
    * with PDF(x1,x2,x3,x4), and the bitmask is "1010", then
    * Marginal will be a 2-dim MVN with PDF(x2,x4).
    * The bitmask should contain exactly as many bits as dim(),
    * otherwise a Dimexc_ exception will be thrown.
    * At least two bits must be set to ON so that Marginal remains
    * a valid Mvn_ object; otherwise a warning is printed
    * and no action will be taken.
    * To get a 1-D marginal distribution,
    * simply obtain the corresponding average coordinate and
    * variance and use a Gaussian_ object.
    * Return value: the dimension of Marginal.
    */
    size_t marginal(const Bits_& Bitmask, Mvn_& Marginal) const;
    
    	// I/O
    
    /**
    * Input from stream In.
    * The ASCII format of an Mvn_ object is:-
    * \verbatim
      MVN
      AVERAGE
      <dbl vector>     (center of distribution)
      COVARIANCE
      <dbl lower triangular matrix>    (covariance matrix)
      ENDMVN
      \endverbatim
    * For the formats of the vectors and matrices, see
    * the corresponding linalg header files.
    * Empty lines are allowed, lines starting with '#' are comments.
    */
    friend istream& operator>>(istream& In, Mvn_& Mvn);
    
    /// Output to stream Out.
    friend ostream& operator<<(ostream& Out, const Mvn_& Mvn);
    
};
// END OF CLASS Mvn_

} // RazorBack

// ==== END OF HEADER Mvn.hh ====

#endif	    // MVN_HEADER
