#ifndef NONLINREG_HEADER
#define NONLINREG_HEADER

// ==== HEADER Nonlinreg.hh ====

/**
 * \file Nonlinreg.hh
 * Nonlinear multiple least-squares regression.
 */

// 19-Feb-2001, Andras Aszodi

// ---- MODULE HEADERS ----

#include "Paramest.hh"
#include "Rangedimexc.hh"
#include "Statexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Nonlinfunc_: this is an abstract base class
 * for function objects that can be used by the Nonlinreg_ class.
 * The nonlinear function is specified by overloading the
 * pure virtual f() method.
 * A virtual jac() method is provided for approximating the Jacobian.
 * Analytical Jacobians may be implemented by overloading jac()
 * in the derived class.
 * Rudimentary error reporting is provided.
 */
class Nonlinfunc_
{
    // data
    private:
	    
    size_t Nx, Np, Ny;	///< vector sizes
    
    // methods
    public:
	    
    /**
     * Init to calculate a nonlinear regression for
     * NX independent variables, NP parameters and
     * NY dependent variables.
     */
    Nonlinfunc_(size_t NX, size_t NP, size_t NY):
	    Nx(NX), Np(NP), Ny(NY) {}
    
    /// Returns the number of the independent variables.
    size_t nx() const { return Nx; }
    
    /// Returns the number of the parameters.
    size_t np() const { return Np; }
    
    /// Returns the number of the dependent variables.
    size_t ny() const { return Ny; }
    
    /**
    * f(X, P, Y): calculates f(X) with parameter vector P
    * and puts the result into Y.
    * Return value: 0 if OK, non-0 on error.
    * Not const because derived classes may modify internal data.
    */
    virtual int f(const Vecdbl_& X, const Vecdbl_& P, Vecdbl_& Y) =0;
    
    /**
    * jac(X, P, Y, J): calculates the Jacobian matrix J=dY/dP
    * by a simple approximation. Y==f(X, P). The Jacobian is
    * calculated subject to P normalisation, i.e.
    * J[i][k]=dY[i]/dP[k]*p[k].
    * Return value: 0 if OK, non-0 on error.
    * Not const because derived classes may modify internal data.
    */
    int jac(const Vecdbl_& X, const Vecdbl_& P, Vecdbl_& Y, Rmatdbl_& J);

};
// END OF CLASS Nonlinfunc_

/**
 * Class Nonlinreg_: this class implements a general-purpose
 * multivariate weighted nonlinear regression with Marquardt's method.
 */
class Nonlinreg_: public Paramest_
{
    // symbolic consts
    public:
	    
    /// Verbosity_: controls std output during the iteration.
    typedef enum {SILENT, TERSE, VERBOSE} Verbosity_;
    
    // data
    protected:
	    
    size_t Nx, Ny;  ///< X and Y vector sizes
    Rmatdbl_ W;	    ///< weight matrix Nm x Ny
    Verbosity_ Verb;	///< progress report verbosity
    
    // methods
    public:
	    
    /**
    * Init to desired sizes if known in advance.
    * Actual regression calls adjust sizes automatically.
    */
    Nonlinreg_(size_t NX, size_t NY, size_t NP, size_t NM):
	    Paramest_(NM, NP), Nx(NX), Ny(NY), Verb(SILENT) {}
    
    /// report_level(V): sets the verbosity level of the progress report.
    void report_level(Verbosity_ V) { Verb=V; }
    
    /**
    * nonlin_reg(Xmeas, Ymeas [,Userwgt], Initpar, Funcobj, Itmax, Steplim):
    * Least-squares multiple nonlinear regression for the function
    * Y=F(X, P) where Y is a Ny-dim. vector, X is a Nx-dim vector,
    * and P is a Np-dim. vector of parameters, F a nonlinear
    * vector-vector function.
    * Routine parameters:-
    * Xmeas: Nm x Nx matrix of independent variables
    * Ymeas: Nm x Ny matrix of dependent variables
    * Userwgt: optional Nm x Ny matrix of weights
    *  	can be omitted if equal or relative weights are used
    * Initpar: Np-dim. vector of the initial parameter estimates
    * Funcobj: a nonlinear function object (instance of class Nonlinfunc_)
    *  	that can calculate F and the Jacobian dF/dP
    * Itmax: on input, the maximal number of iterations,
    *  	on output, the actual number of iterations
    * Steplim: minimal stepsize (default 1e-6).
    * Return value: the residual deviation.
    */
    double nonlin_reg(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas, const Rmatdbl_& Userwgt,
	const Vecdbl_& Initpar, Nonlinfunc_& Funcobj, int& Itmax, double Steplim=1e-6);
    double nonlin_reg(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas,
	const Vecdbl_& Initpar, Nonlinfunc_& Funcobj, int& Itmax, double Steplim=1e-6);
    
    // hidden methods
    protected:
	    
    void adjust_dimensions(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas,
	const Rmatdbl_& Userwgt, const Vecdbl_& Initpar, const Nonlinfunc_& Funcobj);
    void setup_weight(const Rmatdbl_& YW);
    double quality(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas,
	const Vecdbl_& Par, Nonlinfunc_& Funcobj) const;
    void report(ostream& Out, int Itno, double Q, double Lambda,
	    const Vecdbl_& Par) const;
    
};
// END OF CLASS Nonlinreg_

} // RazorBack

// ==== END OF HEADER Nonlinreg.hh ====

#endif	    // NONLINREG_HEADER
