#ifndef ORTHPOLY_HEADER
#define ORTHPOLY_HEADER

// ==== HEADER Orthpoly.hh ====

/**
 * \file Orthpoly.hh
 * Parameter estimation with orthogonal polynomials.
 */

// 19-Sep-2001, Andras Aszodi

// ---- MODULE HEADERS ----

#include "Vector.hh" 	// vector class
#include "Rectmat.hh"	// general MxN matrix class
#include "Symmat.hh" 	// symmetric matrix class
#include "Rangedimexc.hh"
#include "Statexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * This class implements parameter estimation using orthogonal polynomials.
 * The polynomials (up to a desired degree) will be created on-the-fly.
 * You can experiment with various degrees and once you are satisfied
 * you can request a conversion of the orthogonal coefficients into
 * simple polynomial coefficients.
 */
class Orthpoly_
{
    // symbolic consts
    public:
	    
    enum Weighting_ { EQUAL, RELATIVE, POISSON };
    
    // data
    private:
	    
    Vecdbl_ Xmeas, W, Ymeas;   // indep. data points, weights, dep. vars
    size_t Nm, Nf;	    // no. of data points, degrees of freedom
    Smatdbl_ Poly;   // [i][j]: j-th coeff of i-th orthogonal polynomial
    Rmatdbl_ Polyval;	// [i][j]: value of [i]-th orthpoly at Xmeas[j]
    Vecdbl_ Orthcoeff, Varorth, // fitted orthogonal polynomial coeffs & vars
	    Param, Sdev, // "usual" polynomial coeffs & variances
	    Dres, Norm;   	// weighted residuals, squared polynomial norms
    Smatdbl_ Correl;	// Polycoeff's correlation matrix
    size_t Degree, Maxdegree;  	// degree of approximation (maximal so far)
    double Sres;    // residual deviation
    bool Makepolycoeff;     // true if Polycoeff is to be remade
    
    // methods
    public:
	    
    /// Init to empty.
    Orthpoly_():
	    Nm(0), Nf(0), Poly(1), Maxdegree(1), Sres(0.0), Makepolycoeff(true)
	    { }
    
    /**
     * data(Xm, Ym, (Wgt|Wtype)): resets the calling object and
     * stores the data points with the weights in it.
     * If the weighting type is specified instead as the last parameter,
     * then the weight vector will be calculated appropriately from Ym.
     * Return value: the number of data points.
     * Throws Dimexc_ on dim mismatch.
     */
    int data(const Vecdbl_& Xm, const Vecdbl_& Ym,
	const Vecdbl_& Wgt);
    int data(const Vecdbl_& Xm, const Vecdbl_& Ym,
	Weighting_ Wtype=EQUAL);
    
    /**
     * poly_fit(Deg): fits orthogonal polynomials up to degree Deg
     * to the data previously supplied via the data(...) method.
     * Returns the "goodness-of-fit" parameter.
     * Throws Toofewexc_ if there are not enough data points.
     */
    double poly_fit(unsigned int Deg);

    /**
     * make_polycoeff(): once the user is satisfied with the orthogonal fit,
     * then this method should be called to calculate the parameters and
     * standard deviations for the non-orthogonal estimation
     * Y = p[0]+p[1]*x+...+p[Degree]*x^Degree.
     */
    void make_polycoeff();
    
    /**
     * good_fit(): returns a "goodness-of-fit" value that indicates
     * what percentage of Y is "explained" by the fit.
     */
    double good_fit() const
    {
	return (100.0*(1.0-Sres*(Nm-1)/Dres[0]));
    }
    
    /**
     * f_test(): returns the F-statistic variable that can be
     * tested later for significance.
     */
    double f_test() const
    {
	if (!Degree) return 0.0;
	else if (Dres[Degree]<1e-7) return 1e30;
	else return (Dres[Degree-1]-Dres[Degree])/Dres[Degree]*Nf;
    }
    
    /**
     * param():
     * Returns the non-orthogonal parameter coefficients.
     * Invoke make_polycoeff() before calling this method.
     */
    const Vecdbl_& param() const { return Param; }
    
    /**
     * sdev():
     * Returns the standard deviations of the 
     * non-orthogonal parameter coefficients.
     * Invoke make_polycoeff() before calling this method.
     */
    const Vecdbl_& sdev() const { return Sdev; }
    
    /**
     * correl():
     * Returns the correlation (normed covariance) matrix of the
     * non-orthogonal parameter coefficients.
     * Invoke make_polycoeff() before calling this method.
     */
    const Smatdbl_& correl() const { return Correl; }
    
    /// approx(X): approximate Y(X) with the current polynomial.
    double approx(double X) const;

    // hidden methods
    protected:
    
    void setup_weight(const Vecdbl_& Ymeas, Weighting_ Weighting);
    void make_orthcoeff(unsigned int Deg);
    void back_orthcoeff(unsigned int Deg);
    
    static
    int check_dim(const Vecdbl_& Xm, const Vecdbl_& Ym,
	const Vecdbl_& Wgt);
    void reset();
    void resize_orth(unsigned int Maxdeg);
    void resize_param(unsigned int Maxdeg);
    void make_orthpoly(size_t Deg);
};
// END OF CLASS Orthpoly_

} // RazorBack

// ==== END OF HEADER Orthpoly.hh ====

#endif		// ORTHPOLY_HEADER
