#ifndef PARAMEST_HEADER
#define PARAMEST_HEADER

// ==== HEADER Paramest.hh ====

/**
 * \file Paramest.hh
 * Base class for linear and nonlinear parameter estimation.
 */

// 5-Feb-1999, Andras Aszodi

// ---- STANDARD HEADERS ----

#include <stdlib.h>
#include <float.h>
#include <math.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ---- MODULE HEADERS ----

#include "Vector.hh"
#include "Rectmat.hh"
#include "Symmat.hh"

// ==== CLASSES ====

namespace RazorBack {

/// Class Paramest_: base class for parameter estimation.
class Paramest_
{
    // symbolic consts
    public:
    
    /// Weighting_: equal, relative, user-defined
    typedef enum {EQUAL, RELATIVE, USER} Weighting_;
    
    // data
    protected:
	    
    size_t Nm, Np, Nf; ///< no. of measurements, no.of params, degrees of freedom
    Smatdbl_ Xtx;    ///< Fisher's observation matrix
    Vecdbl_ Xty;    ///< rhs of linreg eqn.
    Vecdbl_ Norm;   ///< normalisation vector
    Weighting_ Weighting;   ///< weight type
    Vecdbl_ P, Sdev;	///< parameter values and standard deviations
    Smatdbl_ Correl; 	///< parameter correlation matrix
    
    // methods
    public:
	    
    /// Init to empty
    Paramest_(): 
	    Nm(0), Np(0), Nf(0), Weighting(EQUAL) {}
    
    /**
    * Init to desired sizes if known in advance
    * Actual regression calls may change the settings
    */
    Paramest_(size_t NM, size_t NP):
	    Nm(NM), Np(NP), Weighting(EQUAL)
	    { resize(); }
    
    /// set the weighting scheme
    void weight(Weighting_ W) { Weighting=W; }
    
    /// Returns the estimated parameters.
    const Vecdbl_& param() const { return P; }
    
    /// Returns the estimated standard deviations of the parameters.
    const Vecdbl_& sdev() const { return Sdev; }
    
    /// Returns the estimated correlation matrix between the parameters.
    const Smatdbl_& correl() const { return Correl; }
    
    /**
    * tcrit_95(DOF) returns the value of the t-distribution
    * at 95% significance level with Nf degrees of freedom.
    * Use for confidence intervals.
    */
    float tcrit_95() const;
    
    // hidden methods
    protected:
	    
    void resize()
    {
	Xtx.size(Np); Xty.dim(Np); Norm.dim(Np);
	P.dim(Np); Sdev.dim(Np); Correl.size(Np);
    }
    double normalise();
    double solve(Vecdbl_& Sol, double Lambda0, double Lambdafact);
    static bool posdef_inv(Smatdbl_& A);
    void sdev_correl(double Q);
};
// END OF CLASS Paramest_

} // RazorBack

// ==== END OF HEADER Paramest.hh ====

#endif	    // PARAMEST_HEADER
