#ifndef RANDOM_CLASS
#define RANDOM_CLASS

// ==== HEADER Random.hh ====

/**
 * \file
 * Uniformly distributed random numbers using the 32-bit generator of
 * L'Ecuyer, P. Efficient and portable combined random number
 * generators, C.A.C.M., vol. 31, 742-749, June 1988.
 * Period is very long, about 2.3e18.
 * Plus the Box/Mueller method for standard normal variates.
 * \author Andras Aszodi
 * \date 21-Nov-2001.
 */

// ==== CLASSES ====

namespace RazorBack {

/// Randombase_: random number generator abstract base class.
class Randombase_
{
    private:
    
    // data
    mutable int Seed1, Seed2;
    
    // methods
    public:
	    
    /**
     * Init with the seeds S1, S2.
     * If S1==0 (default), then getpid() is used as seed.
     * If S2==0 (default), then time(NULL) is used as seed.
     */
    explicit Randombase_(int S1=0, int S2=0)
    {
	reset(S1, S2);
    }

    /// destructor
    virtual ~Randombase_() {}
    
    /**
     * reset(S1, S2): resets the generator with S1, S2.
     * If S1==0 (default), then getpid() is used as seed.
     * If S2==0 (default), then time(NULL) is used as seed.
     */
    void reset(int S1=0, int S2=0);

    /**
    * operator(): this call should return a random double 
    * in the derived classes.
    */
    virtual double operator()(void) const =0;
    
    // hidden methods
    protected:
    
    int int_rand(void) const;
    double dbl_rand(void) const;
};
// END OF CLASS Randombase_

/// Randomuni_: generates uniform random numbers.
class Randomuni_: public Randombase_
{
    // methods
    public:
    
    /**
     * Init with the seeds S1, S2.
     * If S1==0 (default), then getpid() is used as seed.
     * If S2==0 (default), then time(NULL) is used as seed.
     */
    explicit Randomuni_(int S1=0, int S2=0):
	    Randombase_(S1, S2) {}
    
    /// operator(): returns a random double value between 0.0 and 1.0.
    double operator()(void) const { return dbl_rand(); }
    
    /// rnd_lim(L, U): returns a random double between L and U. No checks.
    double rnd_lim(double L, double U) const
    {
	return((L-U)*dbl_rand()+U);
    }
};
// END OF CLASS Randomuni_

/**
 * Randomnorm_: random number generator class
 * that produces normally distributed random numbers.
 */
class Randomnorm_: public Randombase_
{
    // methods
    public:
	    
    /**
     * Init with the seeds S1, S2.
     * If S1==0 (default), then getpid() is used as seed.
     * If S2==0 (default), then time(NULL) is used as seed.
     */
    explicit Randomnorm_(int S1=0, int S2=0): 
	    Randombase_(S1, S2) {}

    /**
    * operator(): returns a random double value
    * with zero mean and unit variance (std normal distribution).
    */
    double operator()(void) const;

    /**
    * rnd_msd(M, S): returns a random value from a normal
    * distribution with mean M and standard deviation S. No checks.
    */
    double rnd_msd(double M, double S) const
    {
	return( S*operator()()+M );
    }
};
// END OF CLASS Randomnorm_

} // RazorBack

// ==== END OF HEADER Random.hh ====

#endif	    // RANDOM_CLASS
