#ifndef RECTMAT_TMPL_HEADER
#define RECTMAT_TMPL_HEADER

// ==== HEADER Rectmat.hh ====

/**
 * \file rectmat.hh
 * General NxM rectangular matrix template.
 */

// 8-Dec-2000. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// Exception support is assumed.
// This module depends on the Standard Template Library.
 
// ---- COMPILER COMMAND LINE MACROS ----

// Define the macro INCLUDE_TMPL_DEFS if your compiler
// does not support automatic template source inclusion.

// ---- MODULE HEADERS ----

#include "Vector.hh"
#include "Basemat.hh"
#include "Squarexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Rectmat_: a class of rectangular matrices (Row x Col).
 * Use for general matrices and for non-symmetric square matrices.
 * Derived from Basemat_.
 */
template <class T_>
class Rectmat_: public Basemat_<T_>
{
    // data
    private:
    
    // Layout information:-
    // The matrix elements are stored in row major order.
    //
    // 00 01 02
    // 10 11 12 --> 00 01 02 10 11 12 20 21 22
    // 20 21 22
    //
    size_t R, C;	// number of rows and columns
    
    // methods
    public:
    
    /// default ctor: 3x3 is the default size
    explicit Rectmat_(size_t Row=3, size_t Col=3)
    {
	if (!Row) Row=3; if (!Col) Col=3;
	R=Row; C=Col;
	Matstorage_<T_> *Newmat=new Matstorage_<T_>(Row*Col);
	replace_storage(Newmat);	// alloc storage
    }
    
    /// copy ctor
    Rectmat_(const Rectmat_<T_>& Rm):
	    Basemat_<T_>(Rm), R(Rm.rno()), C(Rm.cno()) {}
    
    /**
    * init Rectmat_ with a conventional array. 
    * It is supposed to have the row-major layout. No checks
    */
    Rectmat_(const T_ *Arr, size_t Row, size_t Col):
	    R(Row), C(Col)
    {
	Matstorage_<T_> *Newmat=new Matstorage_<T_>(R*C);
	replace_storage(Newmat);
	for (register unsigned int i=0; i<R*C; i++)
	    elems(i)=Arr[i];	// direct copy assuming same layout
    }
    
    /// destructor
    virtual ~Rectmat_() {}
    
    	/// Access
    
    /// rno(), cno(): return the number of rows and columns.
    virtual size_t rno() const { return R; }
    virtual size_t cno() const { return C; }
    
    /**
    * The C-style [] operator provides unchecked access.
    * The two-argument function call operator provides optional checked access
    * when Check is true, and FORTRAN-style indexing when Ftnidx is true.
    */
    virtual typename vector<T_>::const_iterator operator[](size_t Idx) const
    	{ return (elems().begin()+Idx*cno()); }
    virtual typename vector<T_>::iterator operator[](size_t Idx)
    	{ clone(); return (elems().begin()+Idx*cno()); }
    virtual const T_& operator()(size_t Ridx, size_t Cidx) const;
    virtual T_& operator()(size_t Ridx, size_t Cidx);
    
    /**
    * size(Rno, Cno): changes the size of the calling object to Rno x Cno.
    * Nothing is done if the size parameters are the same as the current
    * size or if either of them is 0. When the matrix "shrinks", then
    * the upper left corner is preserved, when it grows, then the new
    * rows and/or columns are padded with "zeroes" T_(0).
    */
    void size(size_t Rno, size_t Cno);

    /**
    * row(Idx): returns the Idx-th row as a Vector_ object safely.
    * row(Vec,Idx): sets the elements in the Idx-th row to the elements
    * of the vector Vec, provided Idx is legal and Vec has the right
    * number of elements.
    */
    virtual Vector_<T_> row(size_t Idx) const;
    virtual void row(const Vector_<T_>& Vec, size_t Idx);

    /**
    * col(Idx): returns the Idx-th column as a Vector_ object safely.
    * col(Vec,Idx): sets the elements in the Idx-th column to the elements
    * of the vector Vec, provided Idx is legal and Vec has the right
    * number of elements.
    */
    virtual Vector_<T_> col(size_t Idx) const;
    virtual void col(const Vector_<T_>& Vec, size_t Idx);

    	/// Simple scalar operations
    
    Rectmat_<T_> operator*(const T_& Factor) const;
    Rectmat_<T_> operator/(const T_& Div) const;
    
    	/// Linear algebra
    
    /**
    * Matrix addition and subtraction.
    * Dimexc_ is thrown on dimension mismatches.
    */
    Rectmat_<T_>& operator+=(const Rectmat_<T_>& Mat);
    Rectmat_<T_> operator+(const Rectmat_<T_>& Mat) const;
    Rectmat_<T_>& operator-=(const Rectmat_<T_>& Mat);
    Rectmat_<T_> operator-(const Rectmat_<T_>& Mat) const;

    /**
    * Matrix*matrix and matrix*vector multiplication.
    * Dimexc_ is thrown on dimension mismatches.
    */
    Rectmat_<T_> operator*(const Rectmat_<T_>& Mat) const;
    Vector_<T_> operator*(const Vector_<T_>& Vec) const;

    	/// Transpose
    
    /// get_transpose(): returns the transposed calling object.
    Rectmat_<T_> get_transpose() const;

    /**
    * transpose_inplace(): replaces the calling object with its transpose.
    * Returns a reference to the calling object.
    */
    Rectmat_<T_>& transpose_inplace();

    	/// Diagonals
    
    /**
    * diag(): copies the diagonal into a Vector_ object.
    * diag(Vec): sets the diagonal to the values in the Vector_ object Vec.
    * If the dimensions don't match, a Dimexc_ exception is thrown.
    * If the calling object is not a square matrix, then Squarexc_ is thrown.
    */
    virtual Vector_<T_> diag() const;
    virtual void diag(const Vector_<T_>& Vec);

    /**
    * diag_matrix(Dval): turns the calling object into a diagonal matrix 
    * with the value Dval in all diagonal positions. 
    * If the matrix is not square, a Squarexc_ exception is thrown.
    */
    virtual void diag_matrix(const T_& Dval);

    /**
    * trace(): returns the sum of the diagonal elements.
    * If the matrix is not square, a Squarexc_ exception is thrown.
    */
    virtual T_ trace() const;

    	/// I/O
    virtual void read_matrix(istream& In);
    
    /// hidden methods
    protected:
    
    void print_rows(ostream& Out, size_t Rw, size_t Jbeg, 
	size_t Items, size_t Width, size_t Prec) const;
    
};
// END OF CLASS Rectmat_

// ---- OUT-OF-CLASS FUNCTIONS ----

// These were originally friend templates but EGCS 1.1.2
// did not like them. 12-Jul-1999.

/**
 * Pre-multiplication of a rectangular matrix with a scalar.
 */
template <class T_>
Rectmat_<T_> operator*(const T_& Factor, const Rectmat_<T_>& Bm);

// ---- TYPEDEFS ----

// Float and double matrices (for convenience)
typedef Rectmat_<double> Rmatdbl_;
typedef Rectmat_<float> Rmatflt_;

} // RazorBack

#ifdef INCLUDE_TMPL_DEFS
#include "Rectmat.cc"
#endif

// ==== END OF TEMPLATE HEADER Rectmat.hh ====

#endif	    // RECTMAT_TMPL_HEADER
