#ifndef STAT2_CLASS_HEADER
#define STAT2_CLASS_HEADER

// ==== HEADER Stat2.hh ====

/**
 * \file Stat2.hh
 * 1 and 2-variable statistics (avg,SD,skewness,kurtosis,corr).
 */

// 11-May-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <stdlib.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
using namespace std;
#else
#include <iostream.h>
#endif

// ---- MODULE HEADERS ----

#include "Utilsexc.hh"
#include "Statexc.hh"
#include "Iovar.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Stat_: one-variable statistics. A Stat_ object works
 * much in the same way as the stat operations on a programmable
 * calculator. Double data are accumulated in "stat sums" and
 * the average and SD can be obtained from these.
 */
class Stat_
{
    // data
    static const double BIG_NUMBER;	///< for min,max init
    static const double SMALL_NUMBER;	///< for near-zero testing
    
    double Sx, Sx2, Sx3, Sx4;	///< sums for moments
    double Min, Max;	///< minimal and maximal value
    unsigned int N;	///< number of data
    
    // methods
    public:
    
    /// constructor
    Stat_(): Sx(0.0), Sx2(0.0), Sx3(0.0), Sx4(0.0),
	    Min(BIG_NUMBER), Max(-BIG_NUMBER), N(0) {}
    
    /// clear(): resets the calling object.
    void clear()
    {
	Sx=Sx2=Sx3=Sx4=0.0;
	Min=BIG_NUMBER; Max=-Min; N=0;
    }
    
    /// Stat+=Val: adds a new data point Val to Stat. Returns calling object
    Stat_& operator+=(double Val);
    
    /// Returns the number of data points seen.
    unsigned int data_no() const { return N; }
    
   /**
    * min(), max(): return the minimal/maximal value throws Statexc_ if
    * there were no data items.
    */
    double min() const;
    double max() const;
        
   /**
    * avg(): returns the average. Throws Statexc_ if
    * there were no data items at all.
    */
    double avg() const;
    
   /**
    * sd(): returns the standard deviation. Throws Statexc_ 
    * if there were no data items at all. If N=1, 0.0 is returned automatically.
    */
    double sd() const;
    
    /**
     * skew(): returns the estimated skewness. Throws Statexc_
     * if there were no data items at all. If N==1 happens to be 0.0,
     * then 0.0 is returned. Throws Sdevexc_ if the s.d. was not larger than 0.0.
     */
    double skew() const;
    
    /**
     * kurt(): returns the estimated kurtosis. Note that 3.0 is NOT subtracted
     * from this value (as opposed to the Numerical Recipes convention),
     * so the kurtosis of a Gaussian distribution should be about 3.0.
     * Throws Statexc_ if there were no data items at all.
     * If N==1 happens to be 0.0, then 0.0 is returned.
     * Throws Sdevexc_ if the s.d. was not larger than 0.0.
     */
    double kurt() const;
    
    // I/O (serialization)
    
    /**
     * Prints a Stat_ object to Out. The format is:-
     * \verbatim
       8 STAT
       N = 1000
       MIN = -4.7362
       MAX = 6.30655
       SUM = 957.413
       SUM2 = 4575.58
       SUM3 = 11291.8
       SUM4 = 57072.3
       ENDSTAT
       \endverbatim
     */
    friend ostream& operator<<(ostream& Out, const Stat_& S);
    
    /**
     * Reads a Stat_ object from In. The format is:-
     * \verbatim
       8 STAT
       N = 1000
       MIN = -4.7362
       MAX = 6.30655
       SUM = 957.413
       SUM2 = 4575.58
       SUM3 = 11291.8
       SUM4 = 57072.3
       ENDSTAT
       \endverbatim
     */
    friend istream& operator>>(istream& In, Stat_& S);
};
// END OF CLASS Stat_

/**
 * Class Stat2_: Two-variable simple statistics class. A Stat2_ object
 * contains two Stat_ objects (Xs and Ys) plus a mixed-sum member
 * Sxy for correlation. Data point addition is done by pairs. The
 * one-var stats (min/max/avg, SD) are done via returning Xs or Ys
 * and then using the appropriate Stat_ methods.
 */
class Stat2_
{
    // data
    Stat_ Xs, Ys;   ///< X and Y fields
    double Sxy;	    ///< mixed product sum
    
    // methods
    public:
	    
    /// constructor
    Stat2_(): Xs(), Ys(), Sxy(0.0) {}
    
    /// clear(): clear the calling object.
    void clear() { Xs.clear(); Ys.clear(); Sxy=0.0; }
    
    /// add(): adds a pair of values X,Y to the calling object.
    void add(double X, double Y) { Xs+=X; Ys+=Y; Sxy+=X*Y; }
    
    /// Returns the number of data pairs seen.
    unsigned int data_no() const { return(Xs.data_no()); }
    
   /**
    * xs(), ys(): return the one-stat objects for X and Y. Use
    * the Stat_ methods on these to obtain avg, sd, min, max.
    */
    const Stat_& xs() const { return(Xs); }
    const Stat_& ys() const { return(Ys); }
    
   /**
    * corr(): returns the correlation coefficient between the X and Y
    * points. Throws Statexc_ if there were no data points.
    * Throws Sdevexc_ if the s.d. was not larger than 0.0.
    */
    double corr() const;
};
// END OF CLASS Stat2_

// ==== END OF HEADER Stat2.hh ====

} // RazorBack

#endif	// STAT2_CLASS_HEADER
