#ifndef STATPROB_HEADER
#define STATPROB_HEADER

// ==== CLASS HEADER Statprob.hh ====

/**
 * \file
 * Probability integrals for statistical tests.
 * @author Andras Aszodi
 * @date 28-Nov-2001.
 */

// ==== STANDARD HEADERS ====

#include <stdlib.h>
#include <math.h>

#ifdef _STANDARD_C_PLUS_PLUS
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ==== MODULE HEADERS ====

#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

class Statprob_
{
    // static data members
    private:

    static const unsigned int MAXITER;
    static const double EPSILON;
    
    // all methods are static, too
    public:
	    
    // -- Tests --
	    
    /// Returns the tail of the standard normal distribution at z.
    static double stdnorm_tail(double z);
    
    /**
     * Returns the significance level for the t test.
     * If the value returned by t_test() is low, then the probability is large
     * that t is high.
     */
    static double t_test(unsigned int df, double t);
    
    /**
     * Returns the significance level for the test that the
     * F ratio is larger than 1.0.
     * If the value returned by F_test() is low, then the probability is large
     * that F is high.
     */
    static double F_test(unsigned int df1, unsigned int df2, double F);
    
    /**
     * Returns the significance level for the test that the Chi2 is exceeded
     * for a correct model. The less the probability, the better. :-)
     */
    static double chi2_test(unsigned int df, double Chi2);
    
    /**
     * Returns the Q probability value for the Studentized Range.
     * df is the degrees of freedom, kno is the number of groups.
     */
    static double studrange_prob(double q, unsigned int df, unsigned int kno);
    
    // -- Special functions --
    
    /// The regularized incomplete beta function.
    static double ibeta_regul(double a, double b, double x);
    
    /**
     * The regularized incomplete gamma function
     * (1/Gamma(a))*INTEGRAL[exp(-t)*t^(a-1) dt, 0, x].
     */
    static double igamma_regul(double a, double x);
    
    // hidden methods
    protected:
    
    static double gamma_series(double a, double x);
    static double gamma_contfrac(double a, double x);
    static double beta_contfrac(double a, double b, double x);
};
// END OF CLASS Statprob_

} // RazorBack

// ==== END OF CLASS HEADER Statprob.hh ====

#endif	    // STATPROB_HEADER
