#ifndef PDBREAD_HEADER
#define PDBREAD_HEADER

/* ==== HEADER pdbread.h ==== */

/**
 * \file pdbread.h
 * Protein Data Bank I/O routines.
 * Can read and write PDB files that correspond to the
 * PDB 2.1 format specification. Not all PDB cards are implemented.
 */

/* ANSI C, 1-Oct-1998. Andras Aszodi */

/* The PDB cards supported by this module are marked with a '+'.
 * HEADER +
 * OBSLTE
 * TITLE  +
 * CAVEAT
 * COMPND +
 * SOURCE +
 * KEYWDS
 * EXPDTA +
 * AUTHOR +
 * REVDAT
 * SPRSDE
 * JRNL
 * REMARK 2
 * DBREF
 * SEQADV
 * SEQRES +
 * MODRES
 * HET    +
 * HETNAM +
 * HETSYN
 * FORMUL +
 * HELIX  +
 * SHEET  +
 * TURN   +
 * SSBOND +
 * LINK
 * HYDBND
 * SLTBRG
 * CISPEP
 * SITE   +
 * CRYST1
 * ORIGXn
 * SCALEn
 * MTRIXn
 * TVECT
 * MODEL  +
 * ATOM   +
 * SIGATM
 * ANISOU
 * SIGUIJ
 * TER    +
 * HETATM +
 * ENDMDL +
 * CONECT +
 * MASTER +
 * END    +
 */

/* ---- GLOBAL PDB TYPES ---- */

#ifdef __cplusplus
namespace RazorBack {
#endif

typedef char Str4_[5];  /**< 4-letter words... */
typedef char Str3_[4];	/**< 3-letter words: residue abbreviations */
typedef char Str2_[3];	/**< 2-letter words: element symbols */

/**
 * Flags to indicate macromolecule types: can be ORed together
 * to describe chains which contain different things.
 * PDB_NACID==PDB_RNA|PDB_DNA
 * PDB_BIOPOLY==PDB_PROTEIN|PDB_NACID|PDB_SACCH
 */
typedef enum {PDB_UNKNOWN=0, PDB_PROTEIN=1, PDB_RNA=2, PDB_DNA=4, PDB_NACID=6,
    PDB_SACCH=8, PDB_BIOPOLY=15, PDB_HETERO=16} Typeflags_;

/**
 * These are possible values for the Hetgroup_.Solvent field:
 * PDB_WATER indicates water molecules for which no HET record
 * is provided, PDB_SOLVENT is for other solvents which have
 * a '*' at pos.19 in their FORMUL record and are represented
 * by a HET group (but perhaps with incorrect atom count), 
 * PDB_NOSOLVENT is for all other hetgroups.
 */
typedef enum {PDB_NOSOLVENT=0, PDB_WATER=1, PDB_SOLVENT=2} Solvtype_;

/** a residue or hetgroup */
typedef struct
{
    Str3_ Name;	/**< 3-letter residue code or HET group name */
    char Chid;	/**< chain identifier */
    int Resno;  /**< residue no. within a chain */
    char Rid;   /**< insertion code like in "27A" */    
} Residue_;

/** an atom or heteroatom */
typedef struct
{
    int Atno;	/**< atom serial number as in the PDB entry */
    Str4_ Id;  /**< atom type like CA or OD2: 4 chars max. + \0 */
    char Alt;	/**< alternate location indicator */
    char Res1;     /**< 1-letter residue code or ' ' for OXT or 'X' for unknown etc. */
    Residue_ Res;   /**< residue or hetgroup description */
    float X[3]; /**< the coordinates */
    float Occu, Bfact; /**< occupancy and temperature factor */
    Str4_ Segid;	/**< segment identifier, left justified */
    Str2_ Element;    /**< element symbol, right justified */
    int Charge;	/**< charge on atom */
    long Type;	/**< holds macromolecule typing flag, determined by Res3 */
} Atom_ ;

/** a connection */
typedef struct
{
    int Atomno;	    /**< this little atom is bonded to.... */
    int Bonded[4];  /**< these little atoms... */
    int Hbonded[4]; /**< and hydrogen bonded to these... */
    int Saltbrdg[2];	/**< and salt bridged to these... */
} Connect_;

/** protein secondary structure element types */
typedef enum {PDB_HELIX, PDB_SHEET, PDB_TURN} Sectype_ ;

/** a protein secondary structure element */
typedef struct
{
    Sectype_ Sectype;	/**< type of secondary structure */
    int No;	    /**< serial number of structure */
    Str4_ Id;	    /**< identifier string */
    int Beg, End;   /**< residue nos (1..N) */
    char Chid;	    /**< chain ID, for begin only */
    char Begaa, Endaa;	/**< amino acid 1-letter codes */
    char Begrid, Endrid;    /**< insertion codes */
    int Type;	    /**< 1..10 for helices, -1..1 for sheets, PDB numbering used */
	    /** these entries are for sheets only */
    int Strandno;	/**< number of strands in this sheet */ 
    Str4_ Thisat, Otherat;  /**< atom names for beta registration */
    char Thisaa, Otheraa;   /**< amino acid codes for beta regs */
    int This, Other;	/**< registration AA positions */
    char Thisrid, Otherid, Otherchid;	/**< ins codes and chain ID for the OTHER */
} Secstr_ ;

/** a protein disulfide bridge */
typedef struct
{
    int No; 	    /**< number of disulfide bridge */
    int Pos1, Pos2; /**< CYS seq. positions */
    char Ch1, Ch2, Rid1, Rid2;	/**< chain ID-s and insertion codes */
} Ssbond_ ;

/** a biopolymer chain */
typedef struct
{
    Atom_ *Atoms; /**< array of atoms in this chain */
    int Atomno;   /**< no. of atoms in chain */
    Secstr_ *Secs;    /**< secondary structure info */
    int Secsno;	/**< length of Secs[] */
    Ssbond_ *Ssbs;    /**< array of S-S bonds */
    int Ssbno;	/**< length of Ssbs[] */
    int Resno;     /**< no. of residues in chain */
    char Chid;    /**< chain ID (for multichains) */
    int Modelno;	/**< model number where chain ID is the same */
    long Type;    /**< chain type: ORed Typeflags_ together */
    char *Seq;	/**< real chain sequence in 1-letter code, \0-termin. */
} Chain_ ;

/** a HET group */
typedef struct
{
    Residue_ Het;   /**< hetgroup "residue" description */
    Atom_ *Atoms;   /**< array of atoms:non-NULL for non-biobackbone hets only */
    int Atomno;    /**< no of occurrences as "standalone hetero"=len of Atoms[] */
    int Biopolyno;  /**< no of occurrences in biopolymer chains */
    char Descr[41];   /**< group description string */
    Solvtype_ Solvent;    /**< solvent type */
    int Modelno;    /**< model number */
} Hetgroup_;

/** a HETNAM or FORMUL */
typedef struct
{
    char *Name;		/**< name of hetgroup or formula */
    Str3_ Hetid;	/**< 3-letter het identifier */
    int Compnum;	/**< component number: 0 for HETNAM */
} Hetnamformul_;

/** "theoretical" sequence information */
typedef struct
{
    char Chid;	/**< chain identifier */
    char *Seq;	/**< sequence with 3-letter residue names, length 3*Len+1 */
    int Len;	/**< sequence length */
} Seqinfo_;

/** a site */
typedef struct
{
    Str3_ Name;	/**< site name */
    Residue_ *Res;  /**< array of residues comprising site */
    int Resno;	/**< number of residues in site */
} Site_;

/** a PDB record */
typedef struct
{
    char Header[41];	/**< header info: function description */
    char Date[10];	/**< date of deposition: "20-AUG-64" format */
    Str4_ Pdbcode;	/**< standard PDB code "0XXX" */
    char *Title;	/**< entry title */
    char *Compound;	/**< compound descriptor string */
    char *Source;	/**< species, organ, tissue, mutant */
    char Expdta[61];	/**< experimental technique: default "X-RAY DIFFRACTION" */
    char *Author;	/**< authors: not parsed further */
    float Resol;	/**< resolution; <0 means "NOT APPLICABLE" */
    Seqinfo_ *Seqs;	/**< sequence information for each chain */
    int Seqno;		/**< number of sequence information records */
    Chain_ *Chains;	/**< array of chain records */
    int Chainno;	/**< no. of chains */
    Ssbond_ *Ssbs;	/**< array of interchain S-S bonds */
    int Ssbno;		/**< no, of interchain S-S bonds */
    Hetgroup_ *Hets;	/**< array of hetero groups */
    int Hetsno;		/**< no. of hetero groups */
    Hetnamformul_ *Hetnams; /**< array of HETNAM records */
    int Hetnamsno;	/**< no. of HETNAM records */
    Hetnamformul_ *Hetforms;	/**< array of FORMUL records */
    int Hetformsno;	/**< no. of FORMUL records */
    Site_ *Sites;	/**< array of sites */
    int Siteno;		/**< number of sites */
    Connect_ *Connects;	/**< connection record array */
    int Conno;		/**< number of connection records */
} Pdbentry_ ;

/* ---- PROTOTYPES ---- */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * aa_code31(Aa3):
 * converts 3-letter all-uppercase amino acid abbreviations
 * into 1-char codes. Returns 'X' if it cannot translate Aa3.
 * In addition to the 20 standard AA-s, the following translations
 * are accepted:
 * ALB -> 'A' (beta-alanine)
 * CYH, CSH, CSS, CYX -> 'C' (cysteine and cystine)
 * HYP, PR0, PRZ -> 'P' (hydroxyproline and various Pro synonyms)
 * ILU -> 'I' (isoleucine)
 * TRY-> 'W' (tryptophan)
 * UNK -> 'U' (unknown)
 * any other code is translated to X.
 */
char aa_code31(const char *Aa3);

/**
 * aa_code13(Aa1):
 * converts the 1-letter codes to 3-letter standards.
 * Almost the inverse of aa_code31(). Non-standard translations:
 * J->JJJ, O->OOO, U->UNK, X->XXX
 */
char *aa_code13(char Aa1);

/**
 * nu_code31(Nu3):
 * converts the "3-letter" nucleotide specifications
 * (right-justified space+letter strings like "  A" or " +I")
 * into their 1-letter equivalent. Returns 'X' if it did not work.
 */
char nu_code31(const char *Nu3);

/**
 * read_pdb(Pdbfn):
 * reads a PDB file called Pdbfn and puts the result
 * in a Pdbentry_ which is created inside and returned.
 * If something went wrong then NULL is returned.
 */
Pdbentry_ *read_pdb(const char *Pdbfn);

/**
 * atom_dist(At1, At2):
 * returns the distance between two Atom_ entries.
 */
float atom_dist(const Atom_ *At1, const Atom_ *At2);

/**
 * write_pdb(Pdbfn, Entry, Remarks, Remno):
 * writes the atomic coordinate information in Entry
 * to the file Pdbfn in PDB format. 
 * Remno remarks are supplied in a char array Remarks[].
 */
void write_pdb(const char *Pdbfn, Pdbentry_ *Entry, 
	char *Remarks[], int Remno);

/**
 * alloc_pdbentry(): creates and returns a clean Pdbentry_
 * or NULL if allocation failed.
 */
Pdbentry_ *alloc_pdbentry();

/**
 * free_pdb(Entry): frees up memory allocated to Entry and its arrays.
 * Does not set Entry to NULL. (Destructor in C++)
 */
void free_pdb(Pdbentry_ *Entry);

#ifdef __cplusplus
} }
#endif

/* ==== END OF HEADER pdbread.h ==== */
#endif	/* PDBREAD_HEADER */
