#ifndef STAT2_H
#define STAT2_H

/* ==== HEADER stat2.h ==== */

/**
 * \file stat2.h
 * One- and two-variable statistics.
 * Calculates mean, SD and correlation
 * for double-precision data.
 */

/* ANSI C, 25. Aug. 1997. Andras Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* ---- TYPES ---- */

#ifdef __cplusplus
namespace RazorBack {
#endif

/**
 * Statrec_ : stores the sum and squared sum of X.
 * N is the number of data points.
 */
typedef struct
{
    double Sx, Sx2, Min, Max;
    unsigned long N;
}
Statrec_ ;

/**
 * Stat2rec_ : stores the sums and squared sums and the mixed
 * sum of X and Y. N is the number of data points.
 */
typedef struct
{
    double Sx, Sy, Sx2, Sy2, Sxy, Minx, Maxx, Miny, Maxy;
    unsigned long N;
}
Stat2rec_ ;

/* ---- DEFINITIONS ---- */

/**
 * The Statops_ symbolic constants control the calculation
 * of the avg, SD, Min, Max for the 1-var statistics.
 */
typedef enum {STAT_NOOP=0, STAT_AVG, STAT_SD, STAT_MIN, STAT_MAX} Statops_;

/**
 * The Stat2ops_ symbolic constants control the calculation
 * of the avg's and SD-s for the 2-var statistics.
 */
typedef enum {STAT2_NOOP=0, STAT2_AVGX, STAT2_SDX, STAT2_MINX, STAT2_MAXX, 
	    STAT2_AVGY, STAT2_SDY, STAT2_MINY, STAT2_MAXY, STAT2_CORR} Stat2ops_;

/* ---- PROTOTYPES ---- */

#ifdef __cplusplus
extern "C" {
#endif

/* ---- UNIVARIATE STATISTICS ---- */

/**
 * init_stat(St):
 * initialises the record St.
 */
void init_stat(Statrec_ *St);

/**
 * add_stat(X, St):
 * enters the data X into St.
 */
void add_stat(double X, Statrec_ *St);

/**
 * eval_stat(St, Op):
 * returns the average or SD depending on the Op code.
 */
double eval_stat(const Statrec_ *St, Statops_ Op);

/* ---- BIVARIATE STATISTICS ---- */

/**
 * init_stat2(St):
 * initialises the record St.
 */
void init_stat2(Stat2rec_ *St);

/** 
 * add_stat2(X, Y, St):
 * enters the data pair (X,Y) into St.
 */
void add_stat2(double X, double Y, Stat2rec_ *St);

/**
 * eval_stat2(St, Op):
 * returns the average, SD or correlation values
 * depending on the Op code.
 */
double eval_stat2(const Stat2rec_ *St, Stat2ops_ Op);

#ifdef __cplusplus
} }
#endif

/* ==== END OF HEADER stat2.h ==== */
#endif	/* STAT2_H */
